/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jul 17, 2006
 */
package eu.moonlight3d.ml3d.animation.basic;

import java.util.List;

import com.trolltech.qt.core.QRect;
import com.trolltech.qt.core.QTimer;
import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QBrush;
import com.trolltech.qt.gui.QColor;
import com.trolltech.qt.gui.QComboBox;
import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QIcon;
import com.trolltech.qt.gui.QMouseEvent;
import com.trolltech.qt.gui.QPaintEvent;
import com.trolltech.qt.gui.QPainter;
import com.trolltech.qt.gui.QPushButton;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.preferences.PreferencesGroup;
import eu.moonlight3d.framework.ui.DockWidget;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Manager;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.anim.Animation;
import eu.moonlight3d.ml3d.backend.threedim.anim.AnimationTrack;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;

/**
 * The timeline bar used for navigating within an animation in the
 * main window.
 *
 * @author gregor
 */
public class TimelineView extends DockWidget {
	/**
	 * The drop down which lists all animations in the scene.
	 */
	private QComboBox animationList;
	/**
	 * The start button to jump back to the first frame of the animation
	 */
	private QPushButton startButton;
	/**
	 * The back button to jump back one frame in the animation
	 */
	private QPushButton frameBackButton;
	/**
	 * The play button to start continuous animation playback
	 */
	private QPushButton playButton;
	/**
	 * The stop button to stop animation playback
	 */
	private QPushButton stopButton;
	/**
	 * The forward button to go forward one frame in the animation
	 */
	private QPushButton frameForwardButton;
	/**
	 * The end button to jump to the last frame of the animation
	 */
	private QPushButton endButton;
	/**
	 * The timer used to control animation playback
	 */
	private QTimer playbackTimer;
	
	/**
	 * The animation on which the timeline currently works
	 */
	private Animation animation;
	/**
	 * Start frame of the range to display in the detail area
	 */
	private int detailBeginFrame;
	/**
	 * End frame of the range to display in the detail area
	 */
	private int detailEndFrame;

	/**
	 * Lifecycle listener responsible for making the timeline view react properly
	 * to document lifecycle events. This includes management of animation
	 * manager registration and current animation reference.
	 *
	 * @author gregor
	 */
	private class LifecycleListener implements eu.moonlight3d.framework.document.LifecycleListener {

		/**
		 * Attach the TimelineView's animation manager listener and selection listener 
		 * to the current main document 
		 */
		private void attachListeners() {
			State state=State.getInstance();
			Document document=(Document) state.getDocumentManager().getMainDocument();
			document.getAnimationManager().registerAnimationManagerListener(animationManagerListener);
			try {
				Selection selection=document.getSelectionManager().getSelection("animationSelection");
				selection.registerListener(selectionListener);
			} catch (Exception e) {
				ExceptionDialog.run(e);
			}
		}
		
		/**
		 * Detach the TimelineView's animation manager listener and selection listener
		 * from the current main document 
		 */
		private void detachListeners() {
			State state=State.getInstance();
			Document document=(Document) state.getDocumentManager().getMainDocument();
			document.getAnimationManager().unregisterAnimationManagerListener(animationManagerListener);
			try {
				Selection selection=document.getSelectionManager().getSelection("animationSelection");
				selection.unregisterListener(selectionListener);
			} catch (Exception e) {
				ExceptionDialog.run(e);
			}
		}
		
		public void afterNew() {
			attachListeners();
			rebuildAnimationList();
		}

		public void afterOpen(String filename) {
			attachListeners();			
			rebuildAnimationList();
		}

		public void afterSave(String filename) {
		}

		public void beforeNew() {
			animationList.clear();
		}

		public void beforeOpen(String filename) {
			animationList.clear();
		}

		public void beforeSave(String filename) {
			animationList.clear();
		}

		public void onClose() {
			detachListeners();
			makeAnimationCurrent(null);
			animationList.clear();
		}
		
	}
	
	/**
	 * Listener used internally to keep track of changes in the list of animations.
	 *
	 * @author gregor
	 */
	private class AnimationManagerListener implements eu.moonlight3d.ml3d.backend.threedim.anim.AnimationManagerListener {

		/**
		 * Set the animation on which the timeline works or null to disable
		 * the timeline.
		 * 
		 * @param newAnimation the animation to work on or null
		 */
		private void setAnimation(Animation newAnimation) {
			if(animation!=null) {
				animation.unregisterAnimationListener(animationListener);
			}
			animation=newAnimation;
			if(newAnimation!=null) {
				State state=State.getInstance();
				Document document=(Document) state.getDocumentManager().getMainDocument();
				try {
					Selection selection=document.getSelectionManager().getSelection("animationSelection");
					// set selection type, add this animation
					String animationName="Animation/" + newAnimation.getName();
					selection.clearSelection(Manager.getSelectionType("Animation"));
					selection.add(animationName);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}			
				
				newAnimation.registerAnimationListener(animationListener);
				detailBeginFrame=0;
				detailEndFrame=(int) (newAnimation.getFramesPerSecond()*(newAnimation.getEndTime()-newAnimation.getStartTime()));
				if(detailEndFrame>20) {
					detailEndFrame=20;
				}
			} else {
				// clear animation selection
				State state=State.getInstance();
				Document document=(Document) state.getDocumentManager().getMainDocument();
				try {
					Selection selection=document.getSelectionManager().getSelection("animationSelection");
					selection.clearSelection(Manager.getSelectionType("Animation"));
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}			
			}
			update();
		}

		public void animationAdded(Animation animation) {
			rebuildAnimationList();
			setAnimation(animation);
		}

		public void animationRemoved(Animation animation) {
			rebuildAnimationList();
			// TODO check if currently selected animation is the deleted one
		}
		
	}
	
	/**
	 * Listener used internally to keep track of changes in the currently displayed
	 * animation.
	 *
	 * @author gregor
	 */
	private class AnimationListener implements eu.moonlight3d.ml3d.backend.threedim.anim.AnimationListener {

		/**
		 * Check range of detail area to avoid display of illegal values outside
		 * the animation range.
		 */
		private void checkDetailRange() {
			int animationEndFrame=(int) (animation.getFramesPerSecond()*(animation.getEndTime()-animation.getStartTime()));
			if(detailEndFrame>animationEndFrame) {
				detailEndFrame=animationEndFrame;
			}			
		}

		public void currentTimeUpdated(double oldTime, double newTime) {
			update();
		}

		public void defaultsSet() {
			update();
		}

		public void endTimeUpdated(double oldEndTime, double newEndTime) {
			checkDetailRange();
			update();
		}

		public void startTimeUpdated(double oldStartTime, double newStartTime) {
			checkDetailRange();
			update();
		}

		public void framesPerSecondUpdated(double oldFramesPerSecond, double newFramesPerSecond) {
			checkDetailRange();
			update();
		}
		
		public void trackAdded(AnimationTrack track) {
			// nothing to do here
		}

		public void trackRemoved(AnimationTrack track) {
			// nothing to do here
		}

	}
	
	/**
	 * Selection listener for the current animation selection. Triggers change of
	 * displayed animation on active animation change.
	 *
	 * @author gregor
	 */
	private class SelectionListener implements eu.moonlight3d.ml3d.backend.selection.SelectionListener {

		/**
		 * Update the timeline because the animation selection changed to make the
		 * timeline display the currently active animation at all times
		 */
		private void handleSelectionChange() {
			updateAnimationList();

			Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			try {
				Selection selection=document.getSelectionManager().getSelection("animationSelection");
				
				if(selection.getActiveObject()==null) {
					// TODO handle the case that no animation is active right now
					makeAnimationCurrent(null);
					return;
				}
				
				Animation animation=(Animation) selection.getObjectFromSelectionString(selection.getActiveObject());
				makeAnimationCurrent(animation);
			} catch(Exception e) {
				ExceptionDialog.run(e);
			}
		}
		
		public void elementAdded(String objectName) {
			handleSelectionChange();
		}

		public void elementRemoved(String objectName) {
			handleSelectionChange();
		}

		public void elementUpdated(String oldName, String newName) {
			handleSelectionChange();
		}

		public void elementsAdded(List<String> objectNames) {
			handleSelectionChange();
		}

		public void selectionCleared() {
			handleSelectionChange();
		}
		
	}
	
	/**
	 * Enum denoting the drag mode the TimelineRough widget is currently
	 * in, i.e. which area of the bar was clicked by the user when starting
	 * the drag. 
	 *
	 * @author gregor
	 */
	private enum RoughDragMode {
		/**
		 * Somewhere in the whole bar was clicked (move area)
		 */
		WholeArea,
		/**
		 * The are for the moving the start frame was clicked (move start)
		 */
		Start,
		/**
		 * The are for the moving the end frame was clicked (move end)
		 */
		End
	}
	
	/**
	 * The top bar in the timeline view showing a rough view of the timeline
	 * with the bar to edit the timeline range that is shown in the detail
	 * area below.
	 *
	 * @author gregor
	 */
	private class TimelineCoarse extends QWidget {
		/**
		 * The current drag mode, if any
		 */
		private RoughDragMode roughDragMode;
		/**
		 * Area of the bar inside the widget
		 */
		private QRect barArea;
		/**
		 * Location of the area used to move the start of the bar
		 */
		private QRect start;
		/**
		 * Location of the area used to move the end of the bar
		 */
		private QRect end;
		/**
		 *  X coordinate of start position of dragging operation
		 */
		private int dragStartX;
		/**
		 * saved detailBeginFrame value before start of drag operation
		 */
		private int oldDetailBeginFrame;
		/**
		 * saved detailEndFrame value before start of drag operation
		 */
		private int oldDetailEndFrame;
		
		/**
		 * Default constructor
		 * 
		 * @param parent the widget's parent
		 */
		public TimelineCoarse(QWidget parent) {
			barArea=new QRect();
			start=new QRect();
			end=new QRect();
			setMinimumHeight(16);
		}
		
		/**
		 * Update the on-screen positions of all relevant areas in this widget
		 */
		private void updateAreaPositions() {
			int numFrames=(int) (animation.getFramesPerSecond()*(animation.getEndTime()-animation.getStartTime()));
			double divisionWidth=width()/(float)(numFrames+1);
			int areaBegin=(int)(divisionWidth*detailBeginFrame);
			int areaEnd=(int)(divisionWidth*(detailEndFrame+1));
			
			barArea.setRect(areaBegin,3,areaEnd-areaBegin,height()-6);
			start.setRect(areaBegin+4, 6, 8, height()-12);
			end.setRect(areaEnd-12, 6, 8, height()-12);
		}
		
		@Override
		public void mouseMoveEvent(QMouseEvent event) {
			if(animation==null) {
				return;
			}
			
			int deltaX=event.x()-dragStartX;
			
			int numFrames=(int) (animation.getFramesPerSecond()*(animation.getEndTime()-animation.getStartTime()));
			double divisionWidth=width()/(float)(numFrames+1);

			int deltaFrames=(int)(deltaX/divisionWidth);
			
			// handle drags if dragging (both ends and whole bar)
			if(roughDragMode==RoughDragMode.Start) {
				detailBeginFrame=oldDetailBeginFrame+deltaFrames;
				if(detailBeginFrame>=detailEndFrame) {
					detailBeginFrame=detailEndFrame-1;
				}
			} else if(roughDragMode==RoughDragMode.End) {
				detailEndFrame=oldDetailEndFrame+deltaFrames;
				if(detailEndFrame<=detailBeginFrame) {
					detailEndFrame=detailBeginFrame+1;
				}
			} else if(roughDragMode==RoughDragMode.WholeArea) {
				detailBeginFrame=oldDetailBeginFrame+deltaFrames;
				detailEndFrame=oldDetailEndFrame+deltaFrames;
			}
			
			// range check on detailBeginFrame, detailEndFrame
			if(detailBeginFrame<0) {
				detailBeginFrame=0;				
				detailEndFrame=detailBeginFrame+(oldDetailEndFrame-oldDetailBeginFrame);
			}
			
			int lastFrame=(int) (animation.getFramesPerSecond()*(animation.getEndTime()-animation.getStartTime()));
			if(detailEndFrame>lastFrame) {
				detailEndFrame=lastFrame;
				detailBeginFrame=detailEndFrame-(oldDetailEndFrame-oldDetailBeginFrame);
			}
			
			super.mouseMoveEvent(event);
			update();
			timelineDetail.update();
		}
		
		@Override
		public void mousePressEvent(QMouseEvent event) {
			if(animation==null) {
				return;
			}

			updateAreaPositions();
			
			dragStartX=event.x();
			oldDetailBeginFrame=detailBeginFrame;
			oldDetailEndFrame=detailEndFrame;
			
			// find pressed area
			if(start.contains(event.pos())) {
				roughDragMode=RoughDragMode.Start;
			} else if(end.contains(event.pos())) {
				roughDragMode=RoughDragMode.End;				
			} else if(barArea.contains(event.pos())) {
				roughDragMode=RoughDragMode.WholeArea;		
			}

			super.mousePressEvent(event);
		}
		
		@Override
		public void mouseReleaseEvent(QMouseEvent event) {
			if(animation==null) {
				return;
			}

			if(roughDragMode!=null) {
				roughDragMode=null;
			}
			
			super.mouseReleaseEvent(event);
		}
		
		@Override
		public void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter(this);
			
			if(animation!=null) {
				updateAreaPositions();

				// TODO better widget look
				painter.setPen(QColor.black);
				painter.drawRect(0,0,width()-1,height()-1);
				
				State state=State.getInstance();
				PreferencesGroup timelineViewPrefs=state.getPreferencesManager().getCurrentPreferences().getGroupByName("Timeline View");
				Colour coarseBarColour=timelineViewPrefs.getProperty("timelineCoarseBarColour").getColourValue();
				Colour coarseHandleColour=timelineViewPrefs.getProperty("timelineCoarseHandleColour").getColourValue();
				
				Colour.RGB temp;
				temp=coarseBarColour.getsRGB8();
				QColor roughBarQColor=new QColor(temp.red,temp.green,temp.blue);
				temp=coarseHandleColour.getsRGB8();
				QColor roughHandleQColor=new QColor(temp.red,temp.green,temp.blue);
				
				painter.fillRect(barArea, new QBrush(roughBarQColor));
				
				painter.fillRect(start,new QBrush(roughHandleQColor));
				painter.fillRect(end,new QBrush(roughHandleQColor));				
			}
			
			painter.end();
		}
	}

	/**
	 * The timeline detail area showing individual frames.
	 *
	 * @author gregor
	 */
	private class TimelineDetail extends QWidget {
		/**
		 * Default constructor taking the parent widget
		 * 
		 * @param parent the widget's parent
		 */
		public TimelineDetail(QWidget parent) {
			super(parent);
			setMinimumHeight(16);
		}
		
		@Override
		public void mousePressEvent(QMouseEvent event) {
			if(animation==null) {
				return;
			}
			
			// determine the frame that was clicked
			int numFrames=detailEndFrame-detailBeginFrame;
			double divisionWidth=width()/(float)(numFrames);
			int clickedFrame=(int)(detailBeginFrame+event.x()/divisionWidth);
			
			gotoFrame(clickedFrame);
		}
		
		@Override
		public void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter(this);
			
			painter.setBrush(palette().window());
			painter.fillRect(0, 0, width(), height(), painter.brush());
			
			if(animation!=null) {
				// paint timeline
				int numFrames=detailEndFrame-detailBeginFrame;
				double divisionWidth=width()/(float)(numFrames);
				
				int currentFrame=-1;
				if(animation.getCurrentTime()!=null)  {
					currentFrame=(int) (animation.getStartTime()+animation.getCurrentTime()*animation.getFramesPerSecond());
				}
				
				for(int i=0;i<numFrames;i++) {
					int x=(int)(i*divisionWidth);
					
					if(detailBeginFrame+i==currentFrame) {
						painter.setPen(QColor.red);
					} else {
						painter.setPen(QColor.black);
					}
					
					painter.drawLine(x,0,x,height());
					painter.drawText(x+1, height(), Integer.toString(detailBeginFrame+i+1));
				}
			}
			
			painter.end();
		}
	}

	/**
	 * Instance of own document lifecycle listener
	 */
	private LifecycleListener lifecycleListener;
	/**
	 * Instance of own animation manager listener
	 */
	private AnimationManagerListener animationManagerListener;
	/**
	 * Instance of own animation listener
	 */
	private AnimationListener animationListener;
	/**
	 * Instance of own animation selection listener
	 */
	private SelectionListener selectionListener;

	/**
	 * The timeline coarse bar
	 */
	private TimelineCoarse timelineCoarse;
	/**
	 * The timeline detailed bar
	 */
	private TimelineDetail timelineDetail;

	/**
	 * Default constructor
	 */
	public TimelineView() {
		setDefaultArea(Qt.DockWidgetArea.BottomDockWidgetArea);
		
		setTitle(L10N.translate("Timeline"));
		
		timelineCoarse=new TimelineCoarse(this);
		timelineDetail=new TimelineDetail(this);
		QVBoxLayout timelineLayout=new QVBoxLayout();
		timelineLayout.addWidget(timelineCoarse);
		timelineLayout.addWidget(timelineDetail);
		
		State state=State.getInstance();
		
		animationList=new QComboBox(this);
		animationList.setEditable(false);
		animationList.currentIndexChanged.connect(this,"animationChanged()");

		startButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/start.png"),"");
		startButton.setToolTip(L10N.translate("Go to first frame"));
		startButton.clicked.connect(this,"start()");
		frameBackButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/back.png"),"");
		frameBackButton.setToolTip(L10N.translate("Go back one frame"));
		frameBackButton.clicked.connect(this,"frameBack()");
		playButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/play.png"),"");
		playButton.setToolTip(L10N.translate("Play animation"));
		playButton.clicked.connect(this,"play()");
		stopButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/stop.png"),"");
		stopButton.setEnabled(false);
		stopButton.setToolTip(L10N.translate("Stop animation playback"));
		stopButton.clicked.connect(this,"stop()");
		frameForwardButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/forward.png"),"");
		frameForwardButton.setToolTip(L10N.translate("Go forward one frame"));
		frameForwardButton.clicked.connect(this,"frameForward()");
		endButton=new QPushButton(new QIcon(state.getInstallationDirectory() + "/icons/end.png"),"");
		endButton.setToolTip(L10N.translate("Go to last frame"));
		endButton.clicked.connect(this,"end()");
		
		QHBoxLayout layout=new QHBoxLayout();
		layout.setWidgetSpacing(1);
		layout.addLayout(timelineLayout);
		layout.setStretchFactor(timelineLayout, 2);
		
		QVBoxLayout rightSideLayout=new QVBoxLayout();
		rightSideLayout.addWidget(animationList);
		QHBoxLayout buttonLayout=new QHBoxLayout();
		buttonLayout.addWidget(startButton);
		buttonLayout.addWidget(frameBackButton);
		buttonLayout.addWidget(playButton);
		buttonLayout.addWidget(stopButton);
		buttonLayout.addWidget(frameForwardButton);
		buttonLayout.addWidget(endButton);
		buttonLayout.setMargin(0);
		
		rightSideLayout.addLayout(buttonLayout);
		rightSideLayout.setMargin(0);
		
		layout.addLayout(rightSideLayout);
		layout.setMargin(0);
		setLayout(layout);
		
		lifecycleListener=new LifecycleListener();
		state.getDocumentManager().addLifecycleListener(lifecycleListener);
		
		animationManagerListener=new AnimationManagerListener();
		selectionListener=new SelectionListener();
		/*
		 * When the dock widget is created the first relevant new or open was already missed,
		 * so force registration here
		 */ 
		Document document=(Document) state.getDocumentManager().getMainDocument();
		document.getAnimationManager().registerAnimationManagerListener(animationManagerListener);						
		
		animationListener=new AnimationListener();
		
		playbackTimer=new QTimer(this);
		playbackTimer.timeout.connect(this,"playbackTimeout()");
		playbackTimer.setSingleShot(true);
	}

	/**
	 * Set the time of the animation to the time of the given frame, i.e.
	 * jump to the given frame in the animation
	 * 
	 * @param frame the frame to jump to
	 */
	private void gotoFrame(int frame) {
		double frameTime=animation.getStartTime()+frame/animation.getFramesPerSecond();
		try {
			animation.setToTime(frameTime);
			
			Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			document.getOGManager().update();
			
			// force redraw
			update();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * Move a certain number of frames forward or backward in the animation
	 * 
	 * @param numFrames the number of frames to move
	 * @param wrapAround wrap around if running over the start or end of the animation
	 */
	private void moveNumFrames(int numFrames, boolean wrapAround) {
		int currentFrame=(int)((animation.getCurrentTime()-animation.getStartTime())*animation.getFramesPerSecond());
		int newFrame=currentFrame+numFrames;
		
		int lastFrame=(int)((animation.getEndTime()-animation.getStartTime())*animation.getFramesPerSecond());
		if(wrapAround) {
			if(newFrame<0) {
				while(newFrame<0) {
					newFrame+=lastFrame;
				}
			} else if(newFrame>=lastFrame) {
				while(newFrame>=lastFrame) {
					newFrame-=lastFrame;
				}
			}
		} else {
			if(newFrame<0 || newFrame>=lastFrame) {
				return;
			}
		}
		
		gotoFrame(newFrame);
	}
	
	/**
	 * Rebuild the list of animations inside the animationList drop down
	 */
	private void rebuildAnimationList() {
		animationList.clear();
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		eu.moonlight3d.ml3d.backend.threedim.anim.Manager manager=document.getAnimationManager();
		for(Animation animation : manager.getAnimations()) {
			animationList.addItem(animation.getName());
		}
		
		updateAnimationList();
	}
	
	/**
	 * Make the given animation the current one and update all the timeline
	 * components to reflect that.
	 * 
	 * @param newAnimation the animation to make current
	 */
	private void makeAnimationCurrent(Animation newAnimation) {
		if(animation!=null) {
			animation.unregisterAnimationListener(animationListener);
		}
		animation=newAnimation;
		if(newAnimation!=null) {
			newAnimation.registerAnimationListener(animationListener);
			detailBeginFrame=1;
			detailEndFrame=(int) (newAnimation.getFramesPerSecond()*(newAnimation.getEndTime()-newAnimation.getStartTime()));
			if(detailEndFrame>20) {
				detailEndFrame=20;
			}
		}
		update();		
	}
	
	/**
	 * Update the contents of the animation list combo box.
	 */
	private void updateAnimationList() {
		// TODO update animationList from current animation selection
		State.getInstance().logger().fine("TimelineView#updateAnimationList()");
		
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		try {
			Selection selection=document.getSelectionManager().getSelection("animationSelection");
			
			if(selection.getActiveObject()==null) {
				// TODO handle the case that no animation is active right now
				return;
			}
			
			Animation animation=(Animation) selection.getObjectFromSelectionString(selection.getActiveObject());
			
			int index=-1;
			for(int i=0;i<animationList.count();i++) {
				if(animationList.itemText(i).equals(animation.getName())) {
					index=i;
					break;
				}
			}
			animationList.setCurrentIndex(index);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * Handle the selection of a different animation in the timeline
	 * view
	 */
	protected void animationChanged() {
		State.getInstance().logger().fine("TimelineView#animationChanged()");
		if(animationList.count()==0) {
			return;
		}
		
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		Selection selection;
		try {
			selection = document.getSelectionManager().getSelection("animationSelection");
			selection.clearSelection(eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("Animation"));
			selection.add("Animation/" + animationList.currentText());
		} catch (Exception e) {
			// TODO Auto-generated catch block
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * A slot to handle a click on the start button
	 */
	protected void start() {
		gotoFrame(0);
	}
	
	/**
	 * A slot to handle a click on the frame back button
	 */
	protected void frameBack() {
		moveNumFrames(-1,false);
	}
	
	/**
	 * A slot to handle a click on the play button
	 */
	protected void play() {
		stopButton.setEnabled(true);
		playButton.setEnabled(false);
		// TODO start animation playback in 3D view
		playbackTimer.setInterval((int)(1000.0/animation.getFramesPerSecond()));
		playbackTimer.start();
	}
	
	/**
	 * A slot to handle a click on the stop button
	 */
	protected void stop() {
		stopButton.setEnabled(false);
		playButton.setEnabled(true);		
		// TODO stop animation playback in 3D view
		playbackTimer.stop();
	}
	
	/**
	 * A slot to handle playbackTimer timeouts while
	 * playing animations
	 */
	protected void playbackTimeout() {
		playbackTimer.start();
		moveNumFrames(1,true);
	}
	
	/**
	 * A slot to handlel a click on the frame forward button
	 */
	protected void frameForward() {
		moveNumFrames(1,false);
	}
	
	/**
	 * A slot to handle a click on the end button
	 */
	protected void end() {
		int lastFrame=(int) (animation.getFramesPerSecond()*(animation.getEndTime()-animation.getStartTime()))-1;
		gotoFrame(lastFrame);
	}
}
