/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Nov 21, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.anim;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.helper.NamedObject;
import eu.moonlight3d.ml3d.backend.selection.PathElement;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;

/**
 * One animation in this scene. An animation consists of multiple tracks.
 * Each track controls the value of one property over time. A scene in
 * Moonlight|3D may contain more than one animation.
 *
 * @author gregor
 */
public class Animation implements NamedObject, PathElement {
	/**
	 * The name of the animation.
	 */
	private String name;
	/**
	 * List of animation tracks for operator graph node properties.
	 */
	private ArrayList<AnimationTrack> tracks;
	/**
	 * The animation manager to which this animation belongs.
	 */
	private Manager animationManager;
	/**
	 * Start time of this animation in seconds
	 */
	private double startTime;
	/**
	 * End time of this animation in seconds
	 */
	private double endTime;
	/**
	 * Frames per seconds used for rendering and playback
	 */
	private double framesPerSecond;
	/**
	 * Current time in animation
	 */
	private double currentTime;
	/**
	 * Remembers whether this animation was last reset to defaults or set to a specific time.
	 */
	private boolean defaultsSet;
	/**
	 * List of all registered animation listeners
	 */
	private ArrayList<AnimationListener> animationListeners;
	
	/**
	 * Internal constructor. Use {@link Manager#createAnimation(String)} to
	 * create a new animation.
	 *  
	 * @param manager the manager for this animation
	 */
	Animation(Manager manager) {
		animationManager=manager;
		tracks=new ArrayList<AnimationTrack>();
		animationListeners=new ArrayList<AnimationListener>();
		startTime=0.0;
		endTime=5.0;
		framesPerSecond=24.0;
		currentTime=0.0;
		name="Animation";
		defaultsSet=true;
	}
	
	/**
	 * Set the name of this animation
	 * 
	 * @param name new name for this animation
	 */
	public void setName(String name) {
		this.name=name;
	}
	
	public String getName() {
		return name;
	}

	/**
	 * Register an animation listener. The listener be notified from
	 * this point forward.
	 * 
	 * @param listener the listener to register
	 */
	public void registerAnimationListener(AnimationListener listener) {
		animationListeners.add(listener);
	}
	
	/**
	 * Unregister an animation listener. The listener will no longer
	 * be notified.
	 * 
	 * @param listener the listener to unregister
	 */
	public void unregisterAnimationListener(AnimationListener listener) {
		animationListeners.remove(listener);
	}
	
	/**
	 * Get start time of animation
	 * 
	 * @return start time of animation in seconds
	 */
	public double getStartTime() {
		return startTime;
	}
	
	/**
	 * Set start time of animation
	 * 
	 * @param startTime start time of animation in seconds
	 */
	public void setStartTime(double startTime) {
		double oldStartTime=this.startTime;
		this.startTime=startTime;
		
		for(AnimationListener listener : animationListeners) {
			listener.startTimeUpdated(oldStartTime, startTime);
		}
	}
	
	/**
	 * Get end time of animation
	 * 
	 * @return end time of animation in seconds
	 */
	public double getEndTime() {
		return endTime;
	}
	
	/**
	 * Set end time of animation
	 * 
	 * @param endTime end time of animation in seconds 
	 */
	public void setEndTime(double endTime) {
		double oldEndTime=this.endTime;
		this.endTime=endTime;
		
		for(AnimationListener listener : animationListeners) {
			listener.endTimeUpdated(oldEndTime, endTime);
		}
	}
	
	/**
	 * Get number of frames per second
	 * 
	 * @return number of frames per second
	 */
	public double getFramesPerSecond() {
		return framesPerSecond;
	}
	
	/**
	 * Set number of frames per second
	 * 
	 * @param framesPerSecond number of frames per second
	 */
	public void setFramesPerSecond(double framesPerSecond) {
		double oldFramesPerSecond=this.framesPerSecond;
		this.framesPerSecond=framesPerSecond;
		
		for(AnimationListener listener : animationListeners) {
			listener.framesPerSecondUpdated(oldFramesPerSecond, framesPerSecond);
		}
	}
	
	/**
	 * Add a track to the animaton.
	 * 
	 * @param track the track to add
	 */
	public void addTrack(AnimationTrack track) {
		tracks.add(track);
		
		for(AnimationListener listener : animationListeners) {
			listener.trackAdded(track);
		}
	}
	
	/**
	 * Removee a track from the animation.
	 * 
	 * @param track the track to remove
	 */
	public void removeTrack(AnimationTrack track) {
		tracks.remove(track);
		
		for(AnimationListener listener : animationListeners) {
			listener.trackRemoved(track);
		}
	}

	/**
	 * Get the track for the given node and property if it exists.
	 * 
	 * @param nodeName the node that the track is for
	 * @param propertyName the property that the track is for
	 * @return the track for the given node and property or null if it does not exist
	 */
	public AnimationTrack getTrackByName(String nodeName, String propertyName) {
		for(AnimationTrack track : tracks) {
			if(track.getNodeName().equals(nodeName) && track.getPropertyName().equals(propertyName)) {
				return track;
			}
		}
		
		return null;
	}
	
	/**
	 * Get a list of all tracks in this animation.
	 * 
	 * @return list of all animation tracks in this animation
	 */
	public List<AnimationTrack> getTracks() {
		return (List<AnimationTrack>) tracks.clone();
	}
	
	/**
	 * Set the current position within this animation to the given
	 * time and update the operator graph.
	 * 
	 * @param time time to set
	 * @throws Exception if the operator graph update fails
	 */
	public void setToTime(double time) throws Exception {
		defaultsSet=false;
		eu.moonlight3d.ml3d.backend.threedim.og.Manager ogManager=animationManager.getDocument().getOGManager();
		for(AnimationTrack track : tracks) {
			Node node=ogManager.getNodeByName(track.getNodeName());
			node.getProperty(track.getPropertyName()).setValue(track.getValue(time));
		}
		
		for(AnimationListener listener : animationListeners) {
			listener.currentTimeUpdated(this.currentTime, time);
		}
		
		this.currentTime=time;
	}
	
	/**
	 * Return the current time to which this animation is set.
	 * @return current time or null if defaults are set
	 */
	public Double getCurrentTime() {
		if(defaultsSet) {
			return null;
		} else {
			return currentTime;
		}
	}

	/**
	 * Return the scene to the default values for the animated properties.
	 * This usually is a neutral positions of the animated objects for from
	 * this animation
	 * 
	 * @throws Exception if the operator graph update fails
	 */
	public void resetDefaults() throws Exception {
		defaultsSet=true;
		eu.moonlight3d.ml3d.backend.threedim.og.Manager ogManager=animationManager.getDocument().getOGManager();
		for(AnimationTrack track : tracks) {
			Node node=ogManager.getNodeByName(track.getNodeName());
			node.getProperty(track.getPropertyName()).setValue(track.getDefaultValue());
		}
		
		for(AnimationListener listener : animationListeners) {
			listener.defaultsSet();
		}
	}
	
	public void readXML(Element root) throws Exception {
		startTime=Double.parseDouble(root.getAttribute("starttime"));
		endTime=Double.parseDouble(root.getAttribute("endtime"));
		framesPerSecond=Double.parseDouble(root.getAttribute("framespersecond"));
		defaultsSet=Boolean.parseBoolean(root.getAttribute("defaultsset"));
		currentTime=Double.parseDouble(root.getAttribute("currenttime"));
		
		NodeList tracksList=root.getElementsByTagName("track");
		for(int i=0;i<tracksList.getLength();i++) {
			Element trackElement=(Element) tracksList.item(i);
			String type=trackElement.getAttribute("type");
			
			// FIXME this is a very ugly shortcut; there should really be a factory pattern here
			if(type.equals("keyframe")) {
				KeyframeAnimationTrack track=new KeyframeAnimationTrack();
				track.loadXML(trackElement);
				addTrack(track);
			}
		}
	}
	
	public void saveXML(Document document, Element root) throws DOMException, Exception {
		Element animationNode=document.createElement("animation");
		animationNode.setAttribute("name", name);
		animationNode.setAttribute("starttime",Double.toString(startTime));
		animationNode.setAttribute("endtime",Double.toString(endTime));
		animationNode.setAttribute("framespersecond",Double.toString(framesPerSecond));
		animationNode.setAttribute("defaultsset",Boolean.toString(defaultsSet));
		animationNode.setAttribute("currenttime",Double.toString(currentTime));
		
		for(AnimationTrack track : tracks) {
			Element trackNode=document.createElement("track");
			track.saveXML(document, trackNode);
			animationNode.appendChild(trackNode);
		}
		
		root.appendChild(animationNode);
	}
	
	public PathElement findChildElement(String childName) {
		for(AnimationTrack track : tracks) {
			if(track.getNodeName().equals(childName)) {
				return track;
			}
		}
		return null;
	}
	
	public boolean isValidAppendixName(String appendix) {
		return false;
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}
}
