/*
 * Moonlight|3D
 * Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Nov 16, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.anim;

import java.util.ArrayList;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.selection.PathElement;

/**
 * Keyframe base animation track. It interpolates between
 * various key frames. It is able to use different 
 * interpolation formulas between keyframes.
 *
 * @author gregor
 */
public class KeyframeAnimationTrack implements AnimationTrack {
	
	/**
	 * The basic key/value pair to store the keyframe values.
	 *
	 * @author gregor
	 */
	public class KeyValuePair {
		/**
		 * The time for this key.
		 */
		public double time;
		/**
		 * The value for the key.
		 */
		public Value value;
	}

	/**
	 * The name of the operator graph node that contains
	 * the property that attached to this track.
	 */
	private String nodeName;
	/**
	 * The name of the property that is attached to this
	 * track.
	 */
	private String propertyName;
	/**
	 * The default value of the track, that is, the default
	 * value that should be used if the track is not active.
	 */
	private Value defaultValue;
	/**
	 * The list of key/value pairs used in this track.
	 */
	private ArrayList<KeyValuePair> keys;
	
	/**
	 * Extrapolator that is used before the first key.
	 */
	private KeyframeInterpolator startExtrapolator;
	/**
	 * Extrapolator that is used after the last key.
	 */
	private KeyframeInterpolator endExtrapolator;
	/**
	 * List of interpolators that are used between keys.
	 */
	private ArrayList<KeyframeInterpolator> interpolators;

	public KeyframeAnimationTrack() {
		keys=new ArrayList<KeyValuePair>();
		interpolators=new ArrayList<KeyframeInterpolator>();
	}

	public void setNodeName(String nodeName) {
		this.nodeName=nodeName;
	}
	
	public String getNodeName() {
		return nodeName;
	}
	
	public void setPropertyName(String propertyName) {
		this.propertyName=propertyName;
	}

	public String getPropertyName() {
		return propertyName;
	}

	public void setDefaultValue(Value value) {
		// don't use clone, because the passed value might also be of type Property
		defaultValue=new Value();
		defaultValue.copyFrom(value);
	}
	
	public Value getDefaultValue() {
		return defaultValue;
	}
	
	/**
	 * Insert a key at a given time. The given key/value pair is inserted into the
	 * track and the given interpolator is used as the new interpolator between the
	 * new key and the following key.
	 * 
	 * @param time the time of the new key
	 * @param value the value at the new key time
	 * @param interpolator the new interpolator to use
	 */
	public void insertKey(double time, Value value, KeyframeInterpolator interpolator) {
		KeyValuePair key=new KeyValuePair();
		key.time=time;
		/*
		 * use this instead of clone() because value might be a Property, but 
		 * there must always be a Value referenced in key.value.
		 */
		key.value=new Value();
		key.value.copyFrom(value);

		for(int i=0;i<keys.size();i++) {
			if(keys.get(i).time>time) {
				keys.add(i,key);
				interpolators.add(i,interpolator);
				return;
			}
		}
		keys.add(key);
		interpolators.add(interpolator);
	}
	
	public void removeKey(double time) {
		for(KeyValuePair key : keys) {
			if(key.time==time) {
				keys.remove(key);
				return;
			}
		}
	}
	
	public void removeKey(KeyValuePair key) {
		keys.remove(key);
	}

	public Value getValue(double time) {
		if(keys.size()==0) {
			return defaultValue;
		}
		if(keys.get(0).time>=time) {
			return startExtrapolator.interpolate(this,time);
		}
		if(time>keys.get(keys.size()-1).time) {
			return endExtrapolator.interpolate(this,time);
		}
		for(int i=0;i<keys.size();i++) {
			if(time>keys.get(i).time) {
				return interpolators.get(i).interpolate(this, time);
//				return keys.get(i).value;
			}
		}
		
		return null; // this line should never actually be reached
	}
	
	/**
	 * Get a list of all keys in this track.
	 * 
	 * @return list of keys
	 */
	public ArrayList<KeyValuePair> getKeyList() {
		return keys;
	}
	
	/**
	 * Set the start extrapolator for this animation track. The start
	 * extrapolator is applied to extrapolate from the first keyframe
	 * in the track backwards in time to the desired point in time.
	 * Despite the name, normal interpolators are used for that operation.
	 * It is up to the interpolator implementation to calculate meaningful
	 * values.
	 * 
	 * @param extrapolator the start extrapolator
	 */
	public void setStartExtrapolator(KeyframeInterpolator extrapolator) {
		startExtrapolator=extrapolator;
	}
	
	/**
	 * Return the start extrapolator that is used to interpolate from the
	 * first keyframe backwards in time.
	 * 
	 * @return the start extrapolator
	 */
	public KeyframeInterpolator getStartExtrapolator() {
		return startExtrapolator;
	}
	
	/**
	 * Set the end extrapolator for this animation track. The end
	 * extrapolator is applied to extrapolate from the last keyframe
	 * in the track forward in time to the desired point in time.
	 * Despite the name, normal interpolators are used for that operation.
	 * It is up to the interpolator implementation to calculate meaningful
	 * values.
	 * 
	 * @param extrapolator the end extrapolator
	 */
	public void setEndExtrapolator(KeyframeInterpolator extrapolator) {
		endExtrapolator=extrapolator;
	}

	/**
	 * Return the end extrapolator that is used to interpolate from the
	 * last keyframe forward in time.
	 * 
	 * @return the end extrapolator
	 */
	public KeyframeInterpolator getEndExtrapolator() {
		return endExtrapolator;
	}
	
	/**
	 * Get the proper KeyframeInterpolator instance the type with the given
	 * type name
	 * 
	 * @param type type name to look up
	 * @return a KeyframeInterpolator of the given type or null if type is invalid
	 */
	private KeyframeInterpolator getKeyframeInterpolator(String type) {
		// FIXME this should be a factory pattern
		if(type.equals("LinearInterpolator")) {
			return new LinearValueInterpolator();
		}
		
		return null;
	}
	
	public void loadXML(Element root) throws Exception {
		// some general stuff:
		nodeName=root.getAttribute("nodename");
		propertyName=root.getAttribute("propertyname");		
//		startExtrapolator=getKeyframeInterpolator(root.getAttribute("startextrapolator"));
//		endExtrapolator=getKeyframeInterpolator(root.getAttribute("endextrapolator"));
		startExtrapolator=getKeyframeInterpolator("LinearInterpolator");
		endExtrapolator=getKeyframeInterpolator("LinearInterpolator");
		
		NodeList defaultValueElementList=root.getElementsByTagName("defaultvalue");
		if(defaultValueElementList.getLength()!=1) {
			throw new Exception(L10N.translate("there must be exactly one <defaultvalue> tag in a keyframe animation track"));
		}
		Element defaultValueElement=(Element) defaultValueElementList.item(0);
		
		defaultValue=new Value();
		defaultValue.readFromFile((Element) defaultValueElement.getElementsByTagName("value").item(0), false);
		
		// keys:
		NodeList keysList=root.getElementsByTagName("keys");
		if(keysList.getLength()!=1) {
			throw new Exception(L10N.translate("there must be exactly one <keys> tag in a keyframe animation track"));
		}
		Element keysElement=(Element) keysList.item(0);
		NodeList keyList=keysElement.getElementsByTagName("key");
		for(int i=0;i<keyList.getLength();i++) {
			Element keyNode=(Element) keyList.item(i);
			KeyValuePair keyValuePair=new KeyValuePair();
			keyValuePair.time=Double.parseDouble(keyNode.getAttribute("time"));
			keyValuePair.value=new Value();
			keyValuePair.value.readFromFile((Element) keyNode.getElementsByTagName("value").item(0), false);
			
			keys.add(keyValuePair);
		}
		
		// interpolators:
//		NodeList interpolatorsList=root.getElementsByTagName("interpolators");
//		if(interpolatorsList.getLength()!=1) {
//			throw new Exception(L10N.translate("there must be exactly one <interpolators> tag in a keyframe animation track"));
//		}
//		Element interpolatorsElement=(Element) interpolatorsList.item(0);
//		NodeList interpolatorList=interpolatorsElement.getElementsByTagName("interpolator");
//		for(int i=0;i<interpolatorList.getLength();i++) {
//			Element interpolatorElement=(Element) interpolatorList.item(i);
//			KeyframeInterpolator interpolator=getKeyframeInterpolator(interpolatorElement.getAttribute("type"));
//			
//			interpolators.add(interpolator);
//		}
		for(int i=0;i<keys.size();i++) {
			interpolators.add(getKeyframeInterpolator("LinearInterpolator"));
		}
	}

	public void saveXML(Document document, Element root) throws DOMException, Exception {
		root.setAttribute("type","keyframe");
		root.setAttribute("nodename", nodeName);
		root.setAttribute("propertyname", propertyName);
//		root.setAttribute("startextrapolator",startExtrapolator.getType());
//		root.setAttribute("endextrapolator",endExtrapolator.getType());

		Element defaultValueElement=document.createElement("defaultvalue");
		defaultValue.writeToFile(document, defaultValueElement);
		root.appendChild(defaultValueElement);
		
		Element keysNode=document.createElement("keys");
		
		for(KeyValuePair key : keys) {
			Element keyNode=document.createElement("key");
			keyNode.setAttribute("time", Double.toString(key.time));
			key.value.writeToFile(document, keyNode);
			keysNode.appendChild(keyNode);
		}
		root.appendChild(keysNode);
		
//		Element interpolatorsNode=document.createElement("interpolators");
//		for(KeyframeInterpolator interpolator : interpolators) {
//			Element interpolatorNode=document.createElement("interpolator");
//			interpolatorNode.setAttribute("type", interpolator.getType());
//			interpolatorsNode.appendChild(interpolatorNode);
//		}
//		root.appendChild(interpolatorsNode);
	}

	public PathElement findChildElement(String childName) {
		return null;
	}

	public boolean isValidAppendixName(String appendix) {
		return false;
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}
}
