/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Dec 2, 2007
 */

package eu.moonlight3d.ml3d.backend.threedim.helper;

import eu.moonlight3d.framework.helper.StringTools;

/**
 * This class contains a collection of helper functions that
 * are useful when dealing with selection strings in the
 * specific context of a Moonlight|3D document. These 
 * functions provide easy ways to parse, analyze and build
 * selection strings.
 *
 * @author gregor
 */
public class SelectionHelper {
	/**
	 * Return the scene graph only portion of the given selection
	 * string. 
	 * 
	 * @param selectionString the selection string to select the
	 * scene graph part from.
	 * @return the path or null if the selection string does not
	 * point to a scene graph node
	 */
	public static String getSGPath(String selectionString) {
		if(selectionString==null) {
			return null;
		}
		
		String[] path=selectionString.split("/");
		
		if(path.length<4) {
			return null;
		}
		
		if(!"OperatorGraph".equals(path[0])) {
			return null;
		}
		
		String sgNodePath=path[3];
		for(int i=4;i<path.length;i++) {
			sgNodePath+="/" + path[i];
		}
		sgNodePath=sgNodePath.split(":")[0];
		
		return sgNodePath;
	}
	
	/**
	 * Return the name of the scene graph node that the given
	 * selection string points to. This only returns the name
	 * of the node where the given path ends.
	 * 
	 * 
	 * @param selectionString the selection string to extract 
	 * the scene graph node name from
	 * @return the node name or null if the selection string 
	 * does not point to a scene graph node
	 */
	public static String getSGNodeName(String selectionString) {
		if(selectionString==null) {
			return null;
		}
		
		String[] path=selectionString.split("/");
		
		if(path.length<4) {
			return null;
		}
		
		if(!"OperatorGraph".equals(path[0])) {
			return null;
		}
		
		return path[path.length-1].split(":")[0];
	}
	
	/**
	 * Return the name of the operator graph node that the given
	 * selection string points to
	 * 
	 * @param selectionString the selection string to extract the
	 * operator graph node name from
	 * @return the operator graph node name or null if the given
	 * selection string does not point to one
	 */
	public static String getOGNodeName(String selectionString) {
		if(selectionString==null) {
			return null;
		}
		
		String[] path=selectionString.split("/");
		
		if(path.length<2) {
			return null;
		}
		
		if(!"OperatorGraph".equals(path[0])) {
			return null;
		}
		
		return path[1];
	}

	/**
	 * Return the name of the operator graph slot that the given
	 * selection string points to
	 * 
	 * @param selectionString the selection string to extract the
	 * operator graph slot name from
	 * @return the operator graph slot name or null if the given
	 * selection string does not point to one
	 */
	public static String getOGSlotName(String selectionString) {
		if(selectionString==null) {
			return null;
		}
		
		String[] path=selectionString.split("/");
		
		if(path.length<3) {
			return null;
		}
		
		if(!"OperatorGraph".equals(path[0])) {
			return null;
		}
		
		return path[2];
	}
	
	/**
	 * Strip any appendix - if present - from the given selection
	 * string and return the resulting selection string up to the
	 * node that is actually referenced.
	 * 
	 * @param selectionString the selection string to strip
	 * @return the stripped string up to the last node
	 */
	public static String stripAppendix(String selectionString) {
		return StringTools.split(selectionString, ':')[0];
	}
}
