/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.og;
/**
 * \package eu.moonlight3d.ml3d.backend.og
 * 
 * This is the operator graph base. This packages contains all
 * the base classes for the the operator graph of Moonlight|3D.
 * 
 * This package is part of the core program.
 */

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.NamePool;
import eu.moonlight3d.framework.helper.Observed;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.selection.PathElement;
import eu.moonlight3d.ml3d.backend.selection.RootPathElement;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;


/**
 * This is the operator graph manager class, through which the operator graph 
 * is built and accessed
 * 
 * @author gregor
 */
public class Manager extends Observed implements RootPathElement {
	/**
	 * A list of operator graph node factories.
	 */
    private static ArrayList<eu.moonlight3d.framework.helper.Factory> factories;
    
    static {
        factories=new ArrayList<eu.moonlight3d.framework.helper.Factory>();    	
    }

    /**
     * Register a node factory for a certain kind of operator graph nodes.
     * Nodes will solely be constructed through the registered factory.
     * 
     * @param factory new operator graph node factory
     */
    public static void registerNodeFactory(eu.moonlight3d.framework.helper.Factory factory)
    {
        factories.add(factory);
    }
    
    /**
     * Unregister a node factory for a certain kind of operator graph nodes.
     * Once this happens, no new nodes of this kind will be created.
     * 
     * @param name the name of the operator graph node factory
     * @throws Exception if factory was not found
     */
    public static void unregisterNodeFactory(String name) throws Exception
    {
        for(int i=0;i<factories.size();i++) {
            if(factories.get(i).getName()==name) {
                factories.remove(i);
                return;
            }
        }
        
        throw new Exception(L10N.translate("unable to find factory with given name"));
    }
    
    /**
     * The list of operator graph nodes in this graph.
     */
    protected ArrayList<Node> nodes;
    /**
     * A map between operator graph node names and the respective nodes.
     */
    protected HashMap<String, Node> nodeMap;
    /**
     * The list of links between nodes in this graph.
     */
    protected ArrayList<Link> links;
    /**
     * The pool of names of operator graph nodes used to ensure name uniqueness.
     */
    protected NamePool names;
    /**
     * The list of registered operator graph listeners.
     */
    protected ArrayList<OperatorGraphListener> listeners;
    /**
     * Reference to the document that this operator graph is part of.
     */
    protected Document document;

    /**
     * Default constructor taking the document that contains this graph
     * as an argument.
     * 
     * @param document the document that contains this graph.
     */
    public Manager(Document document) {
        super();
        this.document=document;
        nodes=new ArrayList<Node>();
        nodeMap=new HashMap<String, Node>();
        links=new ArrayList<Link>();
        names=new NamePool();
        listeners=new ArrayList<OperatorGraphListener>();
    }
      
    /**
     * Register an operator graph listener. From that point onwards it will be called
     * whenever a noteworthy change is made in the operator graph.
     * 
     * @param listener the listener to register
     */
    public void registerOperatorGraphListener(OperatorGraphListener listener) {
    	listeners.add(listener);
    }
    
    /**
     * Unregister an operator graph listener. From that point onwards, the listener
     * will no longer be notified of any changes.
     * 
     * @param listener the listener to remove
     */
    public void unregisterOperatorGraphListener(OperatorGraphListener listener) {
    	listeners.remove(listener);
    }
    
    /**
     * Clear operator graph. Used to start out with an fresh, empty scene.
     * Please note that clearing the operator graph may not be sufficient
     * for generating a consistent state in the backend. See the frontend
     * tools on details.
     *
     */
    public void clear()
    {
    		names.clear();
    		nodes.clear();
    		nodeMap.clear();
    		links.clear();
            
            for(OperatorGraphListener listener : listeners) {
            	listener.graphCleared();
            }    		
    }
    
    /**
     * Create a node of the given kind, e.g. "MoveObject" and insert it
     * into the operator graph. You will still have to linke that node's
     * slots afterwards to define the it's input and output. The node
     * will be named after its node type, with eventually a running
     * number added to ensure unique names.
     *  
     * @param name the name of the node type of the new node
     * @return the new node
     * @throws Exception
     */
    public Node createNodeByName(String name) throws Exception
    {
        for(int i=0;i<factories.size();i++) {
            if(factories.get(i).getName().equals(name)) {
                Node node=(Node)factories.get(i).createNew();
                if(node==null) {
                    return null;
                }
                
                node.setName(names.generateUniqueName(name));
                node.setType(name);
                node.setDocument(document);
                 
                addNode(node);
                return node;
            }
        }
        
        throw new Exception(L10N.translate("unable to find factory with given name"));        
    }
  
    /**
     * Create a node of the given kind, e.g. "MoveObject" and insert it
     * into the operator graph. You will still have to linke that node's
     * slots afterwards to define the it's input and output. The node
     * will be given the provided name with eventually a running
     * number added to ensure unique names.
     * 
     * @param typeName the name of the node type of the new node
     * @param name the name of the new node
     * @return the new node
     * @throws Exception
     */
    public Node createNodeByName(String typeName, String name) throws Exception
    {
        if(!names.isNameUnique(name)) {
        	throw new Exception(L10N.translate("attempted to create OG node with existing name"));
        }
        
        for(int i=0;i<factories.size();i++) {
            if(factories.get(i).getName().equals(typeName)) {
                Node node=(Node)factories.get(i).createNew();
                if(node==null) {
                    return null;
                }
                
                node.setName(name);
                node.setType(typeName);
                node.setDocument(document);
                 
                addNode(node);
                return node;
            }
        }
        
        throw new Exception(L10N.translate("unable to find factory with given name"));        
    }

    /**
     * Adds a given node the operator graph. Links between slots have to be created
     * seperately.
     * 
     * @param node operator graph node to add
     * @throws Exception thrown if the name of the node is not unique in this graph
     */
    public void addNode(Node node) throws Exception {
		if (!names.isNameUnique(node.getName())) {
			throw new Exception(L10N.translate("given node does not have a unique name"));
		}
		names.add(node);
		nodes.add(node);
		if(nodeMap.containsKey(node.getName())) {
			throw new Exception(L10N.translate("internal error: key already present in node map"));
		}
		nodeMap.put(node.getName(),node);

		for (OperatorGraphListener listener : listeners) {
			listener.nodeInserted(node);
		}
	}
    
    /**
	 * Remove a certain node from the operator graph. Note that this does not
	 * remove any links to other operator graph nodes that this node may have.
	 * 
	 * @param node
	 *            the node to remove
	 * @throws Exception
	 *             if node is not found in operator graph
	 */
    public void removeNode(Node node) throws Exception
    {
        for(int i=0;i<nodes.size();i++) {
           if(nodes.get(i)==node) {
               node.removeDependencies();
               
               nodes.remove(i);
               names.remove(node);
               nodeMap.remove(node.getName());

               for(OperatorGraphListener listener : listeners) {
            	   listener.nodeRemoved(node);
               }
               
               return;
           }
        }
        
        throw new Exception(L10N.translate("unable to find given node"));
    }
    
    /**
     * Get a list of all nodes in this graph.
     * 
     * @return a list of all operator graph nodes in this graph
     */
    public List<Node> getNodeList()
    {
        return (ArrayList<Node>)nodes.clone();
    }
    
    /**
     * Get the operator graph node with the given name.
     * 
     * @param name the name of the node
     * @return the node with the given name
     * @throws IllegalArgumentException if no node exists with the given name
     */
    public Node getNodeByName(String name) throws IllegalArgumentException {
    	Node node=nodeMap.get(name);
    	if(node!=null) {
    		return node;
    	}
         
    	throw new IllegalArgumentException(L10N.translate("unable to find node with given name"));        
	}
        
    /**
     * adds a link to the node graph
     * 
     * @param output the output slot which will be connected as the link input
     * @param input the input slot which will be connected as the link output
     * @return the new link in the graph
     * @throws Exception
     */
    public Link addLink(Slot output, Slot input) throws Exception
    {
        try {
            Link link=new Link();
            link.setInput(output);
            link.setOutput(input);
            links.add(link);
            
            for(OperatorGraphListener listener : listeners) {
            	listener.linkInserted(link);
            }
            
            return link;
        } catch (Exception e) {
            throw new Exception(L10N.translate("unable to add link to operator graph"),e);
        }
    }
    /**
     * removes a link from the node graph
     * 
     * @param link the link to be removed
     */
    public void removeLink(Link link)
    {
		if (link == null) {
			return;
		}

		links.remove(link);

		link.getInput().setCurrentLink(null);
		link.getOutput().setCurrentLink(null);

		for (OperatorGraphListener listener : listeners) {
			listener.linkRemoved(link);
		}
	}
    /**
	 * returns a list of all links that currently are in the node graph
	 * 
	 * @return a list of links in this operator graph
	 */
    public ArrayList<Link> getLinkList()
    {
        return (ArrayList<Link>)links.clone();
    }
    
    /**
     * Helper function for inserting new nodes into the scene graph. This function
     * will insert the node with the two slots newInput and newOutput after the
     * the slot oldOutput of an existing  node. If oldOutput was connected to Slot
     * oldInput before the operation, then oldOutput will be connected to newInput
     * and newOutput will be connected to oldInput afterwards.
     * 
     * @param oldOutput output slot of a valid OG node
     * @param newInput input slot of a new OG node
     * @param newOutput output slot of a new OG node
     * @throws Exception
     */
    public void insertNodeAfter(Slot oldOutput, Slot newInput, Slot newOutput) throws Exception {
		if (oldOutput.getCurrentLink() == null) {
			addLink(oldOutput, newInput);
		} else {
			Slot oldInput = oldOutput.getCurrentLink().getOutput();
			removeLink(oldOutput.getCurrentLink());
			addLink(oldOutput, newInput);
			addLink(newOutput, oldInput);
		}
	}
  
    /**
	 * Update the whole operator graph
	 * 
	 * @throws Exception
	 */
    public void update() throws Exception {
    	State.getInstance().logger().finer("enter update()");
    	
    	HashSet<Node> unprocessed=new HashSet<Node>();
    	ArrayList<Node> processable=new ArrayList<Node>();
    	HashSet<Node> processed=new HashSet<Node>();
    	ArrayList<Node> errorNodes=new ArrayList<Node>();
    	ArrayList<Exception> errorNodeExceptions=new ArrayList<Exception>();
    	
    	for(Node node : nodes) {
    		if(node.isPureSource()) {
    			processable.add(node);
    		} else {
    			unprocessed.add(node);
    		}
    	}
    	
    	do {
        	ArrayList<Node> nextProcessable=new ArrayList<Node>();
    		
        	for(Node node : processable) {
        		try {
        			State.getInstance().logger().finer("processing node " + node.getName());

        			// update node if required
        			if(node.hasUpdatedProperties()) {
        				node.update();
        				node.clearHasUpdatedProperties();

        				// mark all nodes that receive inputs from this node to require updates
        				List<Slot> slots=node.getSlots();
        				for (int i = 0; i < slots.size(); i++) {
        					if (slots.get(i).getType() == Slot.Type.Output) {
       							if (slots.get(i).getCurrentLink() != null) {
       								slots.get(i).getCurrentLink().getOutput().getNode().setHasUpdatedProperties();
       							}
        					}
        				}
            			
            	    	// send node update notifications to listeners
            	    	for(OperatorGraphListener listener : listeners) {
            	    		listener.nodeUpdated(node);
            	    	}
        			}

        	    	processed.add(node);
        			
        			// check if connected nodes are already processable
        			for(Slot slot : node.getSlots()) {
        				if(slot.getType()!=Slot.Type.Output) {
        					continue;
        				}
        				if(slot.getCurrentLink()!=null) {
        					Node nextNode=slot.getCurrentLink().getOutput().getNode();
        					
        					if(!unprocessed.contains(nextNode)) {
        						// no need to check this node if it has already been taken out of unprocessed
        						continue;
        					}
        					
        					// check if all inputs are connected to nodes in processed
        					boolean isProcessable=true;
        					for(Slot nextSlot : nextNode.getSlots()) {
                				if(nextSlot.getType()!=Slot.Type.Input) {
                					continue;
                				}
        						if(nextSlot.getCurrentLink()==null) {
        							isProcessable=false;
        							break;
        						}
        						if(!processed.contains(nextSlot.getCurrentLink().getInput().getNode())) {
        							isProcessable=false;
        							break;
        						}
        					}
        			
        					// if check succeeded, move node to nextProcessable
        					if(isProcessable) {
        						unprocessed.remove(nextNode);
        						nextProcessable.add(nextNode);
        					}
        				}
        			}
        		} catch(Exception e) {
        			State.getInstance().logger().warning(MessageFormat.format(L10N.translate("failed to evaluate node {0}"),node.getName()));
        			errorNodes.add(node);
        			errorNodeExceptions.add(e);
        		}
        	}
        	
        	processable=nextProcessable;
    	} while(!processable.isEmpty());

		// notify about error nodes with their exceptions
    	for(int i=0;i<errorNodes.size();i++) {
    		Node node=errorNodes.get(i);
    		Exception nodeException=errorNodeExceptions.get(i);
	    	for(OperatorGraphListener listener : listeners) {
	    		listener.nodeUpdateFailed(node, nodeException);
	    	}    		
    	}
    	
		// notify about unprocessed nodes
		for(Node node : unprocessed) {
			Exception nodeException=null;
	    	for(OperatorGraphListener listener : listeners) {
	    		listener.nodeUpdateFailed(node, nodeException);
	    	}
		}
    	
    	// send graph update notifications to listeners
    	for(OperatorGraphListener listener : listeners) {
    		listener.graphUpdated();
    	}

    	notifyObservers();
        
    	State.getInstance().logger().finer("leave update()");
    	
    }
    
	public PathElement findChildElement(String childName) {
		try {
			return getNodeByName(childName);
		} catch (Exception e) {
			return null;
		}
	}

	public boolean isValidAppendixName(String appendix) {
		return false;
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}

	public String getName() {
		return "OperatorGraph";
	}
}
