/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.og;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import eu.moonlight3d.framework.helper.NamedObject;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.ml3d.backend.selection.PathElement;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;


/**
 * This is the base class for all operator graph nodes. New nodes have
 * to be derived from this class.
 * 
 * @author gregor
 */
public abstract class Node extends PropertyContainer implements PathElement, NamedObject {
	/**
	 * The document that this node belongs to
	 */
	private Document document;
	
	/**
	 * The name of this node
	 */
	private String name;

	/**
	 * The list of slots that this node has
	 */
	private ArrayList<Slot> slots;
	/**
	 * A map between slot names and slots
	 */
	private HashMap<String, Slot> slotMap;

	/**
	 * Default constructor
	 */
	public Node() {
		slots = new ArrayList<Slot>();
		slotMap = new HashMap<String, Slot>();
		setHasUpdatedProperties();
	}

	/**
	 * creates a new slot for this node
	 * 
	 * @param name slot name of new slot
	 * @param type type of new slot
	 */
	public void createSlot(String name, Slot.Type type) {
		Slot newSlot = new Slot();

		newSlot.setName(name);
		newSlot.setNode(this);
		newSlot.setType(type);
		slots.add(newSlot);
		slotMap.put(name, newSlot);
	}

	/**
	 * Operator graph node type name
	 */
	private String type;

	/**
	 * This is overloaded by the children of this class to execute
	 * the functionality of the node.
	 * @throws Exception
	 */
	abstract public void update() throws Exception;
	
	/**
	 * This function is is called when an operator graph node is removed
	 * from the scene to clean up any dependencies it may have in other
	 * parts of the program, e.g. the SG node cache.
	 *
	 */
	abstract public void removeDependencies();

	/**
	 * Set node name
	 * 
	 * @param name node name
	 */
	public void setName(String name) {
		this.name = name;
	}

	public String getName() {
		return name;
	}
	
	@Override
	public int hashCode() {
		return name.hashCode();
	}

	/**
	 * Set type name of node
	 * 
	 * @param type node type name
	 */
	public void setType(String type) {
		this.type = type;
	}

	/**
	 * Get type name of node
	 * 
	 * @return node type name
	 */
	public String getType() {
		return type;
	}

	/**
	 * returns the whole array of Slots this Node has. It contains
	 * both Input and Output slots.
	 * 
	 * @return list of slots
	 */
	public List<Slot> getSlots() {
		return (ArrayList<Slot>) slots.clone();
	}

	/**
	 * returns the Slot with the given name assigned to it or 0 if
	 * the requested slot couldn't be found.
	 * 
	 * @param name name of slot
	 * @return Slot the slot with the given name
	 * @throws Exception if slot with given name could not be found
	 */
	public Slot getSlot(String name) throws Exception {
//		for (int i = 0; i < slots.size(); i++) {
//			if (slots.get(i).getName().equals(name)) {
//				return slots.get(i);
//			}
//		}
		Slot slot=slotMap.get(name);
		if(slot!=null) {
			return slot;
		}

		throw new Exception("unable to find slot with given name");
	}

	/**
	 * does this Node have any input Slots?
	 * 
	 * @return true if node is a pure source
	 */
	public boolean isPureSource() {
		for (int i = 0; i != slots.size(); ++i) {
			if (slots.get(i).getType() == Slot.Type.Input) {
				return false;
			}
		}

		return true;
	}

	/**
	 * does this Node have any output Slots?
	 * 
	 * @return true if node is a pure drain
	 */
	public boolean isPureDrain() {
		for (int i = 0; i != slots.size(); ++i) {
			if (slots.get(i).getType() == Slot.Type.Output) {
				return false;
			}
		}

		return true;
	}

	/**
	 * does this Node have any linked input Slots?
	 * 
	 * @return true if node is a source
	 */
	public boolean isSource() {
		for (int i = 0; i != slots.size(); ++i) {
			if (slots.get(i).getType() == Slot.Type.Input && slots.get(i).getCurrentLink() != null) {
				return false;
			}
		}

		return true;

	}

	/**
	 * does this Node have any linked output Slots?
	 * 
	 * @return true if node is a drain
	 */
	public boolean isDrain() {
		for (int i = 0; i != slots.size(); ++i) {
			if (slots.get(i).getType() == Slot.Type.Output && slots.get(i).getCurrentLink() != null) {
				return false;
			}
		}

		return true;
	}

	public PathElement findChildElement(String childName) {
		try {
			return getSlot(childName);
		} catch (Exception e) {
			return null;
		}
	}

	public boolean isValidAppendixName(String appendix) {
		return false;
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}

	/**
	 * Get document that this OG node is attached to
	 * 
	 * @return owning Document
	 */
	public Document getDocument() {
		return document;
	}

	/**
	 * Set owning document of this OG node
	 * 
	 * @param document the owning document
	 */
	final void setDocument(Document document) {
		this.document = document;
	}
}