/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.og;

import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.selection.PathElement;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * This class implements input and output slots for operator graph nodes.
 * Slots are endpoints for Links.
 *
 * @author gregor
 */
public class Slot implements PathElement {
	/**
	 * The slot type. This enum is used for defining the type 
	 * of slots.
	 *
	 * @author gregor
	 */
    public enum Type {
    	/**
    	 * Input slots which receive scene graphs as input for an operator graph node
    	 */
        Input,
        /**
         * Output slots which provide scene graphs as output
         */
        Output
    }
    
    /**
     * The type of this slot
     */
    protected Type type;
    /**
     * The name of this slot. This must be unique within the owning node.
     */
    protected String name;
    /**
     * The operator graph node that is the owner of this slot.
     */
    protected Node node;
    /**
     * The scene graph that this node holds. This is only used in output nodes. 
     */
    protected SceneGraph graph;
    /**
     * The link that this slot is attached to.
     */
    protected Link link;
    
    /**
     * sets the type of this Slot
     * 
     * @param type the type of this slot
     */
    void setType(Type type)
    {
        this.type=type;
    }
    
    /**
     * returns the type of this Slot
     * 
     * @return the type of this slot
     */
    public Type getType()
    {
        return type;
    }
    
    /**
     * sets the name of this Slot
     * 
     * @param name the name of this slot
     */
    void setName(String name)
    {
        this.name=name;
    }
    
    /**
     * returns the name of this Slot
     * 
     * @return the name of this slot
     */
    public String getName()
    {
        return name;
    }
    
    /**
     * sets the Node this Slot is attached to
     * 
     * @param node the node that this slot is attached to
     */
    void setNode(Node node)
    {
        this.node=node;
    }
    
    /**
     * returns the Node this Slot belongs to
     * 
     * @return the node this slot is attached to
     */
    public Node getNode()
    {
        return node;
    }
    
    /**
     * stores a new scene graph in this slot for further processing
     * 
     * @param graph the scene graph to store
     * @throws IllegalStateException if this operation is attempted on an input slot 
     */
    public void setGraph(SceneGraph graph) throws IllegalStateException
    {
        if(type==Type.Input) {
            throw new IllegalStateException(L10N.translate("unable to assign a scene graph to an input node"));
        }

        this.graph=graph;
    }
    
    /**
     * returns the scene graph that is stored in this slot for
     * further processing
     * 
     * @return the stored scene graph or a copy of the scene graph in the linked input slot
     * @throws Exception if the scene graph copying fails
     */
    public SceneGraph getGraph() throws Exception
    {
        try {
            switch(type) {
                case Output:
                    return graph;
                case Input:
                    if(link!=null) {
                        Slot slot=link.getInput();
                        if(slot!=null) {
                            if(slot.graph!=null) {
                                return slot.graph.deepCopy();
                            }
                        }
                    }
                    return null;
            }
            return null;
        } catch (Exception e) {
            throw new Exception(L10N.translate("unable to generate scene graph copy"),e);
        }
    }

    /**
     * Set the the link that this slot is assinged to
     * 
     * @param link the new link
     */
    public void setCurrentLink(Link link)
    {
        this.link=link;
    }
    
    /**
     * Get the link that this slot is assigned to
     * 
     * @return the link that this slot is assigend to
     */
    public Link getCurrentLink()
    {
        return link;
    }

    public PathElement findChildElement(String childName)
    {
        SceneGraph graph;
        
        if(type==Type.Input) {
            if(link!=null) {
                Slot input=link.getInput();
                if(input!=null) {
                    graph=input.graph;
                } else {
                    return null;
                }
            } else {
                return null;
            }
        } else {
            graph=this.graph;
        }
        
        return graph.getNodeByName(childName);
    }

    public boolean isValidAppendixName(String appendix)
    {
        return false;
    }

    public Object getFromAppendix(String appendix)
    {
        return null;
    }
}
