/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Dec 6, 2008
 */

package eu.moonlight3d.ml3d.backend.threedim.resource;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.NamePool;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.selection.PathElement;
import eu.moonlight3d.ml3d.backend.selection.RootPathElement;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;

public class ResourceManager implements RootPathElement {
	public static ArrayList<ResourceFactory> resourceFactories;
	
	static {
		resourceFactories=new ArrayList<ResourceFactory>();
	}
	
	public static void registerResourceFactory(ResourceFactory factory) {
		resourceFactories.add(factory);
	}
	
	public static void unregisterResourceFactory(ResourceFactory factory) {
		resourceFactories.remove(factory);
	}
	
	private Document document;
	
	private ArrayList<Resource> resources;
	private NamePool resourceNames;
	
	private ArrayList<ResourceManagerListener> listeners;
	private ArrayList<ResourceRemoveCheck> removeChecks;
	
	public ResourceManager(Document document) {
		this.document=document;
		resources=new ArrayList<Resource>();
		resourceNames=new NamePool();
		listeners=new ArrayList<ResourceManagerListener>();
		removeChecks=new ArrayList<ResourceRemoveCheck>();
	}
	
	public void registerResourceManagerListener(ResourceManagerListener listener) {
		listeners.add(listener);
	}
	
	public void unregisterResourceManagerListener(ResourceManagerListener listener) {
		listeners.remove(listener);
	}
	
	public void registerResourceRemoveCheck(ResourceRemoveCheck check) {
		removeChecks.add(check);
	}
	
	public void unregisterResourceRemoveCheck(ResourceRemoveCheck check) {
		removeChecks.remove(check);
	}
	
	private void notifyResourceAdded(Resource resource) {
		for(ResourceManagerListener listener : listeners) {
			listener.resourceAdded(resource);
		}
	}
	
	private void notifyResourceRemoved(Resource resource) {
		for(ResourceManagerListener listener : listeners) {
			listener.resourceRemoved(resource);
		}		
	}
	
	public Resource addResource(String type, String name, String filename) {
		for(ResourceFactory factory : resourceFactories) {
			if(factory.getType().equals(type)) {
				name=resourceNames.generateUniqueName(name);
				Resource resource=factory.loadResource(name, filename);
				resource.setDocument(document);
				resources.add(resource);
				resourceNames.add(resource);
				notifyResourceAdded(resource);
				return resource;
			}
		}
		return null;
	}
	
	public boolean canRemoveResource(Resource resource) {
		for(ResourceRemoveCheck check : removeChecks) {
			if(!check.canRemove(resource)) {
				return false;
			}
		}
		
		return true;
	}
	
	public boolean canRemoveResource(String name) {
		Resource resource=getResource(name);
		if(resource!=null) {
			return canRemoveResource(resource);
		}
		
		return true;
	}
	
	public void removeResource(String name) throws Exception {
		for(Resource resource : resources) {
			if(resource.getName().equals(name)) {
				if(!canRemoveResource(resource)) {
					throw new Exception(L10N.translate("cannot remove this resource"));
				}
				resourceNames.remove(resource);
				resources.remove(resource);
				notifyResourceRemoved(resource);
				return;
			}
		}
	}
	
	public Resource getResource(String resourceName) {
		for(Resource resource : resources) {
			if(resource.getName().equals(resourceName)) {
				return resource;
			}
		}
		return null;
	}

	public List<Resource> getResources() {
		return (List<Resource>) resources.clone();
	}
	
	public void write(org.w3c.dom.Document document, Element rootElement) {
		Element resourcesElement=document.createElement("resources");
		
		for(Resource resource : resources) {
			Element resourceElement=document.createElement("resource");
			resourceElement.setAttribute("type", resource.getType());
			resourceElement.setAttribute("name", resource.getName());
			
			resource.write(document, resourceElement);
			
			resourcesElement.appendChild(resourceElement);
		}
		
		rootElement.appendChild(resourcesElement);
	}
	
	public void read(Element rootElement) {
		NodeList resourcesList=rootElement.getElementsByTagName("resources");
		if(resourcesList.getLength()!=1) {
			// this means something is foul with the file format
			State.getInstance().logger().warning(L10N.translate("not exactly one resources tag in the scene data - ignoring resources"));
			return;
		}
		
		Element resourcesElement=(Element) resourcesList.item(0);
		
		NodeList resourceList=resourcesElement.getElementsByTagName("resource");
		
		for(int i=0;i<resourceList.getLength();i++) {
			Element resourceElement=(Element) resourceList.item(i);
			String resourceType=resourceElement.getAttribute("type");
			String resourceName=resourceElement.getAttribute("name");

			for(ResourceFactory factory : resourceFactories) {
				if(factory.getType().equals(resourceType)) {
					Resource resource=factory.readResource(resourceName, resourceElement);
					resource.setDocument(document);
					resources.add(resource);
					resourceNames.add(resource);
					notifyResourceAdded(resource);
					break;
				}
			}
		}
	}

	public PathElement findChildElement(String childName) {
		for(Resource resource : resources) {
			if(resource.getName().equals(childName)) {
				return resource;
			}
		}
		return null;
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}

	public boolean isValidAppendixName(String appendix) {
		return false;
	}

	public String getName() {
		return "Resource";
	}
}
