/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on May 7, 2007
 */

package eu.moonlight3d.ml3d.backend.threedim.sg;

import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Matrix3;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Ray;
import eu.moonlight3d.math.Vector3D;


/**
 * Instance node in the scene graph. This node is used to create instances
 * of geometric objects in the scene graph. Each instance contains a 
 * transformation that is applied to its children. Instances can share children
 * which means that the same shared child objects are visible as children of
 * each of the instance nodes. This makes it possible to display multiple copies
 * of the same object without having to copy the object's complete data.
 *
 * @author gregor
 */
public class Instance extends TransformableNode {
	/**
	 * The transformation matrix which this node contains 
	 */
	private Matrix4 transformationMatrix;
	
	/**
	 * Default constructor.
	 */
	public Instance() {
		super();
		transformationMatrix=new Matrix4();
		transformationMatrix.setIdentity();
	}
	
	@Override
	public void rotate(Quaternion quaternion) {
		transformationMatrix=quaternion.toRotationMatrix4().multiply(transformationMatrix);
	}
	
	@Override
	public void rotate(Vector3D axis, double angle) {
		Quaternion quaternion=new Quaternion();
		quaternion.fromAxisAngle(axis, angle);
		transformationMatrix=quaternion.toRotationMatrix4().multiply(transformationMatrix);
	}
	
	@Override
	public void scale(Vector3D pivot, Vector3D amount)  {
        Matrix4 matrix=new Matrix4();
        
        // recenter pivot on origin
        translate(pivot.multiply(-1));
        
        // build scale matrix
        matrix.clear();
        matrix.X[0][0]=amount.X1;
        matrix.X[1][1]=amount.X2;
        matrix.X[2][2]=amount.X3;
        matrix.X[3][3]=1.0;
        // apply scale matrix
        applyMatrix(matrix);        
        
        // undo translation on pivot
        translate(pivot);
	}
	
	@Override
	public void translate(Vector3D displacement) {
        Matrix4 matrix=new Matrix4();
        
        matrix.clear();
        for(int i=0;i<4;i++) {
            matrix.X[i][i]=1;
        }
        matrix.X[0][3]=displacement.X1;
        matrix.X[1][3]=displacement.X2;
        matrix.X[2][3]=displacement.X3;
        
        applyMatrix(matrix);
	}
	
	@Override
	public void applyMatrix(Matrix3 matrix) {
		transformationMatrix=new Matrix4(matrix).multiply(transformationMatrix);
	}

	@Override
	public void applyMatrix(Matrix4 matrix) {
		transformationMatrix=matrix.multiply(transformationMatrix);
	}

	@Override
	public AxisAlignedBoundingBox getBoundingBox() {
        AxisAlignedBoundingBox res=null;
        List<Node> children=getChildren();
        
        for(Node child : children) {
            if(child instanceof TransformableNode) {
                if(res==null) {
                    res=((TransformableNode)child).getBoundingBox();
                } else {
                    res=res.merge(((TransformableNode)child).getBoundingBox());
                }
            }
        }
        
        if(res==null) {
        	return null;
        }
        
        // apply transformation to all 8 corners and generate new bounding box 
        Vector3D[] corners=res.getAllCorners();
        
        AxisAlignedBoundingBox transformed=new AxisAlignedBoundingBox();
        transformed.min=transformationMatrix.multiply(corners[0]);
        transformed.max=transformed.min.clone();
        
        for(int i=1;i<8;i++) {
        	transformed=transformed.merge(transformationMatrix.multiply(corners[i]));
        }
        
        return transformed;
	}

	@Override
	public Vector3D getIntersectionPoint(Ray ray) {
		// transform ray to from parent to local coordinate system
		Ray transformed=new Ray();
		transformed.start=transformationMatrix.multiply(ray.start);
		transformed.direction=new Matrix3(transformationMatrix).multiply(ray.direction);
		
		// iterate over children and find hit nearest to start
		Vector3D hit=null;
		double distance=0;
		
		for(Node child : getChildren()) {
			if(child instanceof TransformableNode) {
				TransformableNode transformable = (TransformableNode) child;
				Vector3D childHit=transformable.getIntersectionPoint(ray);
				if(childHit!=null) {
					double childDistance=childHit.sub(transformed.start).abs();
					if(childDistance<distance || hit==null) {
						hit=childHit;
						childDistance=distance;
					}
				}
			}
		}
		
		return hit;
	}
	
	/**
	 * Get the translation component of the node's transformation.
	 *. 
	 * @return translation performed by this node
	 */
	public Vector3D getTranslation() {
		return transformationMatrix.getTranslation();
	}
	
	/**
	 * Get the rotation component of the node's transformation.
	 * 
	 * @return rotation performed by this node
	 */
	public Quaternion getRotation() {
		Vector3D x=new Vector3D(1.0,0.0,0.0);
		Vector3D y=new Vector3D(0.0,1.0,0.0);
		Vector3D z=new Vector3D(0.0,0.0,1.0);
		Matrix3 vectorTransformation=new Matrix3(transformationMatrix);
		x=vectorTransformation.multiply(x).norm();
		y=vectorTransformation.multiply(y).norm();
		z=vectorTransformation.multiply(z).norm();
		Quaternion quaternion=new Quaternion();
		quaternion.fromMatrix3(new Matrix3(x,y,z));
		return quaternion;
	}
	
	/**
	 * Get the scale component of the node's transformation.
	 * 
	 * @return scaling performed by this node
	 */
	public Vector3D getScale() {
		// TODO double-check correctness of this method
		Vector3D x=new Vector3D(1.0,0.0,0.0);
		Vector3D y=new Vector3D(0.0,1.0,0.0);
		Vector3D z=new Vector3D(0.0,0.0,1.0);
		Matrix3 vectorTransformation=new Matrix3(transformationMatrix);
		x=vectorTransformation.multiply(x);
		y=vectorTransformation.multiply(y);
		z=vectorTransformation.multiply(z);
		Vector3D scale=new Vector3D();
		scale.X1=x.abs();
		scale.X2=y.abs();
		scale.X3=z.abs();
		return scale;
	}
	
	/**
	 * Get the node's world to local transformation matrix.
	 * 
	 * @return the complete world to local transformation matrix
	 */
	public Matrix4 getTransformationMatrix() {
		return transformationMatrix;
	}
	
	@Override
	public Vector3D transformPointWorldToLocal(Node[] path, Vector3D point) {
		point=transformationMatrix.invert().multiply(point);
		return super.transformPointWorldToLocal(path, point);
	}
	
	@Override
	public Vector3D transformPointLocalToWorld(Node[] path, Vector3D point) {
		point=transformationMatrix.multiply(point);
		return super.transformPointLocalToWorld(path, point);
	}
	
	@Override
	public Vector3D transformVectorWorldToLocal(Node[] path, Vector3D vector) {
		Matrix4 rotateScale=transformationMatrix.clone();
		rotateScale.X[0][3]=0;
		rotateScale.X[1][3]=0;
		rotateScale.X[2][3]=0;
		return rotateScale.invert().multiply(vector);
	}
	
	@Override
	public Vector3D transformVectorLocalToWorld(Node[] path, Vector3D vector) {
		Matrix4 rotateScale=transformationMatrix.clone();
		rotateScale.X[0][3]=0;
		rotateScale.X[1][3]=0;
		rotateScale.X[2][3]=0;
		return rotateScale.multiply(vector);
	}
	
	@Override
	public String getType() {
		return "Instance";
	}

	@Override
	protected void saveXMLData(Document document, Element dataElement) {
		super.saveXMLData(document,dataElement);
		Element matrix=document.createElement("matrix");
		matrix.setAttribute("x00", Double.toString(transformationMatrix.X[0][0]));
		matrix.setAttribute("x01", Double.toString(transformationMatrix.X[0][1]));
		matrix.setAttribute("x02", Double.toString(transformationMatrix.X[0][2]));
		matrix.setAttribute("x03", Double.toString(transformationMatrix.X[0][3]));
		matrix.setAttribute("x10", Double.toString(transformationMatrix.X[1][0]));
		matrix.setAttribute("x11", Double.toString(transformationMatrix.X[1][1]));
		matrix.setAttribute("x12", Double.toString(transformationMatrix.X[1][2]));
		matrix.setAttribute("x13", Double.toString(transformationMatrix.X[1][3]));
		matrix.setAttribute("x20", Double.toString(transformationMatrix.X[2][0]));
		matrix.setAttribute("x21", Double.toString(transformationMatrix.X[2][1]));
		matrix.setAttribute("x22", Double.toString(transformationMatrix.X[2][2]));
		matrix.setAttribute("x23", Double.toString(transformationMatrix.X[2][3]));
		matrix.setAttribute("x30", Double.toString(transformationMatrix.X[3][0]));
		matrix.setAttribute("x31", Double.toString(transformationMatrix.X[3][1]));
		matrix.setAttribute("x32", Double.toString(transformationMatrix.X[3][2]));
		matrix.setAttribute("x33", Double.toString(transformationMatrix.X[3][3]));
		dataElement.appendChild(matrix);
	}

	@Override
	protected void loadXMLData(Element dataElement) throws Exception {
		super.loadXMLData(dataElement);
		Element element=(Element) dataElement.getElementsByTagName("matrix").item(0);
		transformationMatrix.X[0][0]=Double.parseDouble(element.getAttribute("x00"));
		transformationMatrix.X[0][1]=Double.parseDouble(element.getAttribute("x01"));
		transformationMatrix.X[0][2]=Double.parseDouble(element.getAttribute("x02"));
		transformationMatrix.X[0][3]=Double.parseDouble(element.getAttribute("x03"));
		transformationMatrix.X[1][0]=Double.parseDouble(element.getAttribute("x10"));
		transformationMatrix.X[1][1]=Double.parseDouble(element.getAttribute("x11"));
		transformationMatrix.X[1][2]=Double.parseDouble(element.getAttribute("x12"));
		transformationMatrix.X[1][3]=Double.parseDouble(element.getAttribute("x13"));
		transformationMatrix.X[2][0]=Double.parseDouble(element.getAttribute("x20"));
		transformationMatrix.X[2][1]=Double.parseDouble(element.getAttribute("x21"));
		transformationMatrix.X[2][2]=Double.parseDouble(element.getAttribute("x22"));
		transformationMatrix.X[2][3]=Double.parseDouble(element.getAttribute("x23"));
		transformationMatrix.X[3][0]=Double.parseDouble(element.getAttribute("x30"));
		transformationMatrix.X[3][1]=Double.parseDouble(element.getAttribute("x31"));
		transformationMatrix.X[3][2]=Double.parseDouble(element.getAttribute("x32"));
		transformationMatrix.X[3][3]=Double.parseDouble(element.getAttribute("x33"));
	}

	public Object getFromAppendix(String appendix) {
		return null;
	}

	public boolean isValidAppendixName(String appendix) {
		return false;
	}

}
