/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.sg;
/**
 * \package eu.moonlight3d.ml3d.backend.sg
 * 
 * This is the backend scene graph. It contains classes for
 * editable representations of all data in the Moonlight|3D 
 * scenes.
 * 
 * This package is part of the core program.
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.helper.NamedObject;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.selection.PathElement;

/**
 * This is the abstract base class for all scene graph nodes.
 * 
 * @author gregor
 */
public abstract class Node implements Cloneable, PathElement, NamedObject {
    /**
     * Name of this node
     */
    private String name;
    /**
     * List of all parents of this node
     */
    private ArrayList<Node> parents;
    /**
     * List of all children of this node
     */
    private ArrayList<Node> children;
    /**
     * The map of child node names to child nodes.
     */
    private HashMap<String,Node> childrenMap;
    
    /**
     * Reference to the scene graph this node belongs to
     */
    private SceneGraph sceneGraph;
    
    /**
     * Create an empty scene graph nodes. Cannot be used directly because 
     * this class is abstract.
     *
     */
    public Node()
    {
        children=new ArrayList<Node>();
        childrenMap=new HashMap<String, Node>();
        parents=new ArrayList<Node>();
    }
    
    /**
     * Set the name of this scene graph node
     * @param name new name of this node
     */
    public void setName(String name)
    {
        this.name=name;
    }
    
    /**
     * Get the name if this scene graph node
     * @return name of this node
     */
    public String getName()
    {
        return name;
    }
    
    /**
     * Get the scene graph instance to which this node belongs.
     * 
     * @return scene graph which owns this node
     */
	public SceneGraph getSceneGraph() {
		return sceneGraph;
	}

	/**
	 * Set the scene graph instance to which this node belongs.
	 * This operation should only be used within SceneGraph to
	 * ensure correct associations between SceneGraph and Node
	 * objects.
	 * 
	 * @param sceneGraph scene graph which owns this node
	 */
	void setSceneGraph(SceneGraph sceneGraph) {
		this.sceneGraph = sceneGraph;
	}

	/**
     * Add a new reference to a parent node to this scene graph node. 
     * @param node
     */
    void addParent(Node node) {
    	if(parents.contains(node)) {
    		// TODO: error/return
    	}
    	
    	parents.add(node);
    }
    
    /**
     * Remove the reference to the given parent node from this
     * node's parents.
     * 
     * @param node
     */
    void removeParent(Node node) {
    	parents.remove(node);
    }
    
    /**
     * Get all parents of this scene graph node
     * @return parents of this node
     */
    public List<Node> getParents() {
    	return (List<Node>) parents.clone();
    }
    
    /**
     * Add the given node as a new child to the current scene graph node
     * 
     * @param child the node to be added
     * @throws Exception 
     */
    void addChild(Node child) throws Exception
    {
    	if(child==null) {
    		throw new IllegalArgumentException(L10N.translate("child cannot be null"));
    	}
    	
        children.add(child);
        childrenMap.put(child.getName(),child);
    }

    /**
     * Return the child of this scene node with the given name
     * @param name name of the scene node to be queried for
     * @return the node with the requested name
     * @throws Exception if the requested node does not exist.
     */
    public Node getChildByName(String name) throws Exception
    {
    	Node child=childrenMap.get(name);
    	if(child!=null) {
    		return child;
    	}
        
        throw new Exception(L10N.translate("unable to find child with given name"));
    }
    
	/**
	 * @return a copy of the list of children
	 */
	public List<Node> getChildren() {
		return (ArrayList<Node>)children.clone();
	}
	
	/**
	 * Remove the given node from the list of children of this node.
	 * 
	 * @param child the child to remove
	 */
	void removeChild(Node child) {
		children.remove(child);
		childrenMap.remove(child);
	}

    /**
     * Remove a child with a given name if it exists
     *
     * @param name The name of the child to be removed
     */
	public void removeChild(String name)
    {
        Node child=childrenMap.get(name);
        if(child==null) {
        	return;
        }
        children.remove(child);
        childrenMap.remove(child);
    }

	/**
	 * Return a string describing the type of this node.
	 * 
	 * @return the type of this node
	 */
	public abstract String getType();
	
	/**
	 * Load the scene graph node data from the given XML tree.
	 * Each class that wants to add new data to its serialised
	 * form must overload this function and always call the 
	 * superclass implementation as well.
	 * 
	 * @param dataElement the parent element of the XML data for this node
	 * @throws Exception 
	 */
	protected void loadXMLData(Element dataElement) throws Exception {
		// nothing to do here
	}
	
	/**
	 * Load the scene graph node from the given XML tree. This
	 * does the general setup work and calls loadXMLData() for
	 * loading the actual object data.
	 * 
	 * @param parentElement the parent element of the XML tree
	 * @throws Exception 
	 */
	public final void loadFromXML(Element parentElement) throws Exception {
		// load name
		setName(parentElement.getAttribute("name"));
		
		// load data
		NodeList dataList = parentElement.getElementsByTagName("data");
		
		loadXMLData((Element) dataList.item(0));
	}

	/**
	 * Save the scene graph node data to the given XML tree.
	 * Each class that wants to add new data to its serialised
	 * form must overload this function and always call the 
	 * superclass implementation as well.
	 * 
	 * @param document the XML document the data is stored to
	 * @param dataElement the parent element of the XML tree the node data has to be stored to
	 */
	protected void saveXMLData(Document document, Element dataElement) {
		// nothing to do here
	}
	
	/**
	 * Save the scene graph node to the given XML tree.
	 * This method does all the setup work like saving type information
	 * and saving child nodes. For saving the actual node data saveXMLData()
	 * is called.
	 * 
	 * @param document the XML document the data is stored to
	 * @param parentElement the parent element of the XML tree the node has to be stored to
	 */
	public final void saveToXML(Document document, Element parentElement) {
		// save name and type
		parentElement.setAttribute("name",getName());
		parentElement.setAttribute("type",getType());
		
		// save data
		Element dataElement=document.createElement("data");
		saveXMLData(document,dataElement);
		parentElement.appendChild(dataElement);

		// save children
		Element childrenElement=document.createElement("children");
		
		for(Node child : children) {
			Element childElement=document.createElement("child");
			childElement.setAttribute("name", child.getName());
			childrenElement.appendChild(childElement);
		}
		
		parentElement.appendChild(childrenElement);
	}
    
	@Override
	public Node clone() {
		try {
			Node copy=(Node)super.clone();
			copy.children=(ArrayList<Node>) children.clone();
			copy.parents=(ArrayList<Node>) parents.clone();
			copy.childrenMap=(HashMap<String, Node>) childrenMap.clone();
			return copy;
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Create a copy of this node that is completely detached from the
	 * scene graph that contains the original. This means in particular
	 * that all parent and child relationships are removed from the copy.
	 * 
	 * @return the detached copy
	 * @throws Exception if the copying fails for any reason
	 */
	public Node detachedCopy() throws Exception {
		Node copy=this.clone();
		
		copy.childrenMap = new HashMap<String, Node>();
		copy.children = new ArrayList<Node>();
		copy.parents = new ArrayList<Node>();
		copy.sceneGraph = null;

		return copy;
	}
	
    /**
     * Generate a deep copy of this scene node and all of its children.
     * 
     * @param clearParents true if list of parents should be cleared, false if it should be preserved 
     * @return deep copy of this scene node
     * @throws Exception if the deep copy could not be created
     */
    Node deepCopy() throws Exception {
		Node copy = null;

		copy = this.clone();

		copy.childrenMap = new HashMap<String, Node>();
		copy.children = (ArrayList<Node>) children.clone();
		copy.parents = (ArrayList<Node>) parents.clone();

		return copy;
	}
    
    /**
     * Replace the references to a set of nodes with references to other nodes.
     * This affects the list of children and parents of this node.
     * 
     * @param replacementMap the map of old nodes to new nodes to replace them
     */
    void replaceReferences(Map<Node, Node> replacementMap) {
    	ArrayList<Node> newChildren=new ArrayList<Node>();
    	for(Node child : children) {
    		if(replacementMap.get(child)==null) {
    			throw new IllegalStateException(L10N.translate("no replacement found for child node: ") + child.getName());
    		}
    		newChildren.add(replacementMap.get(child));
    		childrenMap.put(child.getName(), replacementMap.get(child));
    	}
    	children=newChildren;
    	ArrayList<Node> newParents=new ArrayList<Node>();
    	for(Node parent : parents) {
    		if(replacementMap.get(parent)==null) {
    			throw new IllegalStateException(L10N.translate("no replacement found for parent node: ") + parent.getName());
    		}
    		newParents.add(replacementMap.get(parent));
    	}
    	parents=newParents;
    }
    
    public PathElement findChildElement(String path)
    {
        try {
            return getChildByName(path);
        } catch(Exception e) {
            return null;
        }
    }
}
