/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Jun 1, 2007
 */

package eu.moonlight3d.ml3d.backend.threedim.sg;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.helper.NamePool;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;

/**
 * The scene graph container class. It manages all nodes in a scene
 * graph including their relationships.
 *
 * @author gregor
 */
public class SceneGraph implements Cloneable {
	/**
	 * The list of scene graph nodes in this scene graph.
	 */
	private ArrayList<Node> sgNodes;
	/**
	 * The name pool used for scene graph nodes to ensure unique node names.
	 */
	private NamePool sgNodeNames;
	/**
	 * The document that this scene graph belongs to.
	 */
	private Document document;
	
	/**
	 * Default constructor for an empty scene graph taking the document it belongs
	 * to as an argument.
	 * 
	 * @param document the document that it belongs to
	 */
	public SceneGraph(Document document) {
		this.document=document;
		sgNodes=new ArrayList<Node>();
		sgNodeNames=new NamePool();
	}
	
	/**
	 * Create a node with the given type and insert it into the scene graph. The
	 * name of the node will be set to preferredName with eventually a number added
	 * to ensure uniqueness of the name.
	 * 
	 * @param typeName the name of the node type
	 * @param preferredName the preferred name of the new node
	 * @return the new node
	 * @throws Exception
	 */
	public Node createNode(String typeName, String preferredName) throws Exception {
		Node node=document.getSGManager().createNode(typeName);
		String uniqueName=sgNodeNames.generateUniqueName(preferredName);
		node.setName(uniqueName);
		sgNodes.add(node);
		sgNodeNames.add(node);
		return node;
	}
	
	/**
	 * Add a node to this scene graph.
	 * 
	 * @param node the node to add
	 * @throws Exception if the node cannot be added
	 */
	public void addNode(Node node) throws Exception {
		for(Node child : node.getChildren()) {
			if(!sgNodes.contains(child)) {
				throw new Exception(MessageFormat.format(L10N.translate("Node {0} references children that are not part of this scene graph."),node.getName()));
			}
		}
		
		for(Node parent : node.getParents()) {
			if(!sgNodes.contains(parent)) {
				throw new Exception(MessageFormat.format(L10N.translate("Node {0} references parents that are not part of this scene graph."),node.getName()));
			}			
		}
		
		if(sgNodes.contains(node)) {
			return;
		}
		
		if(!sgNodeNames.isNameUnique(node.getName())) {
			throw new Exception(MessageFormat.format(L10N.translate("Node name {0} is not unique."),node.getName()));
		}
		
		sgNodeNames.add(node);
		sgNodes.add(node);
		node.setSceneGraph(this);
	}

	/**
	 * Remove a node from the scene graph
	 * 
	 * @param node scene graph node to remove 
	 */
	public void removeNode(Node node) {
		sgNodeNames.remove(node);
		sgNodes.remove(node);
	}
	
	/**
	 * Return the node with the given name.
	 * 
	 * @param name node nmae
	 * @return the node with the given name or null if the node does not exist
	 */
	public Node getNodeByName(String name) {
		for(Node node : sgNodes) {
			if(node.getName().equals(name)) {
				return node;
			}
		}
		
		return null;
	}
	
	/**
	 * Return a list of all nodes in the graph
	 * 
	 * @return flat list of all graph nodes
	 */
	public List<Node> getNodes() {
		return (ArrayList<Node>) sgNodes.clone();
	}
	
	/**
	 * Return a list of all root nodes in the graph. Root
	 * nodes are nodes with no parents.
	 * 
	 * @return a flat list of root nodes 
	 */
	public List<Node> getRootNodes() {
		ArrayList<Node> rootNodes=new ArrayList<Node>();
		
		for(Node node : sgNodes) {
			if(node.getParents().size()==0) {
				rootNodes.add(node);
			}
		}
		
		return rootNodes;
	}
	
	/**
	 * Generate a unique node name for this scene graph based on the given name.
	 * 
	 * @param name the name to use as starting point for the unique name
	 * @return the given name decorated in such a way that it is unique in this graph
	 */
	public String generateUniqueNodeName(String name) {
		return sgNodeNames.generateUniqueName(name);
	}
	
	/**
	 * Add a child to the given scene graph node.
	 * 
	 * @param parent the parent node to which that child is added
	 * @param child the child node that is added to the parent
	 * @throws Exception if a node from another scene graph is passed to this function
	 */
	public void addChild(Node parent, Node child) throws Exception {
		// check if child is in graph
		if(!sgNodes.contains(child)) {
			throw new Exception(L10N.translate("cannot add a non-member to scene graph"));
		}
		
		parent.addChild(child);
		child.addParent(parent);
	}
	
	/**
	 * Remove a child from the given scene graph node.
	 * 
	 * @param parent the parent node from which the child is removed
	 * @param child the child node that is removed from the parent
	 */
	public void removeChild(Node parent, Node child) {
		parent.removeChild(child);
		child.removeParent(parent);
	}
	
	/**
	 * Load the scene graph from below a given parent element
	 * in an XML file.
	 * 
	 * @param parentElement parent element in the source MXL file
	 * @throws Exception if scene graph data is not valid
	 */
	public final void loadFromXML(Element parentElement) throws Exception {
		// load list of nodes, load graph structure
		Element sgElement=(Element) parentElement.getElementsByTagName("scenegraph").item(0);
		NodeList savedNodes=sgElement.getChildNodes();
		
		// first pass: load node data
		for(int i=0;i<savedNodes.getLength();i++) {
			if(savedNodes.item(i).getNodeType()!=org.w3c.dom.Node.ELEMENT_NODE) {
				continue;
			}
			Element element=(Element) savedNodes.item(i);
			if(!"node".equals(element.getNodeName())) {
				continue;
			}
			Node node=createNode(element.getAttribute("type"),element.getAttribute("name"));
			node.loadFromXML(element);
		}
		
		// second pass: assign children
		for(int i=0;i<savedNodes.getLength();i++) {
			if(savedNodes.item(i).getNodeType()!=org.w3c.dom.Node.ELEMENT_NODE) {
				continue;
			}
			Element nodeElement=(Element) savedNodes.item(i);
			if(!"node".equals(nodeElement.getNodeName())) {
				continue;
			}

			NodeList childrenList=((Element) nodeElement.getElementsByTagName("children").item(0)).getChildNodes();

			for(int j=0;j<childrenList.getLength();j++) {
				org.w3c.dom.Node childNode=childrenList.item(j);
				if(childNode.getNodeType()!=org.w3c.dom.Node.ELEMENT_NODE) {
					continue;
				}
				Element childElement=(Element) childNode;
				if(!childElement.getNodeName().equals("child")) {
					continue;
				}
				
				Node node=getNodeByName(nodeElement.getAttribute("name"));
				Node child=getNodeByName(childElement.getAttribute("name"));
				addChild(node,child);
			}
		}
	}
	
	/**
	 * Save the complete scene graph to an XML file below
	 * the given element.
	 * 
	 * @param document XML document to which to save
	 * @param parentElement parent element in the XML document
	 */
	public final void saveToXML(org.w3c.dom.Document document, Element parentElement) {
		Element sgElement=document.createElement("scenegraph");
		for(Node node : sgNodes) {
			Element nodeElement=document.createElement("node");
			// save node data and list of children:
			node.saveToXML(document, nodeElement);
			sgElement.appendChild(nodeElement);
		}
		parentElement.appendChild(sgElement);
	}
	
	@Override
	public SceneGraph clone() {
		SceneGraph sceneGraph;
		try {
			sceneGraph = (SceneGraph) super.clone();
			sceneGraph.sgNodes=(ArrayList<Node>) sgNodes.clone();
			sceneGraph.sgNodeNames=sgNodeNames.clone();
			return sceneGraph;
		} catch (CloneNotSupportedException e) {
			// should not happen here
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Create a complete deep copy of this scene graph. This copy
	 * includes all nodes in this graph including their data.
	 * 
	 * @return a copy of the scene graph.
	 * @throws Exception
	 */
	public SceneGraph deepCopy() throws Exception {
		SceneGraph graph=clone();
		HashMap<Node, Node> replacementMap=new HashMap<Node, Node>();
		
		// clone all nodes
		graph.sgNodes.clear();
		graph.sgNodeNames.clear();
		for(Node node : sgNodes) {
			Node clone=node.deepCopy();
			replacementMap.put(node, clone);
			graph.sgNodes.add(clone);
			graph.sgNodeNames.add(clone);
			clone.setSceneGraph(graph);
		}
		
		// populate children and parents lists of clones with references to cloned nodes
		for(Node node : graph.sgNodes) {
			node.replaceReferences(replacementMap);
		}
		
		return graph;
	}
}
