/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Aug 21, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.sg;

import java.util.Arrays;

import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Matrix3;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Vector3D;

/**
 * Base class for all scene graph nodes that are transformable in 3D
 *
 * @author gregor
 */
abstract public class TransformableNode extends Node {

    /**
     * translate the node contents
     * @param distance the distance and direction in which to translate
     */
    public void translate(Vector3D distance)
    {
        Matrix4 matrix=new Matrix4();
        
        matrix.clear();
        for(int i=0;i<4;i++) {
            matrix.X[i][i]=1;
        }
        matrix.X[0][3]=distance.X1;
        matrix.X[1][3]=distance.X2;
        matrix.X[2][3]=distance.X3;
        
        applyMatrix(matrix);
    }

    /**
     * rotate the node contents
     * @param axis the axis around which to rotate
     * @param angle the angle by which to rotate
     */
    public void rotate(Vector3D axis, double angle)
    {
        Quaternion rotation=new Quaternion();
        Matrix3 result;
        
        rotation.fromAxisAngle(axis,angle);
        result=rotation.toRotationMatrix3();
        applyMatrix(result);
    }
    
    /**
     * rotate the node contents
     * @param quat the quaternion of the rotation
     */
    public void rotate(Quaternion quat)
    {
        Matrix3 result;
        
        result=quat.toRotationMatrix3();
        applyMatrix(result);
    }

    /**
     * scale the node contents
     * @param pivot the pivot point, that is, the point that won't be moved by the operation
     * @param amount the amount by which to scale along each axis
     */
    public void scale(Vector3D pivot, Vector3D amount)
    {
        Matrix4 matrix=new Matrix4();
        
        // recenter pivot on origin
        translate(pivot.multiply(-1));
        
        // build scale matrix
        matrix.clear();
        matrix.X[0][0]=amount.X1;
        matrix.X[0][0]=amount.X2;
        matrix.X[0][0]=amount.X3;
        matrix.X[3][3]=1.0;
        // apply scale matrix
        applyMatrix(matrix);        
        
        // undo translation on pivot
        translate(pivot);
    }

    /**
     * apply a 3x3 transformation matrix to this scene graph node
     * 
     * @param matrix the matrix to be applied
     */
    public abstract void applyMatrix(Matrix3 matrix);

    /**
     * apply a 4x4 transformation matrix to this scene graph node
     * 
     * @param matrix the matrix to be applied
     */
    public abstract void applyMatrix(Matrix4 matrix);

    /**
     * Generate a minimal axis-aligned bounding box around this this scene graph node.
     *  
     * @return a minimal axis-aligned bounding box
     */
    public abstract AxisAlignedBoundingBox getBoundingBox();

    /**
     * Query for intersection of a ray with this scene graph node
     *  
     * @param ray The ray to be checked for intersection
     * @return the coordinates of the intersection point or null, if none exists
     */
    public abstract Vector3D getIntersectionPoint(eu.moonlight3d.math.Ray ray);
    
    /**
     * Resolve the given path through the scene graph to a list of the nodes
     * that this path references.
     * 
     * @param path the path to resolve
     * @return the resolved nodes in order
     */
    private Node[] getNodesFromPath(String path) {
    	String[] nodeNames=path.split("/");
    	Node[] pathNodes=new Node[nodeNames.length];
    	
    	for(int i=0;i<nodeNames.length;i++) {
    		pathNodes[i]=getSceneGraph().getNodeByName(nodeNames[i]);
    	}
    	
    	return pathNodes;
    }

    /**
     * Transform coordinates in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @param point the point to transform
     * @return the transformed point
     */
    public Vector3D transformPointLocalToWorld(Node[] path, Vector3D point) {
    	Vector3D result=point.clone();
    	
    	for(int i=path.length-1;i>=0;i--) {
    		if(path[i] instanceof Instance) {
    			if(i>0) {
    				result=((Instance)path[i]).transformPointLocalToWorld(Arrays.copyOfRange(path, 0, i), point);
    			} else {
    				result=((Instance)path[i]).transformPointLocalToWorld(Arrays.copyOfRange(path, 0, 0), point);    				
    			}
				break;
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Transform coordinates in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @param point the point to transform
     * @return the transformed point
     */
    public final Vector3D transformPointLocalToWorld(String path, Vector3D point) {
    	return transformPointLocalToWorld(getNodesFromPath(path),point);
    }
    
    /**
     * Transform coordinates in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @param point the point to transform
     * @return the transformed point
     */
    public Vector3D transformPointWorldToLocal(Node[] path, Vector3D point) {
    	Vector3D result=point.clone();
    	
    	for(int i=0;i<path.length;i++) {
    		if(path[i] instanceof Instance) {
    			result=((Instance)path[i]).transformPointWorldToLocal(Arrays.copyOfRange(path, i, path.length-1), point);
    			break;
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Transform coordinates in world coordinates to object local coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @param point the point to transform
     * @return the transformed point
     */
    public final Vector3D transformPointWorldToLocal(String path, Vector3D point) {
    	return transformPointWorldToLocal(getNodesFromPath(path),point);
    }
    
    /**
     * Transform a vector in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @param vector the point to transform
     * @return the transformed point
     */
    public Vector3D transformVectorLocalToWorld(Node[] path, Vector3D vector) {
    	Vector3D result=vector.clone();
    	
    	for(int i=path.length-1;i>=0;i--) {
    		if(path[i] instanceof Instance) {
    			result=((Instance)path[i]).transformVectorLocalToWorld(Arrays.copyOfRange(path, 0, i-1), vector);
    			break;
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Transform a vector in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @param vector the point to transform
     * @return the transformed point
     */
    public final Vector3D transformVectorLocalToWorld(String path, Vector3D vector) {
    	return transformVectorLocalToWorld(getNodesFromPath(path),vector);
    }
    
    /**
     * Transform a vector in object local coordinates to world coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @param vector the point to transform
     * @return the transformed point
     */
    public Vector3D transformVectorWorldToLocal(Node[] path, Vector3D vector) {
    	Vector3D result=vector.clone();
    	
    	for(int i=0;i<path.length;i++) {
    		if(path[i] instanceof Instance) {
    			result=((Instance)path[i]).transformVectorWorldToLocal(Arrays.copyOfRange(path, i, path.length-1), vector);
    			break;
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Transform a vector in world coordinates to object local coordinates.
     * The correct transformation depends on the path through the graph to the
     * current node. Different paths to the same node usually mean different
     * transformations.
     * 
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @param vector the point to transform
     * @return the transformed point
     */
    public final Vector3D transformVectorWorldToLocal(String path, Vector3D vector) {
    	return transformVectorWorldToLocal(getNodesFromPath(path),vector);
    }

    /**
     * Calculate the transformation matrix that transforms from world space
     * into the local space of this node. The correct transformation depends
     * on the path through the graph to the current node. Different paths to
     * the same node usually mean different transformations.
     * 
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @return the transformation matrix
     */
    public Matrix4 getWorldToLocalTransformation(Node[] path) {
    	return getLocalToWorldTransformation(path).invert();    	
    }
    
    /**
     * Calculate the transformation matrix that transforms from world space
     * into the local space of this node. The correct transformation depends
     * on the path through the graph to the current node. Different paths to
     * the same node usually mean different transformations.
     *
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @return the transformation matrix
     */
    public final Matrix4 getWorldToLocalTransformation(String path) {
    	return getWorldToLocalTransformation(getNodesFromPath(path));
    }
    
    /**
     * Calculate the transformation matrix that transforms from the local
     * space of this node into world space. The correct transformation depends
     * on the path through the graph to the current node. Different paths to
     * the same node usually mean different transformations.
     *
     * The path is passed as a sorted array of scene graph nodes which need
     * to be passed on the path from the root to this node to get the transformation
     * that should be inverted.
     * 
     * @param path array of all nodes on the path to the node in correct sort order
     * @return the transformation matrix
     */
    public Matrix4 getLocalToWorldTransformation(Node[] path) {
    	Matrix4 result=new Matrix4();
    	result.setIdentity();

    	// FIXME this only works with transformations provided by Instance nodes
    	for(int i=0;i<path.length;i++) {
    		if(path[i] instanceof Instance) {
//    			result=((Instance)path[i]).getTransformationMatrix().multiply(result);
    			result=result.multiply(((Instance)path[i]).getTransformationMatrix());
    		}
    	}
    	
    	return result;
    }
    
    /**
     * Calculate the transformation matrix that transforms from the local
     * space of this node into world space. The correct transformation depends
     * on the path through the graph to the current node. Different paths to
     * the same node usually mean different transformations.
     *
     * The path given must be the path through the scene graph only. No higher
     * level path components (OG nodes, slots, etc...) or appendix may be given.
     * 
     * @param path the path through the scene graph
     * @return the transformation matrix
     */
    public final Matrix4 getLocalToWorldTransformation(String path) {
    	return getLocalToWorldTransformation(getNodesFromPath(path));
    }
}
