/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Nov 3, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.sg.iterator;
/**
 * \package eu.moonlight3d.ml3d.backend.sg.iterator
 * 
 * This is a collection of scene graph iterators and filters
 * for scene graph nodes to make traversal and processing
 * of scene graphs easier.
 * 
 * This package is part of the core program.
 */

import java.util.ArrayList;
import java.util.List;

import eu.moonlight3d.ml3d.backend.threedim.sg.Node;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * An iterator class that walks a scene graph in a flat fashion.
 * Nodes are visited only once and in no guaranteed order. This
 * iterator completely ignores any structure in the graph. It is
 * useful for example to visit nodes of certain types if the 
 * knowledge about the location within the graph itself is 
 * unimportant. The results that are returned by this iterator
 * can be reduced to a subset of the scene graph by applying 
 * filters. Filters should only be added before calling
 * next() or hasNext() for the first time.
 *
 * @author gregor
 */
public class Iterator {
	/**
	 * The list of filters that is applied while walking the
	 * graph.
	 */
	private ArrayList<Filter> filters;
	/**
	 * A copy of the list of nodes in the scene graph.
	 */
	private List<Node> nodes;
	/**
	 * The index marking the current position in the nodes list.
	 */
	private int index;
	
	/**
	 * Create a flat iterator for the given scene graph.
	 * 
	 * @param sceneGraph scene graph to iterate through
	 */
	public Iterator(SceneGraph sceneGraph) {
		filters=new ArrayList<Filter>();
		nodes=sceneGraph.getNodes();
		index=-1;
	}
	
	/** 
	 * Add a filter to the filter queue.
	 * @param filter the filter to add
	 */
	public void addFilter(Filter filter)
	{
		filters.add(filter);
	}
	
	/**
	 * Internal method for testing wether the given node is
	 * accepted by all supplied filters
	 * @param node The node to test
	 * @return true if the Node passed, false otherwise
	 */
	private boolean filtersApply(Node node) {
		for(int i=0;i<filters.size();i++) {
			if(!filters.get(i).check(node)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Internal implementation for hasNext() and next(). Returns the next matching
	 * Node or null if there is none. If peek is true, the current position will not
	 * be updated and the operation will stay reproducable.  If peek is false, the
	 * the pointer be updated and the next call will return the next matching Node.
	 * 
	 * @param peek whether to only peek for the next Node
	 * @return next Node matched by filters
	 */
	private Node getNext(boolean peek)
	{
		int newIndex=index;
		
		while(newIndex<nodes.size()-1) {
			newIndex++;
			if(filtersApply(nodes.get(newIndex))) {
				if(!peek) {
					index=newIndex;
				}
				return nodes.get(newIndex);
			}
		}
		
		return null;
	}
	
	/**
	 * Check if this iterator has a next object that can be returned.
	 * @return true if there is another matching object, false if there isn't
	 */
	public boolean hasNext()
	{
		if(getNext(true)!=null) {
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Return the next matching object from the scene graph. 
	 * @return the matching object or null if there are no more matching objects
	 */
	public Node next()
	{
		return getNext(false);
	}
}
