/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Nov 3, 2005
 */
package eu.moonlight3d.ml3d.backend.threedim.sg.iterator;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import eu.moonlight3d.ml3d.backend.threedim.sg.Node;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * An iterator that walks the scene graph in a tree-like fashion from
 * the root to the leaf nodes. Due to the directed graph structure
 * of the scene graph individual nodes may be visited several times
 * by this iterator. If you wish to visit each scene graph node only
 * once, consider using the {@link Iterator} class. 
 *
 * @author gregor
 */
public class TreeIterator {
	private ArrayList<Filter> filters;
	private Stack<Node> nodes;
	private Stack<List<Node>> children;
	private Stack<Integer> positions;
	
	/**
	 * Create a new TreeIterator on the given scene graph. The iterator
	 * will automatically start at the root of the scene graph and
	 * descent into the graph.
	 * 
	 * @param sceneGraph the graph to iterate over
	 */
	public TreeIterator(SceneGraph sceneGraph) {
		filters=new ArrayList<Filter>();
		nodes=new Stack<Node>();
		children=new Stack<List<Node>>();
		positions=new Stack<Integer>();

		positions.push(0);

		children.push(sceneGraph.getRootNodes());
	}

	/**
	 * Create a new TreeIterator on the given scene graph and start
	 * iterating from the given node. The iterator will automatically
	 * descent into the graph below this node.
	 * 
	 * @param sceneGraph the graph to iterate over
	 * @param parent the start node
	 */
	public TreeIterator(SceneGraph sceneGraph, Node parent) {
		filters=new ArrayList<Filter>();
		nodes=new Stack<Node>();
		children=new Stack<List<Node>>();
		positions=new Stack<Integer>();

		positions.push(0);

		ArrayList<Node> childrenList=new ArrayList<Node>();
		childrenList.add(parent);
		children.push(childrenList);		
	}
	
	/** 
	 * Add a filter to the filter queue.
	 * @param filter the filter to add
	 */
	public void addFilter(Filter filter)
	{
		filters.add(filter);
	}
	
	/**
	 * Internal method for testing wether the given node is
	 * accepted by all supplied filters
	 * @param node The node to test
	 * @return true if the Node passed, false otherwise
	 */
	private boolean filtersApply(Node node) {
		for(int i=0;i<filters.size();i++) {
			if(!filters.get(i).check(node)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Internal implementation for hasNext() and next(). Returns the next matching
	 * Node or null if there is none. If peek is true, the current position will not
	 * be updated and the operation will stay reproducible.  If peek is false, the
	 * the pointer be updated and the next call will return the next matching Node.
	 * 
	 * @param peek whether to only peek for the next Node
	 * @return next Node matched by filters
	 */
	private Node[] getNext(boolean peek)
	{
		boolean doContinue=true;
		Node result[]=null;
		
		if(positions.size()==0) {
			return null;
		}
		
		Stack<Node> nodes;
		Stack<List<Node>> children;
		Stack<Integer> positions;
		
		// if we're only peeking, copy the algorithm state and use that as a working base
		if(peek) {
			nodes=(Stack<Node>)this.nodes.clone();
			children=(Stack<List<Node>>)this.children.clone();
			positions=(Stack<Integer>)this.positions.clone();
		} else {
			nodes=this.nodes;
			children=this.children;
			positions=this.positions;
		}
		
		List<Node> currentChildren=children.peek();
		int currentPosition=positions.peek();

		while(doContinue) {
			// go a level upwards if we've reached the end at that level in that particular node
			if(currentPosition==currentChildren.size()) {
				// abort if no nodes are left
				if(nodes.size()==0) {
					doContinue=false;
					break;
				}
				nodes.pop();
				children.pop();
				positions.pop();
				currentChildren=children.peek();
				currentPosition=positions.peek();
			} else if(currentChildren.get(currentPosition).getChildren().size()!=0) {
				// search depth first, check children before parent node
				nodes.push(currentChildren.get(currentPosition));
				children.push(currentChildren.get(currentPosition).getChildren());
				positions.pop();
				positions.push(currentPosition);
				positions.push(0);

				currentChildren=children.peek();
				currentPosition=0;
				continue;
			}
			
			// test if current node applies and eventually abort
			if(filtersApply(currentChildren.get(currentPosition))) {
				result=new Node[nodes.size()+1];
				for(int i=0;i<nodes.size();i++) {
					result[i]=nodes.get(i);
				}
				result[nodes.size()]=currentChildren.get(currentPosition);
				positions.pop();
				positions.push(currentPosition+1);
				doContinue=false;
			}
			
			currentPosition++;
		}
		
		return result;
	}
	
	/**
	 * Check if this iterator has a next object that can be returned.
	 * @return true if there is another matching object, false if there isn't
	 */
	public boolean hasNext()
	{
		if(getNext(true)!=null) {
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Return the next matching object from the scene graph with a full
	 * list of parent nodes. 
	 * @return the matching object with parents or null if there are no more matching objects
	 */
	public Node[] next()
	{
		return getNext(false);
	}
	
	/**
	 * Return the full path to the next matching object as a String.
	 * 
	 * @return the full path to the next matching object or null if there are no more matches
	 */
	public String nextAsString() {
		Node[] result=next();
		String resultString=null;
		if(result!=null) {
			resultString=result[0].getName();
			for(int i=1;i<result.length;i++) {
				resultString+="/" + result[i].getName();
			}
		}
		return resultString;
	}
}
