/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Apr 15, 2005
 */
package eu.moonlight3d.ml3d.file.ml3d;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMException;
import org.w3c.dom.Element;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.backend.threedim.anim.Animation;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.og.Link;
import eu.moonlight3d.ml3d.backend.threedim.renderer.Manager;
import eu.moonlight3d.ml3d.backend.threedim.renderer.RenderEngine;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * A helper class that writes the contents of the Document to
 * an XML file.
 *
 * @author gregor
 */
public class FileWriter {
	/**
	 * The document to save to the XML file.
	 */
	private Document document;
	/**
	 * The file name of the output file.
	 */
    private String filename;
    
    /**
     * Constructor taking the document to save and the desired 
     * output filename as arguments.
     * 
     * @param document the document to save
     * @param filename the name of the file to save the document to
     */
    public FileWriter(Document document, String filename)
    {
    	this.document=document;
        this.filename=filename;
    }
    
    /**
     * Put the contents of the document's node cache into the XML
     * tree structure.
     * 
     * @param doc the DOM document in which to insert the cache contents
     * @param root the root node below which to insert the cache contents
     */
    private void exportCache(org.w3c.dom.Document doc, Element root) {
    	Element cacheRoot=doc.createElement("cache");
    	for(String nodeName : document.getCacheManager().getSGs().keySet()) {
    		SceneGraph graph=document.getCacheManager().getSG(nodeName);
    		Element nodeElement=doc.createElement("cachenode");
    		nodeElement.setAttribute("name",nodeName);
    		graph.saveToXML(doc,nodeElement);
    		cacheRoot.appendChild(nodeElement);
    	}
    	root.appendChild(cacheRoot);
    }
    
    /**
     * Put the contents of the document's operator graph into the XML
     * tree structure.
     * 
     * @param doc the DOM document in which to insert the OG contents
     * @param root the root node below which to insert the OG contents
     * @throws Exception 
     */
    private void exportOG(org.w3c.dom.Document doc, Element root) throws Exception
    {
        Element ogRoot=doc.createElement("og");
        root.appendChild(ogRoot);
        Element ogNodes=doc.createElement("nodes");
        ogRoot.appendChild(ogNodes);
        Element ogLinks=doc.createElement("links");
        ogRoot.appendChild(ogLinks);
        
        // iterate over OG nodes:
        eu.moonlight3d.ml3d.backend.threedim.og.Manager ogManager=document.getOGManager();
        List<eu.moonlight3d.ml3d.backend.threedim.og.Node> nodeList=ogManager.getNodeList();
        for(int i=0;i<nodeList.size();i++) {
            eu.moonlight3d.ml3d.backend.threedim.og.Node node=nodeList.get(i);
            
            Element docNode=doc.createElement("node");
            docNode.setAttribute("type",node.getType());
            docNode.setAttribute("name",node.getName());
            
            // iterate over properties and save them
            ArrayList<String> propertyList=node.getPropertyList();
            for(int j=0;j<propertyList.size();j++) {
                Property property=node.getProperty(propertyList.get(j));
                property.writeToFile(doc,docNode);
            }
            
            ogNodes.appendChild(docNode);
        }
        
        // iterate over OG links:
        ArrayList<Link> linkList=ogManager.getLinkList();
        for(int i=0;i<linkList.size();i++) {
            Link link=linkList.get(i);
            
            Element docLink=doc.createElement("link");
            Element input=doc.createElement("input");
            input.setAttribute("node",link.getInput().getNode().getName());
            input.setAttribute("slot",link.getInput().getName());
            docLink.appendChild(input);
            Element output=doc.createElement("output");
            output.setAttribute("node",link.getOutput().getNode().getName());
            output.setAttribute("slot",link.getOutput().getName());
            docLink.appendChild(output);
            
            ogLinks.appendChild(docLink);
        }
    }

    /**
     * Put the the document's render settings into the XML
     * tree structure.
     * 
     * @param doc the DOM document in which to insert the render settings
     * @param root the root node below which to insert the render settings
     * @throws DOMException 
     * @throws Exception 
     */
	private void exportRenderSettings(org.w3c.dom.Document doc, Element root) throws DOMException, Exception {
		List<String> renderEngines=eu.moonlight3d.ml3d.backend.threedim.renderer.Manager.getRenderEngines();
		
		Element renderEnginesNode=doc.createElement("renderengines");
		
		for(int i=0;i<renderEngines.size();i++) {
			RenderEngine engine=Manager.getRenderEngine(renderEngines.get(i));
			PropertyContainer engineProperties=document.getRendererManager().getProperties(engine.getName());
			
			Element engineNode=doc.createElement("engine");
			engineNode.setAttribute("name",engine.getName());
			
			ArrayList<String> propertyList=engineProperties.getPropertyList();
            for(int j=0;j<propertyList.size();j++) {
                Property property=engineProperties.getProperty(propertyList.get(j));
                property.writeToFile(doc,engineNode);
            }
            
            renderEnginesNode.appendChild(engineNode);
		}
		
		root.appendChild(renderEnginesNode);
	}
	
    /**
     * Put the contents of the document's animations into the XML
     * tree structure.
     * 
     * @param doc the DOM document in which to insert the animations
     * @param root the root node below which to insert the animations
     * @throws Exception 
     * @throws DOMException 
     */
	private void exportAnimations(org.w3c.dom.Document doc, Element root) throws DOMException, Exception {
		List<Animation> animations=document.getAnimationManager().getAnimations();
		
		Element animationsNode=doc.createElement("animations");
		
		for(Animation animation : animations) {
			animation.saveXML(doc, animationsNode);
		}
		
		root.appendChild(animationsNode);
	}
    
    /**
     * Actually save the file.
     * 
     * @throws Exception if the operations fails
     */
    public void run() throws Exception
    {
        try {
        	org.w3c.dom.Document sceneDoc=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
            
            // generate save file DOM tree here
            Element root=sceneDoc.createElement("moonlight");
            root.setAttribute("fileformatversion","2");
            sceneDoc.appendChild(root);
            
            exportCache(sceneDoc,root);
            exportOG(sceneDoc,root);
            exportAnimations(sceneDoc,root);
            exportRenderSettings(sceneDoc,root);

            Source domSource=new DOMSource(sceneDoc);
            Result streamResult=new StreamResult(new File(filename));
            Transformer transformer=TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty("indent","yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            
            transformer.transform(domSource,streamResult);
        } catch (ParserConfigurationException e) {
            throw new Exception(L10N.translate("Error while saving file"),e);
        } catch (TransformerConfigurationException e) {
            throw new Exception(L10N.translate("Error while saving file"),e);
        } catch (TransformerFactoryConfigurationError e) {
            throw new Exception(L10N.translate("Error while saving file"),e);
        } catch (TransformerException e) {
            throw new Exception(L10N.translate("Error while saving file"),e);
        }
    }
}
