/*
 * Moonlight|3D
 * Copyright (C) 2009 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Jul 19, 2009
 */

package eu.moonlight3d.ml3d.file.obj;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.DrainSlotIterator;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.Node;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.TreeIterator;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.TypeFilter;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;
import eu.moonlight3d.ml3d.rendering.simple.SimpleMaterial;

/**
 * Algorithm object for OBJ file exporter.
 *
 * @author gregor
 */
public class OBJExporter {
	/**
	 * Offset for object vertex indices in OBJ file.
	 */
	private int vertexIndexOffset;
	/**
	 * Offset for object normal indices in OBJ file
	 */
	private int normalIndexOffset;
	
	/**
	 * Export whole scene at once
	 */
	boolean exportWholeScene;
	/**
	 * Transform vertices of selected objects into world space
	 */
	boolean transformToWorldSpace;
	/**
	 * Export material data
	 */
	boolean exportMaterial;
	
	/**
	 * Default constructor
	 * 
	 * @param properties OBJ file format export properties
	 */
	OBJExporter(PropertyContainer properties) {
		vertexIndexOffset=0;
		normalIndexOffset=0;
		
		exportWholeScene=properties.getProperty("exportWholeScene").getBooleanValue();
		transformToWorldSpace=properties.getProperty("transformToWorldSpace").getBooleanValue();
		exportMaterial=properties.getProperty("exportMaterial").getBooleanValue();
	}
	
	/**
	 * Export a given mesh
	 * 
	 * @param output output file
	 * @param mesh mesh to export
	 * @param path path to mesh for transformation to world space or null if no transformation desired
	 * @throws IOException if a write error occurs
	 */
	private void exportMesh(RandomAccessFile output, Mesh mesh, String path) throws IOException {
		if(exportMaterial) {
			// write usemtl line (note: line is always required - object with no material might follow an object with one and there is no reset statement)
			String materialName="__none__";
			for(Node child : mesh.getChildren()) {
				if(child instanceof SimpleMaterial) {
					materialName=child.getName();
				}
			}
			output.writeBytes("usemtl " + materialName + "\n");
		}
		
		Matrix4 transformation=null;
		if(transformToWorldSpace && path!=null) {
			String sgPath=SelectionHelper.getSGPath(path);
			transformation=mesh.getLocalToWorldTransformation(sgPath);
		}
		for(int i=0;i<mesh.getNumVertices();i++) {
			Vector3D vertex=mesh.getVertexPosition(i);
			if(transformation!=null) {
				vertex=transformation.multiply(vertex);
			}
			output.writeBytes("v " + Double.toString(vertex.X1) + " " + Double.toString(vertex.X2) + " " + Double.toString(vertex.X3) + "\n");
		}

		// TODO write vt lines if maps are present

		ArrayList<Integer> faceOffsets=new ArrayList<Integer>();
		int faceOffset=0;
		for(int i=0;i<mesh.getNumFaces();i++) {
			faceOffsets.add(faceOffset);
			for(int j=0;j<mesh.getFaceEdges(i).length;j++) {
				Vector3D normal=mesh.getFaceVertexNormal(i, j);
				output.writeBytes("vn " + normal.X1 + " " + normal.X2 + " " + normal.X3 + "\n");
				faceOffset++;
			}
		}
		
		int totalFaceOffset=faceOffset;

		for(int i=0;i<mesh.getNumFaces();i++) {
			Mesh.Face face=mesh.getFace(i);

			ArrayList<Integer> triangulated=mesh.triangulated(face);
			ArrayList<Integer> vertexIDs=mesh.getVertexIDs(face);

			for(int j=0;j<triangulated.size();j+=3) {
				faceOffset=faceOffsets.get(i);
				int vertexIndex1=vertexIndexOffset+triangulated.get(j)+1;
				int vertexIndex2=vertexIndexOffset+triangulated.get(j+1)+1;
				int vertexIndex3=vertexIndexOffset+triangulated.get(j+2)+1;
				int normalIndex1=normalIndexOffset+faceOffset+vertexIDs.indexOf(triangulated.get(j))+1;
				int normalIndex2=normalIndexOffset+faceOffset+vertexIDs.indexOf(triangulated.get(j+1))+1;
				int normalIndex3=normalIndexOffset+faceOffset+vertexIDs.indexOf(triangulated.get(j+2))+1;
				output.writeBytes("f " + vertexIndex1 + "//" + normalIndex1 + " " + vertexIndex2 + "//" + normalIndex2 + " " + vertexIndex3 + "//" + normalIndex3 + "\n");
			}
		}
		
		vertexIndexOffset+=mesh.getNumVertices();
		normalIndexOffset+=totalFaceOffset;
	}

	/**
	 * Export the selected objects only.
	 * 
	 * @param output output file to write data to
	 * @param document the Moonlight document to process
	 * @throws Exception
	 */
	void exportSelection(RandomAccessFile output, Document document) throws Exception {
		Selection selection=document.getSelectionManager().getSelection("viewSelection");
		if(selection.getType().getName()!="SGNode") {
			State.getInstance().logger().warning(L10N.translate("At least one mesh object needs to be selected"));
			return;
		}
		boolean meshSelected=false;
		ArrayList<Object> selectedObjects=selection.getSelectionObjects();
		for(Object object : selectedObjects) {
			if(object instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh) {
				meshSelected=true;
				break;
			}
		}
		if(!meshSelected) {
			State.getInstance().logger().warning(L10N.translate("At least one mesh object needs to be selected"));
			return;
		}
		
		for(String nodeName : selection.getSelection()) {
			Object object=selection.getObjectFromSelectionString(nodeName);
			
			if(object instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh) {			
				String path=null;
				if(transformToWorldSpace) {
					path=SelectionHelper.getSGPath(nodeName);	
				}
				exportMesh(output,(Mesh)object,path);
			}
		}
	}

	/**
	 * Export the whole scene
	 * 
	 * @param output output file to write data to
	 * @param document the Moonlight document to process
	 * @throws Exception
	 */
	void exportWholeScene(RandomAccessFile output, Document document) throws Exception {
		DrainSlotIterator ogIterator=new DrainSlotIterator(document.getOGManager());
		
		transformToWorldSpace=true; // force world space transform here
		
		while(ogIterator.hasNext()) {
			Slot slot=ogIterator.next();

			String ogNodeName = "OperatorGraph/" + slot.getNode().getName();
			String slotName = ogNodeName + "/" + slot.getName();
			
			TreeIterator sgIterator=new TreeIterator(slot.getGraph());
			sgIterator.addFilter(new TypeFilter("eu.moonlight3d.ml3d.modelling.mesh.Mesh",true));
			
			while(sgIterator.hasNext()) {
				Node[] path = sgIterator.next();
				Mesh node = (Mesh) path[path.length-1];
				String sgNodeName = slotName + "/" + path[0].getName();
				for(int i=1;i<path.length;i++) {
					sgNodeName+="/" + path[i].getName();
				}

				System.out.println(sgNodeName);
				exportMesh(output,node,sgNodeName);
			}
		}
	}
	
	/**
	 * Export material file to go with OBJ file.
	 * 
	 * @param materialFilename file name of material file
	 * @param document document to export
	 * @throws Exception
	 */
	private void exportMaterials(String materialFilename, Document document) throws Exception {
		RandomAccessFile output=new RandomAccessFile(materialFilename,"rw");
		output.writeBytes("# created with Moonlight|3D version " + eu.moonlight3d.framework.State.getInstance().getVersionInformation().getVersionString() + "\n");
		
		DrainSlotIterator ogIterator=new DrainSlotIterator(document.getOGManager());
		
		while(ogIterator.hasNext()) {
			Slot slot=ogIterator.next();
						
			TreeIterator sgIterator=new TreeIterator(slot.getGraph());
			sgIterator.addFilter(new TypeFilter("eu.moonlight3d.ml3d.rendering.simple.SimpleMaterial",true));
			
			while(sgIterator.hasNext()) {
				Node[] path = sgIterator.next();
				SimpleMaterial material = (SimpleMaterial) path[path.length-1];
				output.writeBytes("newmtl " + material.getName() + "\n");
				Colour colour=material.getColour();
				output.writeBytes("Kd " + Double.toString(colour.red) + " " + Double.toString(colour.green) + " " + Double.toString(colour.blue) + "\n");
				Colour specular=material.getSpecular();
				output.writeBytes("Ks " + Double.toString(specular.red) + " " + Double.toString(specular.green) + " " + Double.toString(specular.blue) + "\n");
				Colour ambient=material.getAmbient();
				output.writeBytes("Ka " + Double.toString(ambient.red) + " " + Double.toString(ambient.green) + " " + Double.toString(ambient.blue) + "\n");
			}
		}
		
		output.close();
	}

	/**
	 * Write an OBJ file with given settings
	 * 
	 * @param filename
	 */
	public void exportFile(String filename) {
		Document document=(eu.moonlight3d.ml3d.backend.threedim.document.Document)(State.getInstance().getDocumentManager().getMainDocument());
		try {
			RandomAccessFile output=new RandomAccessFile(filename,"rw");
			output.setLength(0);

			output.writeBytes("# created with Moonlight|3D version " + eu.moonlight3d.framework.State.getInstance().getVersionInformation().getVersionString() + "\n");
			
			if(exportMaterial) {
				String materialFilename=filename.substring(0, filename.length()-3) + "mtl";
				
				exportMaterials(materialFilename,document);
				
				// TODO: strip the path off the filename here
				output.writeBytes("mtllib " + materialFilename + "\n");
			}

			if(!exportWholeScene) {
				exportSelection(output, document);
			} else {
				exportWholeScene(output, document);
			}
			
			output.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
