/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Apr 19, 2006
 */
package eu.moonlight3d.ml3d.file.obj;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

import eu.moonlight3d.common.file.core.FileFormat;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.basic.CopyFromCache;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;

/**
 * Support for the Wavefront .OBJ file format. Most of the
 * features related to polygon meshes are supported.
 *
 * @author gregor
 */
public class OBJFileFormat extends FileFormat {
	/**
	 * Property container that holds exporter settings
	 */
	private PropertyContainer exportProperties;
	
	/**
	 * Default constructor
	 */
	public OBJFileFormat() {
		exportProperties=new PropertyContainer();
		Property exportWholeScene=new Property(exportProperties);
		exportWholeScene.setName("exportWholeScene");
		exportWholeScene.setType(Value.Type.Boolean);
		exportWholeScene.setValue(false);
		Property transformToWorldSpace=new Property(exportProperties);
		transformToWorldSpace.setName("transformToWorldSpace");
		transformToWorldSpace.setType(Value.Type.Boolean);
		transformToWorldSpace.setValue(false);
		Property exportMaterial=new Property(exportProperties);
		exportMaterial.setName("exportMaterial");
		exportMaterial.setType(Value.Type.Boolean);
		exportMaterial.setValue(false);
	}

	@Override
	public FileType getFileType() {
		ArrayList<String> patterns=new ArrayList<String>();
		patterns.add("*.obj");
		return new FileType("Wavefront OBJ",patterns);
	}

	@Override
	public boolean canImport() {
		return true;
	}

	@Override
	public boolean canExport() {
		return true;
	}
	
	@Override
	public PropertyContainer getImportProperties() {
		return null;
	}

	@Override
	public PropertyContainer getExportProperties() {
		return exportProperties;
	}

	@Override
	public void importFile(String filename) throws Exception {
		SceneGraph sceneGraph=new SceneGraph((eu.moonlight3d.ml3d.backend.threedim.document.Document) State.getInstance().getDocumentManager().getMainDocument());
		Instance fileContents=new Instance();
		try {
			RandomAccessFile file=new RandomAccessFile(filename,"r");
			Mesh mesh=new Mesh();
			ArrayList<Vector3D> normals=new ArrayList<Vector3D>();
			ArrayList<double[]> uvCoordinates=new ArrayList<double[]>();
			boolean hasUVMap=false;

			String line;
			while((line=file.readLine())!=null) {
				line=line.trim();
				if(line.equals("")) { // skip empty lines
					continue;
				}
				if(line.charAt(0)=='#') { // skip comment lines
					continue;
				}
				String[] tokens=line.split("[ \t]+");
				if(tokens[0].equals("v")) {
					// parse vertex coodinates
					Vector3D position=new Vector3D();
					position.X1=Double.parseDouble(tokens[1]);
					position.X2=Double.parseDouble(tokens[2]);
					position.X3=Double.parseDouble(tokens[3]);
					mesh.addVertex(position);
				} else if(tokens[0].equals("vt")) {
					// texture coordinates
					double uv[]=new double[2];
					uv[0]=Double.parseDouble(tokens[1]);
					if(tokens.length>=3) {
						uv[1]=Double.parseDouble(tokens[2]);
					} else {
						uv[1]=0;
					}
					uvCoordinates.add(uv);
				} else if(tokens[0].equals("vn")) {
					// vertex normal
					Vector3D normal=new Vector3D();
					normal.X1=Double.parseDouble(tokens[1]);
					normal.X2=Double.parseDouble(tokens[2]);
					normal.X3=Double.parseDouble(tokens[3]);
					normals.add(normal);
				} else if(tokens[0].equals("vp")) {
					// TODO parameter space

				} else if(tokens[0].equals("f")) {
					// face
					ArrayList<Integer> edges=new ArrayList<Integer>();

					int[] face=new int[tokens.length-1];

					for(int i=0;i<tokens.length-1;i++) {
						if(tokens[i+1].contains("/")) {
							face[i]=Integer.parseInt(tokens[i+1].split("/")[0]);
						} else {
							face[i]=Integer.parseInt(tokens[i+1]);
						}
						if(face[i]<0) {
							face[i]=mesh.getNumVertices()-face[i];
						} else {
							face[i]--;
						}
					}

					int faceID=-1;
					try {
						for (int i = 0; i < face.length; i++)
							edges.add(mesh.findOrCreateEdge(face[i], face[(i+1)%face.length]));

						faceID=mesh.addFace(edges);
					} catch (Exception e) {
						// TODO handle exception
					}

					if(tokens[1].contains("/")) {
						// face contains references to normals and maybe uv coordinates
						if(tokens[1].contains("//")) {
							// face only contains references to normals
							for(int i=0;i<tokens.length-1;i++) {
								int normalID=Integer.parseInt(tokens[i+1].split("//")[1]);

								mesh.setFaceVertexNormalMode(faceID, i, Mesh.NORMAL_MODE_LOCKED);
								if(normalID>0) {
									mesh.setFaceVertexNormal(faceID, i, normals.get(normalID-1));
								} else {
									mesh.setFaceVertexNormal(faceID, i, normals.get(normals.size()-normalID));
								}
							}
						} else {
							// face contains references to normals and uv coordinates
							for(int i=0;i<tokens.length-1;i++) {
								String[] splitResult=tokens[i+1].split("/");
								int uvID=Integer.parseInt(splitResult[1]);

								if(!hasUVMap) {
									mesh.addUVMap("UVMap");
									hasUVMap=true;
								}

								if(uvID>0) {
									mesh.setUVCoordinates("UVMap", faceID, i, uvCoordinates.get(uvID-1)[0], 1-uvCoordinates.get(uvID-1)[1]);
								} else {
									mesh.setUVCoordinates("UVMap", faceID, i, uvCoordinates.get(uvCoordinates.size()-uvID)[0], 1-uvCoordinates.get(uvCoordinates.size()-uvID)[1]);
								}

								if(splitResult.length==3) {
									int normalID=Integer.parseInt(splitResult[2]);
									mesh.setFaceVertexNormalMode(faceID, i, Mesh.NORMAL_MODE_LOCKED);
									if(normalID>0) {
										mesh.setFaceVertexNormal(faceID, i, normals.get(normalID-1));
									} else {
										mesh.setFaceVertexNormal(faceID, i, normals.get(normals.size()-normalID));
									}
								}
							}
						}
					}


					edges.clear();
				} else {
					eu.moonlight3d.framework.State.getInstance().logger().warning("unknown command in Wavefront OBJ: " + tokens[0]);
				}
			}

			fileContents.setName("ImportedObjInstance");
			mesh.setName("ImportedOBJ");
			sceneGraph.addNode(fileContents);
			sceneGraph.addNode(mesh);
			sceneGraph.addChild(fileContents,mesh);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}

        Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
        // Add CopyFromCache node to operator graph
        CopyFromCache copyFromCache=(CopyFromCache) document.getOGManager().createNodeByName("CopyFromCache");
        copyFromCache.getProperty("cachedGraph").setValue(copyFromCache.getName());
        // Add root scene graph to persistent cache
        document.getCacheManager().addSG(copyFromCache.getName(),sceneGraph);
        document.getOGManager().update();
	}

	@Override
	public void exportFile(String filename) {
		try {
			OBJExporter exporter=new OBJExporter(exportProperties);
			exporter.exportFile(filename);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
