/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Apr 19, 2006
 */
package eu.moonlight3d.ml3d.file.ogre;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import eu.moonlight3d.common.file.core.FileFormat;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.basic.CopyFromCache;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;

public class OGREFileFormat extends FileFormat {

	@Override
	public FileType getFileType() {
		ArrayList<String> patterns=new ArrayList<String>();
		patterns.add("*.mesh.xml");
		return new FileType("Ogre XML mesh",patterns);
	}

	@Override
	public boolean canImport() {
		return true;
	}

	@Override
	public boolean canExport() {
		return true;
	}

	@Override
	public PropertyContainer getImportProperties() {
		return null;
	}

	@Override
	public PropertyContainer getExportProperties() {
		return null;
	}

	private ArrayList<Vector3D> importVertexPositions(Element vertexBufferNode) {
		NodeList vertexList=vertexBufferNode.getElementsByTagName("vertex");
		ArrayList<Vector3D> vertexPositions=new ArrayList<Vector3D>();

		for(int i=0;i<vertexList.getLength();i++) {
			Element vertexNode=(Element) vertexList.item(i);
			Element positionElement=(Element) vertexNode.getElementsByTagName("position").item(0);
			Vector3D position=new Vector3D();
			position.X1=Double.parseDouble(positionElement.getAttribute("x"));
			position.X2=Double.parseDouble(positionElement.getAttribute("y"));
			position.X3=Double.parseDouble(positionElement.getAttribute("z"));
			vertexPositions.add(position);
		}
		return vertexPositions;
	}

	private ArrayList<Vector3D> importVertexNormals(Element vertexBufferNode) {
		NodeList vertexList=vertexBufferNode.getElementsByTagName("vertex");
		ArrayList<Vector3D> vertexNormals=new ArrayList<Vector3D>();

		for(int i=0;i<vertexList.getLength();i++) {
			Element vertexNode=(Element) vertexList.item(i);
			Element normalElement=(Element) vertexNode.getElementsByTagName("normal").item(0);
			Vector3D normal=new Vector3D();
			normal.X1=Double.parseDouble(normalElement.getAttribute("x"));
			normal.X2=Double.parseDouble(normalElement.getAttribute("y"));
			normal.X3=Double.parseDouble(normalElement.getAttribute("z"));
			vertexNormals.add(normal);
		}
		return vertexNormals;
	}

	private Mesh importSubmesh(Element submeshNode) throws Exception {
		Element facesNode=(Element)submeshNode.getElementsByTagName("faces").item(0);
		NodeList faceList=facesNode.getElementsByTagName("face");

		ArrayList<ArrayList<Integer>> faces;

		faces=new ArrayList<ArrayList<Integer>>();
		for(int i=0;i<faceList.getLength();i++) {
			Element faceNode=(Element) faceList.item(i);
			ArrayList<Integer> face=new ArrayList<Integer>();
			face.add(Integer.parseInt(faceNode.getAttribute("v1")));
			face.add(Integer.parseInt(faceNode.getAttribute("v2")));
			face.add(Integer.parseInt(faceNode.getAttribute("v3")));
			faces.add(face);
		}

		Element geometryNode=(Element)submeshNode.getElementsByTagName("geometry").item(0);
		NodeList vertexBufferList=geometryNode.getElementsByTagName("vertexbuffer");

		ArrayList<Vector3D> vertexPositions=null;
		ArrayList<Vector3D> vertexNormals=null;

		for(int i=0;i<vertexBufferList.getLength();i++) {
			Element vertexBufferNode=(Element) vertexBufferList.item(i);
			if(vertexBufferNode.hasAttribute("positions")) {
				if(vertexBufferNode.getAttribute("positions").equals("true")) {
					// load vertex positions
					vertexPositions=importVertexPositions(vertexBufferNode);
				}
			}
			if(vertexBufferNode.hasAttribute("normals")) {
				if(vertexBufferNode.getAttribute("normals").equals("true")) {
					// load vertex normals
					vertexNormals=importVertexNormals(vertexBufferNode);
				}
			}
		}

		if(vertexPositions==null) {
			throw new Exception(L10N.translate("submesh does not contain vertex positions"));
		}

		// convert submesh to Moonlight mesh
		Mesh mesh=new Mesh();

		HashMap<Integer,Integer> vertexIDMap=new HashMap<Integer,Integer>();

		for(int i=0;i<vertexPositions.size();i++) {
			int vertexID=mesh.addVertex(vertexPositions.get(i));
			vertexIDMap.put(i,vertexID);
		}

		ArrayList<Integer> edges=new ArrayList<Integer>();
		for(ArrayList<Integer> face : faces) {
			// create the edges of the face
			edges.add(mesh.findOrCreateEdge(face.get(0),face.get(1)));
			edges.add(mesh.findOrCreateEdge(face.get(1),face.get(2)));
			edges.add(mesh.findOrCreateEdge(face.get(2),face.get(0)));

			mesh.addFace(edges);

			edges.clear();
		}

		return mesh;
	}

	@Override
	public void importFile(String filename) throws Exception {
		SceneGraph sceneGraph=new SceneGraph((eu.moonlight3d.ml3d.backend.threedim.document.Document) State.getInstance().getDocumentManager().getMainDocument());
        try {
			Document fileDoc=DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(filename);
			Element root=fileDoc.getDocumentElement();

			Element submeshesNode=(Element) root.getElementsByTagName("submeshes").item(0);
			NodeList submeshList=submeshesNode.getElementsByTagName("submesh");

			for(int i=0;i<submeshList.getLength();i++) {
				Element submeshNode=(Element) submeshList.item(i);
				Instance instance=new Instance();
				instance.setName("submesh" + i + "Instance");
				Mesh mesh=importSubmesh(submeshNode);
				mesh.setName("submesh" + i);
				sceneGraph.addNode(instance);
				sceneGraph.addNode(mesh);
				sceneGraph.addChild(instance,mesh);
			}
		} catch (SAXException e) {
			throw new Exception(L10N.translate("error while importing OGRE mesh file"),e);
		} catch (IOException e) {
			throw new Exception(L10N.translate("error while importing OGRE mesh file"),e);
		} catch (ParserConfigurationException e) {
			throw new Exception(L10N.translate("error while importing OGRE mesh file"),e);
		} catch (Exception e) {
			throw new Exception(L10N.translate("error while importing OGRE mesh file"),e);
		}

        eu.moonlight3d.ml3d.backend.threedim.document.Document document;
        document=(eu.moonlight3d.ml3d.backend.threedim.document.Document) State.getInstance().getDocumentManager().getMainDocument();
        // Add CopyFromCache node to operator graph
        CopyFromCache copyFromCache=(CopyFromCache) document.getOGManager().createNodeByName("CopyFromCache");
        copyFromCache.getProperty("cachedGraph").setValue(copyFromCache.getName());
        // Add root scene graph to persistent cache
        document.getCacheManager().addSG(copyFromCache.getName(),sceneGraph);
        document.getOGManager().update();
	}

	private void exportVertices(Document document, Element geometry, Mesh mesh) {
		Element vertexbuffer=document.createElement("vertexbuffer");
		vertexbuffer.setAttribute("positions", "true");
		geometry.appendChild(vertexbuffer);

		for(int i=0;i<mesh.getNumVertices();i++) {
			Vector3D vertex=mesh.getVertexPosition(i);

			Element vertexElement=document.createElement("vertex");
			Element position=document.createElement("position");
			position.setAttribute("x", Double.toString(vertex.X1));
			position.setAttribute("y", Double.toString(vertex.X2));
			position.setAttribute("z", Double.toString(vertex.X3));
			vertexElement.appendChild(position);
			vertexbuffer.appendChild(vertexElement);
		}

		geometry.setAttribute("vertexcount", Integer.toString(mesh.getNumVertices()));
	}

	private void exportFaces(Document document, Element facesElement, Mesh mesh) {
		int numTriangles=0;
		for(int i=0;i<mesh.getNumFaces();i++) {
			Mesh.Face face=mesh.getFace(i);

			ArrayList<Integer> triangulatedFace=mesh.triangulated(face);

			for(int j=0;j<triangulatedFace.size();j+=3) {
				Element faceElement=document.createElement("face");
				faceElement.setAttribute("v1", Integer.toString(triangulatedFace.get(j)));
				faceElement.setAttribute("v2", Integer.toString(triangulatedFace.get(j+1)));
				faceElement.setAttribute("v3", Integer.toString(triangulatedFace.get(j+2)));
				facesElement.appendChild(faceElement);

				numTriangles++;
			}
		}
		facesElement.setAttribute("count",Integer.toString(numTriangles));
	}

	private void exportMesh(Document ogreDocument, Mesh mesh) {
		Element meshElement=ogreDocument.createElement("mesh");
		ogreDocument.appendChild(meshElement);
		Element submeshes=ogreDocument.createElement("submeshes");
		meshElement.appendChild(submeshes);
		Element submesh=ogreDocument.createElement("submesh");
		submeshes.appendChild(submesh);
		submesh.setAttribute("usesharedvertices", "false");
		submesh.setAttribute("material", "");
		submesh.setAttribute("operationtype","triangle_list");
		Element facesElement=ogreDocument.createElement("faces");
		submesh.appendChild(facesElement);
		Element geometry=ogreDocument.createElement("geometry");
		submesh.appendChild(geometry);
		exportVertices(ogreDocument, geometry, mesh);
		exportFaces(ogreDocument, facesElement, mesh);
	}

	@Override
	public void exportFile(String filename) throws Exception {
		eu.moonlight3d.ml3d.backend.threedim.document.Document mlDocument=(eu.moonlight3d.ml3d.backend.threedim.document.Document)(State.getInstance().getDocumentManager().getMainDocument());
		try {
			Selection selection=mlDocument.getSelectionManager().getSelection("viewSelection");
			if(selection.getType()==null) {
				State.getInstance().logger().warning(L10N.translate("A mesh object needs to be selected"));
				return;
			}
			if(selection.getType().getName()!="SGNode") {
				State.getInstance().logger().warning(L10N.translate("A mesh object needs to be selected"));
				return;
			}
			if(!(selection.getObjectFromSelectionString(selection.getActiveObject()) instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh)) {
				State.getInstance().logger().warning(L10N.translate("A mesh object needs to be selected"));
				return;
			}

			Document ogreDocument=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();

			exportMesh(ogreDocument,(Mesh)selection.getObjectFromSelectionString(selection.getActiveObject()));

            Source domSource=new DOMSource(ogreDocument);
            Result streamResult=new StreamResult(new File(filename));
            Transformer transformer=TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty("indent","yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");

            transformer.transform(domSource,streamResult);
		} catch (ParserConfigurationException e) {
			throw new Exception(L10N.translate("error while exporting OGRE mesh file"),e);
		} catch (TransformerConfigurationException e) {
			throw new Exception(L10N.translate("error while exporting OGRE mesh file"),e);
		} catch (TransformerFactoryConfigurationError e) {
			throw new Exception(L10N.translate("error while exporting OGRE mesh file"),e);
		} catch (TransformerException e) {
			throw new Exception(L10N.translate("error while exporting OGRE mesh file"),e);
		} catch (Exception e) {
			throw new Exception(L10N.translate("error while exporting OGRE mesh file"),e);
		}
	}

}
