/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Apr 19, 2006
 */
package eu.moonlight3d.ml3d.file.ply;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;

import eu.moonlight3d.common.file.core.FileFormat;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.FileType;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.basic.CopyFromCache;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;

/**
 * Importer and exporter for Stanford Triangle (PLY) mesh format.
 *
 * @author gregor
 */
public class PLYFileFormat extends FileFormat {

	@Override
	public FileType getFileType() {
		ArrayList<String> patterns=new ArrayList<String>();
		patterns.add("*.ply");
		return new FileType("Stanford PLY",patterns);
	}

	@Override
	public boolean canImport() {
		return true;
	}

	@Override
	public boolean canExport() {
		return true;
	}

	@Override
	public PropertyContainer getImportProperties() {
		return null;
	}

	@Override
	public PropertyContainer getExportProperties() {
		return null;
	}

	@Override
	public void importFile(String filename) throws Exception {
		SceneGraph sceneGraph=new SceneGraph((eu.moonlight3d.ml3d.backend.threedim.document.Document) State.getInstance().getDocumentManager().getMainDocument());
		Instance fileContents=new Instance();
		try {
			RandomAccessFile file=new RandomAccessFile(filename,"r");
			Mesh mesh=new Mesh();
			
			String line;
			
			// test wether first line is "ply"
			line=file.readLine();
			if(!line.toLowerCase().equals("ply")) {
				State.getInstance().logger().severe(L10N.translate("File does not seem to be in PLY format"));
			}
			
			while((line=file.readLine())!=null) {
				line=line.trim();
				if(line.equals("")) { // skip empty lines
					continue;
				}
				String[] tokens=line.split("[ \t]+");
				
				if(tokens[0].toLowerCase().equals("comment")) {
					State.getInstance().logger().info(L10N.translate("PLY comment: ") + line.trim().substring(8));
				} else if(tokens[0].toLowerCase().equals("format")) {
					// TODO support binary format
					if(!tokens[1].equals("ascii")) {
						String message=MessageFormat.format(L10N.translate("Unsupported format variant {0}, aborting import"), line.trim().substring(7));
						State.getInstance().logger().severe(message);
						return;
					}
					if(!tokens[2].equals("1.0")) {
						String message=MessageFormat.format(L10N.translate("Unknown PLY format {0} - file import may fail"), line.trim().substring(7));
						State.getInstance().logger().info(message);
					}
				} else if(tokens[0].toLowerCase().equals("element")) {
					if(tokens[1].toLowerCase().equals("vertex")) {
						
					} else if(tokens[1].toLowerCase().equals("edge")) {
						
					} else if(tokens[1].toLowerCase().equals("face")) {
						
					}
				} else if(tokens[0].toLowerCase().equals("property")) {
					if(tokens[1].equals("char")) {
						
					} else if(tokens[1].equals("uchar")) {
						
					} else if(tokens[1].equals("short")) {
						
					} else if(tokens[1].equals("ushort")) {
						
					} else if(tokens[1].equals("int")) {
						
					} else if(tokens[1].equals("uint")) {
						
					} else if(tokens[1].equals("float")) {
						
					} else if(tokens[1].equals("double")) {
						
					} else if(tokens[1].equals("list")) {
						
					}
					
					// TODO property name, assign property to element
				} else if(tokens[0].toLowerCase().equals("end_header")) {
					// header ends here, start with body
					break;
				}
			}

			while((line=file.readLine())!=null) {
				line=line.trim();
				if(line.equals("")) { // skip empty lines
					continue;
				}
				String[] tokens=line.split("[ \t]+");
				if(tokens[0].equals("v")) {
					// parse vertex coodinates
					Vector3D position=new Vector3D();
					position.X1=Double.parseDouble(tokens[1]);
					position.X2=Double.parseDouble(tokens[2]);
					position.X3=Double.parseDouble(tokens[3]);
					mesh.addVertex(position);
				} else if(tokens[0].equals("vt")) {
					// TODO texture coordinates
					
				} else if(tokens[0].equals("vn")) {
					// TODO vertex normal
					
				} else if(tokens[0].equals("vp")) {
					// TODO parameter space
					
				} else if(tokens[0].equals("f")) {
					// TODO face
					ArrayList<Integer> edges=new ArrayList<Integer>();

					int[] face=new int[tokens.length-1];
					
					for(int i=0;i<tokens.length-1;i++) {
						if(tokens[i+1].contains("/")) {
							face[i]=Integer.parseInt(tokens[i+1].split("/")[0]);	
						} else {
							face[i]=Integer.parseInt(tokens[i+1]);
						}
						if(face[i]<0) {
							face[i]=mesh.getNumVertices()-face[i];
						} else {
							face[i]--;
						}
					}
					
					try {
						int edgeID;
						for (int i = 0; i < face.length; i++) {
							try {
								edgeID = mesh.findEdge(face[i], face[(i+1)%face.length]);
							} catch (Exception e) {
								edgeID = mesh.addEdge(face[i], face[(i+1)%face.length]);
							}
							edges.add(edgeID);
						}

						mesh.addFace(edges);
					} catch (Exception e) {
						// TODO handle exception
					}

					
					edges.clear();
				} else {
					String message=MessageFormat.format(L10N.translate("unknown PLY header element: {0}"), tokens[0]);
					eu.moonlight3d.framework.State.getInstance().logger().warning(message);
				}
			}
			
			fileContents.setName("ImportedPLYInstance");
			mesh.setName("ImportedPLY");
			sceneGraph.addNode(fileContents);
			sceneGraph.addNode(mesh);
			sceneGraph.addChild(fileContents,mesh);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
        Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
        // Add CopyFromCache node to operator graph
        CopyFromCache copyFromCache=(CopyFromCache) document.getOGManager().createNodeByName("CopyFromCache");
        copyFromCache.getProperty("cachedGraph").setValue(copyFromCache.getName());
        // Add root scene graph to persistent cache
        document.getCacheManager().addSG(copyFromCache.getName(),sceneGraph);
        document.getOGManager().update();
	}
	
	private void exportMesh(RandomAccessFile output, Mesh mesh) throws IOException {
		HashMap<Integer, Integer> vertexMap=new HashMap<Integer, Integer>();
		
		output.writeBytes("ply\n");
		output.writeBytes("format ascii 1.0\n");
		output.writeBytes("comment created by " + State.getInstance().getVersionInformation().getApplicationVersionString() + "\n");
		output.writeBytes("element vertex " + mesh.getNumVertices() + "\n");
		output.writeBytes("property x double\n");
		output.writeBytes("property y double\n");
		output.writeBytes("property z double\n");
		output.writeBytes("element face " + mesh.getNumFaces() + "\n");
		output.writeBytes("property list uchar int vertex_index\n");
		output.writeBytes("end_header\n");
		
		for(int i=0;i<mesh.getNumVertices();i++) {
			Vector3D position=mesh.getVertexPosition(i);
			output.writeBytes(Double.toString(position.X1) + " " + Double.toString(position.X2) + " " + Double.toString(position.X3) + "\n");			
		}
		
		for(int i=0;i<mesh.getNumFaces();i++) {
			Mesh.Face face=mesh.getFace(i);
			
			ArrayList<Integer> vertices=mesh.getVertexIDs(face);
			
			output.writeBytes(Integer.toString(vertices.size()));
			
			for(int j=0;j<vertices.size();j++) {
				output.writeBytes(" " + vertices.get(j));
			}
			output.writeBytes("\n");			
		}
	}

	@Override
	public void exportFile(String filename) {
		eu.moonlight3d.ml3d.backend.threedim.document.Document mlDocument=(eu.moonlight3d.ml3d.backend.threedim.document.Document)(State.getInstance().getDocumentManager().getMainDocument());
		try {
			RandomAccessFile output=new RandomAccessFile(filename,"rw");
			output.setLength(0);
			
			output.writeBytes("# created with Moonlight|3D version " + eu.moonlight3d.framework.State.getInstance().getVersionInformation().getVersionString() + "\n");

			Selection selection=mlDocument.getSelectionManager().getSelection("viewSelection");
			if(selection.getType().getName()!="SGNode") {
				State.getInstance().logger().warning(L10N.translate("A mesh object needs to be selected"));
				return;
			}
			if(!(selection.getObjectFromSelectionString(selection.getActiveObject()) instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh)) {
				State.getInstance().logger().warning(L10N.translate("A mesh object needs to be selected"));
				return;
			}
			
			exportMesh(output,(Mesh)selection.getObjectFromSelectionString(selection.getActiveObject()));
			
			output.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
