/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Apr 21, 2008
 */

package eu.moonlight3d.ml3d.modelling.basic;

import java.util.HashMap;
import java.util.Map;

import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * A simple operator graph node that merges the output of two operator
 * graph nodes to one single scene graph.
 * 
 * @author gregor
 */
public class Merge extends Node {
	/**
	 * Map of node names between second input graph and output graph.
	 */
	private HashMap<String, String> nodeNameMap;
	
	/**
	 * Default constructor.
	 */
	public Merge() {
		createSlot("input1", Slot.Type.Input);
		createSlot("input2", Slot.Type.Input);
		createSlot("output", Slot.Type.Output);
	}
	
	/**
	 * the map between original nodes of the second scene graph and their clones
	 * in the merged (first) one.
	 */
	HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node,eu.moonlight3d.ml3d.backend.threedim.sg.Node> nodeMap;
	
	/**
	 * Rebuild the node hierarchy of the original scene graph in the merged one. This
	 * is achieved by descending recursively into the original graph and rebuilding
	 * each parent-child relation ship between the cloned nodes as the original one is
	 * passed.
	 * 
	 * @param graph the merged scene graph in which the node graph structure is rebuilt
	 * @param parent the parent node to iterate over
	 * @throws Exception if the adding of the parent-child relationship fails for any reason
	 */
	private void merge(SceneGraph graph, eu.moonlight3d.ml3d.backend.threedim.sg.Node parent) throws Exception {
		eu.moonlight3d.ml3d.backend.threedim.sg.Node clonedParent=nodeMap.get(parent);
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node child : parent.getChildren()) {
			eu.moonlight3d.ml3d.backend.threedim.sg.Node clonedChild=nodeMap.get(child);
			graph.addChild(clonedParent, clonedChild);
			merge(graph,child);
		}
	}
	
	@Override
	public void update() throws Exception {
		SceneGraph graph1=getSlot("input1").getGraph();
		SceneGraph graph2=getSlot("input2").getGraph();
		
		nodeMap=new HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node,eu.moonlight3d.ml3d.backend.threedim.sg.Node>();
		
		// merge graph2 into graph1
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node node : graph2.getNodes()) {
			eu.moonlight3d.ml3d.backend.threedim.sg.Node clonedNode=node.detachedCopy();
			String uniqueName=graph1.generateUniqueNodeName(clonedNode.getName());
			clonedNode.setName(uniqueName);
			graph1.addNode(clonedNode);
			nodeMap.put(node, clonedNode);
		}
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node node : graph2.getRootNodes()) {
			merge(graph1,node);
		}
		
		nodeNameMap=new HashMap<String, String>();
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node node : nodeMap.keySet()) {
			nodeNameMap.put(node.getName(), nodeMap.get(node).getName());
		}
		
		nodeMap=null;
		
		getSlot("output").setGraph(graph1);
	}

	@Override
	public void removeDependencies() {
	}
	
	/**
	 * Return a mapping of the node names in the second input graph to the node
	 * names in the output graph. This is required by tools inserting Merge so
	 * that they can keep track of renames made by this node.
	 * 
	 * @return node name map between second input and output graph
	 */
	public Map<String, String> getNameMap() {
		return (Map<String, String>) nodeNameMap.clone();
	}
}
