/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on May 7, 2007
 */

package eu.moonlight3d.ml3d.modelling.basic;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.sg.TransformableNode;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.GizmoListener;

/**
 * Base class for gizmo listeners which are used specifically by the
 * Transform Selected mode (see {@link TransformMode}). 
 *
 * @author gregor
 */
public abstract class TransformGizmoListener implements GizmoListener {
	
	/**
	 * Check whether the currently selected objects can all be 
	 * transformed by this gizmo listener.
	 * 
	 * @return true if it is possible, false otherwise
	 */
	public abstract boolean canTransformCurrentSelection();
	
	/**
	 * Determine the initial placement of the gizmo based on the
	 * currently selected objects.
	 * 
	 * @return initial position of gizmo
	 * @throws Exception if initial position cannot be determined
	 */
	public abstract Vector3D getInitialPosition() throws Exception;
	
	/**
	 * Determine the transformation between the world coordinate
	 * system and the local coordinate system of the object that
	 * is edited.
	 * 
	 * @return world to local transformation matrix
	 * @throws Exception if world to local transformation cannot be determined
	 */
	public Matrix4 getLocalToWorldTransformation() throws Exception {
		Selection selection=((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
        String selectionName=selection.getActiveObject();
        String[] selectionParts=selectionName.split("/");
        
        // the object is transformed in its parent's coordinate system
        if(selectionParts.length>4) {
        	String ownerOGNodeName=SelectionHelper.getOGNodeName(selectionName);
        	String ownerOGSlotName=SelectionHelper.getOGSlotName(selectionName);
        	String ownerSGNodePath=selectionParts[3];
        	for(int i=4;i<selectionParts.length-1;i++) {
        		ownerSGNodePath+="/" + selectionParts[i];
        	}
        	ownerSGNodePath=ownerSGNodePath.split(":")[0]; // remove appendix from SG node name
        	String ownerSGNodeName=selectionParts[selectionParts.length-1].split(":")[0];
        	eu.moonlight3d.ml3d.backend.threedim.og.Node ownerOGNode=((Document) State.getInstance().getDocumentManager().getMainDocument()).getOGManager().getNodeByName(ownerOGNodeName);
        	eu.moonlight3d.ml3d.backend.threedim.og.Slot ownerOGSlot=ownerOGNode.getSlot(ownerOGSlotName);
        
        	TransformableNode node=(TransformableNode) ownerOGSlot.getGraph().getNodeByName(ownerSGNodeName);
        
        	return node.getLocalToWorldTransformation(ownerSGNodePath);
        } else {
        	Matrix4 result=new Matrix4();
        	result.setIdentity();
        	return result;
        }
	}
	
}
