/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 21, 2007
 */
package eu.moonlight3d.ml3d.modelling.basic;

import java.util.HashMap;
import java.util.List;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.selection.SelectionListener;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.ui.mainwindow.modes.Mode;
import eu.moonlight3d.ml3d.ui.mainwindow.view.ViewEventHandler;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.Gizmo;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.WidgetViewEventHandler;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.Gizmo.DragMode;

/**
 *
 * This {@link Mode} is used for transforming any kind of object that is
 * currently selected and transformable.
 *
 * @author gregor
 */
public class TransformMode extends Mode {
	/**
	 * True if translation is enabled on the gizmo.
	 */
	private boolean translateEnabled;
	/**
	 * True if rotation is enabled on the gizmo.
	 */
	private boolean rotateEnabled;
	/**
	 * True if scale is enabled on the gizmo.
	 */
	private boolean scaleEnabled;
	/**
	 * The gizmo widget instance.
	 */
	private Gizmo gizmo;
	/**
	 * True if snapping to grid is enabled.
	 */
	private boolean snapToGridEnabled;
	/**
	 * Instance of grid gizmo snapper used when snapping is enabled.
	 */
	private GridGizmoSnapper gridSnapper;
	/**
	 * Indicates whether the selection listener should be active right now.
	 */
	private boolean selectionListenerActive;
	
	/**
	 * A selection listener that is needed to react to selection changes that
	 * happen while a translation mode is active.
	 *
	 * @author gregor
	 */
	private class TransformSelectionListener implements SelectionListener {

		public void elementAdded(String objectName) {
			// restart mode
			restart();
		}

		public void elementRemoved(String objectName) {
			// restart mode
			restart();
		}

		public void elementUpdated(String oldName, String newName) {
			// restart mode
			restart();
		}

		public void elementsAdded(List<String> objectNames) {
			// restart mode
			restart();
		}

		public void selectionCleared() {
			// restart mode
			restart();
		}
		
	}
	
	/**
	 * Wrapper around an actual TransformGizmoListener used to prevent recursions
	 * when the TransformSelectionListener is triggered as a result of an update
	 * of the current selection inside the gizmo listener (e.g. changing the operator
	 * graph node for the current selected objects to the newly inserted node).
	 *
	 * @author gregor
	 */
	private class TransformGizmoListenerWrapper extends TransformGizmoListener {
		/**
		 * The wrapped listener
		 */
		private TransformGizmoListener wrappedListener;

		/**
		 * Default constructor taking the listener that is to be wrapped.
		 * 
		 * @param wrapped the listener to be wrapped
		 */
		public TransformGizmoListenerWrapper(TransformGizmoListener wrapped) {
			this.wrappedListener=wrapped;
		}
		
		@Override
		public boolean canTransformCurrentSelection() {
			return wrappedListener.canTransformCurrentSelection();
		}
		
		@Override
		public Vector3D getInitialPosition() throws Exception {
			return wrappedListener.getInitialPosition();
		}

		public void apply() {
			selectionListenerActive=false;
			wrappedListener.apply();
			selectionListenerActive=true;
		}

		public void begin(DragMode mode) {
			selectionListenerActive=false;
			wrappedListener.begin(mode);
			selectionListenerActive=true;
		}

		public void end() {
			selectionListenerActive=false;
			wrappedListener.end();
			selectionListenerActive=true;
		}

		public void rotate(Vector3D center, Quaternion rotation) {
			selectionListenerActive=false;
			wrappedListener.rotate(center, rotation);
			selectionListenerActive=true;
		}

		public void scale(Vector3D center, Vector3D amount) {
			selectionListenerActive=false;
			wrappedListener.scale(center, amount);
			selectionListenerActive=true;
		}

		public void translate(Vector3D displacment) {
			selectionListenerActive=false;
			wrappedListener.translate(displacment);
			selectionListenerActive=true;
		}
		
	}
	
	/**
	 * The {@link TransformSelectionListener} that is used by the mode if active.
	 */
	private TransformSelectionListener transformSelectionListener;
	
	/**
	 * Default constructor.
	 */
	public TransformMode() {
		super("Transform");
		translateEnabled=true;
		rotateEnabled=true;
		scaleEnabled=true;
		snapToGridEnabled=false;
	}

	/**
	 * A map between selection type names and applicable gizmo listeners.
	 */
	private static HashMap<String, TransformGizmoListener> gizmoListeners;
	
	static {
		gizmoListeners=new HashMap<String, TransformGizmoListener>();
	}
	
	/**
	 * Register a gizmo listener for a certain selection type.
	 * 
	 * @param selectionType the selection type this gizmo listener applies to
	 * @param listener the listener instance to use
	 */
	public static void registerGizmoListener(String selectionType, TransformGizmoListener listener) {
		gizmoListeners.put(selectionType, listener);
	}
	
	/**
	 * Unregister and remove a gizmo listener for a certain selection type
	 * 
	 * @param selectionType the selection type for which the listener should be removed
	 */
	public static void unregisterGizmoListener(String selectionType) {
		gizmoListeners.remove(selectionType);
	}
	
	/**
	 * Enable or disable the translation handles on the gizmo.
	 * 
	 * @param translateEnabled true to enable t translation handles, false to disable
	 */
	public void setTranslateEnabled(boolean translateEnabled) {
		this.translateEnabled=translateEnabled;
		if(gizmo!=null) {
			gizmo.setTranslateEnabled(translateEnabled);
		}
	}
	
	/**
	 * Enable or disable the rotation handles on the gizmo.
	 * 
	 * @param rotateEnabled true to enable the rotation handles, false to disable
	 */
	public void setRotateEnabled(boolean rotateEnabled) {
		this.rotateEnabled=rotateEnabled;
		if(gizmo!=null) {
			gizmo.setRotateEnabled(rotateEnabled);
		}
	}
	
	/**
	 * Enable or disable the scale handles on the gizmo.
	 * 
	 * @param scaleEnabled true to enable the scale handles, false to disable
	 */
	public void setScaleEnabled(boolean scaleEnabled) {
		this.scaleEnabled=scaleEnabled;
		if(gizmo!=null) {
			gizmo.setScaleEnabled(scaleEnabled);
		}
	}

	/**
	 * Enable or disable snapping to grid.
	 * 
	 * @param snapToGridEnabled true when snapping should be enabled
	 */
	public void setSnapToGridEnabled(boolean snapToGridEnabled) {
		this.snapToGridEnabled=snapToGridEnabled;
		if(gizmo!=null) {
			if(snapToGridEnabled) {
				gridSnapper=new GridGizmoSnapper();
				gizmo.addSnapper(gridSnapper);
			} else {
				gizmo.removeSnapper(gridSnapper);
				gridSnapper=null;
			}
		}
	}
	
	/**
	 * Activate the mode. Note that this function is also called when the
	 * mode is reactivated, that is after it is reinitialized because of
	 * a change in the current selection.
	 * 
	 * @param registerSelectionListener true if a new selection listener must
	 * be registered
	 */
	private void activated(boolean registerSelectionListener) {
        Selection selection;
        String selectionType;
        TransformGizmoListener listener;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			if(selection.getType()==null) {
				eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("no current selection"));
				return;
			}
			selectionType=selection.getType().getName();
			
		} catch (Exception e) {
			ExceptionDialog.run(e);
			return;
		}
		
		if(registerSelectionListener) {
			transformSelectionListener=new TransformSelectionListener();
			selection.registerListener(transformSelectionListener);
		}

		listener=gizmoListeners.get(selectionType);
		if(listener==null) {
			eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("the currently selected object is not transformable"));
			return;
		}

		if(!listener.canTransformCurrentSelection()) {
			eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("the current selection is not transformable"));
			return;
		}
		
		if(selection.getSelection().size()==0) {
            eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("something must be selected to transform to it"));
            return;			
		}
		
        WidgetViewEventHandler widgetEventHandler=new WidgetViewEventHandler();
		gizmo=new Gizmo(widgetEventHandler);
		gizmo.setTranslateEnabled(translateEnabled);
		gizmo.setRotateEnabled(rotateEnabled);
		gizmo.setScaleEnabled(scaleEnabled);
		
		if(snapToGridEnabled) {
			gridSnapper=new GridGizmoSnapper();
			gizmo.addSnapper(gridSnapper);
		}
		
		// set gizmo location and transformation matrix
		try {
			gizmo.setPosition(listener.getInitialPosition());
			gizmo.setLocalToWorldTransformation(listener.getLocalToWorldTransformation());
		} catch (Exception e) {
			ExceptionDialog.run(e);
			return;
		}
		
		TransformGizmoListenerWrapper listenerWrapper=new TransformGizmoListenerWrapper(listener);
		selectionListenerActive=true;
		
		gizmo.addListener(listenerWrapper);

		ViewEventHandler.setEventHandler(widgetEventHandler);		
	}
	
	@Override
	public void activated() {
		activated(true);
	}
	
	/**
	 * Deactivate the mode. This is also called before the mode is reactivated
	 * as the result of a change in the current selection. In this case,
	 * the current selection listener will not be removed.
	 *  
	 * @param unregisterSelectionListener true if the selection listener should
	 * be removed (real deactivation) or false when it should stay (reactivation)
	 */
	private void deactivated(boolean unregisterSelectionListener) {
		ViewEventHandler.setEventHandler(ViewEventHandler.getDefaultEventHandler());
		
        Selection selection;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			if(unregisterSelectionListener) {
				selection.unregisterListener(transformSelectionListener);
				transformSelectionListener=null;
			}
		} catch(Exception e) {
			ExceptionDialog.run(e);
		}

		gizmo=null;
		gridSnapper=null;
	}

	@Override
	public void deactivated() {
		deactivated(true);
	}
	
	/**
	 * Restart the mode after some significant change in the scene, e.g. a selection change
	 */
	private void restart() {
		if(selectionListenerActive) {
			deactivated(false);
			activated(false);
		}
	}

}
