/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Aug 16, 2005
 */
package eu.moonlight3d.ml3d.modelling.camera;


import org.w3c.dom.Document;
import org.w3c.dom.Element;

import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Matrix3;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Ray;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.sg.TransformableNode;

/**
 * Camera scene graph node.
 *
 * @author gregor
 */
public class Camera extends TransformableNode {
    private Vector3D position;
    private Vector3D up;
    private Vector3D front;
    private Vector3D right;

    public Camera()
    {
    	super();
        position=new Vector3D();
        position.clear();
        up=new Vector3D();
        up.clear();
        up.X2=1.0;
        front=new Vector3D();
        front.clear();
        front.X3=-1.0;
        right=new Vector3D();
        right.clear();
        right.X1=1.0;
    }
    
    @Override
    public void translate(Vector3D displacement)
    {
        position=position.add(displacement);
    }

    @Override
    public void applyMatrix(Matrix3 matrix)
    {
        position=matrix.multiply(position);
        up=matrix.multiply(up);
        front=matrix.multiply(front);
        right=matrix.multiply(right);
    }

    @Override
    public void applyMatrix(Matrix4 matrix)
    {
        position=matrix.multiply(position);
        up=matrix.multiply(up);
        front=matrix.multiply(front);
        right=matrix.multiply(right);
    }

    @Override
    public AxisAlignedBoundingBox getBoundingBox()
    {
        AxisAlignedBoundingBox boundingBox=new AxisAlignedBoundingBox();
        boundingBox.min=position;
        boundingBox.max=position;
        return boundingBox;
    }

    @Override
    public Vector3D getIntersectionPoint(Ray ray)
    {
        /*
		 * A camera does not have a defined volume associated with it. So we
		 * just define a bounding sphere around it's position which we use for
		 * collision detection. This is a hack to make the camera selection for
		 * the UI work.
		 */
		final double radius = 0.1;
		Vector3D point;
		double a, b, c, det, distance = 0;

		a = ray.direction.multiply(ray.direction);
		b = 2 * ray.direction.multiply(ray.start.sub(position));
		c = ((ray.start.sub(position)).multiply(ray.start.sub(position))) - (radius * radius);
		det = b * b - 4 * a * c;

		if (det == 0) {
			distance = (-b / (2 * a));
		} else if (det > 0) {
			double t1, t2;
			t1 = (-b + Math.sqrt(det)) / (2 * a);
			t2 = (-b - Math.sqrt(det)) / (2 * a);
			if (t1 < t2) {
				distance = t1;
			} else {
				distance = t2;
			}
		} else {
			return null;
		}

		point = ray.start.add(ray.direction.multiply(distance));

		return point;
    }
    
    @Override
	public eu.moonlight3d.ml3d.backend.threedim.sg.Node clone() {
    		Camera c=(Camera)super.clone();
    		
    		c.position=position.clone();
    		c.up=up.clone();
    		c.right=right.clone();
    		c.front=front.clone();
    		
    		return c;
    }

    public Vector3D getPosition()
    {
        return position.clone();
    }
    
    public Vector3D getUp()
    {
        return up.clone();
    }
    
    public Vector3D getRight()
    {
        return right.clone();
    }
    
    public Vector3D getFront()
    {
        return front.clone();
    }

    public boolean isValidAppendixName(String appendix)
    {
        // there is no appendix for a camera right now
        return false;
    }

    public Object getFromAppendix(String appendix)
    {
        // don't try to return anything - we can't
        return null;
    }

	@Override
	public String getType() {
		return "Camera";
	}
	
	@Override
	protected void loadXMLData(Element dataElement) {
		Element positionElement=(Element) dataElement.getElementsByTagName("position").item(0);
		position.X1=Double.parseDouble(positionElement.getAttribute("x"));
		position.X2=Double.parseDouble(positionElement.getAttribute("y"));
		position.X3=Double.parseDouble(positionElement.getAttribute("z"));
		Element upElement=(Element) dataElement.getElementsByTagName("up").item(0);
		up.X1=Double.parseDouble(upElement.getAttribute("x"));
		up.X2=Double.parseDouble(upElement.getAttribute("y"));
		up.X3=Double.parseDouble(upElement.getAttribute("z"));
		Element frontElement=(Element) dataElement.getElementsByTagName("front").item(0);
		front.X1=Double.parseDouble(frontElement.getAttribute("x"));
		front.X2=Double.parseDouble(frontElement.getAttribute("y"));
		front.X3=Double.parseDouble(frontElement.getAttribute("z"));
		Element rightElement=(Element) dataElement.getElementsByTagName("right").item(0);
		right.X1=Double.parseDouble(rightElement.getAttribute("x"));
		right.X2=Double.parseDouble(rightElement.getAttribute("y"));
		right.X3=Double.parseDouble(rightElement.getAttribute("z"));
	}
	
	@Override
	protected void saveXMLData(Document document, Element dataElement) {
		Element positionElement=document.createElement("position");
		positionElement.setAttribute("x",Double.toString(position.X1));
		positionElement.setAttribute("y",Double.toString(position.X2));
		positionElement.setAttribute("z",Double.toString(position.X3));
		dataElement.appendChild(positionElement);
		Element upElement=document.createElement("up");
		upElement.setAttribute("x",Double.toString(up.X1));
		upElement.setAttribute("y",Double.toString(up.X2));
		upElement.setAttribute("z",Double.toString(up.X3));
		dataElement.appendChild(upElement);
		Element frontElement=document.createElement("front");
		frontElement.setAttribute("x",Double.toString(front.X1));
		frontElement.setAttribute("y",Double.toString(front.X2));
		frontElement.setAttribute("z",Double.toString(front.X3));
		dataElement.appendChild(frontElement);
		Element rightElement=document.createElement("right");
		rightElement.setAttribute("x",Double.toString(right.X1));
		rightElement.setAttribute("y",Double.toString(right.X2));
		rightElement.setAttribute("z",Double.toString(right.X3));
		dataElement.appendChild(rightElement);
	}
}
