/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Aug 14, 2005
 */
package eu.moonlight3d.ml3d.modelling.lighting;


import org.w3c.dom.Document;
import org.w3c.dom.Element;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Matrix3;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Ray;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.sg.TransformableNode;

/**
 * A omnidirectional light scene graph node. This scene graph node 
 * represents a light source that radiates in all directions equally.
 *
 * @see eu.moonlight3d.ml3d.backend.sg.TransformableNode eu.moonlight3d.ml3d.modelling.lighting.CreateLight 
 *
 * @author gregor
 */
public class OmniLight extends TransformableNode {
	/**
	 * Light position
	 */
    private Vector3D position;
    /**
     * Light intensity
     */
    private double intensity;
    /**
     * Light colour
     */
    private Colour colour;

    /**
     * Default constructor
     */
    public OmniLight()
    {
        super();
        position=new Vector3D();
        position.clear();
        intensity=100;
        colour=new Colour();
        colour.red=1.0;
        colour.green=1.0;
        colour.blue=1.0;
        colour.alpha=1.0;
    }

    @Override
    public void applyMatrix(Matrix3 matrix)
    {
        position=matrix.multiply(position);
    }

    @Override
    public void applyMatrix(Matrix4 matrix)
    {
        position=matrix.multiply(position);
    }

    @Override
    public AxisAlignedBoundingBox getBoundingBox()
    {
        AxisAlignedBoundingBox boundingBox=new AxisAlignedBoundingBox();
        boundingBox.min=position.clone();
        boundingBox.max=position.clone();
        return boundingBox;
    }

    @Override
    public Vector3D getIntersectionPoint(Ray ray) {
		/*
		 * And now for something completely different: A hack
		 * 
		 * Since light sources traditionally are infinitely small points we have
		 * to play a little trick here to give them a volume that is selectable.
		 * We do this by assuming a bounding sphere around the light source
		 * which is intersected with the ray.
		 */
		final double radius = 0.1;
		Vector3D point;
		double a, b, c, det, distance = 0;

		a = ray.direction.multiply(ray.direction);
		b = 2 * ray.direction.multiply(ray.start.sub(position));
		c = ((ray.start.sub(position)).multiply(ray.start.sub(position))) - (radius * radius);
		det = b * b - 4 * a * c;

		if (det == 0) {
			distance = (-b / (2 * a));
		} else if (det > 0) {
			double t1, t2;
			t1 = (-b + Math.sqrt(det)) / (2 * a);
			t2 = (-b - Math.sqrt(det)) / (2 * a);
			if (t1 < t2) {
				distance = t1;
			} else {
				distance = t2;
			}
		} else {
			return null;
		}

		point = ray.start.add(ray.direction.multiply(distance));

		return point;
    }
    
    @Override
	public eu.moonlight3d.ml3d.backend.threedim.sg.Node clone() {
    	OmniLight l=(OmniLight)super.clone();
		
		l.position=position.clone();
		
   		return l;
    }
    
    /**
     * Return the current light position
     * 
     * @return the light position
     */
    public Vector3D getPosition()
    {
        return position.clone();
    }
    
    /**
     * Return the current light intensity
     * 
     * @return the light intensity
     */
    public double getIntensity() {
    		return intensity;
    }
    
    /**
     * Set the light intensity
     * 
     * @param intensity the new light intensity
     */
    public void setIntensity(double intensity) {
    		this.intensity=intensity;
    }
    
    /**
     * Return the light colour
     * 
     * @return the current light colour
     */
    public Colour getColour() {
    		return colour.clone();
    }
    
    /**
     * Set the light colour
     * 
     * @param colour the new light colour
     */
    public void setColour(Colour colour) {
    		this.colour=colour.clone();
    }

    public boolean isValidAppendixName(String appendix)
    {
        // always return false as there are no valid appendices
        return false;
    }

    public Object getFromAppendix(String appendix)
    {
        // always return null as there are no valid appendices
        return null;
    }

	@Override
	public String getType() {
		return "OmniLight";
	}

	@Override
	protected void loadXMLData(Element dataElement) {
		Element positionElement=(Element) dataElement.getElementsByTagName("position").item(0);
		position.X1=Double.parseDouble(positionElement.getAttribute("x"));
		position.X2=Double.parseDouble(positionElement.getAttribute("y"));
		position.X3=Double.parseDouble(positionElement.getAttribute("z"));
		Element intensityElement=(Element) dataElement.getElementsByTagName("intensity").item(0);
		intensity=Double.parseDouble(intensityElement.getAttribute("value"));
		Element colourElement=(Element) dataElement.getElementsByTagName("colour").item(0);
		colour.red=Double.parseDouble(colourElement.getAttribute("r"));
		colour.green=Double.parseDouble(colourElement.getAttribute("g"));
		colour.blue=Double.parseDouble(colourElement.getAttribute("b"));
		colour.alpha=Double.parseDouble(colourElement.getAttribute("a"));
	}

	@Override
	protected void saveXMLData(Document document, Element dataElement) {
		Element positionElement=document.createElement("position");
		positionElement.setAttribute("x",Double.toString(position.X1));
		positionElement.setAttribute("y",Double.toString(position.X2));
		positionElement.setAttribute("z",Double.toString(position.X3));
		dataElement.appendChild(positionElement);
		Element intensityElement=document.createElement("intensity");
		intensityElement.setAttribute("value",Double.toString(intensity));
		dataElement.appendChild(intensityElement);
		Element colourElement=document.createElement("colour");
		colourElement.setAttribute("r",Double.toString(colour.red));
		colourElement.setAttribute("g",Double.toString(colour.green));
		colourElement.setAttribute("b",Double.toString(colour.blue));
		colourElement.setAttribute("a",Double.toString(colour.alpha));
		dataElement.appendChild(colourElement);
	}
}
