/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Jan 29, 2005
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.Arrays;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * An operator graph node that creates a simple box mesh.
 *
 * @author gregor
 */
public class CreateBox extends Node
{
	private static final int X = 0;
	private static final int Y = 1;
	private static final int Z = 2;

	private Property size;
	private Property name;
	private Property divX, divY, divZ;

	/**
	 * Default constructor.
	 */
	public CreateBox()
	{
		super();
		createSlot("output", Slot.Type.Output);
		size = new Property(this);
		size.setName("size");
		size.setType(Property.Type.Float);
		try {
			size.setValue(1.0);
		} catch (Exception e) {
			// This exception shouldn't occur at all in this limited case
			e.printStackTrace();
		}
		name = new Property(this);
		name.setName("name");
		name.setType(Property.Type.String);
		try	{
			name.setValue("Box");
		} catch (Exception e) {
			// This exception shouldn't occur at all in this limited case
			e.printStackTrace();
		}
		divX = new Property(this);
		divX.setName("divisionsX");
		divX.setType(Property.Type.Integer);
		try	{
			divX.setValue(0);
		} catch (Exception e) {
			// This exception shouldn't occur at all in this limited case
			e.printStackTrace();
		}

		divY = new Property(this);
		divY.setName("divisionsY");
		divY.setType(Property.Type.Integer);
		try {
			divY.setValue(0);
		} catch (Exception e) {
			// This exception shouldn't occur at all in this limited case
			e.printStackTrace();
		}

		divZ = new Property(this);
		divZ.setName("divisionsZ");
		divZ.setType(Property.Type.Integer);
		try {
			divZ.setValue(0);
		} catch (Exception e) {
			// This exception shouldn't occur at all in this limited case
			e.printStackTrace();
		}
	}
	
	/**
	 * Create a properly subdivided edge of the cube. This creates the series of mesh
	 * vertices and edge segments for the given edge.
	 * 
	 * @param mesh
	 * @param startVertex
	 * @param startPos
	 * @param endVertex
	 * @param endPos
	 * @param numDivisions
	 * @return the list of mesh edge IDs of the created edges
	 */
	private int[] createDividedEdge(Mesh mesh, int startVertex, Vector3D startPos, int endVertex, Vector3D endPos, int numDivisions) {
		int[] dividedEdge = new int[numDivisions + 2];
		dividedEdge[0] = startVertex;
		dividedEdge[numDivisions + 1] = endVertex;
		for (int i = 1; i < dividedEdge.length - 1; i++) {
			Vector3D vertexPos=startPos.clone();
			Vector3D.add(vertexPos, endPos.sub(startPos).multiply(i/(float)(numDivisions+1)));
			dividedEdge[i] = mesh.addVertex(vertexPos);
		}
		return dividedEdge;
	}

	private void createFace(int[] edge1, int[] edge2, int[] edge3, int[] edge4, Mesh mesh,
			int edge2Direction, boolean invertedFirstDir, boolean invertedSecondDir) throws Exception {
		// saves the IDs of all Vertices on a face
		int[][] vertexIDs = new int[edge1.length][edge2.length];
		if(!invertedSecondDir) {
			vertexIDs[0] = Arrays.copyOf(edge2, edge2.length);
			vertexIDs[edge1.length - 1] = Arrays.copyOf(edge4, edge4.length);			
		} else {
			for(int i=0;i<edge2.length;i++) {
				vertexIDs[0][i]=edge2[edge2.length-i-1];
			}
			for(int i=0;i<edge4.length;i++) {
				vertexIDs[edge1.length-1][i]=edge4[edge4.length-i-1];
			}
		}
//		vertexIDs[0] = Arrays.copyOf(edge2, edge2.length);
//		vertexIDs[edge1.length - 1] = Arrays.copyOf(edge4, edge4.length);

		for (int i = 0; i < edge1.length; i++) {
			vertexIDs[i][0] = edge1[i];
			vertexIDs[i][edge2.length - 1] = edge3[i];
		}
		try {
			// create all vertices inside the face
			for (int x = 1; x < edge1.length - 1; x++) {
				Vector3D vecX = mesh.getVertexPosition(edge1[x]);
				for (int y = 1; y < edge2.length - 1; y++) {
					Vector3D vert = new Vector3D(vecX);
					if (edge2Direction == X) {
						vert.X1 = mesh.getVertex(edge2[y]).position.X1;
					} else if (edge2Direction == Y) {
						vert.X2 = mesh.getVertex(edge2[y]).position.X2;
					} else if (edge2Direction == Z) {
						vert.X3 = mesh.getVertex(edge2[y]).position.X3;
					}
					if(!invertedSecondDir) {
						vertexIDs[x][y] = mesh.addVertex(vert);
					} else {
						vertexIDs[x][vertexIDs[x].length-y-1] = mesh.addVertex(vert);					
					}
//					vertexIDs[x][y] = mesh.addVertex(vert);
				}
			}
			
			/*
			 * Create the edges and faces. This algorithm sweeps over the face of
			 * the cube and creates the required edges for the faces on the fly.
			 * It only remembers the list of edges in the perpendicular direction
			 * and the last edge used in the parallel direction of the outer loop.
			 * This makes it a bit tricky to read, though. 
			 */

			// create the edges and faces
			int[] rowEdges = new int[edge1.length];
			for (int x = 0; x < edge1.length - 1; x++) {
				// check wether the edge exists before adding it
				rowEdges[x] = mesh.findOrCreateEdge(vertexIDs[x + 1][0], vertexIDs[x][0]);
			}

			for (int y = 0; y < edge2.length - 1; y++) {
				int lastVerticalEdge = mesh.findOrCreateEdge(vertexIDs[0][y], vertexIDs[0][y + 1]);
				for (int x = 0; x < edge1.length - 1; x++) {
					ArrayList<Integer> edges = new ArrayList<Integer>(4);
					edges.add(rowEdges[x]);
					edges.add(lastVerticalEdge);
					// again a check is needed
					rowEdges[x] = mesh.findOrCreateEdge(vertexIDs[x][y + 1], vertexIDs[x + 1][y + 1]);
					lastVerticalEdge = mesh.findOrCreateEdge(vertexIDs[x + 1][y + 1], vertexIDs[x + 1][y]);
					edges.add(rowEdges[x]);
					edges.add(lastVerticalEdge);
					mesh.addFace(edges);
					edges.clear();
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public void createSimpleCube() throws Exception {
    	SceneGraph graph=new SceneGraph(getDocument());
        eu.moonlight3d.ml3d.backend.threedim.sg.Instance instance;
        eu.moonlight3d.ml3d.modelling.mesh.Mesh mesh;
        Vector3D corner=new Vector3D();
        ArrayList<Integer> edges=new ArrayList<Integer>();
        
        instance=new eu.moonlight3d.ml3d.backend.threedim.sg.Instance();
        instance.setName(name.getStringValue() + "Instance");
      
        mesh=new eu.moonlight3d.ml3d.modelling.mesh.Mesh();
        mesh.setName(name.getStringValue());
      
        graph.addNode(instance);
        graph.addNode(mesh);
        graph.addChild(instance,mesh);
      
        // add corners to the box
        corner.X1=corner.X2=corner.X3=size.getFloatValue();
        mesh.addVertex(corner); // +++ (0)
        corner.X1*=-1; 
        mesh.addVertex(corner); // -++ (1)
        corner.X2*=-1; 
        mesh.addVertex(corner); // --+ (2)
        corner.X3*=-1; 
        mesh.addVertex(corner); // --- (3)
        corner.X1*=-1; 
        mesh.addVertex(corner); // +-- (4)
        corner.X2*=-1; 
        mesh.addVertex(corner); // ++- (5)
        corner.X2*=-1;
        corner.X3*=-1;
        mesh.addVertex(corner); // +-+ (6)
        corner.X1*=-1;
        corner.X2*=-1;
        corner.X3*=-1;
        mesh.addVertex(corner); // -+- (7)
        
        // add edges to the box
        mesh.addEdge(0,1); // edgeID 0
        mesh.addEdge(1,2); // edgeID 1
        mesh.addEdge(3,2); // edgeID 2
        mesh.addEdge(3,4); // edgeID 3
        mesh.addEdge(4,5); // edgeID 4
        mesh.addEdge(4,6); // edgeID 5
        mesh.addEdge(3,7); // edgeID 6
        mesh.addEdge(0,6); // edgeID 7
        mesh.addEdge(1,7); // edgeID 8
        mesh.addEdge(2,6); // edgeID 9
        mesh.addEdge(0,5); // edgeID 10
        mesh.addEdge(5,7); // edgeID 11

        // add faces to the box
        edges.add(0); // 0 -> 1
        edges.add(1); // 1 -> 2
        edges.add(9); // 2 -> 6
        edges.add(7); // 6 -> 0
        mesh.addFace(edges); // first face OK

        edges.clear();
        edges.add(0);  // 0 -> 1
        edges.add(10); // 5 -> 0
        edges.add(11); // 7 -> 5
        edges.add(8);  // 1 -> 7
        mesh.addFace(edges); // second face OK

        edges.clear();
        edges.add(7);  // 0 -> 6
        edges.add(5);  // 6 -> 4
        edges.add(4);  // 4 -> 5
        edges.add(10); // 5 -> 0
        mesh.addFace(edges); // third face OK

        edges.clear();
        edges.add(2); // 3 -> 2
        edges.add(1); // 2 -> 1
        edges.add(8); // 1 -> 7
        edges.add(6); // 7 -> 3
        mesh.addFace(edges); // fourth face OK

        edges.clear();
        edges.add(2); // 3 -> 2
        edges.add(3); // 4 -> 3
        edges.add(5); // 6 -> 4
        edges.add(9); // 2 -> 6
        mesh.addFace(edges); // fifth face OK

        edges.clear();
        edges.add(6);  // 3 -> 7
        edges.add(11); // 7 -> 5
        edges.add(4);  // 5 -> 4
        edges.add(3);  // 4 -> 3
        mesh.addFace(edges);

        getSlot("output").setGraph(graph);		
	}

	/* (non-Javadoc)
	 * @see eu.moonlight3d.ml3d.backend.og.Node#update()
	 */
	@Override
	public void update() throws Exception
	{
		if(divX.getIntegerValue()==0 && divY.getIntegerValue()==0 && divZ.getIntegerValue()==0) {
			createSimpleCube();
			return;
		}

		SceneGraph graph = new SceneGraph(getDocument());
		eu.moonlight3d.ml3d.backend.threedim.sg.Instance instance;
		eu.moonlight3d.ml3d.modelling.mesh.Mesh mesh;
		
		instance = new eu.moonlight3d.ml3d.backend.threedim.sg.Instance();
		instance.setName(name.getStringValue() + "Instance");

		mesh = new eu.moonlight3d.ml3d.modelling.mesh.Mesh();
		mesh.setName(name.getStringValue());

		graph.addNode(instance);
		graph.addNode(mesh);
		graph.addChild(instance, mesh);

		double size = this.size.getFloatValue();
		int divX = this.divX.getIntegerValue();
		int divY = this.divY.getIntegerValue();
		int divZ = this.divZ.getIntegerValue();

		int[] corners = new int[8];
		Vector3D[] cornerPositions=new Vector3D[8];
		cornerPositions[0] = new Vector3D(size, size, size);		// +++ (0)
		cornerPositions[1] = new Vector3D(-size, size, size);		// -++ (1)
		cornerPositions[2] = new Vector3D(-size, -size, size);		// --+ (2)
		cornerPositions[3] = new Vector3D(-size, -size, -size);		// --- (3)
		cornerPositions[4] = new Vector3D(size, -size, -size);		// +-- (4)
		cornerPositions[5] = new Vector3D(size, size, -size);		// ++- (5)
		cornerPositions[6] = new Vector3D(size, -size, size);		// +-+ (6)
		cornerPositions[7] = new Vector3D(-size, size, -size);		// -+- (7)
		// Don't change the order of these Lines - it is important for
		// downward compatibility
		corners[0] = mesh.addVertex(cornerPositions[0]);		// +++ (0)
		corners[1] = mesh.addVertex(cornerPositions[1]);		// -++ (1)
		corners[2] = mesh.addVertex(cornerPositions[2]);		// --+ (2)
		corners[3] = mesh.addVertex(cornerPositions[3]);		// --- (3)
		corners[4] = mesh.addVertex(cornerPositions[4]);		// +-- (4)
		corners[5] = mesh.addVertex(cornerPositions[5]);		// ++- (5)
		corners[6] = mesh.addVertex(cornerPositions[6]);		// +-+ (6)
		corners[7] = mesh.addVertex(cornerPositions[7]);		// -+- (7)

		int[][] borderEdges = new int[12][];
		
		// create the edges of the cube - again ordering is important
		borderEdges[0]=createDividedEdge(mesh, corners[1], cornerPositions[1], corners[0], cornerPositions[0], divX);
		borderEdges[1]=createDividedEdge(mesh, corners[2], cornerPositions[2], corners[1], cornerPositions[1], divY);
		borderEdges[2]=createDividedEdge(mesh, corners[3], cornerPositions[3], corners[2], cornerPositions[2], divZ);
		borderEdges[3]=createDividedEdge(mesh, corners[3], cornerPositions[3], corners[4], cornerPositions[4], divX);
		borderEdges[4]=createDividedEdge(mesh, corners[4], cornerPositions[4], corners[5], cornerPositions[5], divY);
		borderEdges[5]=createDividedEdge(mesh, corners[4], cornerPositions[4], corners[6], cornerPositions[6], divZ);
		borderEdges[6]=createDividedEdge(mesh, corners[3], cornerPositions[3], corners[7], cornerPositions[7], divY);
		borderEdges[7]=createDividedEdge(mesh, corners[6], cornerPositions[6], corners[0], cornerPositions[0], divY);
		borderEdges[8]=createDividedEdge(mesh, corners[7], cornerPositions[7], corners[1], cornerPositions[1], divZ);
		borderEdges[9]=createDividedEdge(mesh, corners[2], cornerPositions[2], corners[6], cornerPositions[6], divX);
		borderEdges[10]=createDividedEdge(mesh, corners[5], cornerPositions[5], corners[0], cornerPositions[0], divZ);
		borderEdges[11]=createDividedEdge(mesh, corners[7], cornerPositions[7], corners[5], cornerPositions[5], divX);

		// the 6 faces of the cube are created using a helper function
//		createFace(borderEdges[9], borderEdges[1], borderEdges[0], borderEdges[7], mesh, Y);
		createFace(borderEdges[0], borderEdges[1], borderEdges[9], borderEdges[7], mesh, Y,false,true);
		createFace(borderEdges[11], borderEdges[8], borderEdges[0], borderEdges[10], mesh, Z,false,false);
		createFace(borderEdges[5], borderEdges[4], borderEdges[10], borderEdges[7], mesh, Y,false,false);
//		createFace(borderEdges[2], borderEdges[6], borderEdges[8], borderEdges[1], mesh, Y);
//		createFace(borderEdges[2], borderEdges[1], borderEdges[8], borderEdges[6], mesh, Z);
		createFace(borderEdges[8], borderEdges[6], borderEdges[2], borderEdges[1], mesh, Y,false,true);
		createFace(borderEdges[2], borderEdges[3], borderEdges[5], borderEdges[9], mesh, X,false,false);
		createFace(borderEdges[3], borderEdges[6], borderEdges[11], borderEdges[4], mesh, Y,false,false);

		getSlot("output").setGraph(graph);
	}

	@Override
	public void removeDependencies()
	{
	}
}
