/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on May 1, 2005
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;


/**
 * An operator graph node that creates a pipe mesh.
 *
 * @author gregor
 */
public class CreatePipe extends Node {
    protected Property name;
    protected Property radius;
    protected Property height;
    protected Property thickness;
    protected Property divisions;

    public CreatePipe()
    {
    	super();
        createSlot("output", Slot.Type.Output);
        name=new Property(this);
        name.setName("name");
        name.setType(Property.Type.String);
        try {
            name.setValue("Pipe");
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
        radius=new Property(this);
        radius.setName("radius");
        radius.setType(Property.Type.Float);
        try {
            radius.setValue(1.0);
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
        thickness=new Property(this);
        thickness.setName("thickness");
        thickness.setType(Property.Type.Float);
        try {
            thickness.setValue(0.5);
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
        height=new Property(this);
        height.setName("height");
        height.setType(Property.Type.Float);
        try {
            height.setValue(1.0);
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
        divisions=new Property(this);
        divisions.setName("divisions");
        divisions.setType(Property.Type.Integer);
        try {
            divisions.setValue(32);
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
    }

    @Override
    public void update() throws Exception
    {
    	SceneGraph graph=new SceneGraph(getDocument());
        eu.moonlight3d.ml3d.backend.threedim.sg.Instance instance;
        eu.moonlight3d.ml3d.modelling.mesh.Mesh mesh;
        
        instance=new eu.moonlight3d.ml3d.backend.threedim.sg.Instance();
        instance.setName(name.getStringValue() + "Instance");
        mesh=new eu.moonlight3d.ml3d.modelling.mesh.Mesh();
        
        mesh.setName(name.getStringValue());
        
        graph.addNode(instance);
        graph.addNode(mesh);
        graph.addChild(instance,mesh);
        
        int divisions = this.divisions.getIntegerValue();

        mesh.setAutoNormalsUpdate(false);

		// generate vertices (first inner, then outer vertices)
        Vector3D vertex=new Vector3D();
		for (int i = 0; i < divisions; i++) {
			double phi = 2 * i * Math.PI / divisions;

			vertex.X1 = radius.getFloatValue() * Math.cos(phi);
			vertex.X2 = radius.getFloatValue() * Math.sin(phi);
			vertex.X3 = height.getFloatValue()/2;
			int vertexID1=mesh.addVertex(vertex);
			vertex.X3*=-1;
			int vertexID2=mesh.addVertex(vertex);
			
			mesh.addEdge(vertexID1,vertexID2); // add inner vertical edges
		}
		for (int i = 0; i < divisions; i++) {
			double phi = 2 * i * Math.PI / divisions;
			double outerRadius=radius.getFloatValue()+thickness.getFloatValue();

			vertex.X1 = outerRadius * Math.cos(phi);
			vertex.X2 = outerRadius * Math.sin(phi);
			vertex.X3 = height.getFloatValue()/2;
			int vertexID1=mesh.addVertex(vertex);
			vertex.X3*=-1;
			int vertexID2=mesh.addVertex(vertex);
			
			mesh.addEdge(vertexID1,vertexID2); // add outer vertical edges
		}
		
		// add ring edges (alternating between top and bottom)
		for(int i=0;i<2*divisions;i++) {
			mesh.addEdge(i,(i+2)%(2*divisions)); // inner ring edge
			mesh.addEdge((2*divisions)+i,(2*divisions)+((i+2)%(2*divisions))); // outer ring edge
		}
		// add radial edges
		for(int i=0;i<divisions;i++) {
			mesh.addEdge(2*i,2*(divisions+i)); // upper radial edge
			mesh.addEdge(2*i+1,2*(divisions+i)+1); // lower radial edge
		}
		// add top and bottom faces
		// so far we've created (10*division) edges; we're interested in ring and radial edges 
		for(int i=0;i<divisions;i++) {
			ArrayList<Integer> edgeLoop=new ArrayList<Integer>();
			
			edgeLoop.add(2*divisions+4*i); // inner upper ring edge
			edgeLoop.add(6*divisions+2*i); // first upper radial edge
			edgeLoop.add(2*divisions+(4*i+1)%(4*divisions)); // outer upper ring edge
			edgeLoop.add(6*divisions+((2*i+2)%(2*divisions))); // second upper radial edge
			
			mesh.addFace(edgeLoop);
			
			edgeLoop.clear();
			edgeLoop.add(2*divisions+2+4*i); // inner lower ring edge
			edgeLoop.add(6*divisions+1+2*i); // first lower radial edge
			edgeLoop.add(2*divisions+(4*i+3)%(4*divisions)); // outer lower ring edge
			edgeLoop.add(6*divisions+((2*i+3)%(2*divisions))); // second lower radial edge
			
			mesh.addFace(edgeLoop);
		}
		// add ring faces
		for(int i=0;i<divisions;i++) {
			ArrayList<Integer> edgeLoop=new ArrayList<Integer>();
			
			edgeLoop.add(i); // first inner vertical edge
			edgeLoop.add(2*divisions+4*i); // inner upper ring edge
			edgeLoop.add((i+1)%divisions); // second inner vertical edge
			edgeLoop.add(2*divisions+2+4*i); // inner lower ring edge
			
			mesh.addFace(edgeLoop);
			
			edgeLoop.clear();
			
			edgeLoop.add(divisions+i); // first outer vertical edge
			edgeLoop.add(2*divisions+(4*i+1)%(4*divisions)); // outer upper ring edge
			edgeLoop.add(divisions+(i+1)%divisions); // second outer vertical edge
			edgeLoop.add(2*divisions+(4*i+3)%(4*divisions)); // outer lower ring edge
			
			mesh.addFace(edgeLoop);
		}

        mesh.setAutoNormalsUpdate(true);

        getSlot("output").setGraph(graph);
    }

	@Override
	public void removeDependencies() {
	}
}
