/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jul 23, 2006
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.Arrays;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh.Edge;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh.Face;


/**
 * An operator graph node that splits edges into several segments.
 *
 * @author gregor
 */
public class DivideEdges extends Node {
	/**
	 * The name of the mesh operator graph node to manipulate
	 */
	private Property meshName;
	/**
	 * The IDs of the edges to divide
	 */
	private Property edges;
	/**
	 * The number of subdivisions to create for each edge
	 */
	private Property subdivisions;
	
	/**
	 * Default constructor
	 */
	public DivideEdges() {
		super();
		createSlot("input",Slot.Type.Input);
		createSlot("output",Slot.Type.Output);
		
		meshName=new Property(this);
		meshName.setName("meshName");
		meshName.setType(Value.Type.String);
		try {
			meshName.setValue("none");
		} catch (Exception e) {
			// cannot happen  here
			e.printStackTrace();
		}
		edges=new Property(this);
		edges.setName("edges");
		edges.setType(Value.Type.IntegerArray);
		try {
			edges.setMaxSize(0);
		} catch (Exception e) {
			// cannot happen  here
			e.printStackTrace();
		}
		subdivisions=new Property(this);
		subdivisions.setName("subdivisions");
		subdivisions.setType(Value.Type.Integer);
		try {
			subdivisions.setValue(1);
		} catch (Exception e) {
			// cannot happen  here
			e.printStackTrace();
		}
	}
	
	/**
	 * Rebuild the faces around a certain divided edge.
	 * 
	 * @param mesh the mesh to work on
	 * @param edgeID the ID of the old edge that is going to be deleted
	 * @param edge the old edge
	 * @param newEdges the list of IDs of the new edges to insert into the faces
	 * @throws Exception
	 */
	private void rebuildFaces(Mesh mesh, int edgeID, Edge edge, ArrayList<Integer> newEdges) throws Exception {
		for(int i=0;i<edge.faceIDs.length;i++) {
			Face face=mesh.getFace(edge.faceIDs[i]);
			ArrayList<Integer> newIds=new ArrayList<Integer>();
			// rebuild face loop
//			for(int id : face.edgeIDs) {
			for(int j=0;j<face.edgeIDs.length;j++) {
				int id=face.edgeIDs[j];
				if(id!=edgeID) {
					newIds.add(id);
				} else {
					// check if order of new edges needs to be reversed
//					Edge previousEdge=mesh.getEdge(edge.faceIDs[(i+edge.faceIDs.length-1)%edge.faceIDs.length]);
					Edge previousEdge=mesh.getEdge(face.edgeIDs[(j+face.edgeIDs.length-1)%face.edgeIDs.length]);
					if(previousEdge.vertexID[1]==edge.vertexID[0] || previousEdge.vertexID[0]==edge.vertexID[0]) {
						// add new edges to loop in stored order
						for(int k : newEdges) {
							newIds.add(k);
						}
					} else {
						// reverse order of new edges in edge loop
						for(int k=newEdges.size()-1;k>=0;k--) {
							newIds.add(newEdges.get(k));
						}
					}
				}
			}
			mesh.addFace(newIds);
		}

		int[] faceIDs=Arrays.copyOf(edge.faceIDs,edge.faceIDs.length);
		Arrays.sort(faceIDs);
		for(int i=faceIDs.length-1;i>=0;i--) {
			mesh.removeFace(edge.faceIDs[i]);
		}
	}

	@Override
	public void update() throws Exception {
		SceneGraph graph=getSlot("input").getGraph();
		Mesh mesh=(Mesh) graph.getNodeByName(meshName.getStringValue());
		
		int[] sortedEdges=new int[edges.getSize()];
		for(int i=0;i<edges.getSize();i++) {
			sortedEdges[i]=edges.getArrayIntegerValue(i);
		}
		Arrays.sort(sortedEdges);
		
		for(int i=sortedEdges.length-1;i>=0;i--) {
			int edgeID=sortedEdges[i];
			Edge edge=(Edge) mesh.getEdge(edgeID).clone();
			
			// generate new vertices
			ArrayList<Integer> newVertices=new ArrayList<Integer>();
			Vector3D startPosition=mesh.getVertexPosition(edge.vertexID[0]);
			Vector3D endPosition=mesh.getVertexPosition(edge.vertexID[1]);
			for(int j=0;j<subdivisions.getIntegerValue();j++) {
				Vector3D position;
				
				position=startPosition.multiply(subdivisions.getIntegerValue()-j).add(endPosition.multiply(j+1)).multiply(1.0/(subdivisions.getIntegerValue()+1));
				newVertices.add(mesh.addVertex(position));
			}

			// generate new edges
			ArrayList<Integer> newEdges=new ArrayList<Integer>();
			newEdges.add(mesh.addEdge(edge.vertexID[0],newVertices.get(0)));
			for(int j=1;j<subdivisions.getIntegerValue();j++) {
				newEdges.add(mesh.addEdge(newVertices.get(j-1),newVertices.get(j)));
			}
			newEdges.add(mesh.addEdge(newVertices.get(newVertices.size()-1),edge.vertexID[1]));

			// rebuild adjacent faces
			rebuildFaces(mesh, edgeID, edge, newEdges);
			
			mesh.removeEdge(edgeID);
		}
		
		getSlot("output").setGraph(graph);
	}

	@Override
	public void removeDependencies() {
	}

}
