/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Oct 26, 2005
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;


/**
 * An operator graph node that extrudes a set of faces in a mesh.
 *
 * @author gregor
 */
public class ExtrudeFaces extends Node {
	private Property name;
	private Property faces;
	private Property displacement;

	private ArrayList<Integer> borderEdges;
	private ArrayList<Integer> bridgeEdges;
	private ArrayList<Integer> innerEdges;
	private ArrayList<Integer> borderVertices;
	private ArrayList<Integer> innerVertices;
	private ArrayList<Integer> extrudedBorderVertices;
	private ArrayList<Integer> borderFaces;

	private HashMap<Integer, Integer> facesMap;

	public ExtrudeFaces()
	{
		super();
		createSlot("input",Slot.Type.Input);
		createSlot("output",Slot.Type.Output);

		name=new Property(this);
		name.setName("name");
		name.setType(Property.Type.String);
		try {
			name.setValue("none");
		} catch (Exception e) {
			// cannot happen
			e.printStackTrace();
		}
		faces=new Property(this);
		faces.setName("faces");
		faces.setType(Property.Type.IntegerArray);
		try {
			faces.setMaxSize(0);
		} catch (Exception e) {
			// cannot happen
			e.printStackTrace();
		}
		displacement=new Property(this);
		displacement.setName("displacement");
		displacement.setType(Property.Type.Vector3D);
		try {
			displacement.setValue(new Vector3D());
		} catch (Exception e) {
			// cannot happen
			e.printStackTrace();
		}

		borderEdges=new ArrayList<Integer>();
		bridgeEdges=new ArrayList<Integer>();
		innerEdges=new ArrayList<Integer>();
		borderVertices=new ArrayList<Integer>();
		innerVertices=new ArrayList<Integer>();
		extrudedBorderVertices=new ArrayList<Integer>();
		borderFaces=new ArrayList<Integer>();
	}

	/**
	 * Classifies the list of edges belonging to the given faces into inner edges
	 * and border edges. Border edges are edges that belong to only one face that
	 * is to be extruded. Inner edges are edges that are shared by two or more faces
	 * that are going to be extruded.
	 *
	 * @param mesh the mesh to work on
	 */
	private void classifyEdges(Mesh mesh) {
		for(int i=0;i<faces.getSize();i++) {
			Mesh.Face face=null;
			try {
				face = mesh.getFace(faces.getArrayIntegerValue(i));
			} catch (Exception e) {
				// this cannot happen - i cannot get out of range
				e.printStackTrace();
			}

			for(int j=0;j<face.edgeIDs.length;j++) {
				int edgeID=face.edgeIDs[j];

				if(borderEdges.contains(edgeID)) {
					borderEdges.remove(borderEdges.indexOf(edgeID));
					innerEdges.add(edgeID);
				} else {
					borderEdges.add(edgeID);
				}
			}
		}
	}

	/**
	 * Classifies the list of vertices belonging to the given faces into inner
	 * vertices and border vertices. Border vertices are vertices that are referenced
	 * by at least one border edge. Inner vertices only belong to inner edges.
	 *
	 * @param mesh the mesh to work on
	 */
	private void classifyVertices(Mesh mesh) {
		for(Integer edgeID : borderEdges) {
			Mesh.Edge edge=mesh.getEdge(edgeID);

			for(int i=0;i<2;i++) {
				if(!borderVertices.contains(edge.vertexID[i])) {
					borderVertices.add(edge.vertexID[i]);
				}
			}
		}

		for(Integer edgeID : innerEdges) {
			Mesh.Edge edge=mesh.getEdge(edgeID);

			for(int i=0;i<2;i++) {
				if(!(borderVertices.contains(edge.vertexID[i]) || innerVertices.contains(edge.vertexID[i]))) {
					innerVertices.add(edge.vertexID[i]);
				}
			}
		}
	}

	/**
	 * Further classification of inner edges into bridge edges and real inner edges.
	 * Bridge edges are edges between two border vertices that are not border edges.
	 * Note that this function replaces innerEdges with a new ArrayList.
	 *
	 * @param mesh the mesh to work on
	 */
	private void classifyInnerEdges(Mesh mesh) {
		ArrayList<Integer> newInnerEdges=new ArrayList<Integer>();

		for(Integer edgeID : innerEdges) {
			Mesh.Edge edge=mesh.getEdge(edgeID);

			if(borderVertices.contains(edge.vertexID[0]) && borderVertices.contains(edge.vertexID[1])) {
				bridgeEdges.add(edgeID);
			} else {
				newInnerEdges.add(edgeID);
			}
		}

		innerEdges=newInnerEdges;
	}

	/**
	 * Classifies the list of faces into inner faces and border faces. Border
	 * faces have at least one border edge or bridge edge while inner faces
	 * only consist of inner edges.
	 *
	 * @param mesh the mesh to work on
	 */
	private void classifyFaces(Mesh mesh) {
		for(int i=0;i<faces.getSize();i++) {
			int faceID=-1;
			boolean isBorderFace=false;
			Mesh.Face face;

			try {
				faceID = faces.getArrayIntegerValue(i);
			} catch (Exception e) {
				// this cannot happen - i cannot get out of range
				e.printStackTrace();
			}
			face = mesh.getFace(faceID);

			for(int j=0;j<face.edgeIDs.length;j++) {
				int edgeID=face.edgeIDs[j];
				if(borderEdges.contains(edgeID) || bridgeEdges.contains(edgeID)) {
					isBorderFace=true;
					break;
				}
			}

			if(isBorderFace) {
				borderFaces.add(faceID);
			}
		}
	}

	/**
	 * Extrude inner vertices. This just moves the inner vertices by displacment.
	 * No other actions are needed for inner vertices
	 *
	 * @param mesh the mesh to work on
	 */
	private void extrudeInnerVertices(Mesh mesh) {
		Vector3D displacement=this.displacement.getVectorValue();
		for(Integer vertexID : innerVertices) {
			Vector3D position=mesh.getVertexPosition(vertexID);
			Vector3D.add(position,displacement);
			mesh.setVertexPosition(vertexID, position);
		}
	}

	/**
	 * Extrude border Vertices by creating new vertices at the displaced position
	 * and generating edges between old and new vertices.
	 *
	 * @param mesh the mesh to work on
	 */
	private void extrudeBorderVertices(Mesh mesh) {
		for(Integer vertexID : borderVertices) {
			int newVertexID=-1;
			try {
				newVertexID = mesh.addVertex(mesh.getVertexPosition(vertexID).add(displacement.getVectorValue()));
			} catch (Exception e) {
				// this cannot happen as displacement is always of type Vector3D
				e.printStackTrace();
			}
			extrudedBorderVertices.add(newVertexID);
			try {
				mesh.addEdge(vertexID,newVertexID);
			} catch (Exception e) {
				// this cannot happen as both vertex IDs are guaranteed to be valid
				e.printStackTrace();
			}
		}
	}

	/**
	 * Create a quad from the given four vertex IDs.
	 *
	 * @param mesh the mesh to work on
	 * @param vertices the list of vertices that form the corners of the quad
	 * @return the face ID of the quad
	 */
	private int createQuad(Mesh mesh, int vertices[]) {
		int faceID=1;
		ArrayList<Integer> edges=new ArrayList<Integer>();

		for(int i=0;i<4;i++) {
				edges.add(mesh.findOrCreateEdge(vertices[i],vertices[(i+1)%4]));
		}

		try {
			faceID=mesh.addFace(edges);
		} catch (Exception e) {
			// this cannot happen as the edge loop is valid
			e.printStackTrace();
		}

		return faceID;
	}

	/**
	 * Extrude a border edge by creating a new edge between the corresponding extruded
	 * vertices and create a quad between the border edge and the new edge to close off
	 * the extruded area.
	 *
	 * @param mesh the mesh to work on
	 */
	private void extrudeBorderEdges(Mesh mesh) {
		int[] newVertices=new int[2];
		int[] allVertices=new int[4];
		for(Integer edgeID : borderEdges) {
			Mesh.Edge edge=mesh.getEdge(edgeID);
			for(int i=0;i<2;i++) {
				newVertices[i]=extrudedBorderVertices.get(borderVertices.indexOf(edge.vertexID[i]));
			}
			try {
				mesh.addEdge(newVertices[0],newVertices[1]);
			} catch (Exception e) {
				// this cannot happen here as the vertex ids are legal and there is no edge between them yet
				e.printStackTrace();
			}
			allVertices[0]=newVertices[0];
			allVertices[1]=newVertices[1];
			allVertices[2]=edge.vertexID[1];
			allVertices[3]=edge.vertexID[0];
			createQuad(mesh,allVertices);
		}
	}

	/**
	 * Extrude border faces by generating faces that use the extruded border vertices instead
	 * of the old vertices.
	 *
	 * @param mesh the mesh to work on
	 */
	private void extrudeBorderFaces(Mesh mesh) {
		facesMap=new HashMap<Integer, Integer>();
		for(Integer faceID : borderFaces) {
			Mesh.Face face=mesh.getFace(faceID);
			ArrayList<Integer> vertexIDs=mesh.getVertexIDs(face);
			ArrayList<Integer> newVertexIDs=new ArrayList<Integer>();
			ArrayList<Integer> edges=new ArrayList<Integer>();

			for(Integer vertexID : vertexIDs) {
				if(borderVertices.contains(vertexID)) {
					newVertexIDs.add(extrudedBorderVertices.get(borderVertices.indexOf(vertexID)));
				} else {
					newVertexIDs.add(vertexID);
				}
			}

			for(int i=0;i<newVertexIDs.size();i++) {
				edges.add(mesh.findOrCreateEdge(newVertexIDs.get(i),newVertexIDs.get((i+1)%newVertexIDs.size())));
			}

			try {
				int newFaceID=mesh.addFace(edges);
				facesMap.put(faceID,newFaceID);
			} catch (Exception e) {
				// this cannot happen as the edge loop is guaranteed to be valid
				e.printStackTrace();
			}
		}
	}

	/**
	 * Remove obsolete edges. Edges are obsoleted when the face they belong to
	 * has been replaced. This happens to edges in border faces that have only
	 * one border vertex and to all bridge edges.
	 *
	 * @param mesh the mesh to work on
	 */
	private void removeObsoleteEdges(Mesh mesh) {
		ArrayList<Integer> obsoleteEdges=(ArrayList<Integer>) bridgeEdges.clone();

		for(Integer faceID : borderFaces) {
			Mesh.Face face = mesh.getFace(faceID);

			for(int i=0;i<face.edgeIDs.length;i++) {
				Mesh.Edge edge=mesh.getEdge(face.edgeIDs[i]);

				int obsoleteVertexCount=0;
				for(int j=0;j<2;j++) {
					if(borderVertices.contains(edge.vertexID[j])) {
						obsoleteVertexCount++;
					}
				}
				if(obsoleteVertexCount==1) {
					if(!obsoleteEdges.contains(edge.edgeID)) {
						obsoleteEdges.add(edge.edgeID);
					}
				}
			}
		}

		Collections.sort(obsoleteEdges,Collections.reverseOrder());
		for(Integer edgeID : obsoleteEdges) {
			try {
				mesh.removeEdge(edgeID);
			} catch (Exception e) {
				// this cannot happen as edge IDs are guaranteed to be valid
				e.printStackTrace();
			}
		}
	}

	@Override
	public void update() throws Exception {
		SceneGraph graph=getSlot("input").getGraph();
		eu.moonlight3d.ml3d.backend.threedim.sg.Node node=graph.getNodeByName(name.getStringValue());

		if(node==null) {
			throw new Exception(L10N.translate("given node does not exist"));
		}

		if(!(node instanceof Mesh)) {
			throw new Exception(L10N.translate("given object is not a mesh"));
		}

		eu.moonlight3d.ml3d.modelling.mesh.Mesh mesh=(Mesh)node;

		borderEdges.clear();
		bridgeEdges.clear();
		innerEdges.clear();
		borderVertices.clear();
		innerVertices.clear();
		extrudedBorderVertices.clear();
		borderFaces.clear();

		classifyEdges(mesh);
		classifyVertices(mesh);
		classifyInnerEdges(mesh);
		classifyFaces(mesh);
		extrudeInnerVertices(mesh);
		extrudeBorderVertices(mesh);
		extrudeBorderEdges(mesh);
		extrudeBorderFaces(mesh);
		removeObsoleteEdges(mesh);

		getSlot("output").setGraph(graph);
	}

	public HashMap<Integer, Integer> getFacesMap() {
		return (HashMap<Integer, Integer>) facesMap.clone();
	}

	@Override
	public void removeDependencies() {
	}
}
