/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 21, 2007
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.List;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.selection.SelectionListener;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.modelling.basic.GridGizmoSnapper;
import eu.moonlight3d.ml3d.ui.mainwindow.modes.Mode;
import eu.moonlight3d.ml3d.ui.mainwindow.modes.ModeManager;
import eu.moonlight3d.ml3d.ui.mainwindow.view.ViewEventHandler;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.Gizmo;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.WidgetViewEventHandler;
import eu.moonlight3d.ml3d.ui.mainwindow.widgets.Gizmo.DragMode;

public class ExtrudeMode extends Mode {
	/**
	 * The gizmo widget instance.
	 */	
	private Gizmo gizmo;
	/**
	 * Reference to selection listener that deactives this mode.
	 */
	private ExtrudeSelectionListener extrudeSelectionListener;
	/**
	 * Indicates whether the selection listener should be active right now.
	 */
	private boolean selectionListenerActive;
	/**
	 * True if snapping to grid is enabled.
	 */
	private boolean snapToGridEnabled;
	/**
	 * Instance of grid gizmo snapper used when snapping is enabled.
	 */
	private GridGizmoSnapper gridSnapper;
	
	private class ExtrudeSelectionListener implements SelectionListener {

		public void elementAdded(String objectName) {
			forceDeactivation();
		}

		public void elementRemoved(String objectName) {
			forceDeactivation();
		}

		public void elementUpdated(String oldName, String newName) {
			forceDeactivation();
		}

		public void elementsAdded(List<String> objectNames) {
			forceDeactivation();
		}

		public void selectionCleared() {
			forceDeactivation();
		}
		
	}
	
	private class ExtrudeGizmoListenerWrapper extends ExtrudeGizmoListener {
		private ExtrudeGizmoListener wrappedListener;
		
		public ExtrudeGizmoListenerWrapper(ExtrudeGizmoListener wrapped) {
			wrappedListener=wrapped;
		}

		@Override
		public Vector3D getInitialPosition() throws Exception {
			return wrappedListener.getInitialPosition();
		}

		public void apply() {
			selectionListenerActive=false;
			wrappedListener.apply();
			selectionListenerActive=true;
		}

		public void begin(DragMode mode) {
			selectionListenerActive=false;
			wrappedListener.begin(mode);
			selectionListenerActive=true;
		}

		public void end() {
			selectionListenerActive=false;
			wrappedListener.end();
			selectionListenerActive=true;
		}

		public void rotate(Vector3D center, Quaternion rotation) {
			selectionListenerActive=false;
			wrappedListener.rotate(center, rotation);
			selectionListenerActive=true;
		}

		public void scale(Vector3D center, Vector3D amount) {
			selectionListenerActive=false;
			wrappedListener.scale(center, amount);
			selectionListenerActive=true;
		}

		public void translate(Vector3D displacment) {
			selectionListenerActive=false;
			wrappedListener.translate(displacment);
			selectionListenerActive=true;
		}
		
	}
	
	/**
	 * Default constructor
	 */
	public ExtrudeMode() {
		super("Extrude");
		snapToGridEnabled=false;
	}

	/**
	 * Enable or disable snapping to grid.
	 * 
	 * @param snapToGridEnabled true when snapping should be enabled
	 */
	public void setSnapToGridEnabled(boolean snapToGridEnabled) {
		this.snapToGridEnabled=snapToGridEnabled;
		if(gizmo!=null) {
			if(snapToGridEnabled) {
				gridSnapper=new GridGizmoSnapper();
				gizmo.addSnapper(gridSnapper);
			} else {
				gizmo.removeSnapper(gridSnapper);
				gridSnapper=null;
			}
		}
	}

	@Override
	public void activated() {
        Selection selection;
        ExtrudeGizmoListener listener=null;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			if(selection.getType()==null) {
				eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("no current selection"));
				return;
			}
			String selectionType=selection.getType().getName();
			
			if(selectionType.equals("MeshVertex")) {
				listener=new ExtrudeVerticesGizmoListener(); 
			} else if(selectionType.equals("MeshEdge")) {
				listener=new ExtrudeEdgesGizmoListener();
			} else if(selectionType.equals("MeshFace")) {
				listener=new ExtrudeFacesGizmoListener();
			} else {
				eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("the current selection is not extrudable"));
				return;				
			}
			
		} catch (Exception e) {
			ExceptionDialog.run(e);
			return;
		}
		
		if(selection.getSelection().size()==0) {
            eu.moonlight3d.framework.State.getInstance().logger().warning(L10N.translate("a whole object must be active to apply Transform Object to it"));
            return;			
		}
		
		extrudeSelectionListener=new ExtrudeSelectionListener();
		selection.registerListener(extrudeSelectionListener);
		
        WidgetViewEventHandler widgetEventHandler=new WidgetViewEventHandler();
		gizmo=new Gizmo(widgetEventHandler);
		gizmo.setTranslateEnabled(true);
		gizmo.setRotateEnabled(false);
		gizmo.setScaleEnabled(false);
		
		if(snapToGridEnabled) {
			gridSnapper=new GridGizmoSnapper();
			gizmo.addSnapper(gridSnapper);
		}

		// set gizmo location and transformation matrix
		try {
			gizmo.setPosition(listener.getInitialPosition());
			gizmo.setLocalToWorldTransformation(listener.getLocalToWorldTransformation());
		} catch (Exception e) {
			ExceptionDialog.run(e);
			return;
		}

		// set gizmo location
		try {
			gizmo.setPosition(listener.getInitialPosition());			
		} catch (Exception e) {
			ExceptionDialog.run(e);
			return;
		}
		
		gizmo.addListener(listener);

		ViewEventHandler.setEventHandler(widgetEventHandler);		
	}

	@Override
	public void deactivated() {
		ViewEventHandler.setEventHandler(ViewEventHandler.getDefaultEventHandler());
        
		Selection selection;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			selection.unregisterListener(extrudeSelectionListener);
			extrudeSelectionListener=null;
		} catch(Exception e) {
			// TODO proper error handling
			e.printStackTrace();
		}

		gizmo=null;
		gridSnapper=null;
	}

	
	/**
	 * Force deactivation of the mode after some significant change in the 
	 * scene, e.g. a selection change
	 */
	private void forceDeactivation() {
		if(selectionListenerActive) {
			ModeManager.getInstance().setMode((Mode)null);
		}
	}

}
