/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 20, 2007
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;

import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.OperatorGraphHelper;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh.Edge;
import eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool;

/**
 * A modeless tool that inserts a new face between selected edges of a mesh.
 * 
 * @see eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool
 *
 * @author gregor
 */
public class MakeFaceTool extends Tool {
	/**
	 * Default constructor.
	 */
	public MakeFaceTool() {
		super();
		setName("MakeFaceTool");
		setDisplayName(L10N.translate("Make Face"));
		setCategory("Edit/Mesh");
	}

	@Override
	public void activated() {
        try {
            State state=eu.moonlight3d.framework.State.getInstance();
            
            // don't do anything if selected object is not a SG node
            Selection selection = ((Document) state.getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
            if(selection.getType()!=eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("MeshEdge")) {
                state.logger().warning(L10N.translate("current selection must be at least three mesh edges"));
                return;
            }
            if(selection.getActiveObject()==null) {
            	State.getInstance().logger().info(L10N.translate("no active object to work on"));
            	return;
            }
            
            String selectionName=selection.getActiveObject();
            String ogNodeName=SelectionHelper.getOGNodeName(selectionName);
            String sgNodeName=SelectionHelper.getSGNodeName(selectionName);
            

            if(selection.getSelection().size() <= 2) {
                state.logger().warning(L10N.translate("make face aborted: at least 3 edges need to be selected"));
                return;
            }

            // sort list of edges into an edge loop
            ArrayList<Object> selectedObjects=selection.getSelectionObjects();
            ArrayList<Mesh.Edge> sortedEdges=new ArrayList<Mesh.Edge>();
            sortedEdges.add((Edge) selectedObjects.get(0));
            selectedObjects.remove(0);
            do {
            	boolean foundNextEdge=false;
            	Edge lastEdge=sortedEdges.get(sortedEdges.size()-1);
            	for(int i=0;i<selectedObjects.size();i++) {
            		Edge currentEdge=(Edge) selectedObjects.get(i);
            		if(lastEdge.vertexID[0]==currentEdge.vertexID[0] || lastEdge.vertexID[0]==currentEdge.vertexID[1] ||
            		   lastEdge.vertexID[1]==currentEdge.vertexID[0] || lastEdge.vertexID[1]==currentEdge.vertexID[1]) {
            			sortedEdges.add(currentEdge);
            			selectedObjects.remove(i);
            			lastEdge=currentEdge;
            			foundNextEdge=true;
            			break;
            		}
            	}
            	if(!foundNextEdge) {
            		state.logger().warning(L10N.translate("make face aborted: cannot make a proper loop out of given edges"));
            		return;
            	}
            } while(selectedObjects.size()>0);
            
            // check if first and last edge match
            Edge firstEdge=sortedEdges.get(0);
            Edge lastEdge=sortedEdges.get(sortedEdges.size()-1);
    		if(firstEdge.vertexID[0]!=lastEdge.vertexID[0] && firstEdge.vertexID[0]!=lastEdge.vertexID[1] &&
         	   firstEdge.vertexID[1]!=lastEdge.vertexID[0] && firstEdge.vertexID[1]!=lastEdge.vertexID[1]) {
    			state.logger().warning(L10N.translate("make face aborted: cannot make a proper loop out of given edges"));
    			return;
    		}
            
            Manager ogManager = ((Document) state.getDocumentManager().getMainDocument()).getOGManager();
            Node ogNode = OperatorGraphHelper.insert(selectionName, "MakeFace", "input", "output");
            ogNode.getProperty("name").setValue(sgNodeName);
            
            for(int i=0;i<sortedEdges.size();i++) {
                ogNode.getProperty("edges").setArrayValue(i,sortedEdges.get(i).edgeID);
            }

			ogManager.update();
	        selection.updateBase(ogNodeName,ogNode.getName());
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}

	@Override
	public void deactivated() {
	}

	@Override
	public QWidget getOptionEditor(QWidget parent) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public boolean hasOptionEditor() {
		// TODO Auto-generated method stub
		return false;
	}

}
