/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Apr 25, 2008
 */

package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.List;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.StringTools;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * Merge a list of mesh objects given by name into the first one
 * in the list.
 *
 * @author gregor
 */
public class MergeMeshes extends Node {
	/**
	 * A list of names of all the mesh objects to merge
	 */
	private Property meshes;
	
	/**
	 * Default constructor.
	 */
	public MergeMeshes() {
		createSlot("input", Slot.Type.Input);
		createSlot("output", Slot.Type.Output);
		
		meshes=new Property(this);
		meshes.setName("meshes");
		meshes.setType(Value.Type.StringArray);
		meshes.setMaxSize(0);
	}
	
	/**
	 * Get the name of the scene graph node referenced by the given path.
	 *  
	 * @param path scene graph path
	 * @return scene graph node name
	 */
	private String getNodeName(String path) {
		String[] pathElements=StringTools.split(path, '/');
		return pathElements[pathElements.length-1];
	}
	
	/**
	 * Get the name of the parent of the scene graph node referenced by
	 * the given path.
	 * 
	 * @param path scene  graph path
	 * @return parent name or null if parent does not exist
	 */
	private String getParent(String path) {
		String[] pathElements=StringTools.split(path, '/');
		if(pathElements.length==1) {
			return null;
		}
		return pathElements[pathElements.length-2];		
	}
	
	@Override
	public void update() throws Exception {
		SceneGraph graph=getSlot("input").getGraph();
		
		if(meshes.getSize()==0) {
			throw new Exception(L10N.translate("MergeMeshes must have at least one mesh in meshes list"));
		}
		
		Mesh firstMesh=(Mesh) graph.getNodeByName(getNodeName(meshes.getArrayStringValue(0))).clone();
		
		for(int i=1;i<meshes.getSize();i++) {
			Mesh secondMesh=(Mesh) graph.getNodeByName(getNodeName(meshes.getArrayStringValue(i)));
			
			int vertexIDOffset=firstMesh.getNumVertices();
			
			Matrix4 transformation=firstMesh.getWorldToLocalTransformation(meshes.getArrayStringValue(0));
			transformation=transformation.multiply(secondMesh.getLocalToWorldTransformation(meshes.getArrayStringValue(i)));
			for(int j=0;j<secondMesh.getNumVertices();j++) {
				firstMesh.addVertex(transformation.multiply(secondMesh.getVertexPosition(j)));
			}
			
			int edgeIDOffset=firstMesh.getNumEdges();

			for(int j=0;j<secondMesh.getNumEdges();j++) {
				int[] edgeVertices=secondMesh.getEdgeVertices(j);
				firstMesh.addEdge(edgeVertices[0]+vertexIDOffset, edgeVertices[1]+vertexIDOffset);
			}
			
			for(int j=0;j<secondMesh.getNumFaces();j++) {
				int faceEdges[]=secondMesh.getFaceEdges(j);
				ArrayList<Integer> list=new ArrayList<Integer>();
				for(int k=0;k<faceEdges.length;k++) {
					list.add(faceEdges[k]+edgeIDOffset);
				}
				firstMesh.addFace(list);
			}
			
			
			String parentName=getParent(meshes.getArrayStringValue(i));
			if(parentName!=null) {
				eu.moonlight3d.ml3d.backend.threedim.sg.Node parent=graph.getNodeByName(parentName);
				graph.removeChild(parent, secondMesh);
				// TODO parent removal should be recursive
				if(parent instanceof Instance) {
					if(parent.getChildren().size()==0) {
						for(eu.moonlight3d.ml3d.backend.threedim.sg.Node parentParent : parent.getParents()) {
							graph.removeChild(parentParent,parent);
						}
						graph.removeNode(parent);
					}
				}
			}
		}

		// replace original copy of firstMesh with merged one and remove merged meshes
		Mesh originalFirst=(Mesh) graph.getNodeByName(getNodeName(meshes.getArrayStringValue(0)));
		List<eu.moonlight3d.ml3d.backend.threedim.sg.Node> children=originalFirst.getChildren();
		List<eu.moonlight3d.ml3d.backend.threedim.sg.Node> parents=originalFirst.getParents();
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node child : children) {
			graph.removeChild(originalFirst, child);
		}
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node parent : parents) {
			graph.removeChild(parent, originalFirst);
		}
		graph.removeNode(originalFirst);
		for(int i=1;i<meshes.getSize();i++) {
			Mesh secondMesh=(Mesh) graph.getNodeByName(getNodeName(meshes.getArrayStringValue(i)));
			// merged meshes may already be gone because of instancing
			if(secondMesh!=null) {
				graph.removeNode(secondMesh);
			}
		}
		graph.addNode(firstMesh);
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node child : children) {
			graph.removeChild(firstMesh, child);
			graph.addChild(firstMesh, child);
		}
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node parent : parents) {
			graph.removeChild(parent, firstMesh);
			graph.addChild(parent, firstMesh);
		}
		
		getSlot("output").setGraph(graph);
	}

	@Override
	public void removeDependencies() {
		// nothing to be done here
	}

}
