/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Apr 25, 2008
 */

package eu.moonlight3d.ml3d.modelling.mesh;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.modelling.basic.Merge;
import eu.moonlight3d.ml3d.modelling.basic.OGOutputMerger;
import eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool;

/**
 *
 * Tool to merge two or meshes into one single mesh object
 * containg all elements of all meshes.
 *
 * @author gregor
 */
public class MergeMeshesTool extends Tool {

	/**
	 * Default constructor.
	 */
	public MergeMeshesTool() {
		super();
		setName("MergeMeshesTool");
		setDisplayName(L10N.translate("Merge Meshes"));
		setCategory("Edit/Mesh");
	}
	
	@Override
	public void activated() {
        try {
            State state=State.getInstance();
            
            // don't do anything if selected object is not a SG node
            Selection selection = ((Document) state.getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
            if(selection.getType()!=eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("SGNode")) {
                state.logger().warning(L10N.translate("current selection must be a scene graph node"));
                return;
            }
            
            if(selection.getActiveObject()==null) {
            	state.logger().warning(L10N.translate("no active object"));
            	return;
            }
            
            String path=selection.getActiveObject();
            String ogNodeName=SelectionHelper.getOGNodeName(path);
            String slotName=SelectionHelper.getOGSlotName(path);
            
            // create OGOutputMerger helper object
            OGOutputMerger ogOutputMerger=new OGOutputMerger();
            
            List<String> selectedObjects = new ArrayList<String>();
            
            // remove anything from current selection that is not a mesh:
            int numMeshes=0;
            for(String name : selection.getSelection()) {
            	Object selectedObject=selection.getObjectFromSelectionString(name);
            	if(!(selectedObject instanceof Mesh)) {
            		state.logger().info(MessageFormat.format(L10N.translate("selected object {0} is not a mesh and will not be merged"),name));
            		continue;
            	}
            	selectedObjects.add(name);
            	numMeshes++;
            }
            if(numMeshes<2) {
            	state.logger().warning(L10N.translate("at least 2 meshes must be selected for merging"));
            	return;
            }
            
            List<String> ogNodes=ogOutputMerger.getOGNodeNames(selectedObjects);
            
            Manager ogManager=((Document) state.getDocumentManager().getMainDocument()).getOGManager();
            if(ogNodes.size()>1) {
            	// merge output slots
            	ogNodeName=ogOutputMerger.mergeOGOutputSlots(ogNodes);
            	            	
            	slotName="output";
            	
            	// update selection list with renamed objects from merge node output
            	selectedObjects=ogOutputMerger.getMergedSGNodeNames(selectedObjects);

            	// run OG update to get the name mappings
            	ogManager.update();
            }
            
            // insert MergeMeshes OG node
            Node ogNode=ogManager.getNodeByName(ogNodeName);
            Slot slot=ogNode.getSlot(slotName);
            ogNode=ogManager.createNodeByName("MergeMeshes");
            
            int i=0;
            for(String name : selectedObjects) {
            	String nodePath=SelectionHelper.getSGPath(name);
                ogNode.getProperty("meshes").setArrayValue(i,nodePath);
                i++;
            }
            
            // set links
            ogManager.insertNodeAfter(slot,ogNode.getSlot("input"),ogNode.getSlot("output"));

            // force update of operator graph
			ogManager.update();

			// select output of MergeMeshes
            path=selection.getSelection().get(0);
            selection.clearSelection(eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("SGNode"));
            selection.add("OperatorGraph/" + ogNode.getName( )+ "/output/" + SelectionHelper.getSGPath(path));
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}		
	}

	@Override
	public void deactivated() {
		// nothing to be done here

	}

	@Override
	public QWidget getOptionEditor(QWidget parent) {
		// this tool has no options
		return null;
	}

	@Override
	public boolean hasOptionEditor() {
		// this tool has no options
		return false;
	}

}
