/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Jan 5, 2005
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.BooleanArrayBuffer;
import eu.moonlight3d.framework.helper.BooleanBuffer;
import eu.moonlight3d.framework.helper.ByteArrayBuffer;
import eu.moonlight3d.framework.helper.DoubleArrayBuffer;
import eu.moonlight3d.framework.helper.DoubleBuffer;
import eu.moonlight3d.framework.helper.IntArrayBuffer;
import eu.moonlight3d.framework.helper.IntBuffer;
import eu.moonlight3d.framework.helper.Vector3DArrayBuffer;
import eu.moonlight3d.framework.helper.Vector3DBuffer;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Line2D;
import eu.moonlight3d.math.Matrix3;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Plane;
import eu.moonlight3d.math.Ray;
import eu.moonlight3d.math.Vector2D;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.sg.GeometryNode;

/**
 * This class implements the mesh scene graph node type.
 *
 * @author gregor
 */
public class Mesh extends GeometryNode
{

	/**
	 * Buffer for vertex positions
	 */
	private Vector3DBuffer vertexPosition;
	/**
	 * Array buffer for vertex edges list. It contains a list of those edges for each vertex that
	 * have this vertex as an end point.
	 */
	private IntArrayBuffer vertexEdges;

	/**
	 * Buffer for edge vertices list. It contains the pair of corner vertex IDs for each edge.
	 */
	private IntBuffer edgeVertices;
	/**
	 * Array buffer for edge faces list. It contains a list of the faces for each edge that this
	 * edge is part of.
	 */
	private IntArrayBuffer edgeFaces;

	/**
	 * Array buffer for face edges list. It contains a list of the edges in the face loop for every
	 * face.
	 */
	private IntArrayBuffer faceEdges;

	/**
	 * Array buffer for face vertex UV coordinates. It contains a pair of UV coordinates for each
	 * vertex of each face.
	 */
	private HashMap<String, DoubleArrayBuffer> uvMaps;

	/**
	 * The locked mode for transforming face vertex normals which means that the specific normal
	 * vector will not be transformed automatically.
	 */
	public static final byte NORMAL_MODE_LOCKED = 0;
	/**
	 * The hard mode for transforming face vertex normals which means that the specific normal
	 * vector will always be set to the calculated normal of the face that this normal vector
	 * belongs to. This is used to create the impression of hard edges.
	 */
	public static final byte NORMAL_MODE_HARD = 1;
	/**
	 * The soft mode for transforming face vertex normals which means that the specific normal
	 * vector will always be set to the average of the calculated normals of the faces that the
	 * underlying vertex belongs to. This is used to create the impression of soft edges and smooth
	 * surfaces.
	 */
	public static final byte NORMAL_MODE_SMOOTH = 2;
	/**
	 * Array buffer for the modes of the face vertex normals. It contains the mode for the normal of
	 * each vertex of each face. The mode defines how normals are transformed when the underlying
	 * vertex is transformed.
	 */
	private ByteArrayBuffer normalMode;
	/**
	 * Array buffer for face vertex normals. It contains a normal vector for each vertex of each
	 * face. These normal vectors are updated depending on the modes that are set for them.
	 */
	private Vector3DArrayBuffer normals;
	/**
	 * A boolean indicating whether normals should be updated automatically on each individual
	 * change. Operations that modify a large number of vertices should disable this until they are
	 * done, activating a more efficient lazy evaluation behaviour.
	 */
	private boolean autoNormalsUpdate;

	private HashMap<String, IntBuffer> vertexIntData;
	private HashMap<String, BooleanBuffer> vertexBooleanData;
	private HashMap<String, DoubleBuffer> vertexDoubleData;
	private HashMap<String, Vector3DBuffer> vertexVector3DData;

	private HashMap<String, IntBuffer> edgeIntData;
	private HashMap<String, BooleanBuffer> edgeBooleanData;
	private HashMap<String, DoubleBuffer> edgeDoubleData;
	private HashMap<String, Vector3DBuffer> edgeVector3DData;

	private HashMap<String, IntBuffer> faceIntData;
	private HashMap<String, BooleanBuffer> faceBooleanData;
	private HashMap<String, DoubleBuffer> faceDoubleData;
	private HashMap<String, Vector3DBuffer> faceVector3DData;

	private HashMap<String, IntArrayBuffer> faceVertexIntData;
	private HashMap<String, BooleanArrayBuffer> faceVertexBooleanData;
	private HashMap<String, DoubleArrayBuffer> faceVertexDoubleData;
	private HashMap<String, Vector3DArrayBuffer> faceVertexVector3DData;

	/**
	 * Vertex data structure used in eu.moonlight3d.ml3d.backend.sg.Mesh
	 *
	 * @author gregor
	 */
	public static class Vertex implements Cloneable
	{
		/**
		 * Unique ID of this vertex
		 */
		public int vertexID;
		/**
		 * Position of this vertex in world coordinates
		 */
		public Vector3D position;
		/**
		 * The edges that this vertex is part of
		 */
		public int[] edgeIDs;

		@Override
		public Object clone()
		{
			Vertex v = null;
			try
			{
				v = (Vertex) super.clone();
			} catch (CloneNotSupportedException e)
			{
				// cannot happen
			}

			v.position = position.clone();
			if (edgeIDs != null)
			{
				v.edgeIDs = edgeIDs.clone();
			} else
			{
				v.edgeIDs = null;
			}

			return v;
		}
	}

	/**
	 * Edge data structure used to represent edges in eu.moonlight3d.ml3d.backend.sg.Mesh
	 *
	 * @author gregor
	 */
	public static class Edge implements Cloneable
	{
		/**
		 * Unique ID of this edge
		 */
		public int edgeID;
		/**
		 * Array containing the IDs of the two end points of this edge
		 */
		public int vertexID[];
		/**
		 * Array containing the IDs of the manifold edges that are bundled in this edge.
		 */
		public int[] faceIDs;

		/**
		 * Constructor creating an empty edge class. At least vertexID has to be filled in for this
		 * class to be useable.
		 */
		Edge()
		{
			vertexID = new int[2];
			faceIDs = null;
		}

		@Override
		public Object clone()
		{
			Edge e = null;
			try
			{
				e = (Edge) super.clone();
			} catch (CloneNotSupportedException e1)
			{
				// cannot happen
			}

			e.vertexID = new int[2];
			e.vertexID[0] = vertexID[0];
			e.vertexID[1] = vertexID[1];

			if (faceIDs != null)
			{
				e.faceIDs = faceIDs.clone();
			} else
			{
				e.faceIDs = null;
			}

			return e;
		}
	}

	/**
	 * Face data structure for eu.moonlight3d.ml3d.backend.sg.Mesh. This class defines a single face
	 * in the mesh by providing a starting point into an edge loop.
	 *
	 * @author gregor
	 */
	public static class Face implements Cloneable
	{
		/**
		 * Unique ID of this face
		 */
		public int faceID;
		/**
		 * Unique ID of start manifold edge of defining edge loop
		 */
		public int[] edgeIDs;
		/**
		 * List of triangle corner vertices for triangulated version of face.
		 */
		public ArrayList<Integer> triangulated;

		@Override
		public Object clone() throws CloneNotSupportedException
		{
			return super.clone();
		}
	}

	/**
	 * utility function returning the indices of the components of the given vector with the biggest
	 * absolute values
	 *
	 * @param vector the vector to find the biggest components in
	 * @return the indices of the two absolute biggest components
	 */
	private int[] getMaxComponents(Vector3D vector)
	{
		int[] indices = new int[2];
		// 1. find biggest component
		int maxComponent = 0;
		if (Math.abs(vector.X2) > Math.abs(vector.X1))
		{
			maxComponent = 1;
		}
		if (maxComponent == 1)
		{
			if (Math.abs(vector.X3) > Math.abs(vector.X2))
			{
				maxComponent = 2;
			}
		} else
		{
			if (Math.abs(vector.X3) > Math.abs(vector.X1))
			{
				maxComponent = 2;
			}
		}

		// 2. find indices of components to keep (admittedly, the formulation is a bit tricky :-)
		indices[0] = 0;
		indices[1] = 1;

		if (indices[0] == maxComponent)
		{
			indices[0]++;
			indices[1]++;
		} else if (indices[1] == maxComponent)
		{
			indices[1]++;
		}

		return indices;
	}

	/**
	 * Project a face given by an array of vertex IDs onto a coordinate plane in such a way that the
	 * resulting coordinates span the largest possible range.
	 *
	 * @param vertexIDs the vertex IDs of the corner vertices of the face
	 * @return the projected coordinates of the corner vertices
	 */
	private Vector2D[] projectFace(ArrayList<Integer> vertexIDs)
	{
		// 1. calculate plane equation
		Plane plane = new Plane();
		plane.fromPoints(getVertexPosition(vertexIDs.get(0)), getVertexPosition(vertexIDs.get(1)),
			getVertexPosition(vertexIDs.get(2)));

		// 2. find biggest component of plane normal
		int indices[] = getMaxComponents(plane.normal);

		// 3. kick that coordinate from all the face's vertices to project onto a coordinate plane
		Vector2D[] points = new Vector2D[vertexIDs.size()];

		for (int i = 0; i < vertexIDs.size(); i++)
		{
			Vector3D vertex = getVertexPosition(vertexIDs.get(i));
			points[i] = new Vector2D();
			if (indices[0] == 0)
			{
				points[i].X1 = vertex.X1;
			} else if (indices[0] == 1)
			{
				points[i].X1 = vertex.X2;
			} else
			{
				points[i].X1 = vertex.X3;
			}
			if (indices[1] == 0)
			{
				points[i].X2 = vertex.X1;
			} else if (indices[1] == 1)
			{
				points[i].X2 = vertex.X2;
			} else
			{
				points[i].X2 = vertex.X3;
			}
		}

		return points;
	}

	/**
	 * Determine whether the given point is inside the the projected face given by a two-dimensional
	 * point list as returned from projectFace().
	 *
	 * @param projectedFace a list of 2D coordinates of the corner points of the 2D face
	 * @param point the point to test
	 * @return true if the given point is inside the projected face
	 */
	private boolean isInsideProjectedFace(Vector2D[] projectedFace, Vector2D point)
	{
		boolean inside = false;

		for (int i = 0; i < projectedFace.length; i++)
		{
			if (projectedFace[i].X2 < point.X2
					&& projectedFace[(i + 1) % projectedFace.length].X2 < point.X2)
			{
				continue; // edge is entirely below the test line
			}
			if (projectedFace[i].X2 > point.X2
					&& projectedFace[(i + 1) % projectedFace.length].X2 > point.X2)
			{
				continue; // edge is entirely above the test line
			}
			// check, if intersection is on the left of the test line
			if (projectedFace[i].X1 + (point.X2 - projectedFace[i].X2)
					/ (projectedFace[(i + 1) % projectedFace.length].X2 - projectedFace[i].X2)
					* (projectedFace[(i + 1) % projectedFace.length].X1 - projectedFace[i].X1) < point.X1)
			{
				inside = !inside;
			}
		}

		return inside;
	}

	/**
	 * Constructor for new Meshes. Creates a new empty mesh.
	 */
	public Mesh()
	{
		super();

		vertexPosition = new Vector3DBuffer();
		vertexEdges = new IntArrayBuffer();

		edgeVertices = new IntBuffer();
		edgeFaces = new IntArrayBuffer();

		faceEdges = new IntArrayBuffer();

		uvMaps = new HashMap<String, DoubleArrayBuffer>();

		normals = new Vector3DArrayBuffer();
		normalMode = new ByteArrayBuffer();

		autoNormalsUpdate = true;
	}

	private void transformFaceVertexNormal(int faceID, int vertexIndex)
	{
		if (!autoNormalsUpdate)
		{
			return;
		}

		int normalMode = this.normalMode.getArrayElement(faceID, vertexIndex);
		ArrayList<Integer> vertexIDs;
		Vector3D side1;
		Vector3D side2;
		Vector3D normal;

		switch (normalMode)
		{
		case NORMAL_MODE_LOCKED:
			// do nothing here - normal direction is locked
			break;
		case NORMAL_MODE_HARD:
			// calculate and assign face normal
			vertexIDs = getVertexIDs(faceID);

			side1 = getVertexPosition(vertexIDs.get(1)).sub(getVertexPosition(vertexIDs.get(0)));
			side2 = getVertexPosition(vertexIDs.get(2)).sub(getVertexPosition(vertexIDs.get(0)));
			normal = side1.crossMultiply(side2).norm();

			normals.setArrayElement(faceID, vertexIndex, normal);
			break;
		case NORMAL_MODE_SMOOTH:
			// calculate average of adjacent face normals
			int vertexID = getVertexIDs(faceID).get(vertexIndex);
			ArrayList<Integer> faceIDs = new ArrayList<Integer>();
			for (int edgeID : getVertexEdges(vertexID))
			{
				for (int adjacentFaceID : getEdgeFaces(edgeID))
				{
					if (!faceIDs.contains(adjacentFaceID))
					{
						// only consider a face normal if the transition should be smooth on that
						// side of the vertex
						if (this.normalMode.getArrayElement(adjacentFaceID, getFaceVertexIndex(
							adjacentFaceID, vertexID)) == NORMAL_MODE_SMOOTH)
						{
							faceIDs.add(adjacentFaceID);
						}
					}
				}
			}

			Vector3D averagedNormal = new Vector3D(0, 0, 0);
			for (int adjacentFaceID : faceIDs)
			{
				vertexIDs = getVertexIDs(getFace(adjacentFaceID));
				side1 = getVertexPosition(vertexIDs.get(1))
						.sub(getVertexPosition(vertexIDs.get(0)));
				side2 = getVertexPosition(vertexIDs.get(2))
						.sub(getVertexPosition(vertexIDs.get(0)));
				normal = side1.crossMultiply(side2).norm();
				Vector3D.add(averagedNormal, normal);
			}
			Vector3D.multiply(averagedNormal, 1.0 / faceIDs.size());

			normals.setArrayElement(faceID, vertexIndex, averagedNormal);
			break;
		}
	}

	private void transformFaceVertexNormals(int vertexID)
	{
		if (!autoNormalsUpdate)
		{
			return;
		}

		ArrayList<Integer> faceIDs = new ArrayList<Integer>();
		for (int edgeID : getVertexEdges(vertexID))
		{
			for (int adjacentFaceID : getEdgeFaces(edgeID))
			{
				if (!faceIDs.contains(adjacentFaceID))
				{
					faceIDs.add(adjacentFaceID);
				}
			}
		}

		for (int adjacentFaceID : faceIDs)
		{
			for (int i = 0; i < getVertexIDs(getFace(adjacentFaceID)).size(); i++)
			{
				transformFaceVertexNormal(adjacentFaceID, i);
			}
		}
	}

	/**
	 * Transform, that is, update all face vertex normals in the mesh unless they are locked.
	 */
	private void transformFaceVertexNormals()
	{
		if (!autoNormalsUpdate)
		{
			return;
		}

		Vector3D side1;
		Vector3D side2;
		Vector3D normal = new Vector3D();
		Vector3DBuffer faceNormals = new Vector3DBuffer();
		faceNormals.setReserve(faceEdges.getNumArrays());
		for (int i = 0; i < getNumFaces(); i++)
		{
			ArrayList<Integer> vertexIDs = getVertexIDs(i);

			side1 = getVertexPosition(vertexIDs.get(1)).sub(getVertexPosition(vertexIDs.get(0)));
			side2 = getVertexPosition(vertexIDs.get(2)).sub(getVertexPosition(vertexIDs.get(0)));
			normal = side1.crossMultiply(side2).norm();

			faceNormals.append(normal);
		}

		for (int i = 0; i < getNumFaces(); i++)
		{
			ArrayList<Integer> vertexIDs = getVertexIDs(i);
			for (int j = 0; j < vertexIDs.size(); j++)
			{
				if (normalMode.getArrayElement(i, j) == NORMAL_MODE_SMOOTH)
				{
					int vertexID = vertexIDs.get(j);
					ArrayList<Integer> faceIDs = new ArrayList<Integer>();
					for (int edgeID : getVertexEdges(vertexID))
					{
						for (int adjacentFaceID : getEdgeFaces(edgeID))
						{
							if (!faceIDs.contains(adjacentFaceID))
							{
								// only consider a face normal if the transition should be smooth on
								// that side of the vertex
								if (normalMode.getArrayElement(adjacentFaceID, getFaceVertexIndex(
									adjacentFaceID, vertexID)) == NORMAL_MODE_SMOOTH)
								{
									faceIDs.add(adjacentFaceID);
								}
							}
						}
					}
					normal.X1 = 0;
					normal.X2 = 0;
					normal.X3 = 0;
					for (int k = 0; k < faceIDs.size(); k++)
					{
						Vector3D.add(normal, faceNormals.getVector(faceIDs.get(k)));
					}
					Vector3D.multiply(normal, 1.0 / faceIDs.size());
					normals.setArrayElement(i, j, normal);
				} else if (normalMode.getArrayElement(i, j) == NORMAL_MODE_HARD)
				{
					normals.setArrayElement(i, j, faceNormals.getVector(i));
				}
			}
		}
	}

	/**
	 * translate each vertex of the mesh by distance
	 */
	@Override
	public void translate(Vector3D distance)
	{
		double[] vertexPositionData = vertexPosition.getRawData();

		for (int i = 0; i < vertexPosition.getLength(); i++)
		{
			vertexPositionData[3 * i] += distance.X1;
			vertexPositionData[3 * i + 1] += distance.X2;
			vertexPositionData[3 * i + 2] += distance.X3;
		}

		transformFaceVertexNormals();
	}

	/**
	 * scale the position of each vertex along a straight through the vertex and the pivot by amount
	 */
	@Override
	public void scale(Vector3D pivot, Vector3D amount)
	{
		double[] vertexPositionData = vertexPosition.getRawData();
		for (int i = 0; i < vertexPosition.getLength(); i++)
		{
			vertexPositionData[3 * i] = (vertexPositionData[3 * i] - pivot.X1) * amount.X1
					+ pivot.X1;
			vertexPositionData[3 * i + 1] = (vertexPositionData[3 * i + 1] - pivot.X2) * amount.X2
					+ pivot.X2;
			vertexPositionData[3 * i + 2] = (vertexPositionData[3 * i + 2] - pivot.X3) * amount.X3
					+ pivot.X3;
		}

		transformFaceVertexNormals();
	}

	/**
	 * apply a transformation matrix to each vertex of the mesh
	 */
	@Override
	public void applyMatrix(Matrix3 matrix)
	{
		double[] vertexPositionData = vertexPosition.getRawData();
		Vector3D temp = new Vector3D();

		for (int i = 0; i < vertexPosition.getLength(); i++)
		{
			temp.X1 = vertexPositionData[3 * i];
			temp.X2 = vertexPositionData[3 * i + 1];
			temp.X3 = vertexPositionData[3 * i + 2];

			vertexPositionData[3 * i] = temp.X1 * matrix.X[0][0] + temp.X2 * matrix.X[0][1]
					+ temp.X3 * matrix.X[0][2];
			vertexPositionData[3 * i + 1] = temp.X1 * matrix.X[1][0] + temp.X2 * matrix.X[1][1]
					+ temp.X3 * matrix.X[1][2];
			vertexPositionData[3 * i + 2] = temp.X1 * matrix.X[2][0] + temp.X2 * matrix.X[2][1]
					+ temp.X3 * matrix.X[2][2];

		}

		transformFaceVertexNormals();
	}

	/**
	 * apply a transformation matrix to each vertex of the mesh
	 */
	@Override
	public void applyMatrix(Matrix4 matrix)
	{
		double[] vertexPositionData = vertexPosition.getRawData();
		Vector3D temp = new Vector3D();
		double res4;

		for (int i = 0; i < vertexPosition.getLength(); i++)
		{

			temp.X1 = vertexPositionData[3 * i];
			temp.X2 = vertexPositionData[3 * i + 1];
			temp.X3 = vertexPositionData[3 * i + 2];

			res4 = matrix.X[3][0] * temp.X1 + matrix.X[3][1] * temp.X2 + matrix.X[3][2] * temp.X3
					+ matrix.X[3][3];

			vertexPositionData[3 * i] = temp.X1 * matrix.X[0][0] + temp.X2 * matrix.X[0][1]
					+ temp.X3 * matrix.X[0][2] + matrix.X[0][3];
			vertexPositionData[3 * i] /= res4;
			vertexPositionData[3 * i + 1] = temp.X1 * matrix.X[1][0] + temp.X2 * matrix.X[1][1]
					+ temp.X3 * matrix.X[1][2] + matrix.X[1][3];
			vertexPositionData[3 * i + 1] /= res4;
			vertexPositionData[3 * i + 2] = temp.X1 * matrix.X[2][0] + temp.X2 * matrix.X[2][1]
					+ temp.X3 * matrix.X[2][2] + matrix.X[2][3];
			vertexPositionData[3 * i + 2] /= res4;

		}

		transformFaceVertexNormals();
	}

	/**
	 * calculate and return bounding axis-alligned bounding box
	 */
	@Override
	public AxisAlignedBoundingBox getBoundingBox()
	{
		double[] vertexPositionData = vertexPosition.getRawData();
		AxisAlignedBoundingBox bBox = new AxisAlignedBoundingBox();

		bBox.min.X1 = vertexPositionData[0];
		bBox.min.X2 = vertexPositionData[1];
		bBox.min.X3 = vertexPositionData[2];
		bBox.max = bBox.max.clone();

		for (int i = 3; i < vertexPosition.getLength(); i++)
		{
			if (vertexPositionData[3 * i] < bBox.min.X1)
			{
				bBox.min.X1 = vertexPositionData[3 * i];
			}
			if (vertexPositionData[3 * i + 1] < bBox.min.X2)
			{
				bBox.min.X2 = vertexPositionData[3 * i + 1];
			}
			if (vertexPositionData[3 * i + 2] < bBox.min.X3)
			{
				bBox.min.X3 = vertexPositionData[3 * i + 2];
			}
			if (vertexPositionData[3 * i] > bBox.max.X1)
			{
				bBox.max.X1 = vertexPositionData[3 * i];
			}
			if (vertexPositionData[3 * i + 1] > bBox.max.X2)
			{
				bBox.max.X2 = vertexPositionData[3 * i + 1];
			}
			if (vertexPositionData[3 * i + 2] > bBox.max.X3)
			{
				bBox.max.X3 = vertexPositionData[3 * i + 2];
			}
		}
		return bBox;
	}

	/**
	 * Test a ray for intersection with this face. Currently this works only for planar faces.
	 *
	 * @param face the face to intersect with
	 * @param ray ray that should be tested against this face
	 * @return coordinates of intersection point or null if none exists
	 */
	public Vector3D getIntersectionPoint(Face face, Ray ray)
	{
		/*
		 * roadmap:
		 * 1. find plane in which polygon lies
		 *    1.1 get three vertices out of polygon
		 *    1.2 calculate plane equation
		 * 2. find intersection with plane
		 * 3. project into 2d coordinate system
		 *    3.1 find biggest component of plane normal
		 *    3.2 kick that coordinate from all the face's vertices to project onto a coordinate plane
		 * 4. determine wether point is inside the face:
		 */

		// 1.1 get vertices out of polygon:
		ArrayList<Integer> vertexIDs = getVertexIDs(face);

		// 1.2 calculate plane equation
		Plane plane = new Plane();
		plane.fromPoints(getVertexPosition(vertexIDs.get(0)), getVertexPosition(vertexIDs.get(1)),
			getVertexPosition(vertexIDs.get(2)));

		// 2. find intersection with plane
		Vector3D intersectionPoint = plane.getIntersectionPoint(ray);
		if (intersectionPoint == null)
		{ // abandon search if no point was returned
			// eu.moonlight3d.ml3d.core.State.getInstance().logger().info(getName() +
			// ": no intersection point returned in plane intersection");
			return null;
		}

		// 3.1 find biggest component of plane normal
		int indices[] = getMaxComponents(plane.normal);

		// 3.2 kick that coordinate from intersection point
		Vector2D intersection = new Vector2D();

		if (indices[1] == 0)
		{
			intersection.X2 = intersectionPoint.X1;
		} else if (indices[1] == 1)
		{
			intersection.X2 = intersectionPoint.X2;
		} else
		{
			intersection.X2 = intersectionPoint.X3;
		}
		if (indices[0] == 0)
		{
			intersection.X1 = intersectionPoint.X1;
		} else if (indices[0] == 1)
		{
			intersection.X1 = intersectionPoint.X2;
		} else
		{
			intersection.X1 = intersectionPoint.X3;
		}

		Vector2D[] points = projectFace(vertexIDs);

		if (isInsideProjectedFace(points, intersection))
		{
			// eu.moonlight3d.ml3d.core.State.getInstance().logger().info(getName() +
			// ": intersection point found");
			return intersectionPoint;
		} else
		{
			// eu.moonlight3d.ml3d.core.State.getInstance().logger().info(getName() +
			// ": no intersection with polygon");
			return null;
		}
	}

	@Override
	public Vector3D getIntersectionPoint(eu.moonlight3d.math.Ray ray)
	{
		Vector3D intersection = null;
		double distance = 0;

		for (int i = 0; i < getNumFaces(); i++)
		{
			if (intersection == null)
			{
				intersection = getIntersectionPoint(getFace(i), ray);
				if (intersection != null)
				{
					distance = intersection.sub(ray.start).abs();
				}
			} else
			{
				Vector3D newIntersection = getIntersectionPoint(getFace(i), ray);
				if (newIntersection != null)
				{
					if (distance > newIntersection.sub(ray.start).abs())
					{
						intersection = newIntersection;
						distance = intersection.sub(ray.start).abs();
					}
				}
			}
		}

		return intersection;
	}

	@Override
	public eu.moonlight3d.ml3d.backend.threedim.sg.Node clone()
	{
		State.getInstance().logger().finer("enter clone()");

		Mesh mesh = null;
		mesh = (Mesh) super.clone();

		mesh.vertexPosition = vertexPosition.clone();
		mesh.vertexEdges = vertexEdges.clone();

		mesh.edgeVertices = edgeVertices.clone();
		mesh.edgeFaces = edgeFaces.clone();

		mesh.faceEdges = faceEdges.clone();

		mesh.normals = normals.clone();
		mesh.normalMode = normalMode.clone();

		mesh.uvMaps = new HashMap<String, DoubleArrayBuffer>();
		for (String key : uvMaps.keySet())
		{
			mesh.uvMaps.put(key, uvMaps.get(key).clone());
		}

		State.getInstance().logger().finer("leave clone()");

		return mesh;
	}

	@Override
	public String getType()
	{
		return "Mesh";
	}

	/**
	 * Return the number of vertices in the mesh.
	 *
	 * @return number of vertices
	 */
	public int getNumVertices()
	{
		return vertexPosition.getLength();
	}

	/**
	 * returns the vertex given by the vertex number
	 *
	 * @param vertexID the ID of the vertex
	 * @return the vertex with the given ID
	 */
	public Vertex getVertex(int vertexID)
	{
		Vertex vertex = new Vertex();
		vertex.vertexID = vertexID;
		vertex.position = vertexPosition.getVector(vertexID);
		vertex.edgeIDs = vertexEdges.getArray(vertexID);

		return vertex;
	}

	/**
	 * Get the position of the vertex with the given ID
	 *
	 * @param vertexID the ID of the vertex
	 * @return the position of the vertex
	 */
	public Vector3D getVertexPosition(int vertexID)
	{
		return vertexPosition.getVector(vertexID);
	}

	/**
	 * Get get a list of edges that contain the vertex with the given ID.
	 *
	 * @param vertexID the ID of the vertex
	 * @return the list of edges (by ID) which contain the given vertex
	 */
	public int[] getVertexEdges(int vertexID)
	{
		return vertexEdges.getArray(vertexID);
	}
	
	/**
	 * Get the raw buffer containing the vertex position data. This buffer
	 * may not be modified directly.
	 * 
	 * @return the vector buffer containing the vertex positions.
	 */
	public Vector3DBuffer getVertexBuffer() {
		return vertexPosition;
	}

	/**
	 * sets the vertex with the given number to the given position
	 *
	 * @param vertexID the ID of the vertex
	 * @param position the new position of the vertex
	 */
	public void setVertexPosition(int vertexID, Vector3D position)
	{
		vertexPosition.setVector(vertexID, position);
		transformFaceVertexNormals(vertexID);
	}

	/**
	 * create a new vertex at the given position
	 *
	 * @param position the position of the vertex
	 * @return the ID of the new vertex
	 */
	public int addVertex(Vector3D position)
	{
		vertexPosition.append(position);
		int[] vertexEdgeIDs = new int[0];
		vertexEdges.append(vertexEdgeIDs);

		return vertexPosition.getLength() - 1;
	}

	/**
	 * create a new vertex at the given position
	 *
	 * @param x the x-value of the vertex
	 * @param y the y-value of the vertex
	 * @param z the z-value of the vertex
	 * @return the ID of the new vertex
	 */
	public int addVertex(double x, double y, double z)
	{
		return addVertex(new Vector3D(x, y, z));
	}

	/**
	 * remove the vertex with the given number. This will also remove all edges, manifold edges and
	 * faces, where this vertex is currently part of.
	 *
	 * @param vertexID the ID of the vertex to remove
	 * @throws Exception
	 */
	public void removeVertex(int vertexID) throws Exception
	{
		if (vertexID < 0 || vertexID >= vertexPosition.getLength())
		{
			throw new Exception(L10N.translate("requested vertex does not exist"));
		}

		// remove edges depending on vertex
		int[] edgeIDs = vertexEdges.getArray(vertexID);
		Arrays.sort(edgeIDs);
		for (int i = edgeIDs.length - 1; i >= 0; i--)
		{
			removeEdge(edgeIDs[i]);
		}

		// remove vertex data from arrays
		vertexEdges.remove(vertexID);
		vertexPosition.remove(vertexID);

		// renumber edge vertex IDs
		int vertexIDs[] = edgeVertices.getRawData();
		for (int i = 0; i < edgeVertices.getLength(); i++)
		{
			if (vertexIDs[i] > vertexID)
			{
				vertexIDs[i]--;
			}
		}
	}

	/**
	 * Generate a list of all faces that touch this vertex.
	 *
	 * @param vertexID the unique ID of a vertex
	 * @return Array containing the unique IDs of all faces that touch the vertex, in no particular
	 *         order
	 * @throws Exception
	 */
	public ArrayList<Integer> getAdjacentFaces(int vertexID) throws Exception
	{
		ArrayList<Integer> faceIDs = new ArrayList<Integer>();

		int[] vertexEdgeIDs = getVertexEdges(vertexID);
		if (vertexID < 0 || vertexID > vertexPosition.getLength())
		{
			throw new IllegalArgumentException(L10N
					.translate("there is no vertex with the given ID in the mesh"));
		}
		if (vertexEdgeIDs == null)
		{
			return faceIDs;
		}
		for (int edgeID : vertexEdgeIDs)
		{
			Edge edge = getEdge(edgeID);
			if (edge.faceIDs == null)
			{
				continue;
			}
			for (int faceID : edge.faceIDs)
			{
				if (faceIDs.contains(faceID))
				{
					continue;
				}
				faceIDs.add(faceID);
			}
		}

		return faceIDs;
	}


	/**
	 * Add an edge defined by the IDs of it's end point vertices to the mesh.
	 *
	 * @param vertexID1 ID of start vertex
	 * @param vertexID2 ID of end vertex
	 * @return ID of newly created edge
	 * @throws IllegalArgumentException if an ID is not valid or edge already exists
	 */
	public int addEdge(int vertexID1, int vertexID2) throws IllegalArgumentException
	{
		// check if given vertices do indeed exist
		if (vertexID1 < 0 || vertexID1 > vertexPosition.getLength() || vertexID2 < 0
				|| vertexID2 > vertexPosition.getLength())
		{
			throw new IllegalArgumentException(L10N
					.translate("at least one of the given vertices does not exist"));
		}

		// check if edge already exists
		if (findEdge(vertexID1, vertexID2) >= 0)
			throw new IllegalArgumentException(L10N
					.translate("cannot add an edge that already exists in the mesh"));

		// append vertices to edge vertices list
		int edgeID = edgeVertices.append(vertexID1) / 2;
		edgeVertices.append(vertexID2);

		// append empty faces array to allocate the edge ID in that buffer
		int faceIDs[] = new int[0];
		edgeFaces.append(faceIDs);

		vertexEdges.resize(vertexID1, vertexEdges.getArrayLength(vertexID1) + 1);
		vertexEdges.resize(vertexID2, vertexEdges.getArrayLength(vertexID2) + 1);
		vertexEdges.setArrayElement(vertexID1, vertexEdges.getArrayLength(vertexID1) - 1, edgeID);
		vertexEdges.setArrayElement(vertexID2, vertexEdges.getArrayLength(vertexID2) - 1, edgeID);

		return edgeID;
	}

	/**
	 * Find an edge between two specific vertices.
	 *
	 * @param vertexID1 ID of first vertex defining the edge
	 * @param vertexID2 ID of second vertex defining the edge
	 * @return ID of edge between the two vertices or -1 no edge could be found
	 */
	public int findEdge(int vertexID1, int vertexID2)
	{
		// search for edge
		int[] vertex1EdgeIDs = getVertexEdges(vertexID1);
		if (vertex1EdgeIDs != null)
		{
			int[] edgeVertices = new int[2];
			for (Integer i : vertex1EdgeIDs)
			{
				getEdgeVertices(i, edgeVertices);
				if (edgeVertices[0] == vertexID2 || edgeVertices[1] == vertexID2)
				{
					return i;
				}
			}
		}
		return -1;
	}

	/**
     * Search for an edge between the two given vertices and create it if
     * it does not exist yet.
     *
     * @param mesh the mesh to search in
     * @param vertexID1 the ID of the first vertex
     * @param vertexID2 the ID of the second vertex
     * @return the ID of the edge between the vertices
     */
    public int findOrCreateEdge(int vertexID1, int vertexID2) {
    	int edge;
    	return (edge = findEdge(vertexID1, vertexID2)) >= 0 ?
    			edge : addEdge(vertexID1, vertexID2);
    }

	/**
	 * Remove the given edge and all faces which contain these edges.
	 *
	 * @param edgeID the edge ID of the edge to remove
	 * @param dependenciesOnly true if the edge itself should not be deleted
	 * @throws IllegalArgumentException if given edge ID is invalid
	 */
	private void removeEdge(int edgeID, boolean dependenciesOnly) throws IllegalArgumentException
	{
		/*
		 * delete any face using this edge:
		 *   for each linked face:
		 *     delete face by given face ID
		 * remove edge reference from vertices
		 * delete edge
		 */
		if (getEdge(edgeID) == null)
		{
			throw new IllegalArgumentException(L10N.translate("requested edge does not exist"));
		}

		Edge edge = getEdge(edgeID);

		// copy edge so that we can safely iterate over the face id vector
		if (edge.faceIDs != null)
		{
			int[] faceIDs = edge.faceIDs.clone();
			Arrays.sort(faceIDs);
			for (int i = faceIDs.length - 1; i >= 0; i--)
			{
				removeFace(faceIDs[i]);
			}
		}

		// remove edge reference from vertices
		for (int i = 0; i < 2; i++)
		{
			int[] edgeIDs = vertexEdges.getArray(edge.vertexID[i]);
			int[] newEdgeIDs = new int[edgeIDs.length - 1];
			int index = 0;
			for (int j = 0; j < edgeIDs.length; j++)
			{
				if (edgeIDs[j] != edgeID)
				{
					newEdgeIDs[index] = edgeIDs[j];
					index++;
				}
			}
			vertexEdges.setArray(edge.vertexID[i], newEdgeIDs);
		}

		if (!dependenciesOnly)
		{
			// remove both vertices from edge vertices list
			edgeFaces.remove(edgeID);
			edgeVertices.remove(edgeID * 2 + 1);
			edgeVertices.remove(edgeID * 2);

			// renumber references to edges with higher edge IDs in vertices
			int vertexEdgesData[][] = vertexEdges.getRawData();
			for (int i = 0; i < vertexEdges.getNumArrays(); i++)
			{
				for (int j = 0; j < vertexEdgesData[i].length; j++)
				{
					if (vertexEdgesData[i][j] > edgeID)
					{
						vertexEdgesData[i][j]--;
					}
				}
			}

			// renumber references to edges with higher edge IDs in faces
			int faceEdgesData[][] = faceEdges.getRawData();
			for (int i = 0; i < faceEdges.getNumArrays(); i++)
			{
				for (int j = 0; j < faceEdgesData[i].length; j++)
				{
					if (faceEdgesData[i][j] > edgeID)
					{
						faceEdgesData[i][j]--;
					}
				}
			}
		}
	}

	/**
	 * Remove an edge from the mesh.
	 *
	 * @param edgeID ID of the edge to remove
	 * @throws IllegalArgumentException if given edge ID is invalid
	 */
	public void removeEdge(int edgeID) throws IllegalArgumentException
	{
		removeEdge(edgeID, false);
	}

	/**
	 * return the number of edges in this mesh
	 *
	 * @return number of edges in this mesh
	 */
	public int getNumEdges()
	{
		return edgeVertices.getLength() / 2;
	}

	/**
	 * return the edge with the given edge ID
	 *
	 * @param edgeID the ID of the edge
	 * @return the data of the edge
	 */
	public Edge getEdge(int edgeID)
	{
		Edge edge = new Edge();

		if (edgeID < 0 || edgeID >= edgeFaces.getNumArrays())
		{
			return null;
		}

		edge.edgeID = edgeID;
		edge.vertexID = new int[2];
		edge.vertexID[0] = edgeVertices.getElement(edgeID * 2);
		edge.vertexID[1] = edgeVertices.getElement(edgeID * 2 + 1);
		edge.faceIDs = edgeFaces.getArray(edgeID);

		return edge;
	}

	/**
	 * Get the corner vertices for the edge with the given ID.
	 *
	 * @param edgeID the ID of the edge to get the corner vertices for
	 * @return an array with the IDs of the corner vertices
	 */
	public int[] getEdgeVertices(int edgeID)
	{
		int[] vertices = new int[2];
		vertices[0] = edgeVertices.getElement(edgeID * 2);
		vertices[1] = edgeVertices.getElement(edgeID * 2 + 1);
		return vertices;
	}

	/**
	 * Get the corner vertices for the edge with the given ID.
	 *
	 * @param edgeID the ID of the edge to get the corner vertices for
	 * @param vertices an array with at least two elements in which the IDs of the corner vertices
	 *        are returned
	 */
	public void getEdgeVertices(int edgeID, int[] vertices)
	{
		vertices[0] = edgeVertices.getElement(edgeID * 2);
		vertices[1] = edgeVertices.getElement(edgeID * 2 + 1);
	}
	
	/**
	 * Get the buffer in which the edge vertex index data is stored. This
	 * buffer may not be modified directly.
	 * 
	 * @return the edge vertex index buffer
	 */
	public IntBuffer getEdgeVerticesBuffer()
	{
		return edgeVertices;
	}

	/**
	 * Get the list of faces of which the edge with the given ID is part of.
	 *
	 * @param edgeID the ID of the edge to get the list of faces for
	 * @return an array with the IDs of the faces that the given edge is part of
	 */
	public int[] getEdgeFaces(int edgeID)
	{
		return edgeFaces.getArray(edgeID);
	}

	/**
	 * Run all the required checks to verify that the given list of edges forms a closed loop that
	 * can contain a face.
	 *
	 * @param edges the list of edges to check
	 * @throws Exception if the requirements are not met
	 */
	private void checkFaceEdgeLoop(ArrayList<Integer> edges) throws Exception
	{
		/*
		 * check if there are enough edges to form a face
		 * check if all edge IDs are valid
		 * check if edges form a loop:
		 *   for each edge in edges vector:
		 *     check if edge shares a vertex with both next and previous edge
		 */
		if (edges.size() < 3)
		{
			throw new Exception("cannot generate a face from less than 3 edges");
		}

		// check if all edge IDs are valid
		for (int i = 0; i < edges.size(); i++)
		{
			if (edges.get(i) < 0 || edges.get(i) >= edgeFaces.getNumArrays())
			{
				throw new Exception(L10N
						.translate("at least one of the given edge IDs is not valid"));
			}
		}

		// check wether given edges form a correct single loop
		// (the first and last edge in this array are special cases)
		for (int i = 0; i < edges.size(); i++)
		{
			Edge edge = getEdge(edges.get(i));
			Edge prev, next;
			if (i == 0)
			{
				prev = getEdge(edges.get(edges.size() - 1));
				next = getEdge(edges.get(1));
			} else if (i == edges.size() - 1)
			{
				prev = getEdge(edges.get(edges.size() - 2));
				next = getEdge(edges.get(0));
			} else
			{
				prev = getEdge(edges.get(i - 1));
				next = getEdge(edges.get(i + 1));
			}
			boolean edgeOK = false;

			if (edge.vertexID[0] == prev.vertexID[0] || edge.vertexID[0] == prev.vertexID[1])
			{
				if (edge.vertexID[1] == next.vertexID[0] || edge.vertexID[1] == next.vertexID[1])
				{
					edgeOK = true;
				}
			} else if (edge.vertexID[1] == prev.vertexID[0] || edge.vertexID[1] == prev.vertexID[1])
			{
				if (edge.vertexID[0] == next.vertexID[0] || edge.vertexID[0] == next.vertexID[1])
				{
					edgeOK = true;
				}
			}

			if (!edgeOK)
			{
				throw new Exception(L10N.translate("the list of given edges does not form a loop"));
			}
		}
	}

	/**
	 * Add a face defined as an edge loop to the mesh.
	 *
	 * @param edges edge loop defining the face
	 * @param verify true if the given edge loop should be verified as such before inserting
	 * @return the ID of the new face
	 * @throws Exception
	 */
	public int addFace(ArrayList<Integer> edges, boolean verify) throws Exception
	{
		/*
		 * check if given edge loop creates a valid face
		 * for each edge in edges vector:
		 *   add face ID to face list
		 * insert face into faces hash_map
		 */
		if (verify)
		{
			checkFaceEdgeLoop(edges);
		}

		int[] edgeIDs = new int[edges.size()];
		for (int i = 0; i < edges.size(); i++)
		{
			edgeIDs[i] = edges.get(i);
		}
		int faceID = faceEdges.append(edgeIDs);

		// actually begin adding the face:
		for (int i = 0; i < edges.size(); i++)
		{
			int edgeID = edges.get(i);
			int length = edgeFaces.getArrayLength(edgeID);
			edgeFaces.resize(edgeID, length + 1);
			edgeFaces.setArrayElement(edgeID, length, faceID);
		}

		// add UV map coordinates for this face:
		double uvCoords[] = new double[2 * edges.size()];
		for (String mapName : uvMaps.keySet())
		{
			uvMaps.get(mapName).append(uvCoords);
		}

		// add normals for the face vertices
		byte[] normalModes = new byte[edges.size()];
		for (int i = 0; i < edges.size(); i++)
		{
			normalModes[i] = NORMAL_MODE_SMOOTH;
		}
		normalMode.append(normalModes);

		Vector3D[] normals = new Vector3D[edges.size()];
		for (int i = 0; i < edges.size(); i++)
		{
			normals[i] = new Vector3D(0, 0, 0);
		}
		this.normals.append(normals);
		// initialise normals
		// this is an ugly hack relying on other functions operating on faces
		// before the face is fully there:
		ArrayList<Integer> vertexIDs = getVertexIDs(getFace(faceID));
		for (Integer vertexID : vertexIDs)
		{
			transformFaceVertexNormals(vertexID);
		}

		return faceID;
	}

	/**
	 * Add a face defined as an edge loop to the mesh.
	 *
	 * @param edges edge loop defining the face
	 * @return the ID of the new face
	 * @throws Exception
	 */
	public int addFace(ArrayList<Integer> edges) throws Exception
	{
		return addFace(edges, true);
	}

	/**
	 * Return number of faces in this mesh.
	 *
	 * @return number of faces
	 */
	public int getNumFaces()
	{
		return faceEdges.getNumArrays();
	}

	/**
	 * return the face corresponding to the given face ID
	 *
	 * @param faceID the ID of the face
	 * @return the data for the face
	 */
	public Face getFace(int faceID)
	{
		Face face = new Face();

		if (faceID < 0 || faceID >= faceEdges.getNumArrays())
		{
			return null;
		}

		face.faceID = faceID;
		face.edgeIDs = faceEdges.getArray(faceID);

		return face;
	}

	/**
	 * Get the list of edges that form the loop that defines the face with the given ID
	 *
	 * @param faceID the ID of the face to get the edges for
	 * @return a list of edges that define the bounds of the given face
	 */
	public int[] getFaceEdges(int faceID)
	{
		return faceEdges.getArray(faceID);
	}

	/**
	 * Return the face vertex normal mode of the given face vertex
	 *
	 * @param faceID the ID of the face to get the mode for
	 * @param vertexIndex the index of the vertex to get the mode for
	 * @return the normal mode of the face vertex
	 */
	public byte getFaceVertexNormalMode(int faceID, int vertexIndex)
	{
		return normalMode.getArrayElement(faceID, vertexIndex);
	}

	/**
	 * Set the face vertex normal mode of the given face vertex
	 *
	 * @param faceID the ID of the face to get the mode for
	 * @param vertexIndex the index of the vertex to get the mode for
	 * @param mode the normal mode of the face vertex
	 */
	public void setFaceVertexNormalMode(int faceID, int vertexIndex, byte mode)
	{
		normalMode.setArrayElement(faceID, vertexIndex, mode);
		if (mode != NORMAL_MODE_LOCKED)
		{
			if (mode == NORMAL_MODE_HARD)
			{
				transformFaceVertexNormal(faceID, vertexIndex);
			} else
			{
				int vertexID = getVertexIDs(faceID).get(vertexIndex);
				transformFaceVertexNormals(vertexID);
			}
		}
	}

	/**
	 * Get the normal for the given face vertex
	 *
	 * @param faceID the ID of the face to get the normal for
	 * @param vertexIndex the index of the vertex to get the normal for
	 * @return the normal of the given face vertex
	 */
	public Vector3D getFaceVertexNormal(int faceID, int vertexIndex)
	{
		return normals.getArrayElement(faceID, vertexIndex);
	}

	/**
	 * Set the normal for the given face vertex. This method can only be used for face vertex
	 * normals that have the mode set to NORMAL_MODE_LOCKED.
	 *
	 * @param faceID the ID of the face to set the normal for
	 * @param vertexIndex the index of the vertex to set the normal for
	 * @param normal the new normal
	 * @throws IllegalStateException if the normal mode of the face vertex is incorrect
	 */
	public void setFaceVertexNormal(int faceID, int vertexIndex, Vector3D normal)
	{
		if (normalMode.getArrayElement(faceID, vertexIndex) != NORMAL_MODE_LOCKED)
		{
			throw new IllegalStateException(
					L10N
							.translate("vertex normals that are in modes other than locked cannot be set to specific values"));
		}
		normals.setArrayElement(faceID, vertexIndex, normal);
	}

	/**
	 * Get the index of the vertex in the face vertex data arrays of the given face.
	 *
	 * @param faceID the ID of the face to get the face vertex index for
	 * @param vertexID the ID of the vertex to get the index for
	 * @return the index of the vertex in the face vertex data arrays
	 */
	public int getFaceVertexIndex(int faceID, int vertexID)
	{
		int edgeID = faceEdges.getArrayElement(faceID, 0);
		// first edge is slightly special as both vertices need to be respected:
		int[] edgeVertices = getEdgeVertices(edgeID);
		// determine orientation of first edge by searching for shared vertex
		int[] nextEdgeVertices = getEdgeVertices(faceEdges.getArrayElement(faceID, 0));
		if (nextEdgeVertices[0] == edgeVertices[0] || nextEdgeVertices[1] == edgeVertices[0])
		{
			if (edgeVertices[1] == vertexID)
			{
				return 0;
			} else if (edgeVertices[0] == vertexID)
			{
				return 1;
			}
		} else
		{
			if (edgeVertices[0] == vertexID)
			{
				return 0;
			} else if (edgeVertices[1] == vertexID)
			{
				return 1;
			}
		}

		int index = 2;
		int[] lastEdgeVertices = new int[2];
		for (int i = 1; i < faceEdges.getArrayLength(faceID) - 1; i++)
		{
			edgeID = faceEdges.getArrayElement(faceID, i);

			lastEdgeVertices[0] = edgeVertices[0];
			lastEdgeVertices[1] = edgeVertices[1];
			getEdgeVertices(edgeID, edgeVertices);
			if (edgeVertices[0] == lastEdgeVertices[0] || edgeVertices[0] == lastEdgeVertices[1])
			{
				if (edgeVertices[1] == vertexID)
				{
					return index;
				}
			} else
			{
				if (edgeVertices[0] == vertexID)
				{
					return index;
				}
			}
			index++;
		}

		// if the loop ran through without returning, the vertex isn't part of the face
		throw new IllegalArgumentException(L10N
				.translate("the given vertex is not part of the given face"));
	}

	/**
	 * Generate a list of all vertices that are corners of this face.
	 *
	 * @param face the face to generate the list of corner vertices for
	 * @return Array containing the Unique IDs of all vertices that are part of this face, in order
	 *         of appearence in edge list
	 */
	public ArrayList<Integer> getVertexIDs(Face face)
	{
		ArrayList<Integer> vertexIDs = new ArrayList<Integer>();

		int edgeID = face.edgeIDs[0];
		// first edge is slightly special as both vertices need to be respected:
		int[] edgeVertices = getEdgeVertices(edgeID);
		// determine orientation of first edge by searching for shared vertex
		int[] nextEdgeVertices = getEdgeVertices(face.edgeIDs[1]);
		if (nextEdgeVertices[0] == edgeVertices[0] || nextEdgeVertices[1] == edgeVertices[0])
		{
			vertexIDs.add(edgeVertices[1]);
			vertexIDs.add(edgeVertices[0]);
		} else
		{
			vertexIDs.add(edgeVertices[0]);
			vertexIDs.add(edgeVertices[1]);
		}

		int[] lastEdgeVertices = new int[2];
		for (int i = 1; i < face.edgeIDs.length - 1; i++)
		{
			edgeID = face.edgeIDs[i];

			lastEdgeVertices[0] = edgeVertices[0];
			lastEdgeVertices[1] = edgeVertices[1];
			getEdgeVertices(edgeID, edgeVertices);
			if (edgeVertices[0] == lastEdgeVertices[0] || edgeVertices[0] == lastEdgeVertices[1])
			{
				vertexIDs.add(edgeVertices[1]);
			} else
			{
				vertexIDs.add(edgeVertices[0]);
			}
		}

		return vertexIDs;
	}

	/**
	 * Generate a list of all vertices that are corners of this face.
	 *
	 * @param faceID the ID of the face to generate the list of corner vertices for
	 * @return Array containing the Unique IDs of all vertices that are part of this face, in order
	 *         of appearence in edge list
	 */
	public ArrayList<Integer> getVertexIDs(int faceID)
	{
		ArrayList<Integer> vertexIDs = new ArrayList<Integer>();

		int edgeID = faceEdges.getArrayElement(faceID, 0);
		// first edge is slightly special as both vertices need to be respected:
		int[] edgeVertices = getEdgeVertices(edgeID);
		// determine orientation of first edge by searching for shared vertex
		int[] nextEdgeVertices = getEdgeVertices(faceEdges.getArrayElement(faceID, 1));
		if (nextEdgeVertices[0] == edgeVertices[0] || nextEdgeVertices[1] == edgeVertices[0])
		{
			vertexIDs.add(edgeVertices[1]);
			vertexIDs.add(edgeVertices[0]);
		} else
		{
			vertexIDs.add(edgeVertices[0]);
			vertexIDs.add(edgeVertices[1]);
		}

		int[] lastEdgeVertices = new int[2];
		for (int i = 1; i < faceEdges.getArrayLength(faceID) - 1; i++)
		{
			edgeID = faceEdges.getArrayElement(faceID, i);

			lastEdgeVertices[0] = edgeVertices[0];
			lastEdgeVertices[1] = edgeVertices[1];
			getEdgeVertices(edgeID, edgeVertices);
			if (edgeVertices[0] == lastEdgeVertices[0] || edgeVertices[0] == lastEdgeVertices[1])
			{
				vertexIDs.add(edgeVertices[1]);
			} else
			{
				vertexIDs.add(edgeVertices[0]);
			}
		}

		return vertexIDs;
	}

	private int findEar(ArrayList<Vector2D> points)
	{
		Vector2D[] pointsArray = new Vector2D[points.size()];
		pointsArray = points.toArray(pointsArray);
		for (int i = 0; i < points.size(); i++)
		{
			// get neighbouring vertices in polygon
			Vector2D previous, next;

			if (i - 1 < 0)
			{
				previous = points.get(points.size() + i - 1);
			} else
			{
				previous = points.get(i - 1);
			}
			next = points.get((i + 1) % points.size());

			// check if connecting line of neighboring vertices intersects another edge
			Line2D connection = new Line2D(previous, next);
			boolean isEar = true;
			for (int j = 0; j < points.size(); j++)
			{
				// do not check against vertices in potential ear
				if (points.get(j) == previous || points.get(j) == next
						|| points.get((j + 1) % points.size()) == previous
						|| points.get((j + 1) % points.size()) == next)
				{
					continue;
				}

				Line2D line = new Line2D(points.get(j), points.get((j + 1) % points.size()));
				if (connection.getIntersectionPoint(line) != null)
				{
					isEar = false;
					break;
				}
			}
			// check if midpoint of connecting line is inside polygon
			if (isEar)
			{
				Vector2D midpoint = previous.add(next).multiply(0.5);
				isEar = isInsideProjectedFace(pointsArray, midpoint);
			}

			if (isEar)
			{
				return i;
			}
		}
		/*
		 * If we fall through to this line there is no ear left, which probably means that the polygon
		 * is self-intersecting in strange ways. Triangulating such a beast is nontrivial and therefore
		 * this algorithm gives up and pretends that the first vertex given is part of an ear in the
		 * hope that the result will at least be partly reasonable and not something other algorithms
		 * might not expect.
		 */
		return 0;
	}

	private void splitEar(ArrayList<Vector2D> points, ArrayList<Integer> vertexIDs,
			ArrayList<Integer> triangles)
	{
		// handle trivial case:
		if (points.size() == 3)
		{
			for (int i = 0; i < 3; i++)
			{
				triangles.add(vertexIDs.get(i));
			}
			return;
		}

		// split a recognised ear from polygon:
		int earIndex = findEar(points);

		int previousIndex = earIndex - 1;
		if (previousIndex < 0)
		{
			previousIndex = points.size() - 1;
		}
		int nextIndex = (earIndex + 1) % points.size();

		triangles.add(vertexIDs.get(previousIndex));
		triangles.add(vertexIDs.get(earIndex));
		triangles.add(vertexIDs.get(nextIndex));

		points.remove(earIndex);
		vertexIDs.remove(earIndex);

		splitEar(points, vertexIDs, triangles);
	}

	/**
	 * Return a triangulation of the given face. The returned array contains 3-tuples of vertex IDs
	 * which form the triangles.
	 *
	 * @param face the face to triangulate
	 * @return 3-tuples of vertex IDs, each forming a triangle
	 */
	public ArrayList<Integer> triangulated(Face face)
	{
		if (face.triangulated != null)
		{
			return face.triangulated;
		}

		ArrayList<Integer> triangles = new ArrayList<Integer>();
		ArrayList<Integer> vertexIDs = getVertexIDs(face);

		Vector2D[] points = projectFace(vertexIDs);

		ArrayList<Vector2D> pointList = new ArrayList<Vector2D>();

		for (int i = 0; i < points.length; i++)
		{
			pointList.add(points[i]);
		}

		splitEar(pointList, vertexIDs, triangles);

		return triangles;
	}

	/**
	 * Remove the face with the given face ID.
	 *
	 * @param faceID the ID of the face to remove
	 * @throws IllegalArgumentException
	 */
	public void removeFace(int faceID) throws IllegalArgumentException
	{
		/*
		 * remove face references in edges
		 * remove face from faces HashMap
		 */

		if (faceID < 0 || faceID > faceEdges.getNumArrays())
		{
			throw new IllegalArgumentException(L10N.translate("given face ID is invalid"));
		}

		int[] edgeIDs = faceEdges.getArray(faceID);
		for (int i = 0; i < edgeIDs.length; i++)
		{
			int numFaces = edgeFaces.getArrayLength(edgeIDs[i]);
			int index = 0;
			for (int j = 0; j < numFaces; j++)
			{
				int currentFaceID = edgeFaces.getArrayElement(edgeIDs[i], j);
				if (currentFaceID != faceID)
				{
					edgeFaces.setArrayElement(edgeIDs[i], index, currentFaceID);
					index++;
				}
			}
			edgeFaces.resize(edgeIDs[i], numFaces - 1);
		}

		faceEdges.remove(faceID);

		// renumber the face IDs in all edges if they are greater than faceID
		// careful: this code depends on the (internal) allocation behaviour of IntArrayBuffer
		int[][] faceIDs = edgeFaces.getRawData();
		for (int i = 0; i < edgeFaces.getNumArrays(); i++)
		{
			for (int j = 0; j < faceIDs[i].length; j++)
			{
				if (faceIDs[i][j] > faceID)
				{
					faceIDs[i][j]--;
				}
			}
		}

		// remove UV coordinates for face
		for (String mapName : uvMaps.keySet())
		{
			uvMaps.get(mapName).remove(faceID);
		}

		// remove face vertex normals for this face
		normalMode.remove(faceID);
		normals.remove(faceID);
	}

	/**
	 * Add a new UV map with the given name to the mesh.
	 *
	 * @param mapName the name of the UV map to add
	 */
	public void addUVMap(String mapName)
	{
		if (uvMaps.get(mapName) != null)
		{
			throw new IllegalArgumentException(L10N
					.translate("a UV map with the given name already exists"));
		}

		DoubleArrayBuffer map = new DoubleArrayBuffer();
		double dummy[] = new double[0];
		for (int i = 0; i < faceEdges.getNumArrays(); i++)
		{
			map.append(dummy);
			map.resize(i, 2 * faceEdges.getArrayLength(i));
		}
		// TODO check if buffer needs to be initialised
		uvMaps.put(mapName, map);
	}

	/**
	 * Remove an UV map from the mesh.
	 *
	 * @param mapName the name of the UV map to remove
	 */
	public void removeUVMap(String mapName)
	{
		if (uvMaps.get(mapName) == null)
		{
			throw new IllegalArgumentException(L10N
					.translate("there is no UV map with the given name"));
		}
		uvMaps.remove(mapName);
	}

	/**
	 * Return a list of the names of all UV maps in the mesh.
	 *
	 * @return a list containing the names of all UV maps in the mesh
	 */
	public ArrayList<String> getUVMaps()
	{
		ArrayList<String> maps = new ArrayList<String>();

		for (String mapName : uvMaps.keySet())
		{
			maps.add(mapName);
		}

		return maps;
	}

	/**
	 * Set UV coordinates for a face vertex in a given UV map
	 *
	 * @param mapName the name of the UV map to alter
	 * @param faceID the ID of the face to alter
	 * @param vertex the vertex index of the vertex to alter
	 * @param u the new U coordinate to be assigned to the face vertex
	 * @param v the new V coordinate to be assigned to the face vertex
	 */
	public void setUVCoordinates(String mapName, int faceID, int vertex, double u, double v)
	{
		if (uvMaps.get(mapName) == null)
		{
			throw new IllegalArgumentException(L10N
					.translate("there is no UV map with the given name"));
		}

		uvMaps.get(mapName).setArrayElement(faceID, 2 * vertex, u);
		uvMaps.get(mapName).setArrayElement(faceID, 2 * vertex + 1, v);
	}

	/**
	 * Return the UV coordinates for a face vertex in a given UV map
	 *
	 * @param mapName the name of the UV map
	 * @param faceID the ID of the face
	 * @param vertex the vertex index of vertex
	 * @return an array containing the the two elements of the requested UV coordinates
	 */
	public double[] getUVCoordinates(String mapName, int faceID, int vertex)
	{
		if (uvMaps.get(mapName) == null)
		{
			throw new IllegalArgumentException(L10N
					.translate("there is no UV map with the given name"));
		}

		double uvCoords[] = new double[2];
		uvCoords[0] = uvMaps.get(mapName).getArrayElement(faceID, 2 * vertex);
		uvCoords[1] = uvMaps.get(mapName).getArrayElement(faceID, 2 * vertex + 1);

		return uvCoords;
	}

	/**
	 * Set the amount of reserve storage that the mesh should hold for the upcoming operations. This
	 * does not alter the behaviour of the Mesh data structure, but enables sometimes significant
	 * speedups by reducing the numbers of expensive reallocation operations that are made
	 * internally.
	 *
	 * @param numVertices
	 * @param numEdges
	 * @param numFaces
	 * @param numFaceEdges
	 */
	public void setReserves(int numVertices, int numEdges, int numFaces, int numFaceEdges)
	{
		// System.out.println("setting reserves");
		vertexPosition.setReserve(numVertices);
		vertexEdges.setReserve(numVertices, 2 * numEdges);
		edgeVertices.setReserve(2 * numEdges);
		edgeFaces.setReserve(numEdges, numFaceEdges);
		faceEdges.setReserve(numFaces, numFaceEdges);
		for (String mapName : uvMaps.keySet())
		{
			uvMaps.get(mapName).setReserve(numFaces, 2 * numFaceEdges);
		}
		// System.out.println("reserves set");
	}

	/**
	 * Indicate, whether normals should be updated automatically on each individual change.
	 * Operations that modify a large number of vertices should disable this until they are done,
	 * activating a more efficient lazy evaluation behaviour. Enabling automatic update will force a
	 * recalculation of all normals.
	 *
	 * @param autoNormalsUpdate true if normals should be updated automatically
	 */
	public void setAutoNormalsUpdate(boolean autoNormalsUpdate)
	{
		if (this.autoNormalsUpdate == autoNormalsUpdate)
		{
			// no change, do nothing
			return;
		}
		this.autoNormalsUpdate = autoNormalsUpdate;
		if (autoNormalsUpdate)
		{
			// enforce an update of all normals as some updates may have been skipped
			transformFaceVertexNormals();
		}
	}

	/**
	 * Check if automatic updating of normals is activated.
	 *
	 * @return true if automatic updating is enabled, false otherwise
	 */
	public boolean getAutoNormalsUpdate()
	{
		return autoNormalsUpdate;
	}

	/**
	 * This resets the amount of reserve storage that is kept to 0. It is recommended you call this
	 * after you have performed all operations on a mesh for which you reserved storage.
	 */
	public void resetReserves()
	{
		setReserves(0, 0, 0, 0);
	}

	public boolean isValidAppendixName(String appendix)
	{
		if (appendix.substring(0, 5).equals("faces"))
		{
			int index = Integer.parseInt(appendix.substring(6, appendix.length() - 1));
			if (getFace(index) != null)
			{
				return true;
			}
		} else if (appendix.substring(0, 5).equals("edges"))
		{
			int index = Integer.parseInt(appendix.substring(6, appendix.length() - 1));
			if (getEdge(index) != null)
			{
				return true;
			}
		} else if (appendix.substring(0, 8).equals("vertices"))
		{
			int index = Integer.parseInt(appendix.substring(9, appendix.length() - 1));
			if (index >= 0 && index < vertexPosition.getLength())
			{
				return true;
			}
		} else if (appendix.substring(0, 2).equals("uv"))
		{
			int nameEnd = appendix.indexOf(']');
			String mapName = appendix.substring(3, nameEnd);
			if (uvMaps.get(mapName) != null)
			{
				// TODO check face and vertex indices
				return true;
			}
		}

		return false;
	}

	public Object getFromAppendix(String appendix)
	{
		if (appendix.substring(0, 5).equals("faces"))
		{
			int index = Integer.parseInt(appendix.substring(6, appendix.length() - 1));
			return getFace(index);
		} else if (appendix.substring(0, 5).equals("edges"))
		{
			int index = Integer.parseInt(appendix.substring(6, appendix.length() - 1));
			return getEdge(index);
		} else if (appendix.substring(0, 8).equals("vertices"))
		{
			int index = Integer.parseInt(appendix.substring(9, appendix.length() - 1));
			return getVertex(index);
		}

		return null;
	}

	@Override
	protected void loadXMLData(Element dataElement) throws Exception
	{
		// load vertices
		NodeList vertexList = dataElement.getElementsByTagName("vertex");
		Vector3D position = new Vector3D();
		int[] vertexEdgeData = new int[0];
		int vertexListLength = vertexList.getLength();
		for (int i = 0; i < vertexListLength; i++)
		{
			Element vertexElement = (Element) vertexList.item(i);

			int vertexID = Integer.parseInt(vertexElement.getAttribute("id"));

			if (i != vertexID)
			{
				throw new Exception(L10N.translate("unsorted vertex elements in mesh data"));
			}

			position.X1 = Double.parseDouble(vertexElement.getAttribute("x"));
			position.X2 = Double.parseDouble(vertexElement.getAttribute("y"));
			position.X3 = Double.parseDouble(vertexElement.getAttribute("z"));

			vertexPosition.append(position);
			vertexEdges.append(vertexEdgeData);

		}
		// load edges and rebuild vertex edgeID lists
		NodeList edgeList = dataElement.getElementsByTagName("edge");
		int edgeListLength = edgeList.getLength();
		for (int i = 0; i < edgeListLength; i++)
		{
			Element edgeElement = (Element) edgeList.item(i);
			int edgeID = Integer.parseInt(edgeElement.getAttribute("id"));
			int[] vertexIDs = new int[2];
			vertexIDs[0] = Integer.parseInt(edgeElement.getAttribute("vertex1"));
			vertexIDs[1] = Integer.parseInt(edgeElement.getAttribute("vertex2"));
			edgeVertices.append(vertexIDs[0]);
			edgeVertices.append(vertexIDs[1]);

			for (int j = 0; j < 2; j++)
			{
				vertexEdges.resize(vertexIDs[j], vertexEdges.getArrayLength(vertexIDs[j]) + 1);
				vertexEdges.setArrayElement(vertexIDs[j],
					vertexEdges.getArrayLength(vertexIDs[j]) - 1, edgeID);

			}

			String faceIDsString = edgeElement.getAttribute("faceIDs");
			int[] faceIDs;
			if (faceIDsString.equals(""))
			{
				faceIDs = new int[0];
			} else
			{
				String[] faceIDStrings = faceIDsString.split(";");
				faceIDs = new int[faceIDStrings.length];
				for (int j = 0; j < faceIDs.length; j++)
				{
					faceIDs[j] = Integer.parseInt(faceIDStrings[j]);
				}
			}
			if (edgeFaces.append(faceIDs) != edgeID)
			{
				throw new Exception(L10N.translate("unsorted edge elements in mesh data"));
			}
		}
		// load faces and recalculate triangulation
		NodeList faceList = dataElement.getElementsByTagName("face");
		int faceListLength = faceList.getLength();
		for (int i = 0; i < faceListLength; i++)
		{
			Element faceElement = (Element) faceList.item(i);
			int faceID;
			int[] faceEdgeIDs;

			faceID = Integer.parseInt(faceElement.getAttribute("id"));

			String edgeIDsString = faceElement.getAttribute("edgeIDs");
			if (edgeIDsString.equals(""))
			{
				throw new Exception(L10N.translate("invalid face with no edges found in mesh data"));
			} else
			{
				String[] edgeIDs = edgeIDsString.split(";");
				faceEdgeIDs = new int[edgeIDs.length];
				for (int j = 0; j < edgeIDs.length; j++)
				{
					faceEdgeIDs[j] = Integer.parseInt(edgeIDs[j]);
				}
			}

			if (faceEdges.append(faceEdgeIDs) != faceID)
			{
				throw new Exception(L10N.translate("unsorted face elements in mesh data"));
			}
		}

		// initialise normals array buffer
		for (int i = 0; i < getNumFaces(); i++)
		{
			Vector3D[] normalsArray = new Vector3D[faceEdges.getArrayLength(i)];
			for (int j = 0; j < normalsArray.length; j++)
			{
				normalsArray[j] = new Vector3D(0, 0, 0);
			}
			normals.append(normalsArray);
			byte[] normalModeArray = new byte[faceEdges.getArrayLength(i)];
			for (int j = 0; j < normalModeArray.length; j++)
			{
				normalModeArray[j] = NORMAL_MODE_SMOOTH;
			}
			normalMode.append(normalModeArray);
		}

		// load normals if present in file
		NodeList normalsList = dataElement.getElementsByTagName("normals");
		int normalsListLength = normalsList.getLength();
		if (normalsListLength > 0)
		{
			for (int i = 0; i < normalsListLength; i++)
			{
				Element normalElement = (Element) normalsList.item(i);

				int faceID = Integer.parseInt(normalElement.getAttribute("faceID"));
				int vertexIndex = Integer.parseInt(normalElement.getAttribute("vertexIndex"));
				byte mode = Byte.parseByte(normalElement.getAttribute("mode"));
				Vector3D normal = new Vector3D();
				normal.X1 = Double.parseDouble(normalElement.getAttribute("x"));
				normal.X2 = Double.parseDouble(normalElement.getAttribute("y"));
				normal.X3 = Double.parseDouble(normalElement.getAttribute("z"));

				normalMode.setArrayElement(faceID, vertexIndex, mode);
				normals.setArrayElement(faceID, vertexIndex, normal);
			}
		} else
		{
			// no normals present in file, recalculate from scratch
			transformFaceVertexNormals();
		}

		// load uv coordinates
		NodeList uvMapsList = dataElement.getElementsByTagName("uvmap");
		int uvMapsListLength = uvMapsList.getLength();
		if (uvMapsListLength > 0)
		{
			for (int i = 0; i < uvMapsListLength; i++)
			{
				Element uvMapElement = (Element) uvMapsList.item(i);

				String mapName = uvMapElement.getAttribute("name");
				if (mapName.equals(""))
				{
					throw new Exception(L10N.translate("found an UV map with no name"));
				}

				addUVMap(mapName);

				NodeList uvList = uvMapElement.getElementsByTagName("uv");
				int uvListLength = uvList.getLength();
				for (int j = 0; j < uvListLength; j++)
				{
					Element uvElement = (Element) uvList.item(j);
					int faceID = Integer.parseInt(uvElement.getAttribute("faceID"));
					int vertexIndex = Integer.parseInt(uvElement.getAttribute("vertexIndex"));
					double u = Double.parseDouble(uvElement.getAttribute("u"));
					double v = Double.parseDouble(uvElement.getAttribute("v"));
					setUVCoordinates(mapName, faceID, vertexIndex, u, v);
				}
			}
		}
	}

	@Override
	protected void saveXMLData(Document document, Element dataElement)
	{
		// save vertices (don't save edgeIDs, though as these are rebuilt during load)
		double[] vertexPositionData = vertexPosition.getRawData();
		for (int i = 0; i < vertexPosition.getLength(); i++)
		{
			Element vertexElement = document.createElement("vertex");
			vertexElement.setAttribute("id", Integer.toString(i));
			vertexElement.setAttribute("x", Double.toString(vertexPositionData[3 * i]));
			vertexElement.setAttribute("y", Double.toString(vertexPositionData[3 * i + 1]));
			vertexElement.setAttribute("z", Double.toString(vertexPositionData[3 * i + 2]));

			dataElement.appendChild(vertexElement);
		}
		for (int i = 0; i < edgeFaces.getNumArrays(); i++)
		{
			Element edgeElement = document.createElement("edge");
			edgeElement.setAttribute("id", Integer.toString(i));
			edgeElement.setAttribute("vertex1", Integer.toString(edgeVertices.getElement(i * 2)));
			edgeElement.setAttribute("vertex2", Integer
					.toString(edgeVertices.getElement(i * 2 + 1)));

			if (edgeFaces.getArrayLength(i) == 0)
			{
				edgeElement.setAttribute("faceIDs", "");
			} else
			{
				int[] faceIDs = edgeFaces.getArray(i);
				String faceIDsString = Integer.toString(faceIDs[0]);
				for (int j = 1; j < faceIDs.length; j++)
				{
					faceIDsString = faceIDsString + ";" + Integer.toString(faceIDs[j]);
				}
				edgeElement.setAttribute("faceIDs", faceIDsString);
			}

			dataElement.appendChild(edgeElement);
		}
		// save faces (the face triangulation will be recalculated on load)
		for (int i = 0; i < faceEdges.getNumArrays(); i++)
		{
			Element faceElement = document.createElement("face");
			faceElement.setAttribute("id", Integer.toString(i));

			int[] edgeIDs = faceEdges.getArray(i);
			String edgeIDsString = Integer.toString(edgeIDs[0]);
			for (int j = 1; j < edgeIDs.length; j++)
			{
				edgeIDsString = edgeIDsString + ";" + Integer.toString(edgeIDs[j]);
			}
			faceElement.setAttribute("edgeIDs", edgeIDsString);

			dataElement.appendChild(faceElement);
		}

		// save face vertex normals
		for (int i = 0; i < getNumFaces(); i++)
		{
			for (int j = 0; j < faceEdges.getArrayLength(i); j++)
			{
				Element normalElement = document.createElement("normal");
				normalElement.setAttribute("faceID", Integer.toString(i));
				normalElement.setAttribute("vertexIndex", Integer.toString(j));
				normalElement.setAttribute("mode", Integer.toString(normalMode
						.getArrayElement(i, j)));
				normalElement.setAttribute("x", Double.toString(normals.getArrayElement(i, j).X1));
				normalElement.setAttribute("y", Double.toString(normals.getArrayElement(i, j).X2));
				normalElement.setAttribute("z", Double.toString(normals.getArrayElement(i, j).X3));

				dataElement.appendChild(normalElement);
			}
		}

		// save UV maps
		for (String mapName : uvMaps.keySet())
		{
			Element uvMapElement = document.createElement("uvmap");

			uvMapElement.setAttribute("name", mapName);

			DoubleArrayBuffer uvMap = uvMaps.get(mapName);

			for (int i = 0; i < getNumFaces(); i++)
			{
				for (int j = 0; j < faceEdges.getArrayLength(i); j++)
				{
					Element uvElement = document.createElement("uv");
					uvElement.setAttribute("faceID", Integer.toString(i));
					uvElement.setAttribute("vertexIndex", Integer.toString(j));
					uvElement.setAttribute("u", Double.toString(uvMap.getArrayElement(i, 2 * j)));
					uvElement.setAttribute("v", Double
							.toString(uvMap.getArrayElement(i, 2 * j + 1)));

					uvMapElement.appendChild(uvElement);
				}
			}

			dataElement.appendChild(uvMapElement);
		}
	}
}
