/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 31, 2006
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.nio.DoubleBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import javax.media.opengl.GL;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.preferences.Preferences;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.selection.Manager;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.ui.mainwindow.view.RenderableNode;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Scene;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Viewport;

/**
 * A schematic representation of a mesh in the 3D view scene
 * graph. This class draws a representation of the mesh in the
 * 3D view.
 * 
 * @see eu.moonlight3d.ml3d.ui.mainwindow.view.Renderable
 *
 * @author gregor
 */
public class MeshRenderable extends RenderableNode {
	/**
	 * The colour to render selected mesh elements in
	 */
	private Colour selectedColour;
	/**
	 * The colour to render active mesh elements in
	 */
	private Colour activeColour;
	/**
	 * True if the mesh object as a whole is selected. 
	 */
    private boolean selected;
    /**
     * True if the mesh object as a whole is active.
     */
    private boolean active;
    
    /**
     * The mesh to render.
     */
    private Mesh mesh;
    
    /**
     * True if the display lists were already created,
     * false otherwise.
     */
    private boolean displayListsRendered;
    /**
     * The OpenGL name of the mesh preview display list
     */
    private int previewList;
    /**
     * The OpenGL name of the display lists for the.
     * selected mesh elements
     */
    private int selectionList;

    /**
     * Array containing all buffer IDs that are used by this mesh.
     */
    int bufferIDs[];
    /**
     * Name of vertex buffer for this mesh.
     */
    int vertexBufferID;
    /**
     * Name of edge index buffer for this mesh.
     */
    int edgeVertexBufferID;
    
    /**
     * Default constructor.
     * 
     * @param mesh the mesh to display
     */
    public MeshRenderable(Mesh mesh) {
    	this.mesh=mesh;
		Preferences currentPreferences=eu.moonlight3d.framework.State.getInstance().getPreferencesManager().getCurrentPreferences();
		try {
			selectedColour=currentPreferences.getGroupByName("3D View").getProperty("selectedColour").getColourValue();
		} catch (Exception e) {
			// this cannot happen
		}
		try {
			activeColour=currentPreferences.getGroupByName("3D View").getProperty("activeColour").getColourValue();
		} catch (Exception e) {
			// this cannot happen
		}
		
		active=false;
		selected=false;
    }
    
    /**
     * Create OpenGL buffer objects used to draw the mesh. This function accesses the buffer objects that
     * Mesh uses internally so that no explicit copying of any memory is required.
     * 
     * @param gl the OpenGL context to create the buffers in
     * @param mesh the mesh to create the buffers for
     */
    private void createGLBuffers(GL gl, Mesh mesh) {
    	DoubleBuffer vertexBuffer;
    	IntBuffer edgeVertexBuffer;
    	
    	vertexBuffer=DoubleBuffer.wrap(mesh.getVertexBuffer().getRawData());
    	edgeVertexBuffer=IntBuffer.wrap(mesh.getEdgeVerticesBuffer().getRawData());
    	
    	bufferIDs=new int[2];
    	gl.glGenBuffers(2, bufferIDs, 0)
    	;
    	vertexBufferID=bufferIDs[0];
    	edgeVertexBufferID=bufferIDs[1];
    	
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexBufferID);
    	gl.glBufferData(GL.GL_ARRAY_BUFFER, mesh.getNumVertices()*3*Double.SIZE/8, vertexBuffer, GL.GL_STATIC_DRAW);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, edgeVertexBufferID);
    	gl.glBufferData(GL.GL_ARRAY_BUFFER, mesh.getNumEdges()*2*Integer.SIZE/8, edgeVertexBuffer, GL.GL_STATIC_DRAW);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
    }
    
    /**
     * Draw the mesh vertices using the previously created OpenGL buffer objects.
     * 
     * @param gl the OpenGL context to draw the buffers in
     * @param mesh the mesh to paint the vertices for
     */
    private void drawVertices(GL gl, Mesh mesh) {
    	gl.glEnable(GL.GL_VERTEX_ARRAY);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexBufferID);
    	gl.glVertexPointer(3, GL.GL_DOUBLE,	0, 0);
    	gl.glDrawArrays(GL.GL_POINTS, 0, mesh.getNumVertices());
    	gl.glDisable(GL.GL_VERTEX_ARRAY);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
    }
    
    /**
     * Draw the mesh edges using the previously created OpenGL buffer objects.
     * 
     * @param gl the OpenGL context to draw the buffers in
     * @param mesh the mesh to paint the edges for
     */
    private void drawEdges(GL gl, Mesh mesh) {
    	gl.glEnable(GL.GL_VERTEX_ARRAY);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexBufferID);
    	gl.glVertexPointer(3, GL.GL_DOUBLE,	0, 0);
    	gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, edgeVertexBufferID);
    	gl.glDrawElements(GL.GL_LINES,2*mesh.getNumEdges(),GL.GL_UNSIGNED_INT,0);
    	gl.glDisable(GL.GL_VERTEX_ARRAY);
    	gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
    	gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);    	
    }
    
    /**
     * Render the mesh faces of the given mesh in the given
     * OpenGL context. This function is used to build display
     * lists for the given mesh.
     * 
     * @param gl the OpenGL context to render the faces into
     * @param mesh the mesh to render the faces for
     */
    private void renderPreviewMesh(GL gl, Mesh mesh) {
    	gl.glBegin(GL.GL_TRIANGLES);
		for(int i=0;i<mesh.getNumFaces();i++) {
    		ArrayList<Integer> triangulated=mesh.triangulated(mesh.getFace(i));
    		ArrayList<Integer> vertexIDs=mesh.getVertexIDs(mesh.getFace(i));
    		for (int j = 0; j < triangulated.size(); j+=3) {
    			Vector3D pos1 = mesh.getVertexPosition(triangulated.get(j));
    			Vector3D pos2 = mesh.getVertexPosition(triangulated.get(j+1));
    			Vector3D pos3 = mesh.getVertexPosition(triangulated.get(j+2));
    			
    			Vector3D normal;
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			gl.glVertex3d(pos1.X1, pos1.X2, pos1.X3);
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j+1)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			gl.glVertex3d(pos2.X1, pos2.X2, pos2.X3);
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j+2)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			gl.glVertex3d(pos3.X1, pos3.X2, pos3.X3);
    		}
    	}
		gl.glEnd();
    }
    
    /**
     * Render the mesh faces of the given mesh in the given
     * OpenGL context. This function is used to build display
     * lists for the given mesh.
     * 
     * @param gl the OpenGL context to render the faces into
     * @param mesh the mesh to render the faces for
     * @param uvMapName
     */
    private void renderPreviewMesh(GL gl, Mesh mesh, String uvMapName) {
    	gl.glBegin(GL.GL_TRIANGLES);
		for(int i=0;i<mesh.getNumFaces();i++) {
    		ArrayList<Integer> triangulated=mesh.triangulated(mesh.getFace(i));
    		ArrayList<Integer> vertexIDs=mesh.getVertexIDs(mesh.getFace(i));
    		for (int j = 0; j < triangulated.size(); j+=3) {
    			Vector3D pos1 = mesh.getVertexPosition(triangulated.get(j));
    			Vector3D pos2 = mesh.getVertexPosition(triangulated.get(j+1));
    			Vector3D pos3 = mesh.getVertexPosition(triangulated.get(j+2));
    			
    			Vector3D normal;
    			double[] uv;
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			uv=mesh.getUVCoordinates(uvMapName, i, vertexIDs.indexOf(triangulated.get(j)));
    			gl.glTexCoord2d(uv[0], uv[1]);
    			gl.glVertex3d(pos1.X1, pos1.X2, pos1.X3);
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j+1)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			uv=mesh.getUVCoordinates(uvMapName, i, vertexIDs.indexOf(triangulated.get(j+1)));
    			gl.glTexCoord2d(uv[0], uv[1]);
    			gl.glVertex3d(pos2.X1, pos2.X2, pos2.X3);
    			normal=mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangulated.get(j+2)));    			
    			gl.glNormal3d(normal.X1,normal.X2,normal.X3);
    			uv=mesh.getUVCoordinates(uvMapName, i, vertexIDs.indexOf(triangulated.get(j+2)));
    			gl.glTexCoord2d(uv[0], uv[1]);
    			gl.glVertex3d(pos3.X1, pos3.X2, pos3.X3);
    		}
    	}
		gl.glEnd();
    }
    
    /**
     * Render the OpenGL preview of the given mesh into
     * a display list.
     * 
     * @param gl the OpenGL context to render elements to
     * @param mesh the mesh to render structure elements for
     */
    private void renderPreviewList(GL gl, Mesh mesh) {
    	gl.glNewList(previewList,GL.GL_COMPILE);
		gl.glPushAttrib(GL.GL_ALL_ATTRIB_BITS);

		boolean hasUVCoordinates=false;
		if(mesh.getUVMaps().size()>=1) {
			hasUVCoordinates=true;
		}
		
		if(getMaterial()==null) {
			gl.glColor3d(1.0, 1.0, 1.0);
			renderPreviewMesh(gl, mesh);
	 	} else {
			for(int i=0;i<getMaterial().getNumPasses(gl);i++) {
				getMaterial().doPass(gl,i);
				if(hasUVCoordinates) {
					// render with UV coordinates if requested and the given map exists
					String uvMapName=getMaterial().getUVMapForPass(gl, i);
					if(uvMapName!=null && mesh.getUVMaps().contains(uvMapName)) {
						renderPreviewMesh(gl, mesh, uvMapName);		
					} else {
						renderPreviewMesh(gl, mesh);
					}
				} else {
					renderPreviewMesh(gl, mesh);
				}
			}
			getMaterial().reset(gl);
		}

		gl.glPopAttrib();
    	gl.glEndList();    	
    }
    
    /**
     * Render the display list containing all selected mesh elements.
     * 
     * @param gl the OpenGL context to render the elements to
     * @param mesh the mesh to render the selected elements for
     */
    private void renderSelectionList(GL gl, Mesh mesh) {
    	gl.glNewList(selectionList,GL.GL_COMPILE);
		try {
	        eu.moonlight3d.ml3d.backend.selection.Manager selectionManager=((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager();
	        eu.moonlight3d.ml3d.backend.selection.Selection selection;
			selection = selectionManager.getSelection("viewSelection");

	        ArrayList<String> selectedObjects;
	        selectedObjects=selection.getSelection();
	        
	        if(selectedObjects.size()>0) {
        		if(selection.getType()==Manager.getSelectionType("MeshVertex")) {
		    		// paint selected vertices:
		    		gl.glPointSize((float) 3.0);

		    		gl.glDisable(GL.GL_DEPTH_TEST);
		    		Colour.RGB sRGB=selectedColour.getsRGB8();
		    		gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);

		    		gl.glBegin(GL.GL_POINTS);
			    	for(int i=0;i<selectedObjects.size();i++) {
				    	String nodeName=SelectionHelper.stripAppendix(selectedObjects.get(i));
				    	if(selection.getObjectFromSelectionString(nodeName)==mesh) {
				    		Vector3D position=((eu.moonlight3d.ml3d.modelling.mesh.Mesh.Vertex)selection.getObjectFromSelectionString(selectedObjects.get(i))).position;
				    		gl.glVertex3d(position.X1,position.X2,position.X3);
				    	}
			    	}

			    	if(selection.getActiveObject()!=null && selection.getObjectFromSelectionString(SelectionHelper.stripAppendix(selection.getActiveObject()))==mesh) {
			    		sRGB=activeColour.getsRGB8();
			    		gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
			    		Vector3D position=((eu.moonlight3d.ml3d.modelling.mesh.Mesh.Vertex)selection.getObjectFromSelectionString(selection.getActiveObject())).position;
			    		gl.glVertex3d(position.X1,position.X2,position.X3);
			    	}
			    	gl.glEnd();
			    } else if(selection.getType()==Manager.getSelectionType("MeshEdge")) {
			        // paint selected edges:
		    		gl.glDisable(GL.GL_DEPTH_TEST);
		    		Colour.RGB sRGB=selectedColour.getsRGB8();
		    		gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
		    		gl.glBegin(GL.GL_LINES);
		    		for(int i=0;i<selectedObjects.size();i++) {
				    	String nodeName=SelectionHelper.stripAppendix(selectedObjects.get(i));
				    	if(selection.getObjectFromSelectionString(nodeName)==mesh) {
				    		Mesh.Edge edge;
				    		edge=(Mesh.Edge)selection.getObjectFromSelectionString(selectedObjects.get(i));
				    		Vector3D vertex1=mesh.getVertexPosition(edge.vertexID[0]);
				    		Vector3D vertex2=mesh.getVertexPosition(edge.vertexID[1]);
				    		gl.glVertex3d(vertex1.X1,vertex1.X2,vertex1.X3);
				    		gl.glVertex3d(vertex2.X1,vertex2.X2,vertex2.X3);
				    	}
		    		}
			        if(selection.getActiveObject()!=null && selection.getObjectFromSelectionString(SelectionHelper.stripAppendix(selection.getActiveObject()))==mesh) {
						sRGB=activeColour.getsRGB8();
						gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
			            Mesh.Edge edge;
			            edge=(Mesh.Edge)selection.getObjectFromSelectionString(selection.getActiveObject());
			            Vector3D vertex1=mesh.getVertexPosition(edge.vertexID[0]);
			            Vector3D vertex2=mesh.getVertexPosition(edge.vertexID[1]);				        	
			            gl.glVertex3d(vertex1.X1,vertex1.X2,vertex1.X3);
			            gl.glVertex3d(vertex2.X1,vertex2.X2,vertex2.X3);
			        }
			        gl.glEnd();
							
			        // paint edge endpoints again as points
			        sRGB=selectedColour.getsRGB8();
					gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
					gl.glPointSize(5.0f);
			    	gl.glBegin(GL.GL_POINTS);
			        for(int i=0;i<selectedObjects.size();i++) {
				    	String nodeName=SelectionHelper.stripAppendix(selectedObjects.get(i));
				    	if(selection.getObjectFromSelectionString(nodeName)==mesh) {
				    		Mesh.Edge edge;
				    		edge=(Mesh.Edge)selection.getObjectFromSelectionString(selectedObjects.get(i));
				    		Vector3D vertex1=mesh.getVertexPosition(edge.vertexID[0]);
				    		Vector3D vertex2=mesh.getVertexPosition(edge.vertexID[1]);
				    		gl.glVertex3d(vertex1.X1,vertex1.X2,vertex1.X3);
				    		gl.glVertex3d(vertex2.X1,vertex2.X2,vertex2.X3);
				    	}
			        }
			        if(selection.getActiveObject()!=null && selection.getObjectFromSelectionString(SelectionHelper.stripAppendix(selection.getActiveObject()))==mesh) {
						sRGB=activeColour.getsRGB8();
						gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);

						Mesh.Edge edge;
						edge=(Mesh.Edge)selection.getObjectFromSelectionString(selection.getActiveObject());
						Vector3D vertex1=mesh.getVertexPosition(edge.vertexID[0]);
						Vector3D vertex2=mesh.getVertexPosition(edge.vertexID[1]);				        	
						gl.glVertex3d(vertex1.X1,vertex1.X2,vertex1.X3);
						gl.glVertex3d(vertex2.X1,vertex2.X2,vertex2.X3);
			        }
			        gl.glEnd();
			    } else if(selection.getType()==Manager.getSelectionType("MeshFace")) {
			        // add selected / active faces as such to the GLMesh
		    		Colour.RGB sRGB=selectedColour.getsRGB8();
		    		gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
		    		gl.glBegin(GL.GL_TRIANGLES);
		    		for(int i=0;i<selectedObjects.size();i++) {
				    	String nodeName=SelectionHelper.stripAppendix(selectedObjects.get(i));
				    	if(selection.getObjectFromSelectionString(nodeName)==mesh) {
			    			ArrayList<Integer> triangulated=mesh.triangulated((eu.moonlight3d.ml3d.modelling.mesh.Mesh.Face)selection.getObjectFromSelectionString(selectedObjects.get(i)));
			    			for (int j = 0; j < triangulated.size(); j+=3) {
			    				Vector3D pos1 = mesh.getVertexPosition(triangulated.get(j));
			    				Vector3D pos2 = mesh.getVertexPosition(triangulated.get(j+1));
			    				Vector3D pos3 = mesh.getVertexPosition(triangulated.get(j+2));
			    				Vector3D normal=pos2.sub(pos1).crossMultiply(pos3.sub(pos1)).norm();
			    				gl.glNormal3d(normal.X1,normal.X2,normal.X3);
			    				gl.glVertex3d(pos1.X1, pos1.X2, pos1.X3);
			    				gl.glNormal3d(normal.X1,normal.X2,normal.X3);
			    				gl.glVertex3d(pos2.X1, pos2.X2, pos2.X3);
			    				gl.glNormal3d(normal.X1,normal.X2,normal.X3);
			    				gl.glVertex3d(pos3.X1, pos3.X2, pos3.X3);
			    			}	
			    		}
			    	}
			        if(selection.getActiveObject()!=null) {
				    	String nodeName=SelectionHelper.stripAppendix(selection.getActiveObject());
				    	if(selection.getObjectFromSelectionString(nodeName)==mesh) {
				    		sRGB=activeColour.getsRGB8();
							gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
							ArrayList<Integer> triangulated=mesh.triangulated((eu.moonlight3d.ml3d.modelling.mesh.Mesh.Face)selection.getObjectFromSelectionString(selection.getActiveObject()));
							for (int j = 0; j < triangulated.size(); j+=3) {
								Vector3D pos1 = mesh.getVertexPosition(triangulated.get(j));
								Vector3D pos2 = mesh.getVertexPosition(triangulated.get(j+1));
								Vector3D pos3 = mesh.getVertexPosition(triangulated.get(j+2));
								Vector3D normal=pos2.sub(pos1).crossMultiply(pos3.sub(pos1)).norm();
								gl.glNormal3d(normal.X1,normal.X2,normal.X3);
								gl.glVertex3d(pos1.X1, pos1.X2, pos1.X3);
								gl.glNormal3d(normal.X1,normal.X2,normal.X3);
								gl.glVertex3d(pos2.X1, pos2.X2, pos2.X3);
								gl.glNormal3d(normal.X1,normal.X2,normal.X3);
								gl.glVertex3d(pos3.X1, pos3.X2, pos3.X3);
							}
				    	}
			        }
			        gl.glEnd();
			    } else if(selection.getType()==Manager.getSelectionType("SGNode")) {
			    	// whole mesh selected/active
			    	for(int i=0;i<selectedObjects.size();i++) {
			    		if(selection.getObjectFromSelectionString(selectedObjects.get(i))==mesh) {
			    			
							Colour.RGB sRGB;
							if(selection.getObjectFromSelectionString(selection.getActiveObject())==mesh) {
								active=true;
								sRGB=activeColour.getsRGB8();
							} else {
								selected=true;
								sRGB=selectedColour.getsRGB8();
							}
							gl.glColor3d(sRGB.red/256.0,sRGB.green/256.0,sRGB.blue/256.0);
							
			    			renderPreviewMesh(gl, mesh);
			    			break;
			    		}
			    	}
			    }                
	        }
		} catch (Exception e1) {
			// this can happen if the selection contains elements that no longer exist - should be totally ignored
			e1.printStackTrace();
		}
        
    	gl.glEndList();    	
    }
    
    /**
     * Render display lists for the given mesh.
     * 
     * @param gl reference to the GL context to build display lists in
     * @param mesh the mesh to build display lists for
     */
    private void renderDisplayLists(GL gl, Mesh mesh) {
    	if(displayListsRendered) {
    		return;
    	}
    	
    	createGLBuffers(gl, mesh);
    	
    	previewList=gl.glGenLists(2);
    	selectionList=previewList+1;

    	renderPreviewList(gl,mesh);
    	renderSelectionList(gl,mesh);
    	
    	displayListsRendered=true;
    }
    
    @Override
	public void dispose(GL gl) {
    	if(displayListsRendered) {
    		gl.glDeleteLists(previewList, 2);
    		gl.glDeleteBuffers(2, bufferIDs, 0);
    	}
    }
    
	@Override
	public void drawPreview(Viewport view, Scene scene) {
		// Skip preview if we're in any other rendering mode than Solid.
		// The structure pass will draw all edges and vertices, anyway.
		if (view.getCurrentPreviewMode() == Viewport.PreviewMode.None
				|| view.getCurrentPreviewMode() == Viewport.PreviewMode.Wireframe) {
			return;
		}

		if (active || selected) {
			return;
		}

		GL gl = view.getGL();
		renderDisplayLists(gl, mesh);
		gl.glCallList(previewList);
	}

	@Override
	public void drawVisibleStructure(Viewport view, Scene scene) {
		GL gl = view.getGL();

		renderDisplayLists(gl, mesh);
		
		if(view.getCurrentStructureMode()==Viewport.StructureMode.Full || view.getCurrentStructureMode()==Viewport.StructureMode.Occluded) {
			gl.glEnable(GL.GL_DEPTH_TEST);
			gl.glDepthMask(false);

			// draw visible vertices and edges in white:
			gl.glColor3d(1.0, 1.0, 1.0);
			drawVertices(gl, mesh);
			drawEdges(gl, mesh);
			
			gl.glDisable(GL.GL_DEPTH_TEST);
			gl.glDepthMask(true);

		}
		gl.glCallList(selectionList);
	}
	
	@Override
	public void drawOccludedStructure(Viewport view, Scene scene) {
		GL gl = view.getGL();

		renderDisplayLists(gl, mesh);
		
		if(view.getCurrentStructureMode()==Viewport.StructureMode.Full) {
			gl.glEnable(GL.GL_DEPTH_TEST);
			gl.glDepthMask(false);

			gl.glDepthFunc(GL.GL_GREATER);
			
			// draw occluded vertices in black:
			gl.glColor3d(0.0, 0.0, 0.0);
			drawVertices(gl, mesh);
			drawEdges(gl,mesh);

			gl.glDepthFunc(GL.GL_LESS);
			gl.glDisable(GL.GL_DEPTH_TEST);
			gl.glDepthMask(true);
		}
	}

	@Override
	public void drawOverlay(Viewport view, Scene scene) {
		// no overlay to draw here
	}

	@Override
	public void drawOverlay2D(Viewport view, Scene scene) {
		// no overlay to draw here
	}
}
