/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Jan 15, 2007
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;

import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.OperatorGraphHelper;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool;

/**
 * A modeless tool that splits the selected egdges in a mesh into
 * several segments.
 *
 * @see eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool
 *
 * @author gregor
 */
public class SplitFaceTool extends Tool {
	/**
	 * Default constructor
	 */
	public SplitFaceTool() {
		super();
		setName("SplitFaceTool");
		setDisplayName(L10N.translate("Split Face"));
		setCategory("Edit/Mesh");
	}

	private ArrayList<Integer> getFaces(Mesh mesh, int edges[]) {
		ArrayList<Integer> faces=new ArrayList<Integer>();

		for(int i=0;i<edges.length;i++) {
			int edgeFaces[]=mesh.getEdgeFaces(edges[i]);
			for(int j=0;j<edgeFaces.length;j++) {
				if(!faces.contains(edgeFaces[j])) {
					faces.add(edgeFaces[j]);
				}
			}
		}

		return faces;
	}

	@Override
	public void activated() {
        try {
            State state=eu.moonlight3d.framework.State.getInstance();

            // don't do anything if selected object is not a SG node
    		Selection selection = ((Document) state.getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
    		if (selection.getType() != eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("MeshVertex") ||
    			selection.getSelection().size()!=2) {
    			state.logger().warning(L10N.translate("current selection must consist of two mesh vertices"));
    			return;
    		}

    		String selectionName = selection.getSelection().get(0);
    		String sgNodeName = SelectionHelper.getSGNodeName(selectionName);

    		if(!SelectionHelper.getSGNodeName(selection.getSelection().get(0)).equals(SelectionHelper.getSGNodeName(selection.getSelection().get(0)))) {
    			state.logger().warning(L10N.translate("the selected vertices do not belong to the same mesh"));
    			return;
    		}

    		Manager ogManager = ((Document) state.getDocumentManager().getMainDocument()).getOGManager();

            ArrayList<Object> selectedObjects=selection.getSelectionObjects();

            // get vertex ids
            int vertices[]=new int[2];
            vertices[0]=((Mesh.Vertex)selectedObjects.get(0)).vertexID;
            vertices[1]=((Mesh.Vertex)selectedObjects.get(1)).vertexID;
            // find mesh
            Mesh mesh=(Mesh) selection.getObjectFromSelectionString(selectionName.split(":")[0]);
            // check if an edge already exists between the vertices
        	if (mesh.findEdge(vertices[0], vertices[1]) >= 0)
        	{
        		// if findEdge does not throw an exception there is already an edge betweent the two vertices - abort
        		state.logger().warning(L10N.translate("there already exists an edge between the two selected vertices"));
        		return;
        	}

            // find the common face of the two vertices
            int edges1[]=mesh.getVertexEdges(vertices[0]);
            int edges2[]=mesh.getVertexEdges(vertices[1]);

            ArrayList<Integer> faces1=getFaces(mesh,edges1);
            ArrayList<Integer> faces2=getFaces(mesh,edges2);

            int faceID=-1;
            for(int i=0;i<faces1.size();i++) {
            	if(faces2.contains(faces1.get(i))) {
            		faceID=faces1.get(i);
            		break;
            	}
            }

            if(faceID==-1) {
            	state.logger().warning(L10N.translate("selected vertices do not have a common face"));
            	return;
            }

    		// set OG node properties
    		Node ogNode = OperatorGraphHelper.insert(selectionName, "SplitFace", "input", "output");
    		ogNode.getProperty("meshName").setValue(sgNodeName);
    		ogNode.getProperty("face").setValue(faceID);
    		ogNode.getProperty("vertices").setArrayValue(0, vertices[0]);
    		ogNode.getProperty("vertices").setArrayValue(1, vertices[1]);

            // clear selection to prevent rogue selection items
			ogManager.update();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}

	@Override
	public void deactivated() {
	}

	@Override
	public QWidget getOptionEditor(QWidget parent) {
		return null;
	}

	@Override
	public boolean hasOptionEditor() {
		return false;
	}

}
