/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 16, 2007
 */
package eu.moonlight3d.ml3d.modelling.mesh;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;


/**
 * An operator graph node that applies Catmull-Clark subdivision to a
 * mesh.
 *
 * @author gregor
 */
public class SubdivideMesh extends Node {
	/**
	 * The name of the mesh to subdivide
	 */
	private Property name;
	/**
	 * The number of subdivision iterations
	 */
	private Property iterations;
	
	/**
	 * Default constructor.
	 */
	public SubdivideMesh() {
		super();
		
		createSlot("input",Slot.Type.Input);
		createSlot("output",Slot.Type.Output);

        name=new Property(this);
        name.setName("name");
        name.setType(Property.Type.String);
        try {
            name.setValue("Nothing");
        } catch (Exception e) {
            // This exception shouldn't occur at all in this limited case
            e.printStackTrace();
        }
        iterations=new Property(this);
        iterations.setName("iterations");
        iterations.setType(Property.Type.Integer);
        try {
            iterations.setValue(1);
        } catch (Exception e1) {
            // This exception shouldn't occur at all in this limited case
            e1.printStackTrace();
        }
	}
	
	/**
	 * Generate all face vertices by averaging all vertex coodinates of all vertices of a face
	 * to find the face midpoint.
	 * 
	 * @param mesh the mesh to work on
	 * @param faceVertices the array of face vertices that is filled in by this function
	 * @param faceVertexMap the map of existing faces to their face vertices that is filled in
	 * by this function
	 */
	private void generateFaceVertices(Mesh mesh, ArrayList<Integer> faceVertices, HashMap<Integer,Integer> faceVertexMap) {
		for(int i=0;i<mesh.getNumFaces();i++) {
			// generate face vertex for given face by averaging the vertex coodinates
			Mesh.Face face=mesh.getFace(i);
			ArrayList<Integer> vertices=mesh.getVertexIDs(face);
			Vector3D position=new Vector3D();
			position.clear();
			for(int vertexID : vertices) {
				position=position.add(mesh.getVertexPosition(vertexID));
			}
			position=position.multiply(1.0/vertices.size());
			
			int faceVertexID=mesh.addVertex(position);
			faceVertices.add(faceVertexID);
			faceVertexMap.put(face.faceID, faceVertexID);
		}
	}
	
	/**
	 * Generate edge vertices for the given mesh by averaging the edge's endpoint vertices and
	 * all surrounding face vertices of the edge to find the position of the new vertex.
	 * 
	 * @param mesh the mesh to work on
	 * @param faceVertexMap the mapping from faces to face vertices created by generateFaceVertices
	 * @param edgeVertices the array of edge vertices that is filled in by this function
	 * @param edgeVertexMap the map of existing edges to their edge vertices that is filled in 
	 * by this function
	 */
	private void generateEdgeVertices(Mesh mesh, HashMap<Integer,Integer> faceVertexMap, ArrayList<Integer> edgeVertices, HashMap<Integer, Integer> edgeVertexMap) {
		for(int i=0;i<mesh.getNumEdges();i++) {
			// generate edge vertex for given edge
			Mesh.Edge edge=mesh.getEdge(i);
			Vector3D position=new Vector3D();
			position.clear();
			// average over neighbouring face midpoints as well as the edge endpoints
			int count=2;
			if(edge.faceIDs!=null) {
				for(int faceID : edge.faceIDs) {
					Vector3D faceVertex=mesh.getVertexPosition(faceVertexMap.get(faceID));
					position=position.add(faceVertex);
				}
				count+=edge.faceIDs.length;
			}
			position=position.add(mesh.getVertexPosition(edge.vertexID[0]).add(mesh.getVertexPosition(edge.vertexID[1])));
			position=position.multiply(1.0/count);
			
			int edgeVertexID=mesh.addVertex(position);
			edgeVertices.add(edgeVertexID);
			edgeVertexMap.put(edge.edgeID, edgeVertexID);
		}
	}
	
	/**
	 * Move all original vertices of the mesh to their final positions. This is done by
	 * first creating the average of the face vertex positions \f$\vec{v}_{F}\f$ of the
	 * \f$n_{F}\f$ adjacent faces and the average of the edge vertex positions 
	 * \f$\vec{v}_{E}\f$ of the \f$n_{E}\f$ adjacent edges and then calculating the final 
	 * vertex position \f$\vec{v'}\f$ using the formula 
	 * \f$\vec{v'}=\frac{\vec{v}+2\vec{v}_{e}+(n_{E}-3) \vec{v} + \vec{v}_{F}}{n_{E}}.\f$.  
	 * 
	 * @param mesh the mesh to work on
	 * @param numOriginalVertices the number of original vertices in the mesh
	 * @param faceVertexMap the face to face vertex map as created by generateFaceVertices
	 * @param edgeVertexMap the face to face vertex map as created by generateEdgeVertices
	 */
	private void moveOriginalVertices(Mesh mesh, int numOriginalVertices, HashMap<Integer,Integer> faceVertexMap, HashMap<Integer,Integer> edgeVertexMap) {
		HashMap<Integer, Vector3D> newVertexPositions=new HashMap<Integer, Vector3D>();
		
		for(int i=0;i<numOriginalVertices;i++) {
			Mesh.Vertex vertex=mesh.getVertex(i);
			// find neighbouring faces
			ArrayList<Integer> adjacentFaces=null;
			try {
				adjacentFaces=mesh.getAdjacentFaces(vertex.vertexID);
			} catch (Exception e) {
				// this exception shouldn't happen here - it's a controlled environment
				e.printStackTrace();
			}
			// find neighbouring face vertices and edge vertices
			ArrayList<Integer> adjacentFaceVertices=new ArrayList<Integer>();
			for(int faceID : adjacentFaces) {
				adjacentFaceVertices.add(faceVertexMap.get(faceID));
			}
			// calculate face point average
			Vector3D averageFaceVertexPosition=new Vector3D();
			averageFaceVertexPosition.clear();
			for(int vertexID : adjacentFaceVertices) {
				averageFaceVertexPosition=averageFaceVertexPosition.add(mesh.getVertexPosition(vertexID));
			}
			averageFaceVertexPosition=averageFaceVertexPosition.multiply(1.0/adjacentFaceVertices.size());
			// calculate edge point average
			Vector3D averageEdgeVertexPosition=new Vector3D();
			averageEdgeVertexPosition.clear();
			if(vertex.edgeIDs!=null) {
				for(int edgeID : vertex.edgeIDs) {
					Mesh.Edge edge=mesh.getEdge(edgeID);
					averageEdgeVertexPosition=averageEdgeVertexPosition.add(mesh.getVertexPosition(edge.vertexID[0]).multiply(1.0/(2*vertex.edgeIDs.length)));
					averageEdgeVertexPosition=averageEdgeVertexPosition.add(mesh.getVertexPosition(edge.vertexID[1]).multiply(1.0/(2*vertex.edgeIDs.length)));
				}
			}
			Vector3D newVertexPosition;
			// calculate new vertex position
			newVertexPosition=averageFaceVertexPosition.clone();
			newVertexPosition=newVertexPosition.add(averageEdgeVertexPosition.multiply(2));
			newVertexPosition=newVertexPosition.add(vertex.position.multiply(vertex.edgeIDs.length-3));
			newVertexPosition=newVertexPosition.multiply(1.0/vertex.edgeIDs.length);

			newVertexPositions.put(vertex.vertexID,newVertexPosition);
		}
		
		for(int vertexID : newVertexPositions.keySet()) {
			mesh.setVertexPosition(vertexID, newVertexPositions.get(vertexID));
		}
	}
	
	/**
	 * Build the new edges and faces for the subdivided geometry.
	 * 
	 * @param mesh the mesh to work on
	 * @param edgeVertexMap the edge to edge vertex map as created by generateEdgeVertices
	 * @param faceVertexMap the face to face vertex map as created by generateFaceVertices
	 * @param oldEdges a list of old egdes which must be removed that is created
	 * by this function
	 */
	private void buildNewEdges(Mesh mesh, HashMap<Integer, Integer> edgeVertexMap, HashMap<Integer,Integer> faceVertexMap, ArrayList<Integer> oldEdges) {
		ArrayList<Integer> newEdges=new ArrayList<Integer>();
		ArrayList<ArrayList<Integer> > newFaces=new ArrayList<ArrayList<Integer> >();
		
		for(int i=0;i<mesh.getNumEdges();i++) {
			Mesh.Edge edge=mesh.getEdge(i);
			
			int edgeVertex=edgeVertexMap.get(edge.edgeID);
			
			newEdges.add(edge.vertexID[0]);
			newEdges.add(edgeVertex);
			newEdges.add(edge.vertexID[1]);
			newEdges.add(edgeVertex);
			
			if(edge.faceIDs!=null) {
				for(int faceID : edge.faceIDs) {
					int faceVertex=faceVertexMap.get(faceID);
					newEdges.add(edgeVertex);
					newEdges.add(faceVertex);
				}
			}
			
			oldEdges.add(edge.edgeID);
		}
		
		for(int i=0;i<newEdges.size();i+=2) {
			try {
				mesh.addEdge(newEdges.get(i), newEdges.get(i+1));
			} catch (Exception e) {
				// should not happen here
				e.printStackTrace();
			}
		}
		
		for(int j=0;j<mesh.getNumFaces();j++) {
			Mesh.Face face=mesh.getFace(j);
			int faceVertexID=faceVertexMap.get(face.faceID);
			for(int i=0;i<face.edgeIDs.length;i++) {
				Mesh.Edge oldFirstEdge=mesh.getEdge(face.edgeIDs[i]);
				Mesh.Edge oldSecondEdge=mesh.getEdge(face.edgeIDs[(i+1)%face.edgeIDs.length]);
				int firstEdgeVertex=edgeVertexMap.get(face.edgeIDs[i]);
				int secondEdgeVertex=edgeVertexMap.get(face.edgeIDs[(i+1)%face.edgeIDs.length]);
				int commonVertexID=-1;
				if(oldFirstEdge.vertexID[0]==oldSecondEdge.vertexID[1] || oldFirstEdge.vertexID[0]==oldSecondEdge.vertexID[0]) {
					commonVertexID=oldFirstEdge.vertexID[0];
				} else {
					commonVertexID=oldFirstEdge.vertexID[1];
				}
				try {
					ArrayList<Integer> newFace=new ArrayList<Integer>();
					int firstEdge=mesh.findEdge(commonVertexID, firstEdgeVertex);
					int secondEdge=mesh.findEdge(commonVertexID, secondEdgeVertex);
					int firstRadialEdge=mesh.findEdge(faceVertexID,firstEdgeVertex);
					int secondRadialEdge=mesh.findEdge(faceVertexID,secondEdgeVertex);
					newFace.add(firstRadialEdge);
					newFace.add(firstEdge);
					newFace.add(secondEdge);
					newFace.add(secondRadialEdge);
					newFaces.add(newFace);
				} catch (Exception e) {
					// this should not happen
					e.printStackTrace();
				}
			}
		}
		
		for(ArrayList<Integer> newFace : newFaces) {
			try {
				mesh.addFace(newFace);
			} catch (Exception e) {
				// this should not happen
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Delete the given old edges, that is, those that were replaced with
	 * subdivided ones, from the given mesh.
	 * 
	 * @param mesh the mesh to remove the edges from
	 * @param oldEdges the list of old edges to remove
	 */
	private void deleteOldEdges(Mesh mesh, ArrayList<Integer> oldEdges) {
		Integer[] sortedEdges=new Integer[0];
		sortedEdges=oldEdges.toArray(sortedEdges);
		Arrays.sort(sortedEdges);
		
		for(int i=sortedEdges.length-1;i>=0;i--) {
			try {
				mesh.removeEdge(sortedEdges[i]);
			} catch (Exception e) {
				// should not happen here
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public void update() throws Exception {
		SceneGraph graph;
        eu.moonlight3d.ml3d.backend.threedim.sg.Node node;
        
        graph=getSlot("input").getGraph();
        node=graph.getNodeByName(name.getStringValue());
        if(!(node instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh)) {
            throw new Exception(L10N.translate("SubdivideMesh can only operate on mesh nodes"));
        }
        Mesh mesh=(Mesh)node;
        
        mesh.setAutoNormalsUpdate(false);
        
        for(int i=0;i<iterations.getIntegerValue();i++) {
        	long startTime=System.currentTimeMillis();
        	
			// set reserves - this is an estimate only for the first iteration
			int vertexReserve=mesh.getNumEdges()+mesh.getNumFaces();
			int edgeReserve=2*mesh.getNumEdges()+4*mesh.getNumFaces();
			int faceReserve=4*mesh.getNumFaces();
			int faceEdgesReserve=3*mesh.getNumFaces();
			mesh.setReserves(vertexReserve,	edgeReserve, faceReserve, faceEdgesReserve);
			
			int numOriginalVertices=mesh.getNumVertices();
			
        	ArrayList<Integer> faceVertices=new ArrayList<Integer>();
        	HashMap<Integer,Integer> faceVertexMap=new HashMap<Integer,Integer>();
        	ArrayList<Integer> edgeVertices=new ArrayList<Integer>();
        	HashMap<Integer,Integer> edgeVertexMap=new HashMap<Integer,Integer>();
    		ArrayList<Integer> oldEdges=new ArrayList<Integer>();
    		
        	generateFaceVertices(mesh,faceVertices,faceVertexMap);
        	generateEdgeVertices(mesh,faceVertexMap,edgeVertices,edgeVertexMap);
        	moveOriginalVertices(mesh,numOriginalVertices,faceVertexMap,edgeVertexMap);
        	buildNewEdges(mesh,edgeVertexMap,faceVertexMap,oldEdges);
        	deleteOldEdges(mesh,oldEdges);
        	
        	long endTime=System.currentTimeMillis();
        	System.out.println("iteration " + i + ": " + (endTime-startTime));
        }

        mesh.resetReserves();
        mesh.setAutoNormalsUpdate(true);
        
        getSlot("output").setGraph(graph);
	}

	@Override
	public void removeDependencies() {
	}
}
