/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Oct 19, 2007
 */

package eu.moonlight3d.ml3d.modelling.objects;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.backend.threedim.sg.TransformableNode;

public class DuplicateObjects extends Node {

	private Property name;
	private Property displacement;
	private Property rotationAxis;
	private Property rotationAngle;
	private Property scale;
	private Property copies;
	private Property copyChildren;
	
	public DuplicateObjects() {
		createSlot("input",Slot.Type.Input);
		createSlot("output",Slot.Type.Output);
		
		name=new Property(this);
		name.setName("name");
		name.setType(Property.Type.StringArray);
		
		displacement=new Property(this);
		displacement.setName("displacement");
		displacement.setType(Property.Type.Vector3D);
		try {
			Vector3D blank=new Vector3D();
			blank.clear();
			displacement.setValue(blank);
		} catch (Exception e) {
			// This exception cannot happen here
			e.printStackTrace();
		}
		
		rotationAxis=new Property(this);
		rotationAxis.setName("rotationAxis");
		rotationAxis.setType(Property.Type.Vector3D);
		try {
			Vector3D axis=new Vector3D();
			axis.clear();
			axis.X1=1;
			rotationAxis.setValue(axis);
		} catch (Exception e) {
			// This exception cannot happen here
			e.printStackTrace();
		}
		
		rotationAngle=new Property(this);
		rotationAngle.setName("rotationAngle");
		rotationAngle.setType(Property.Type.Float);
		try {
			rotationAngle.setValue(0.0);
		} catch(Exception e) {			
			// This exception cannot happen here
			e.printStackTrace();
		}
		
		scale=new Property(this);
		scale.setName("scale");
		scale.setType(Property.Type.Vector3D);
		try {
			Vector3D scaleValue=new Vector3D();
			scaleValue.X1=scaleValue.X2=scaleValue.X3=1;
			scale.setValue(scaleValue);
		} catch(Exception e) {
			// This exception cannot happen here
			e.printStackTrace();			
		}
		
		copies=new Property(this);
		copies.setName("copies");
		copies.setType(Property.Type.Integer);
		try {
			copies.setValue(1);
		} catch (Exception e) {
			// This exception cannot happen here
			e.printStackTrace();			
		}
		
		copyChildren=new Property(this);
		copyChildren.setName("copyChildren");
		copyChildren.setType(Property.Type.Boolean);
		try {
			copyChildren.setValue(false);
		} catch (Exception e) {
			// This exception cannot happen here
			e.printStackTrace();			
		}
	}
	
	@Override
	public void update() throws Exception {
		SceneGraph sceneGraph=getSlot("input").getGraph();

		for(int i=0;i<name.getSize();i++) {
			String nodeName=name.getArrayStringValue(i);
			eu.moonlight3d.ml3d.backend.threedim.sg.Node sgNode=sceneGraph.getNodeByName(nodeName);
			
			if(sgNode instanceof TransformableNode) {
				for(int j=0;j<copies.getIntegerValue();j++) {
					TransformableNode copy=(TransformableNode) sgNode.clone();
					if(copyChildren.getBooleanValue()) {
						// create deep copy of children for duplicate:
						for(eu.moonlight3d.ml3d.backend.threedim.sg.Node child : sgNode.getChildren()) {
							sceneGraph.removeChild(copy, child);
							eu.moonlight3d.ml3d.backend.threedim.sg.Node clonedChild=child.clone();
							String cloneName=sceneGraph.generateUniqueNodeName(child.getName());
							clonedChild.setName(cloneName);
							sceneGraph.addNode(clonedChild);
							sceneGraph.removeChild(sgNode,clonedChild);
							sceneGraph.addChild(copy, clonedChild);
						}
					} else {
						// add existing children as children to new node:
						for(eu.moonlight3d.ml3d.backend.threedim.sg.Node child : sgNode.getChildren()) {
							/* HACK ALERT: The duplicated node also has a copy of the list of
							 * children of the original, but the children do not have the new copy
							 * in their list of parents. Since we cannot manipulate the lists of children
							 * and parents of the nodes directly, we fix them up here by first removing the
							 * parent-child-relationship and then adding it back. This works because trying
							 * to remove nodes from the parents and children list that were not there in the
							 * first place is not considered to be an error. 
							 */
							sceneGraph.removeChild(copy, child);
							sceneGraph.addChild(copy, child);
						}
					}
					
					// transform copy:
					for(int k=0;k<j+1;k++) {
						copy.translate(displacement.getVectorValue());
						copy.rotate(rotationAxis.getVectorValue(), rotationAngle.getFloatValue());
						AxisAlignedBoundingBox boundingBox=copy.getBoundingBox();
						copy.scale(boundingBox.min.add(boundingBox.max).multiply(0.5),scale.getVectorValue());
					}
					String copyName=sgNode.getName() + "Dup" + j;
					copyName=sceneGraph.generateUniqueNodeName(copyName);
					copy.setName(copyName);
					sceneGraph.addNode(copy);
					
					for(eu.moonlight3d.ml3d.backend.threedim.sg.Node parent : sgNode.getParents()) {
						sceneGraph.addChild(parent, copy);
					}
				}
			}
		}
		
		getSlot("output").setGraph(sceneGraph);
	}

	@Override
	public void removeDependencies() {
	}

}
