/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Dec 10, 2005
 */
package eu.moonlight3d.ml3d.modelling.textures;

import java.util.ArrayList;

import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.math.AxisAlignedBoundingBox;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;


public class GenerateUVMap extends Node {
	private Property parentName;
	private Property mapName;
	private Property projectionType;
	
	public GenerateUVMap() {
		super();
		
		createSlot("input",Slot.Type.Input);
		createSlot("output",Slot.Type.Output);
		
		parentName=new Property(this);
		parentName.setName("parentName");
		parentName.setType(Value.Type.String);
		try {
			parentName.setValue("None");
		} catch (Exception e) {
			// this cannot happen as value type is correct
			e.printStackTrace();
		}
		mapName=new Property(this);
		mapName.setName("mapName");
		mapName.setType(Value.Type.String);
		try {
			mapName.setValue("UVMap");
		} catch (Exception e) {
			// this cannot happen as value type is correct
			e.printStackTrace();
		}
		projectionType=new Property(this);
		projectionType.setName("projectionType");
		projectionType.setType(Value.Type.Integer);
		try {
			projectionType.setValue(0);
		} catch (Exception e) {
			// this cannot happen as value type is correct
			e.printStackTrace();
		}
	}

	@Override
	public void update() throws Exception {
		SceneGraph graph= getSlot("input").getGraph();
		
		Mesh parentNode=(Mesh)graph.getNodeByName(parentName.getStringValue());

		String mapName=this.mapName.getStringValue();
		parentNode.addUVMap(mapName);
		
		AxisAlignedBoundingBox boundingBox=parentNode.getBoundingBox();

		switch(projectionType.getIntegerValue()) {
		case 0: // planar mapping (use x-y-plane for now, needs more flexibility)
			double width=boundingBox.max.X1-boundingBox.min.X1;
			double height=boundingBox.max.X2-boundingBox.min.X2;
			
			for(int i=0;i<parentNode.getNumFaces();i++) {
				Mesh.Face face=parentNode.getFace(i);
				ArrayList<Integer> vertexIDs=parentNode.getVertexIDs(face);
				
				for(int j=0;j<vertexIDs.size();j++) {
					Vector3D position=parentNode.getVertexPosition(vertexIDs.get(j));
					double u=(position.X1-boundingBox.min.X1)/width;
					double v=(position.X2-boundingBox.min.X2)/height;
					
					parentNode.setUVCoordinates(mapName, face.faceID, j, u, v);
				}
			}
			
			break;
		case 1: // spherical mapping
			Vector3D center=boundingBox.min.add(boundingBox.max.sub(boundingBox.min).multiply(0.5));
			
			for(int i=0;i<parentNode.getNumFaces();i++) {
				Mesh.Face face=parentNode.getFace(i);
				ArrayList<Integer> vertexIDs=parentNode.getVertexIDs(face);
				
				for(int j=0;j<vertexIDs.size();j++) {
					Vector3D position=parentNode.getVertexPosition(vertexIDs.get(j));
					Vector3D radial=position.sub(center).norm();
					double theta=Math.PI/2-Math.atan2(radial.X3, Math.sqrt(radial.X1*radial.X1+radial.X2*radial.X2));
					double phi;
					if(radial.X1!=0 && radial.X2!=0) {
						phi=Math.acos(radial.X1 / Math.sqrt(radial.X1*radial.X1+radial.X2*radial.X2));
					} else {
						phi=0;
					}
					if(radial.X2<0) { // acos alone is ambiguous
						phi=2*Math.PI-phi;
					}
					
					if(phi==0) {
						/*
						 * This case is ambigous; phi=0 is the same line as phi=2*PI and sometimes
						 * the later is the more anticipated value. We cheat here by assuming that
						 * polygons that are entirely below the X1-X3-plane (that is, they have X2<0)
						 * should get phi=2*PI. This trick makes textures wrap around spheres
						 * correctly without having backflipped faces.    
						 */
						boolean substitute=true;
						for(int k : vertexIDs) {
							if(parentNode.getVertexPosition(k).X2>0) {
								substitute=false;
								break;
							}
						}
						if(substitute) {
							phi=2*Math.PI;
						}
					}
					
					// map phi to the u coordinate, theta to the v coordinate and renormalize to [0, 1]
					
					double u=phi/(2*Math.PI);
					double v=theta/(Math.PI);
					
					parentNode.setUVCoordinates(mapName, face.faceID, j, u, v);
				}
			}
			break;
		case 2: // TODO: cubic mapping
			break;
		}
		
		getSlot("output").setGraph(graph);
	}

	@Override
	public void removeDependencies() {
	}

}
