/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Mar 27, 2007
 */
package eu.moonlight3d.ml3d.modelling.textures;

import java.util.ArrayList;
import java.util.List;

import com.trolltech.qt.core.QPoint;
import com.trolltech.qt.core.QRect;
import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QColor;
import com.trolltech.qt.gui.QComboBox;
import com.trolltech.qt.gui.QImage;
import com.trolltech.qt.gui.QMouseEvent;
import com.trolltech.qt.gui.QPaintEvent;
import com.trolltech.qt.gui.QPainter;
import com.trolltech.qt.gui.QPolygon;
import com.trolltech.qt.gui.QScrollArea;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWheelEvent;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.common.image.core.Channel;
import eu.moonlight3d.common.image.core.Framebuffer;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.helper.StringTools;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.DrainSlotIterator;
import eu.moonlight3d.ml3d.backend.threedim.og.Link;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.resource.Resource;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.TreeIterator;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.TypeFilter;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;

/**
 * An editor for UV mappings
 *
 * @author gregor
 */
public class UVEditor extends QWidget {
	private static UVEditor instance=null;
	
	public static UVEditor getInstance() {
		return instance;
	}
	
	private class UVCanvas extends QWidget {
		private final static int NORMAL_SIZE=500;
		private double zoom;
		private boolean mouseDragged;
		private boolean mousePressed;
		private int mouseLastX;
		private int mouseLastY;
		private int selectedFaceID=-1;
		private int selectedVertexIndex=-1;

		public UVCanvas(QWidget parent) {
			super(parent);
			setFixedSize(NORMAL_SIZE+1,NORMAL_SIZE+1);
			zoom=1.0;
			mouseDragged=false;
			mousePressed=false;
		}

		@Override
		protected void paintEvent(QPaintEvent event) {
			QPainter painter=new QPainter();
			painter.begin(this);
			
			if(image!=null) {
				// paint background image if available
				painter.drawImage(rect(),image,image.rect());
			}
			
			// draw a grid
			painter.setPen(QColor.white);
			for(int i=0;i<=5;i++) {
				int x=(int) (zoom*NORMAL_SIZE*i/5);
				painter.drawLine(x,0,x,(int) (zoom*NORMAL_SIZE));
			}
			for(int i=0;i<=5;i++) {
				int y=(int) (zoom*NORMAL_SIZE*i/5);
				painter.drawLine(0,y,(int) (zoom*NORMAL_SIZE),y);
			}
			
			if(currentUVSet==null) {
				painter.end();
				return;
			}
			
			painter.setPen(QColor.black);
			
			// draw schematic of UV coordinate layout
			QPoint selectedPoint=null;
			for(int i=0;i<currentMesh.getNumFaces();i++) {
				ArrayList<Integer> vertices = currentMesh.getVertexIDs(currentMesh.getFace(i));
				QPolygon polygon=new QPolygon();
				for(int j=0;j<vertices.size();j++) {
					double uv[]=currentMesh.getUVCoordinates(currentUVSet, i, j);
					QPoint point=new QPoint((int)(zoom*uv[0]*NORMAL_SIZE),(int)(zoom*uv[1]*NORMAL_SIZE));
					polygon.append(point);
					
					if(selectedFaceID==i && selectedVertexIndex==j) {
						selectedPoint=point;
					}
				}
				painter.drawPolygon(polygon);
			}
			
			// this needs to be down here to properly paint over polygons: 
			if(selectedPoint!=null) {
				painter.setPen(QColor.red);
				QPoint delta=new QPoint(1,1);
				painter.drawRect(new QRect(selectedPoint.subtract(delta),selectedPoint.add(delta)));
			}

			painter.end();
		}
		
		@Override
		protected void mousePressEvent(QMouseEvent event) {
			if(event.button()==Qt.MouseButton.LeftButton) {
				mousePressed=true;
				mouseDragged=false;
				mouseLastX=event.x();
				mouseLastY=event.y();
			}
		}
		
		@Override
		protected void mouseMoveEvent(QMouseEvent event) {
			if(mousePressed) {
				mouseDragged=true;
			}
			
			Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			document.update();

			// lookup or create TweakUVMap node
			if(tweakUVMap==null) {
				// try to locate existing tweak node for current UV map
				String ogNodeName=SelectionHelper.getOGNodeName(meshName);
				boolean foundTweak=false;
				if(document.getOGManager().getNodeByName(ogNodeName) instanceof TweakUVMap) {
					tweakUVMap=(TweakUVMap) document.getOGManager().getNodeByName(ogNodeName);
					if(tweakUVMap.getProperty("meshName").getStringValue().equals(currentMesh.getName()) &&
					   tweakUVMap.getProperty("mapName").getStringValue().equals(currentUVSet)) {
						foundTweak=true;
					}
				}
				if(!foundTweak) {
					// need to create new TweakUVMap node
					try {
						tweakUVMap=(TweakUVMap) document.getOGManager().createNodeByName("TweakUVMap");
						String oldSlotName=SelectionHelper.getOGSlotName(meshName);
						Slot oldOutput=document.getOGManager().getNodeByName(ogNodeName).getSlot(oldSlotName);
						document.getOGManager().insertNodeAfter(oldOutput, tweakUVMap.getSlot("input"), tweakUVMap.getSlot("output"));
						tweakUVMap.getProperty("meshName").setValue(SelectionHelper.getSGNodeName(meshName));
						tweakUVMap.getProperty("mapName").setValue(currentUVSet);
					} catch (Exception e) {
						e.printStackTrace();
					}
					
				}
			}

			int deltaX=event.x()-mouseLastX;
			int deltaY=event.y()-mouseLastY;
			double deltaU=deltaX/(zoom*NORMAL_SIZE);
			double deltaV=deltaY/(zoom*NORMAL_SIZE);
			mouseLastX=event.x();
			mouseLastY=event.y();
			
			// TODO check if properties of tweak node are animated 
			
			// lookup current face vertex in properties, add it if it does not exist
			Property faceIDs=tweakUVMap.getProperty("faceIDs");
			Property vertexIDs=tweakUVMap.getProperty("vertexIDs");
			int uvVertexIndex=-1;
			for(int i=0;i<faceIDs.getSize();i++) {
				if(faceIDs.getArrayIntegerValue(i)==selectedFaceID && vertexIDs.getArrayIntegerValue(i)==selectedVertexIndex) {
					uvVertexIndex=i;
					break;
				}
			}
			// create uv vertex entry if it does not exist
			if(uvVertexIndex==-1) {
				// create entry for uv vertex
				faceIDs.setArrayValue(faceIDs.getSize(), selectedFaceID);
				vertexIDs.setArrayValue(vertexIDs.getSize(), selectedVertexIndex);
				uvVertexIndex=faceIDs.getSize()-1;
			}

			// update values
			double uv[]=currentMesh.getUVCoordinates(currentUVSet, selectedFaceID, selectedVertexIndex);
			uv[0]+=deltaU;
			uv[1]+=deltaV;
			tweakUVMap.getProperty("coordinates").setArrayValue(2*uvVertexIndex, uv[0]);
			tweakUVMap.getProperty("coordinates").setArrayValue(2*uvVertexIndex+1, uv[1]);

			// order OG update (will trigger repaint)
			document.update();
		}
		
		@Override
		protected void mouseReleaseEvent(QMouseEvent event) {
			if(event.button()==Qt.MouseButton.LeftButton) {
				if(!mouseDragged) {
					double mouseU=event.x()/(zoom*NORMAL_SIZE);
					double mouseV=event.y()/(zoom*NORMAL_SIZE);
					
					State.getInstance().logger().finer("mouseU: " + mouseU + " mouseV: " + mouseV);
					
					double minDistance=1;
					boolean firstHit=true;
					int hitFaceID=-1;
					int hitVertexIndex=-1;
					
					for(int i=0;i<currentMesh.getNumFaces();i++) {
						ArrayList<Integer> vertices = currentMesh.getVertexIDs(currentMesh.getFace(i));
						for(int j=0;j<vertices.size();j++) {
							double uv[]=currentMesh.getUVCoordinates(currentUVSet, i, j);
							
							double distance=Math.sqrt((mouseU-uv[0])*(mouseU-uv[0])+(mouseV-uv[1])*(mouseV-uv[1]));
							if(firstHit) {
								minDistance=distance;
								hitFaceID=i;
								hitVertexIndex=j;
								firstHit=false;
							} else {
								if(distance<minDistance) {
									minDistance=distance;
									hitFaceID=i;
									hitVertexIndex=j;
								}
							}
						}
					}
					
					// mark resulting vertex as selected
					selectedFaceID=hitFaceID;
					selectedVertexIndex=hitVertexIndex;
					
					State.getInstance().logger().info("selected face: " + selectedFaceID + " vertex: " + selectedVertexIndex);
					
					update();
				}
			}
		}
		
		@Override
		protected void wheelEvent(QWheelEvent event) {
			if(event.delta()>0) {
				for(int i=0;i<event.delta()/32;i++) {
					zoom*=1.1;
				}
			} else {
				for(int i=0;i>event.delta()/32;i--) {
					zoom/=1.1;
				}
			}
			setFixedSize(((int) (zoom*NORMAL_SIZE))+1, ((int) (zoom*NORMAL_SIZE))+1);
			update();
		}
	}
	
	/**
	 * An operator graph listener implementation for the UV editor.
	 *
	 * @author gregor
	 */
	private class OperatorGraphListener implements eu.moonlight3d.ml3d.backend.threedim.og.OperatorGraphListener {

		public void graphCleared() {
			updateUVSets();
		}

		public void graphUpdated() {
			updateUVSets();
		}

		public void linkInserted(Link link) {
		}

		public void linkRemoved(Link link) {
		}

		public void nodeInserted(Node node) {
			updateUVSets();
		}

		public void nodeRemoved(Node node) {
			updateUVSets();
		}

		public void nodeUpdated(Node node) {
		}

		public void nodeUpdateFailed(Node node, Exception exception) {
		}
		
	}

	/**
	 * Actual instance of the operator graph listener.
	 */
	private OperatorGraphListener operatorGraphListener;

	/**
	 * Resource manager listener used by the UV editor.
	 *
	 * @author gregor
	 */
	private class ResourceManagerListener implements eu.moonlight3d.ml3d.backend.threedim.resource.ResourceManagerListener {
		public void resourceAdded(Resource resource) {
			updateTextures();
		}

		public void resourceRemoved(Resource resource) {
			updateTextures();
		}		
	}
	
	/**
	 * Actual instance of the resource manager listener.
	 */
	private ResourceManagerListener resourceManagerListener;
	
	/**
	 * Document lifecycle listener used by the UV editor.
	 *
	 * @author gregor
	 */
	private class LifecycleListener implements eu.moonlight3d.framework.document.LifecycleListener {

		/**
		 * Default constructor
		 */
		public LifecycleListener() {
			registerListeners();
		}
		
		/**
		 * Register the UV editor's operator graph listener and resource manager listener
		 * with the current main document.
		 */
		private void registerListeners() {
			State state=State.getInstance();
			operatorGraphListener=new OperatorGraphListener();
			resourceManagerListener=new ResourceManagerListener();
			Document document=(Document) state.getDocumentManager().getMainDocument();
			document.getOGManager().registerOperatorGraphListener(operatorGraphListener);
			document.getResourceManager().registerResourceManagerListener(resourceManagerListener);
		}
		
		public void afterNew() {
			registerListeners();
		}

		public void afterOpen(String filename) {
			registerListeners();
		}

		public void afterSave(String filename) {
		}

		public void beforeNew() {
		}

		public void beforeOpen(String filename) {
		}

		public void beforeSave(String filename) {
		}
		
		/**
		 * Unregister listeners from the current main document.
		 */
		public void unregisterListeners() {
			State state=State.getInstance();
			Document document=(Document) state.getDocumentManager().getMainDocument();
			document.getOGManager().unregisterOperatorGraphListener(operatorGraphListener);
			document.getResourceManager().unregisterResourceManagerListener(resourceManagerListener);
			operatorGraphListener=null;
			resourceManagerListener=null;
		}

		public void onClose() {
			unregisterListeners();
			image=null;
		}
		
	}
	
	/**
	 * Actual instance of the lifecycle listener implementation.
	 */
	private LifecycleListener lifecycleListener;
	
	private QImage image;
    private QScrollArea scrollArea;
    private QWidget canvas;
    private QComboBox uvSets;
    private QComboBox textures;
    private String currentUVSet;
    private Mesh currentMesh;
    private String meshName;
    
    private TweakUVMap tweakUVMap;

    /**
     * Default constructor
     * 
     * @param parent parent widget
     */
	public UVEditor(QWidget parent) {
		super(parent);
		
		instance=this;

		uvSets=new QComboBox(this);
		uvSets.setEditable(false);
		uvSets.currentIndexChanged.connect(this,"uvSetChanged()");
		textures=new QComboBox(this);
		textures.setEditable(false);
		textures.currentIndexChanged.connect(this,"textureChanged()");
		
		scrollArea=new QScrollArea(this);
		
//        canvas=new ImageCanvas(scrollArea);
        canvas=new UVCanvas(scrollArea);
        
        scrollArea.setWidget(canvas);
        
        QVBoxLayout layout=new QVBoxLayout();
        layout.addWidget(uvSets);
        layout.addWidget(textures);
        layout.addWidget(scrollArea);
        layout.setMargin(0);
        setLayout(layout);

        lifecycleListener=new LifecycleListener();
		State.getInstance().getDocumentManager().addLifecycleListener(lifecycleListener);
		
		updateUVSets();
		updateTextures();
	}
	
	private void setImage(Framebuffer framebuffer)
	{
		image=new QImage(framebuffer.getWidth(),framebuffer.getHeight(),QImage.Format.Format_RGB32);
		Channel red, green, blue;
		Colour colour=new Colour();
		
		red=framebuffer.getChannel("red");
		green=framebuffer.getChannel("green");
		blue=framebuffer.getChannel("blue");
		
		for(int j=0;j<framebuffer.getHeight();j++) {
			for(int i=0;i<framebuffer.getWidth();i++) {
				
				colour.red=red.getPixel(i,j);
				colour.green=green.getPixel(i,j);
				colour.blue=blue.getPixel(i,j);
				
				Colour.RGB rgb=colour.getsRGB8();

				image.setPixel(i, j, new QColor(rgb.red,rgb.green,rgb.blue).rgb());
			}
		}
	}

	/**
	 * Update the list of known UV coordinate sets from the current scene graph.
	 */
	private void updateUVSets() {
		uvSets.clear();
		tweakUVMap=null;
		DrainSlotIterator drainSlotIterator=new DrainSlotIterator(((Document) State.getInstance().getDocumentManager().getMainDocument()).getOGManager());
		while(drainSlotIterator.hasNext()) {
			Slot slot=drainSlotIterator.next();
			
			try {
				SceneGraph graph=slot.getGraph();
				String slotName="OperatorGraph/" + slot.getNode().getName() + "/" + slot.getName();
				
				if(graph!=null) {
					// ignore top-level root node here
					TreeIterator iterator=new TreeIterator(graph);
					iterator.addFilter(new TypeFilter("eu.moonlight3d.ml3d.modelling.mesh.Mesh",true));
					
					while(iterator.hasNext()) {
						eu.moonlight3d.ml3d.backend.threedim.sg.Node nodes[]=iterator.next();
						
						Mesh mesh=(Mesh) nodes[nodes.length-1];
						for(String uvMapName : mesh.getUVMaps()) {
							String fullName=nodes[0].getName();
							for(int i=1;i<nodes.length;i++) {
								fullName+="/" + nodes[i].getName() + ":uv[" + uvMapName + "]";
							}
							uvSets.addItem(fullName,slotName + "/" + fullName);							
						}
						
					}
				}
			} catch (Exception e) {
				// this should not happen
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Update the list of known textures from the resource manager.
	 */
	private void updateTextures() {
		textures.clear();
		
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		List<Resource> resources=document.getResourceManager().getResources();
		
		for(Resource resource : resources) {
			if(resource.getType().equals("Texture")) {
				textures.addItem(resource.getName(),"Resource/" + resource.getName());
			}
		}
	}
	
	/**
	 * Slot reacting to changes in the uv set dropdown of the editor. 
	 */
	protected void uvSetChanged() {
		tweakUVMap=null;
		meshName= (String) uvSets.itemData(uvSets.currentIndex());
		if(meshName==null) {
			currentUVSet=null;
			return;
		}
		Selection selection;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			
			String splitName[]=StringTools.split(meshName, ':');
			meshName=splitName[0];
			String appendix=splitName[1];
			currentUVSet=appendix.substring(3,appendix.indexOf(']'));
			
			currentMesh=(Mesh) selection.getObjectFromSelectionString(meshName);
			canvas.repaint();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * Slot reacting to changes in the texture dropdown of the editor
	 */
	protected void textureChanged() {
		String textureName= (String) textures.itemData(textures.currentIndex());
		Texture currentTexture;
		if(textureName==null) {
			currentTexture=null;
			image=null;
			return;
		}
		Selection selection;
		try {
			selection = ((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			currentTexture = (Texture) selection.getObjectFromSelectionString(textureName);
			setImage(currentTexture.getImage());
			canvas.repaint();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	public Mesh getCurrentMesh() {
		return currentMesh;
	}
	
	public String getUVMapName() {
		return currentUVSet;
	}
	
	@Override
	protected void disposed() {
		lifecycleListener.unregisterListeners();
		State.getInstance().getDocumentManager().removeLifecycleListener(lifecycleListener);
		instance=null;
	}
}
