/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on May 3, 2006
 */
package eu.moonlight3d.ml3d.rendering.glsl;

import java.text.MessageFormat;
import java.util.HashMap;

import javax.media.opengl.GL;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.ml3d.ui.mainwindow.view.GLHelper;
import eu.moonlight3d.ml3d.ui.mainwindow.view.MaterialNode;

public class GLSLMaterial extends MaterialNode {
	private String vertexProgram;
	private String fragmentProgram;
	private HashMap<String,Uniform> uniformParameters;
	private int programID;
	private int vertexShader;
	private int fragmentShader;
	private boolean compiledShader;
	private boolean compiledCorrectly;
	
	public GLSLMaterial() {
		compiledShader=false;
		compiledCorrectly=true;
		programID=0;
	}

	@Override
	public int getNumPasses(GL gl) {
		// try to compile program here
		if (!compiledShader) {
			compiledShader=true;
			programID = gl.glCreateProgram();
			vertexShader = gl.glCreateShader(GL.GL_VERTEX_SHADER);
			fragmentShader = gl.glCreateShader(GL.GL_FRAGMENT_SHADER);

			String[] shaderBuffer = new String[1];

			shaderBuffer[0] = vertexProgram;
			gl.glShaderSource(vertexShader, 1, shaderBuffer, null);
			shaderBuffer[0] = fragmentProgram;
			gl.glShaderSource(fragmentShader, 1, shaderBuffer, null);

			byte[] errorBuffer = new byte[8192];
			int[] messageLength = new int[1];

			gl.glCompileShader(vertexShader);
			gl.glGetShaderInfoLog(vertexShader, 8192, messageLength, 0, errorBuffer, 0);
			if (messageLength[0] > 0) {
				eu.moonlight3d.framework.State.getInstance().logger().warning(
						L10N.translate("GLSL vertex program compiler messages: ") + new String(errorBuffer, 9, messageLength[0]));
				compiledCorrectly=false;
				programID=0;
			}
			gl.glCompileShader(fragmentShader);
			gl.glGetShaderInfoLog(fragmentShader, 8192, messageLength, 0, errorBuffer, 0);
			if (messageLength[0] > 0) {
				eu.moonlight3d.framework.State.getInstance().logger().warning(
						L10N.translate("GLSL fragment program compiler messages: ") + new String(errorBuffer, 0, messageLength[0]));
				compiledCorrectly=false;
				programID=0;
			}

			gl.glAttachShader(programID, vertexShader);
			gl.glAttachShader(programID, fragmentShader);
			gl.glLinkProgram(programID);
			gl.glGetProgramInfoLog(programID, 8192, messageLength, 0, errorBuffer, 0);
			if (messageLength[0] > 0) {
				eu.moonlight3d.framework.State.getInstance().logger().warning(
						L10N.translate("GLSL program linker messages: ") + new String(errorBuffer, 0, messageLength[0]));
				compiledCorrectly=false;
				programID=0;
			}
			
			// set uniform shader parameters
			for(Uniform uniform : uniformParameters.values()) {
				int uniformIndex=gl.glGetUniformLocation(programID,uniform.getName());
				if(uniformIndex==-1) {
					String message=MessageFormat.format(L10N.translate("uniform shader parameter {0} does not exist"), uniform.getName());
					State.getInstance().logger().warning(message);
					continue;
				}
				if(uniform.getGLSLType()==Uniform.GLSLType.Bool) {
					try {
						if(uniform.getBooleanValue()==true) {
							gl.glUniform1i(uniformIndex,GL.GL_TRUE);						
						} else {
							gl.glUniform1i(uniformIndex,GL.GL_FALSE);
						}
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Int) {
					try {
						gl.glUniform1i(uniformIndex,uniform.getIntegerValue());						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Float) {
					try {
						gl.glUniform1f(uniformIndex,(float)uniform.getFloatValue());						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Vec2) {
					try {
						float[] floatArray=new float[2];
						for(int i=0;i<2;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniform2fv(uniformIndex,1,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Vec3) {
					try {
						float[] floatArray=new float[3];
						for(int i=0;i<3;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniform3fv(uniformIndex,1,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Vec4) {
					try {
						float[] floatArray=new float[4];
						for(int i=0;i<4;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniform4fv(uniformIndex,1,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.BVec2) {
					try {
						int[] intArray=new int[2];
						for(int i=0;i<2;i++) {
							if(uniform.getBooleanValue()==true) {
								intArray[i]=GL.GL_TRUE;						
							} else {
								intArray[i]=GL.GL_FALSE;						
							}
						}
						gl.glUniform2iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.BVec3) {
					try {
						int[] intArray=new int[3];
						for(int i=0;i<3;i++) {
							if(uniform.getBooleanValue()==true) {
								intArray[i]=GL.GL_TRUE;						
							} else {
								intArray[i]=GL.GL_FALSE;						
							}
						}
						gl.glUniform3iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.BVec4) {
					try {
						int[] intArray=new int[3];
						for(int i=0;i<3;i++) {
							if(uniform.getBooleanValue()==true) {
								intArray[i]=GL.GL_TRUE;						
							} else {
								intArray[i]=GL.GL_FALSE;						
							}
						}
						gl.glUniform3iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.IVec2) {
					try {
						int[] intArray=new int[2];
						for(int i=0;i<2;i++) {
							intArray[i]=uniform.getArrayIntegerValue(i);
						}
						gl.glUniform2iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.IVec3) {
					try {
						int[] intArray=new int[3];
						for(int i=0;i<2;i++) {
							intArray[i]=uniform.getArrayIntegerValue(i);
						}
						gl.glUniform3iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.IVec4) {
					try {
						int[] intArray=new int[4];
						for(int i=0;i<4;i++) {
							intArray[i]=uniform.getArrayIntegerValue(i);
						}
						gl.glUniform4iv(uniformIndex,1,intArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Mat2) {
					try {
						float[] floatArray=new float[4];
						for(int i=0;i<4;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniformMatrix2fv(uniformIndex,1,false,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Mat3) {
					try {
						float[] floatArray=new float[9];
						for(int i=0;i<9;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniformMatrix3fv(uniformIndex,1,false,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Mat4) {
					try {
						float[] floatArray=new float[16];
						for(int i=0;i<16;i++) {
							floatArray[i]=(float) uniform.getArrayFloatValue(i);
						}
						gl.glUniformMatrix4fv(uniformIndex,1,false,floatArray,0);						
					} catch (Exception e) {
						// cannot happen here
						e.printStackTrace();
					}
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Sampler1D) {
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Sampler2D) {
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Sampler3D) {
				} else if(uniform.getGLSLType()==Uniform.GLSLType.SamplerCube) {
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Sampler1DShadow) {
				} else if(uniform.getGLSLType()==Uniform.GLSLType.Sampler2DShadow) {
				}
			}
		}

		// there is always only one pass for a GLSL material
		return 1;
	}
	
	@Override
	public String getUVMapForPass(GL gl, int numPass) {
		return null;
	}

	@Override
	public void doPass(GL gl, int numPass) {
		if(compiledCorrectly) {
			gl.glUseProgram(programID);
		} else {
			gl.glUseProgram(0);			
		}
		GLHelper.checkAndLogGLError(gl,"glUseProgram");
	}

	@Override
	public void reset(GL gl) {
		gl.glUseProgram(0);
	}
	
	public void setVertexProgram(String vertexProgram) {
		this.vertexProgram=vertexProgram;
		compiledShader=false;
	}
	
	public void setFragmentProgram(String fragmentProgram) {
		this.fragmentProgram=fragmentProgram;
		compiledShader=false;
	}

	public void setUniformParameters(HashMap<String, Uniform> uniformParameters) {
		this.uniformParameters=uniformParameters;
		compiledShader=false;
	}
}
