/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 20, 2007
 */
package eu.moonlight3d.ml3d.rendering.simple;


import java.util.List;

import com.trolltech.qt.gui.QComboBox;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.SelectionHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.resource.Resource;
import eu.moonlight3d.ml3d.backend.threedim.sg.GeometryNode;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;
import eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool;

public class CreateTextureMaterialTool extends Tool {
	private QComboBox uvSets;
	private QComboBox textures;
	
	private class SelectionListener implements eu.moonlight3d.ml3d.backend.selection.SelectionListener {

		public void elementAdded(String objectName) {
			updateUVList();
		}

		public void elementRemoved(String objectName) {
			updateUVList();
		}

		public void elementUpdated(String oldName, String newName) {
			updateUVList();
		}

		public void elementsAdded(List<String> objectNames) {
			updateUVList();
		}

		public void selectionCleared() {
			updateUVList();
		}
		
	}
	
	private SelectionListener selectionListener;
	
	private class ResourceManagerListener implements eu.moonlight3d.ml3d.backend.threedim.resource.ResourceManagerListener {

		public void resourceAdded(Resource resource) {
			updateTextureList();
		}

		public void resourceRemoved(Resource resource) {
			updateTextureList();
		}
		
	}
	
	private ResourceManagerListener resourceManagerListener;
	
	private class LifecycleListener implements eu.moonlight3d.framework.document.LifecycleListener {
		
		public LifecycleListener() {
			registerListeners();
		}
		
		private void registerListeners() {
			Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			selectionListener=new SelectionListener();
			try {
				document.getSelectionManager().getSelection("viewSelection").registerListener(selectionListener);
			} catch (Exception e) {
				// this exception should never occur
				ExceptionDialog.run(e);
			}
			resourceManagerListener=new ResourceManagerListener();
			document.getResourceManager().registerResourceManagerListener(resourceManagerListener);
		}
		
		private void unregisterListeners() {
			Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			try {
				document.getSelectionManager().getSelection("viewSelection").unregisterListener(selectionListener);
			} catch (Exception e) {
				ExceptionDialog.run(e);
			}
			document.getResourceManager().unregisterResourceManagerListener(resourceManagerListener);
			selectionListener=null;
			resourceManagerListener=null;
		}

		public void afterNew() {
			registerListeners();
		}

		public void afterOpen(String filename) {
			registerListeners();
		}

		public void afterSave(String filename) {
		}

		public void beforeNew() {
		}

		public void beforeOpen(String filename) {
		}

		public void beforeSave(String filename) {
		}

		public void onClose() {
			unregisterListeners();
		}
		
	}
	
	private LifecycleListener lifecycleListener;
	
	public CreateTextureMaterialTool() {
		setName("CreateTextureMaterialTool");
		setDisplayName("Create Texture Material");
		setCategory("Create/Material");
	}
	
	@Override
	public void activated() {
        try {
            State state=eu.moonlight3d.framework.State.getInstance();
            Manager ogManager = ((Document) state.getDocumentManager().getMainDocument()).getOGManager();

            // don't do anything if selected object is not a SG node
            Selection selection=((Document) state.getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
            if(selection.getType()!=eu.moonlight3d.ml3d.backend.selection.Manager.getSelectionType("SGNode")) {
                state.logger().warning(L10N.translate("current selection must be a scene graph object"));
                return;
            }
            
            if(!(selection.getObjectFromSelectionString(selection.getActiveObject()) instanceof GeometryNode)) {
            	state.logger().warning(L10N.translate("materials can only be assigned to scene graph nodes which contain visible geometry (e.g. Meshes)"));
            	return;
            }
            
            if(uvSets.count()==0) {
            	state.logger().warning(L10N.translate("currently active object does not have any UV sets"));
            	return;
            }
            
            if(textures.count()==0) {
            	state.logger().warning(L10N.translate("there are currently no textures loaded"));
            	return;
            }
            
            String ogNodeName=SelectionHelper.getOGNodeName(selection.getActiveObject());
            String slotName=SelectionHelper.getOGSlotName(selection.getActiveObject());
            String sgNodeName=SelectionHelper.getSGNodeName(selection.getActiveObject());
            
            Node ogNode=ogManager.getNodeByName(ogNodeName);
            Slot slot=ogNode.getSlot(slotName);
            Node createTextureMaterial=ogManager.createNodeByName("CreateTextureMaterial");
            
            // set node properties
            createTextureMaterial.getProperty("ownerName").setValue(sgNodeName);
            
            String textureName=textures.itemText(textures.currentIndex());
            String uvSetName=uvSets.itemText(uvSets.currentIndex());
            createTextureMaterial.getProperty("textureName").setValue(textureName);
            createTextureMaterial.getProperty("uvSetName").setValue(uvSetName);
            
            // set links
            ogManager.insertNodeAfter(slot,createTextureMaterial.getSlot("input"),createTextureMaterial.getSlot("output"));
			ogManager.update();
	        selection.updateBase(ogNodeName,createTextureMaterial.getName());
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}

	@Override
	public void deactivated() {
	}

	@Override
	public QWidget getOptionEditor(QWidget parent) {
		QWidget widget=new QWidget(parent);
		
		QLabel uvListLabel=new QLabel(widget);
		uvListLabel.setText(L10N.translate("UV set:"));
		uvSets=new QComboBox(widget);
		QLabel texturesLabel=new QLabel(widget);
		texturesLabel.setText(L10N.translate("Texture:"));
		textures=new QComboBox(widget);
		
		QVBoxLayout layout=new QVBoxLayout(widget);
		layout.addWidget(uvListLabel);
		layout.addWidget(uvSets);
		layout.addWidget(texturesLabel);
		layout.addWidget(textures);
		widget.setLayout(layout);
		
        State state=eu.moonlight3d.framework.State.getInstance();
        lifecycleListener=new LifecycleListener();
        state.getDocumentManager().addLifecycleListener(lifecycleListener);

        return widget;
	}

	@Override
	public boolean hasOptionEditor() {
		return true;
	}
	
	private void updateUVList() {
		uvSets.clear();
		
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		
		try {
			Selection selection=document.getSelectionManager().getSelection("viewSelection");
			
			if(selection.getType().getName().equals("SGNode")) {
				if(selection.getActiveObject()==null) {
					return;
				}
				
				Object object=selection.getObjectFromSelectionString(selection.getActiveObject());
				if(object instanceof eu.moonlight3d.ml3d.modelling.mesh.Mesh) {
					Mesh mesh=(Mesh) object;
					for(String mapName : mesh.getUVMaps()) {
						uvSets.addItem(mapName);
					}
				}
			}
		} catch (Exception e1) {
			ExceptionDialog.run(e1);
		}
	}
	
	private void updateTextureList() {
		textures.clear();
		
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		List<Resource> resources=document.getResourceManager().getResources();
		
		for(Resource resource : resources) {
			if(resource.getType().equals("Texture")) {
				textures.addItem(resource.getName(),"Resource/" + resource.getName());
			}
		}
	}

}
