/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 18, 2006
 */
package eu.moonlight3d.ml3d.rendering.simple;

import java.util.HashMap;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.preferences.PreferencesGroup;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.modelling.textures.Texture;

/**
 * Cache for OpenGL textures. This class manages uploaded OpenGL
 * textures. New textures get uploaded while those that have gone
 * unused for longer than a certain period of time are discarded.
 *
 * @author gregor
 */
public class TextureCache {

	/**
	 * Reference to single global instance of TextureCache
	 */
	private static TextureCache instance=null;
	
	/**
	 * Get the TextureCache singleton instance. The singleton
	 * is created on the first call.
	 * 
	 * @return TextureCache singleton
	 */
	static TextureCache getInstance() {
		if(instance==null) {
			instance=new TextureCache();
		}
		
		return instance;
	}
	
	/**
	 * Cleanup thread that is responsible for discarding unused OpenGL
	 * textures.
	 *
	 * @author gregor
	 */
	private class CacheCleanupThread implements Runnable {

		public void run() {
			PreferencesGroup textureCachePreferences=State.getInstance().getPreferencesManager().getCurrentPreferences().getGroupByName("Texture Cache");
			while (true) {
				long currentTimeMillis=System.currentTimeMillis();
				long gracePeriod;
				try {
					gracePeriod = textureCachePreferences.getProperty("textureGraceTime").getIntegerValue()*1000;
				} catch (Exception e1) {
					gracePeriod=600*1000;
				}
				
				boolean textureRemoved=false;
				do {
					for(String name : textures.keySet()) {
						GLTextureMaterial material=textures.get(name);
						if(material.getLastUsedMillis()+gracePeriod<currentTimeMillis) {
							textures.remove(name);
							material.disposeGLTexture();
							textureRemoved=true;
							break;
						}
					}
				} while(textureRemoved);
				
				try {
					long delay;
					try {
						delay=textureCachePreferences.getProperty("cleanupInterval").getIntegerValue()*1000;
					} catch (Exception e1) {
						delay=60*1000;
					} 
					synchronized(this) {
						wait(delay);
					}
				} catch (InterruptedException e) {
				}
			}
		}
		
	}
	
	/**
	 * Map of texture names to textures
	 */
	private HashMap<String,GLTextureMaterial> textures;
	
	/**
	 * Default constructor. This constructor starts the cleanup thread
	 * as a daemon thread.
	 */
	private TextureCache() {
		textures=new HashMap<String,GLTextureMaterial>();
		Thread cacheCleanupThread=new Thread(new CacheCleanupThread());
		cacheCleanupThread.setDaemon(true);
		cacheCleanupThread.start();
	}
	
	/**
	 * Get the OpenGL texture for a given internal texture.
	 * 
	 * @param textureMaterial the texture material to get the uploaded OpenGL texture for
	 * @return the internal texture to lookup the OpenGL texture for
	 */
	GLTextureMaterial getTextureMaterial(TextureMaterial textureMaterial) {
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		Texture texture=(Texture) document.getResourceManager().getResource(textureMaterial.getTextureName());
		if(textures.get(texture.getFilename())!=null) {
			return textures.get(texture.getFilename());
		}
		
		GLTextureMaterial glTextureMaterial=new GLTextureMaterial(texture.getImage(),"UVMap");
		textures.put(texture.getFilename(),glTextureMaterial);
		return glTextureMaterial;
	}
}
