/*
 * Moonlight|3D
 * Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Sep 5, 2007
 */

package eu.moonlight3d.ml3d.rendering.sunflow;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.sunflow.SunflowAPI;
import org.sunflow.math.Point3;
import org.sunflow.math.Vector3;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.common.image.core.Framebuffer;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.NamePool;
import eu.moonlight3d.framework.helper.NamedObject;
import eu.moonlight3d.framework.helper.PropertyContainer;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.og.DrainSlotIterator;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.Group;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;
import eu.moonlight3d.ml3d.backend.threedim.sg.MaterialNode;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.Iterator;
import eu.moonlight3d.ml3d.backend.threedim.sg.iterator.TypeFilter;
import eu.moonlight3d.ml3d.modelling.camera.Camera;
import eu.moonlight3d.ml3d.modelling.lighting.OmniLight;
import eu.moonlight3d.ml3d.modelling.lighting.SphereLight;
import eu.moonlight3d.ml3d.modelling.lighting.SpotLight;
import eu.moonlight3d.ml3d.modelling.mesh.Mesh;
import eu.moonlight3d.ml3d.modelling.textures.Texture;
import eu.moonlight3d.ml3d.rendering.simple.SimpleMaterial;
import eu.moonlight3d.ml3d.rendering.simple.TextureMaterial;

/**
 * A converter that populates a Sunflow instance with all the data of
 * the current scene. This is done in three stages: first, materials
 * and light sources are exported. Then geometry data is exported
 * which may reference the previously exported shaders. Finally,
 * object instances are placed at their proper positions in the
 * scene.
 *
 * @author gregor
 */
public class SunflowSceneConverter {
	
	private class SunflowNodeName implements NamedObject {
		public String name;

		public String getName() {
			return name;
		}
	}
	
	/**
	 * Reference to the SunflowAPI interface of the target Sunflow 
	 * instance.
	 */
    private SunflowAPI sunflowAPI;
    
    private NamePool namePool;
    private HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node, String> nameMap;
    private HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node, String> instanceNameMap;

    /**
     * Constructor.
     * 
     * @param sunflowAPI the Sunflow instance to populate with the current 
     * scene.
     */
    public SunflowSceneConverter(SunflowAPI sunflowAPI) {
    	this.sunflowAPI=sunflowAPI;
    	namePool=new NamePool();
    	nameMap=new HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node, String>();
    	instanceNameMap=new HashMap<eu.moonlight3d.ml3d.backend.threedim.sg.Node, String>();
    }
    
    private String mapNodeName(eu.moonlight3d.ml3d.backend.threedim.sg.Node node, String nodePath) {
    	if(nameMap.get(node)!=null) {
    		return nameMap.get(node);
    	}
    	
    	String mappedPath=namePool.generateUniqueName(nodePath);
    	SunflowNodeName sunflowNodeName=new SunflowNodeName();
    	sunflowNodeName.name=mappedPath;
    	namePool.add(sunflowNodeName);
    	nameMap.put(node,mappedPath);
    	return mappedPath;
    }
    
    private String mapInstanceNodeName(eu.moonlight3d.ml3d.backend.threedim.sg.Node node, String nodePath) {
    	String mappedPath=namePool.generateUniqueName(nodePath);
    	SunflowNodeName sunflowNodeName=new SunflowNodeName();
    	sunflowNodeName.name=mappedPath;
    	namePool.add(sunflowNodeName);
    	instanceNameMap.put(node,mappedPath);
    	return mappedPath;
    }
    
    /**
     * Convert all materials in the scene. Unfortunately this function only
     * knows about a fixed set of shaders that is not extensible, although
     * Sunflow would have facilities that would allow exporting almost any
     * material.
     */
	private void convertMaterials() {
		Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
		DrainSlotIterator iterator=new DrainSlotIterator(document.getOGManager());
		
		while(iterator.hasNext()) {
			Slot slot=iterator.next();
			String slotName="OperatorGraph/" + slot.getNode().getName() + "/" + slot.getName();
			
			Iterator sgIterator=null;
			try {
				sgIterator = new Iterator(slot.getGraph());
			} catch (Exception e) {
				// cannot happen
				e.printStackTrace();
			}
			
			sgIterator.addFilter(new TypeFilter("eu.moonlight3d.ml3d.backend.threedim.sg.MaterialNode",true));
			
			while(sgIterator.hasNext()) {
				MaterialNode materialNode=(MaterialNode) sgIterator.next();
				String materialNodeName=slotName + "/" + materialNode.getName();
				
				materialNodeName=mapNodeName(materialNode, materialNodeName);
				
				if(materialNode.getType().equals("SimpleMaterial")) {
					SimpleMaterial material=(SimpleMaterial)materialNode;
					
					Colour colour=material.getColour();
					sunflowAPI.parameter("diffuse", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					colour=material.getSpecular();
					sunflowAPI.parameter("specular", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.parameter("power", (float) 2);
					sunflowAPI.parameter("samples", 4);
					sunflowAPI.shader(materialNodeName, "phong");
				} else if(materialNode.getType().equals("TextureMaterial")) {
					TextureMaterial textureMaterial=(TextureMaterial)materialNode;
					String textureName=textureMaterial.getTextureName() + ".moonlight";

					// register texture framebuffer in BitmapReader
					Texture texture=(Texture) document.getResourceManager().getResource(textureMaterial.getTextureName());
					Framebuffer framebuffer=texture.getImage();
					BitmapReader.add(textureName, framebuffer);
					
					sunflowAPI.parameter("texture", textureName);
					sunflowAPI.shader(materialNodeName,"textured_shiny_diffuse");
				} else if(materialNode.getType().equals("AmbientOcclusionMaterial")) {
					AmbientOcclusionMaterial material=(AmbientOcclusionMaterial)materialNode;
					
					Colour colour=material.getLightColour();					
					sunflowAPI.parameter("bright", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					colour=material.getDarkColour();
					sunflowAPI.parameter("dark", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.parameter("samples", material.getNumSamples());
					sunflowAPI.parameter("maxDist", (float) material.getMaxDistance());
					sunflowAPI.shader(materialNodeName, "ambient_occlusion");
				} else if(materialNode.getType().equals("ConstantMaterial")) {
					ConstantMaterial material=(ConstantMaterial)materialNode;
					Colour colour=material.getColour();

					sunflowAPI.parameter("color", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.shader(materialNodeName, "constant");
				} else if(materialNode.getType().equals("DiffuseMaterial")) {
					DiffuseMaterial material=(DiffuseMaterial)materialNode;
					Colour colour=material.getColour();
					
					sunflowAPI.parameter("diffuse", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.shader(materialNodeName, "diffuse");
				} else if(materialNode.getType().equals("GlassMaterial")) {
					GlassMaterial material=(GlassMaterial)materialNode;
					Colour colour=material.getColour();
					
					sunflowAPI.parameter("color", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.parameter("eta", (float) material.getEta());
					// TODO move thess parameters to GlassMaterial
//					sunflowAPI.parameter("absorption.distance", 10.0f);
//					sunflowAPI.parameter("absorption.color", new Color());
					sunflowAPI.shader(materialNodeName, "glass");
				} else if(materialNode.getType().equals("MirrorMaterial")) {
					MirrorMaterial material=(MirrorMaterial)materialNode;
					Colour colour=material.getReflectedColour();
					
					sunflowAPI.parameter("color", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.shader(materialNodeName, "mirror");
				} else if(materialNode.getType().equals("PhongMaterial")) {
					PhongMaterial material=(PhongMaterial)materialNode;
				
					Colour colour=material.getDiffuseColour();
					sunflowAPI.parameter("diffuse", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					colour=material.getSpecularColour();
					sunflowAPI.parameter("specular", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.parameter("power", (float) material.getPower());
					sunflowAPI.parameter("samples", material.getNumRays());
					sunflowAPI.shader(materialNodeName, "phong");
				} else if(materialNode.getType().equals("ShinyDiffuseMaterial")) {
					ShinyDiffuseMaterial material=(ShinyDiffuseMaterial)materialNode;
					Colour colour=material.getDiffuseColour();
					
					sunflowAPI.parameter("diffuse", "sRGB linear", (float)colour.red,(float)colour.green,(float)colour.blue);
					sunflowAPI.parameter("shiny", (float) material.getReflectivity());
					sunflowAPI.shader(materialNodeName, "shiny_diffuse");
				} else {
					eu.moonlight3d.framework.State.getInstance().logger().warning("Unknown material type " + materialNode.getType() + " found in scene. Rendered image may not be correct");
				}
			}
		}
	}
	
    /**
     * Convert a mesh geometry object to the corresponding Sunflow 
     * representation.
     *  
     * @param ogSlot the parent operator graph slot
     * @param mesh the mesh to convert
     */
    private void convertMesh(Slot ogSlot, Mesh mesh) {
		// convert mesh to sunflow Mesh
		float[] vertices=new float[3*mesh.getNumVertices()];
		ArrayList<Integer> faces=new ArrayList<Integer>();
		ArrayList<Vector3D> faceVertexNormals=new ArrayList<Vector3D>();

		int index = 0;
		for(int i=0;i<mesh.getNumVertices();i++) {
			Vector3D vertex=mesh.getVertexPosition(i);
			vertices[index++]=(float) vertex.X1;
			vertices[index++]=(float) vertex.X2;
			vertices[index++]=(float) vertex.X3;
		}
		
		for(int i=0;i<mesh.getNumFaces();i++) {
			Mesh.Face face = mesh.getFace(i);
			ArrayList<Integer> triangles = mesh.triangulated(face);
			ArrayList<Integer> vertexIDs = mesh.getVertexIDs(face);
			for (int j = 0; j < triangles.size(); j++) {
				faces.add(triangles.get(j));
				faceVertexNormals.add(mesh.getFaceVertexNormal(i, vertexIDs.indexOf(triangles.get(j))));
			}			
		}
		
		int[] facesArray=new int[faces.size()];
		for(int i=0;i<facesArray.length;i++) {
			facesArray[i]=faces.get(i);
		}
		float[] normals=new float[3*faceVertexNormals.size()];
		for(int i=0;i<faceVertexNormals.size();i++) {
			normals[3*i]=(float) faceVertexNormals.get(i).X1;
			normals[3*i+1]=(float) faceVertexNormals.get(i).X2;
			normals[3*i+2]=(float) faceVertexNormals.get(i).X3;
		}

		// search children for an uv map reference:
		float uv[]=null;
		for(eu.moonlight3d.ml3d.backend.threedim.sg.Node node : mesh.getChildren()) {
			if(node instanceof TextureMaterial) {
				TextureMaterial material=(TextureMaterial) node;
				String uvMapName=material.getUVSetName();
				
				uv=new float[2*faces.size()];
				index=0;
				for(int i=0;i<mesh.getNumFaces();i++) {
					Mesh.Face face = mesh.getFace(i);
					ArrayList<Integer> triangles = mesh.triangulated(face);
					ArrayList<Integer> vertexIDs = mesh.getVertexIDs(face);
					for (int j = 0; j < triangles.size(); j++) {
						int vertexIndex=vertexIDs.indexOf(triangles.get(j));
						double[] uvPair=mesh.getUVCoordinates(uvMapName, i, vertexIndex);
						
						uv[index++]= (float) uvPair[0];
						uv[index++]= (float) uvPair[1];
					}
				}							
			}
		}
		
//		System.out.println("faces " + faces.size());
//		System.out.println("normals " + normals.length);
//		System.out.println("uv coordinates " + index);
//		System.out.println("uv array length " + uv.length);
		
		sunflowAPI.parameter("points","point","vertex",vertices);
		sunflowAPI.parameter("triangles", facesArray);
		sunflowAPI.parameter("normals","vector","facevarying",normals);
		if(uv!=null) {
			sunflowAPI.parameter("uvs", "texcoord", "facevarying", uv);
		}
		
		String meshName=mapNodeName(mesh, mesh.getName());
		
		sunflowAPI.geometry(meshName, "triangle_mesh");
    }

    /**
     * Convert scene geometry data.
     */
    private void convertGeometry() {
		DrainSlotIterator iterator=new DrainSlotIterator(((Document) State.getInstance().getDocumentManager().getMainDocument()).getOGManager());
		
		while(iterator.hasNext()) {
			Slot slot=iterator.next();
			
			Iterator sgIterator=null;
			try {
				sgIterator = new Iterator(slot.getGraph());
			} catch (Exception e) {
				// cannot happen
				e.printStackTrace();
			}
			
			
			while(sgIterator.hasNext()) {
				eu.moonlight3d.ml3d.backend.threedim.sg.Node geometryNode=sgIterator.next();
				if(geometryNode instanceof Mesh) {
					convertMesh(slot, (Mesh) geometryNode);
				}
			}
		}
		
	}

	/**
	 * Convert an omnidirectional light source to a corresponding Sunflow 
	 * light source.
	 * 
	 * @param transformation transformation into local coordinate system of this object
	 * @param light the light to convert
	 * @param name the name to assign to the exported sunflow object
	 */
	private void convertOmniLight(Matrix4 transformation, OmniLight light, String name) {
        Colour colour=light.getColour();
        Vector3D position=light.getPosition();
        position=transformation.multiply(position);
        double intensity=light.getIntensity();

        Point3 pos=new Point3((float) position.X1, (float) position.X2, (float) position.X3);
        
        sunflowAPI.parameter("center", pos);
        sunflowAPI.parameter("power", "sRGB linear",(float) (intensity*colour.red), (float) (intensity*colour.green), (float) (intensity*colour.blue));
        sunflowAPI.light(name,"point");
    }
    
	/**
	 * Convert a spherical light source to a corresponding Sunflow light
	 * source.
	 * 
	 * @param transformation transformation into local coordinate system of this object
	 * @param light the light to convert
	 * @param name the name to assign to the exported sunflow object
	 */
	private void convertSphereLight(Matrix4 transformation, SphereLight light, String name) {
        Colour colour=light.getColour();
        Vector3D position=light.getPosition();
        position=transformation.multiply(position);
        double intensity=light.getIntensity();

        Point3 pos=new Point3((float) position.X1, (float) position.X2, (float) position.X3);
        
        sunflowAPI.parameter("center", pos);
        sunflowAPI.parameter("radiance", "sRGB linear",(float) (intensity*colour.red), (float) (intensity*colour.green), (float) (intensity*colour.blue));
        sunflowAPI.parameter("radius", (float) light.getRadius());
        sunflowAPI.parameter("samples", light.getSamples());
        sunflowAPI.light(name,"sphere");
    }
    
	/**
	 * Convert a spot light to a corresponding Sunflow light source.
	 * 
	 * @param transformation transformation into local coordinate system of this object
	 * @param light the light to convert
	 * @param name the name to assign to the exported sunflow object
	 */
	private void convertSpotLight(Matrix4 transformation, SpotLight light, String name) {
        Colour colour=light.getColour();
        Vector3D position=light.getPosition();
        position=transformation.multiply(position);
        Vector3D direction=light.getDirection();
        Matrix4 directionTransform=transformation.clone();
        for(int i=0;i<3;i++) {
        	directionTransform.X[i][3]=0;
        }
        direction=directionTransform.multiply(direction);
        double intensity=light.getIntensity();

        Point3 pos=new Point3((float) position.X1, (float) position.X2, (float) position.X3);
        Point3 dir=new Point3((float) direction.X1, (float) direction.X2, (float) direction.X3);
        
        sunflowAPI.parameter("source", pos);
        sunflowAPI.parameter("dir",dir);
        sunflowAPI.parameter("radiance", "sRGB linear",(float) (intensity*colour.red), (float) (intensity*colour.green), (float) (intensity*colour.blue));
        sunflowAPI.parameter("radius", (float) light.getAngle());
        sunflowAPI.light(name,"directional");
    }

	/**
	 * Convert a camera to the corresponding representation inside Sunflow, if
	 * it is the camera that is to be used in the current rendering.
	 * 
	 * @param properties the renderer settings
	 * @param transformation transformation into local coordinate system of this object
	 * @param camera the camera to convert
	 * @param name the name to assign to the exported sunflow object
	 */
    private void convertCamera(SunflowProperties properties, Matrix4 transformation, Camera camera, String name) {
		if(!camera.getName().equals(properties.getCameraName())) {
			return;
		}
    	
        float fov=45.0f;

        Vector3D position=camera.getPosition();
		Vector3D front = camera.getFront();
		Vector3D right = camera.getRight();
		Vector3D up = camera.getUp();
		
		Matrix4 cameraTransformation=new Matrix4();
		
		cameraTransformation.fromTranslation(position);
		
		cameraTransformation.X[0][0]=right.X1;
		cameraTransformation.X[1][0]=right.X2;
		cameraTransformation.X[2][0]=right.X3;
		cameraTransformation.X[0][1]=up.X1;
		cameraTransformation.X[1][1]=up.X2;
		cameraTransformation.X[2][1]=up.X3;
		cameraTransformation.X[0][2]=-front.X1;
		cameraTransformation.X[1][2]=-front.X2;
		cameraTransformation.X[2][2]=-front.X3;
		
		cameraTransformation=transformation.multiply(cameraTransformation);

		sunflowAPI.parameter("fov", fov);
		sunflowAPI.parameter("aspect", properties.getResX()/(float)properties.getResY());
		org.sunflow.math.Matrix4 sunflowMatrix;
		sunflowMatrix=new org.sunflow.math.Matrix4((float)cameraTransformation.X[0][0],
				(float)cameraTransformation.X[0][1],
				(float)cameraTransformation.X[0][2],
				(float)cameraTransformation.X[0][3],
				(float)cameraTransformation.X[1][0],
				(float)cameraTransformation.X[1][1],
				(float)cameraTransformation.X[1][2],
				(float)cameraTransformation.X[1][3],
				(float)cameraTransformation.X[2][0],
				(float)cameraTransformation.X[2][1],
				(float)cameraTransformation.X[2][2],
				(float)cameraTransformation.X[2][3]);
//		sunflowMatrix=new org.sunflow.math.Matrix4((float)transformation.X[0][0],
//				(float)transformation.X[0][1],
//				(float)transformation.X[0][2],
//				(float)transformation.X[0][3],
//				(float)transformation.X[1][0],
//				(float)transformation.X[1][1],
//				(float)transformation.X[1][2],
//				(float)transformation.X[1][3],
//				(float)transformation.X[2][0],
//				(float)transformation.X[2][1],
//				(float)transformation.X[2][2],
//				(float)transformation.X[2][3]);
		sunflowAPI.parameter("transform", sunflowMatrix);
		sunflowAPI.camera(camera.getName(),"pinhole");
    }
    
    /**
     * Instantiate a mesh for a given occurence in the scene graph. The
     * scene graph semantics allow the same mesh to be referenced by
     * parent nodes at different positions in the scene graph, denoting
     * different instances of the same mesh.
     *  
     * @param ogSlot the parent operator graph slot
     * @param transformation the local to world transformation matrix for
     * this instance
     * @param mesh the mesh to convert
	 * @param name the name to assign to the exported sunflow object
     */
    private void convertMeshInstance(Slot ogSlot, Matrix4 transformation, Mesh mesh, String name) {
		// search children for reference to material:
		String shaderName[]=null;
		for (eu.moonlight3d.ml3d.backend.threedim.sg.Node child : mesh.getChildren()) {
			if(child instanceof MaterialNode) {
				shaderName=new String[1];
				shaderName[0]=nameMap.get(child);
			}
		}
		
		if(shaderName!=null) {
			sunflowAPI.parameter("shaders",shaderName);
		}
		org.sunflow.math.Matrix4 sunflowMatrix;
		sunflowMatrix=new org.sunflow.math.Matrix4((float)transformation.X[0][0],
				(float)transformation.X[0][1],
				(float)transformation.X[0][2],
				(float)transformation.X[0][3],
				(float)transformation.X[1][0],
				(float)transformation.X[1][1],
				(float)transformation.X[1][2],
				(float)transformation.X[1][3],
				(float)transformation.X[2][0],
				(float)transformation.X[2][1],
				(float)transformation.X[2][2],
				(float)transformation.X[2][3]);
		sunflowAPI.parameter("transform", sunflowMatrix);
		
		sunflowAPI.instance("Instance" + name, nameMap.get(mesh));
    }
    
    /**
     * Convert a scene graph segment to its Sunflow representation.
     * 
     * @param properties the current renderer settings
     * @param ogSlot the operator graph slot that contains this scene graph segment
     * @param transformation recursively calculated local to world transformation matrix
     * @param node the root node of this scene graph segment
	 * @param name the complete path through the scene graph to the current node
     */
    private void convertSG(SunflowProperties properties, Slot ogSlot, Matrix4 transformation, eu.moonlight3d.ml3d.backend.threedim.sg.Node node, String name) {
        List<eu.moonlight3d.ml3d.backend.threedim.sg.Node> children=node.getChildren();
                
        if("".equals(name)) {
        	name=node.getName();
        } else {
        	name+="/" + node.getName();
        }
        
    	name=mapInstanceNodeName(node, name);

    	// guess node type:
        if(node.getType()=="Mesh") {
            convertMeshInstance(ogSlot, transformation, (eu.moonlight3d.ml3d.modelling.mesh.Mesh)node,name);
        } else if(node.getType()=="OmniLight") {
            convertOmniLight(transformation,(eu.moonlight3d.ml3d.modelling.lighting.OmniLight)node,name);
        } else if(node.getType()=="SphereLight") {
            convertSphereLight(transformation,(eu.moonlight3d.ml3d.modelling.lighting.SphereLight)node,name);
        } else if(node.getType()=="SpotLight") {
            convertSpotLight(transformation,(eu.moonlight3d.ml3d.modelling.lighting.SpotLight)node,name);
        } else if(node.getType()=="Camera") {
            convertCamera(properties, transformation, (eu.moonlight3d.ml3d.modelling.camera.Camera)node,name);
        } else if(node.getType()=="Instance") {
        	Instance instance = (Instance) node;
        	transformation=transformation.multiply(instance.getTransformationMatrix());
        } else if(node.getType()=="Group") {
        	Group group = (Group) node;
        	transformation=transformation.multiply(group.getTransformationMatrix());
        } else {
            //State.getInstance().logger().warning("unknown scene graph node type encountered. rendered image may not contain all scene objects");
        }
        
        for(int i=0;i<children.size();i++) {
            convertSG(properties, ogSlot, transformation, children.get(i),name);
        }
    }

    /**
     * Convert all scene geometry to its sunflow representation.
     * @param properties
     */
    private void convertCurrentSG(SunflowProperties properties) {
        eu.moonlight3d.ml3d.backend.threedim.og.Manager ogManager=((Document) State.getInstance().getDocumentManager().getMainDocument()).getOGManager();
        List<Node> ogNodes;
        
        ogNodes=ogManager.getNodeList();
        
        for(int i=0;i<ogNodes.size();i++) {
            if(ogNodes.get(i).isDrain()) {
                List<Slot> slots=ogNodes.get(i).getSlots();
                for(int j=0;j<slots.size();j++) {
                    if(slots.get(j).getType()==Slot.Type.Output) {
                        try {
                        	for(eu.moonlight3d.ml3d.backend.threedim.sg.Node root : slots.get(j).getGraph().getRootNodes()) {
                        		Matrix4 transformation=new Matrix4();
                        		transformation.setIdentity();
                        		convertSG(properties, slots.get(j), transformation, root, "");
                        	}
                        } catch (Exception e) {
                            // this should not happen
                            e.printStackTrace();
                        }
                    }
                }
            }
        }
    }

    /**
     * Convert the operator graph of the current main document as it is now
     * to a Sunflow representation in memory that can be rendered out by
     * that instance of the  Sunflow renderer.
     * 
     * @param rendererProperties the current renderer settings
     */
	public void convertCurrentOG(PropertyContainer rendererProperties) {
		SunflowProperties properties=(SunflowProperties) rendererProperties;
		
		BitmapReader.clear();
        
        convertMaterials();
        convertGeometry();
        convertCurrentSG(properties);
        
        sunflowAPI.parameter("threads", properties.getThreads());
        sunflowAPI.parameter("camera", properties.getCameraName());
        sunflowAPI.parameter("aa.min",properties.getAntiAliasMin());
        sunflowAPI.parameter("aa.max",properties.getAntiAliasMax());
        sunflowAPI.parameter("samples",properties.getSuperSample());
        switch(properties.getFilter()) {
        case 0:
        	sunflowAPI.parameter("filter","box");
        	break;
        case 1:
        	sunflowAPI.parameter("filter","gaussian");
        	break;
        case 2:
        	sunflowAPI.parameter("filter","mitchell");
        	break;
        case 3:
        	sunflowAPI.parameter("filter","catmull-rom");
        	break;
        case 4:
        	sunflowAPI.parameter("filter","blackman-harris");
        	break;
        case 5:
        	sunflowAPI.parameter("filter","sinc");
        	break;
        case 6:
        	sunflowAPI.parameter("filter","lanczos");
        	break;
        case 7:
        	sunflowAPI.parameter("filter","triangle");
        	break;
        }
        sunflowAPI.parameter("resolutionX",properties.getResX());
        sunflowAPI.parameter("resolutionY",properties.getResY());
        sunflowAPI.parameter("depths.diffuse",properties.getDiffuseTraceDepth());
        sunflowAPI.parameter("depths.reflection",properties.getReflectionTraceDepth());
        sunflowAPI.parameter("depths.refraction",properties.getRefractionTraceDepth());
        
        if(properties.getCausticsEnable()) {
        	sunflowAPI.parameter("caustics.emit", properties.getCausticsEmit());
        	sunflowAPI.parameter("caustics", "kd");
        	sunflowAPI.parameter("caustics.gather", properties.getCausticsGather());
        	sunflowAPI.parameter("caustics.radius", (float) properties.getCausticsRadius());
        	sunflowAPI.parameter("caustics.filter", (float) properties.getCausticsFilter());
        }
        
        switch(properties.getGIMethod()) {
        case 0:
        	// No GI engine, so no parameters to set
        	break;
        case 1:
        	// Fake GI engine
        	{
        		Vector3D up=properties.getFakeUp();
        		Vector3 sunflowUp=new Vector3((float) up.X1, (float) up.X2, (float) up.X3);
        		Colour sky=properties.getFakeSky();
        		Colour ground=properties.getFakeGround();
        		sunflowAPI.parameter("gi.engine","fake");
        		sunflowAPI.parameter("gi.fake.up", sunflowUp);
        		sunflowAPI.parameter("gi.fake.sky", "sRGB linear", (float) sky.red, (float) sky.green, (float) sky.blue);
        		sunflowAPI.parameter("gi.fake.ground", "sRGB linear",(float) ground.red, (float) ground.green, (float) ground.blue);
        	}
        	break;
        case 2:
        	// Irradiance cache GI engine
        	sunflowAPI.parameter("gi.engine","irr-cache");
        	sunflowAPI.parameter("gi.irr-cache.samples",properties.getIrrCacheSamples());
        	sunflowAPI.parameter("gi.irr-cache.tolerance", (float) properties.getIrrCacheTolerance());
        	sunflowAPI.parameter("gi.irr-cache.min_spacing", (float) properties.getIrrCacheMinSpacing());
        	sunflowAPI.parameter("gi.irr-cache.max_spacing", (float) properties.getIrrCacheMaxSpacing());
        	switch(properties.getIrrCacheGMap()) {
        	case 0:
            	sunflowAPI.parameter("gi.irr-cache.gmap","none");
        		break;
        	case 1:
            	sunflowAPI.parameter("gi.irr-cache.gmap","kd");
        		break;
        	case 2:
            	sunflowAPI.parameter("gi.irr-cache.gmap","grid");
        		break;
        	}
        	sunflowAPI.parameter("gi.irr-cache.gmap.emit", properties.getIrrCacheEmit());
        	sunflowAPI.parameter("gi.irr-cache.gmap.gather", properties.getIrrCacheGather());
        	sunflowAPI.parameter("gi.irr-cache.gmap.radius", (float) properties.getIrrCacheRadius());
        	break;
        case 3:
        	// Path tracing GI engine
        	sunflowAPI.parameter("gi.engine","path");
        	sunflowAPI.parameter("gi.path.samples", properties.getPathSamples());
        	break;
        case 4:
        	// Ambient occlusion GI engine
        	{
        		Colour bright=properties.getFakeSky();
        		Colour dark=properties.getFakeGround();
            	sunflowAPI.parameter("gi.engine","ambocc");
        		sunflowAPI.parameter("gi.ambocc.bright", "sRGB linear", (float) bright.red, (float) bright.green, (float) bright.blue);
        		sunflowAPI.parameter("gi.ambocc.dark", "sRGB linear", (float) dark.red, (float) dark.green, (float) dark.blue);
        		sunflowAPI.parameter("gi.ambocc.samples", properties.getAmbOccSamples());
        		sunflowAPI.parameter("gi.ambocc.maxdist", (float) properties.getAmbOccMaxDistance());
        	}
        	break;
        case 5:
        	// Instant GI engine
        	sunflowAPI.parameter("gi.engine","igi");
        	sunflowAPI.parameter("gi.igi.samples", properties.getInstantSamples());
        	sunflowAPI.parameter("gi.igi.sets", properties.getInstantSets());
        	sunflowAPI.parameter("gi.igi.c", (float) properties.getInstantBiasBound());
        	sunflowAPI.parameter("gi.igi.bias_samples", properties.getInstantBiasRays());
        	break;
        }
        sunflowAPI.options(SunflowAPI.DEFAULT_OPTIONS);
	}

}
