/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Mar 18, 2007
 */
package eu.moonlight3d.ml3d.ui.mainwindow;

import java.util.ArrayList;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.ui.mainwindow.ogeditor.ModelGenerator;
import eu.moonlight3d.ml3d.ui.mainwindow.ogeditor.OGListener;
import eu.moonlight3d.ml3d.ui.mainwindow.outline.OutlineView;
import eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.PropertyEditorView;
import eu.moonlight3d.ml3d.ui.mainwindow.view.OpenGLConverter;

/**
 * The document lifecycle listener for the core main window components. It is
 * responsible for registering the various listeners of the main window components
 * in new documents.
 *
 * @author gregor
 */
public class LifecycleListener implements eu.moonlight3d.framework.document.LifecycleListener {
	/**
	 * The mode manager's selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.modes.SelectionListener modeSelectionListener;
	/**
	 * The operatr graoh editor's selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.ogeditor.SelectionListener ogEditorSelectionListener;
	/**
	 * The operator graph editor's operator graph listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.ogeditor.OGListener ogEditorOGListener;
	/**
	 * the outline's selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.outline.SelectionListener outlineSelectionListener;
	/**
	 * The outline's operator graph listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.outline.OperatorGraphListener outlineOGListener;
	/**
	 * The property editor's operator graph observer
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.OperatorGraphObserver propertyEditorOGObserver;
	/**
	 * The property editor's selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.SelectionListener propertyEditorSelectionListener;
	/**
	 * The property editor's animation selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.AnimationSelectionListener propertyEditorAnimationSelectionListener;
	/**
	 * The 3D view's selection listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.view.SelectionListener viewSelectionListener;
	/**
	 * The 3D view's operator graph listener
	 */
	private eu.moonlight3d.ml3d.ui.mainwindow.view.OGListener viewOGListener;
	
	/**
	 * Default constructor.
	 */
	public LifecycleListener() {
	}
	
	/**
	 * Create the selection named "viewSelection" which most main window components 
	 * and tools expect.
	 * 
	 * @param document the document to create the selection in
	 */
	private void createViewSelection(Document document) {
        try {
			document.getSelectionManager().addSelection("viewSelection");
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}		
	}
	
	
	/**
	 * Create the selection named "animationSelection" which is used to denote the
	 * current animation that is edited for example through in the property editor.
	 * 
	 * @param document the document to create the selection in
	 */
	private void createAnimationSelection(Document document) {
        try {
			document.getSelectionManager().addSelection("animationSelection");
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}		
	}
	
	/**
	 * Register the selection listener for the mode management.
	 * 
	 * @param document the document to register the selection listener in
	 */
	private void registerModeSelectionListener(Document document) {
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			modeSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.modes.SelectionListener(viewSelection);
			viewSelection.registerListener(modeSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}		
	}
	
	/**
	 * Register the operator graph and selection listener for the operator graph editor.
	 * 
	 * @param document the document to register the listeners in
	 */
	private void registerOGEditorListeners(Document document) {
		ogEditorOGListener=new OGListener();
		document.getOGManager().registerOperatorGraphListener(ogEditorOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			ogEditorSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.ogeditor.SelectionListener();
			viewSelection.registerListener(ogEditorSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * Register the operator graph and selection listeners of the outline.
	 * 
	 * @param document the document to register the listeners in
	 */
	private void registerOutlineListeners(Document document) {
		outlineOGListener=new eu.moonlight3d.ml3d.ui.mainwindow.outline.OperatorGraphListener();
		document.getOGManager().registerOperatorGraphListener(outlineOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			outlineSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.outline.SelectionListener();
			viewSelection.registerListener(outlineSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}
	
	/**
	 * Register the operator graph observer and selection listener for the property editor
	 * 
	 * @param document the document to register the listeners in
	 */
	private void registerPropertyEditorListeners(Document document) {
		propertyEditorOGObserver=new eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.OperatorGraphObserver();
		document.getOGManager().registerObserver(propertyEditorOGObserver);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			propertyEditorSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.SelectionListener();
			viewSelection.registerListener(propertyEditorSelectionListener);

			Selection animationSelection=document.getSelectionManager().getSelection("animationSelection");
			propertyEditorAnimationSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor.AnimationSelectionListener();
			animationSelection.registerListener(propertyEditorAnimationSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}				
	}
	
	/**
	 * Register the operator graph and selection listeners for the 3D view
	 * 
	 * @param document the document to register the listeners in
	 */
	private void registerViewListeners(Document document) {
		viewOGListener=new eu.moonlight3d.ml3d.ui.mainwindow.view.OGListener();
		document.getOGManager().registerOperatorGraphListener(viewOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelectionListener=new eu.moonlight3d.ml3d.ui.mainwindow.view.SelectionListener();
			viewSelection.registerListener(viewSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}				
	}
	
	/**
	 * Register all required listeners in the current main document.
	 */
	private void registerAll() {
		State state=State.getInstance();
		Document document=((Document) state.getDocumentManager().getMainDocument());
		createViewSelection(document);
		createAnimationSelection(document);
		registerModeSelectionListener(document);
		registerOGEditorListeners(document);
		registerOutlineListeners(document);
		registerPropertyEditorListeners(document);
		registerViewListeners(document);		
	}
	
	/**
	 * Unregister the selection listner for the mode manager
	 * 
	 * @param document the document to unregister the listener in
	 */
	private void unregisterModeSelectionListener(Document document) {
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelection.unregisterListener(modeSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
		// explicitely unreference listener to get objects garbage collected
		modeSelectionListener=null;
	}

	/**
	 * Unregister the operator graph listeners.
	 * 
	 * @param document the document to unregister the listener in
	 */
	private void unregisterOGEditorListeners(Document document) {
		document.getOGManager().unregisterOperatorGraphListener(ogEditorOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelection.unregisterListener(ogEditorSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
		// explicitely unreference listeners to get objects garbage collected
		ogEditorOGListener=null;
		ogEditorSelectionListener=null;
	}

	/**
	 * Unregister the outline listeners
	 * 
	 * @param document the document to unregister the listeners in
	 */
	private void unregisterOutlineListeners(Document document) {
		document.getOGManager().unregisterOperatorGraphListener(outlineOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelection.unregisterListener(outlineSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
		// explicitely unreference listeners to get objects garbage collected
		outlineOGListener=null;
		outlineSelectionListener=null;
	}
	
	/**
	 * Unregister the property editor listeners
	 * 
	 * @param document the document to unregister the listeners in
	 */
	private void unregisterPropertyEditorListeners(Document document) {
		document.getOGManager().unregisterObserver(propertyEditorOGObserver);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelection.unregisterListener(propertyEditorSelectionListener);

			Selection animationSelection=document.getSelectionManager().getSelection("animationSelection");
			animationSelection.unregisterListener(propertyEditorAnimationSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}				
		// explicitely unreference listeners to get objects garbage collected
		propertyEditorOGObserver=null;
		propertyEditorSelectionListener=null;
		propertyEditorAnimationSelectionListener=null;
	}
	
	/**
	 * Unregister the 3D view listeners
	 * 
	 * @param document the document to unregister the listeners in
	 */
	private void unregisterViewListeners(Document document) {
		document.getOGManager().unregisterOperatorGraphListener(viewOGListener);
		try {
			Selection viewSelection=document.getSelectionManager().getSelection("viewSelection");
			viewSelection.unregisterListener(viewSelectionListener);
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
		// explicitely unreference listeners to get objects garbage collected
		viewOGListener=null;
		viewSelectionListener=null;
	}
	
	/**
	 * Unregister all listeners in the current main document
	 */
	void unregisterAll() {
		State state=State.getInstance();
		Document document=((Document) state.getDocumentManager().getMainDocument());
		unregisterModeSelectionListener(document);
		unregisterOGEditorListeners(document);		
		unregisterOutlineListeners(document);
		unregisterPropertyEditorListeners(document);
		unregisterViewListeners(document);		
	}
	
	/**
	 * Clear the all of the data displayed in any main window components.
	 */
	private void clearAll() {
		OpenGLConverter.getInstance().clearScene();
		OpenGLConverter.getInstance().redrawViews();
		ModelGenerator generator=ModelGenerator.getInstance();
		generator.clearGraph();
		OutlineView outlineView=OutlineView.getInstance();
		outlineView.clear();
		ArrayList<PropertyEditorView> propertyEditors=PropertyEditorView.getPropertyEditors();
		for(PropertyEditorView propertyEditor : propertyEditors) {
			propertyEditor.clear();
		}
	}
	
	public void afterNew() {
		registerAll();
	}

	public void afterOpen(String filename) {
		registerAll();
		ModelGenerator generator=ModelGenerator.getInstance();
//		generator.updateGraph();
	}

	public void afterSave(String filename) {
	}

	public void beforeNew() {
	}

	public void beforeOpen(String filename) {
	}

	public void beforeSave(String filename) {
	}

	public void onClose() {
		unregisterAll();
		clearAll();
	}

}
