/*
 * Moonlight|3D Copyright (C) 2007 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Feb 20, 2007
 */
package eu.moonlight3d.ml3d.ui.mainwindow.ogeditor;

import java.util.ArrayList;

import com.trolltech.qt.gui.QMessageBox;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.cache.Manager;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.og.Link;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;
import eu.moonlight3d.ml3d.modelling.basic.CopyFromCache;
import eu.moonlight3d.ml3d.ui.mainwindow.toolchest.Tool;

public class TruncateHistoryTool extends Tool {
	public TruncateHistoryTool() {
		setName("TruncateHistoryTool");
		setDisplayName("Truncate History");
		setCategory("Edit/Operator Graph");
	}
	
	private boolean removeNodes(Node ogNode, ArrayList<Node> candidates, ArrayList<Node> newCandidates) throws Exception {
		State state=eu.moonlight3d.framework.State.getInstance();

		// if the node still has output slots connected it must not be removed
		if(!ogNode.isDrain()) {
			return false;
		}
		
		for(Slot slot : ogNode.getSlots()) {
			if(slot.getType()==Slot.Type.Input) {
				Link link=slot.getCurrentLink();
				if(link!=null) {
					link.getInput().setCurrentLink(null);
					slot.setCurrentLink(null);
					((Document) state.getDocumentManager().getMainDocument()).getOGManager().removeLink(link);
					
					if(link.getInput().getNode().isDrain()) {
						removeNodes(link.getInput().getNode(),candidates,newCandidates);
					} else {
						newCandidates.add(link.getInput().getNode());						
					}
				}
			}
		}
		((Document) state.getDocumentManager().getMainDocument()).getOGManager().removeNode(ogNode);
		return true;
	}

	@Override
	public void activated() {
		try {
			State state=eu.moonlight3d.framework.State.getInstance();
			Selection selection=((Document) state.getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
			ArrayList<Node> candidates=new ArrayList<Node>();
			
			if(!selection.getType().getName().equals("OGNode")) {
				state.logger().info("History truncation can only be performed on operator graph nodes");
				return;
			}
			
			if(selection.getSelection().size()==0) {
				state.logger().info("No operator graph node selected");
				return;
			}

			String title=L10N.translate("Confirm Truncate");
			String message=L10N.translate("Do you really want to truncate the node history at this point? All previous nodes will be removed permanently.");
			QMessageBox.StandardButton result=QMessageBox.question(state.getUIManager().getActiveApplicationWindow(), title, message,new QMessageBox.StandardButtons(QMessageBox.StandardButton.Yes,QMessageBox.StandardButton.No));
			if(result!=QMessageBox.StandardButton.Yes) {
				return;
			}
			
			ArrayList<Object> selectedObjects=selection.getSelectionObjects();
			selection.clearSelection(selection.getType());
			
			for(Object ogNodeObject : selectedObjects) {
				Node ogNode=(Node)ogNodeObject;
				
				if(ogNode.isSource()) {
					state.logger().info("Node " + ogNode.getName() + " is a source node without removeable history");
					continue;
				}
				for(Slot slot : ogNode.getSlots()) {
					if(slot.getType()==Slot.Type.Output) {
						SceneGraph graph=slot.getGraph();
						
						CopyFromCache copyFromCache=(CopyFromCache) ((Document) state.getDocumentManager().getMainDocument()).getOGManager().createNodeByName("CopyFromCache");
						
						Manager cacheManager=((Document) state.getDocumentManager().getMainDocument()).getCacheManager();
						cacheManager.addSG("OperatorGraph/" + copyFromCache.getName(),graph);
						
						copyFromCache.getProperty("cachedGraph").setValue("OperatorGraph/" + copyFromCache.getName());
						
						// relink graph
						Link link=slot.getCurrentLink();
						if(link!=null) {
							link.setInput(copyFromCache.getSlot("output"));
						}
					}
				}

				if(ogNode.isDrain()) {
					candidates.add(ogNode);
				}
			}
			
			// now delete old scene history
			// Roadmap:
			// 1. iterate until no more node removed:
			// 1.1 recursively for all nodes in list:
			// 1.1.1 if input is connected to another OG node with exactly one output, remove current OG node, make other OG node current
			// 1.1.2 finish if current OG node is source node or if og node has connected output
			
			boolean removedNode=false;
			
			do {
				removedNode=false;
				ArrayList<Node> newCandidates=new ArrayList<Node>();
				for(Node ogNode : candidates) {
					if(removeNodes(ogNode,candidates,newCandidates)) {
						removedNode=true;
					}
				}
				
				candidates=newCandidates;
				if(candidates.size()==0) {
					break;
				}
			} while(removedNode);
			
			((Document) state.getDocumentManager().getMainDocument()).getOGManager().update();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
	}

	@Override
	public void deactivated() {
	}

	@Override
	public QWidget getOptionEditor(QWidget parent) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public boolean hasOptionEditor() {
		// TODO Auto-generated method stub
		return false;
	}

}
