/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Apr 8, 2008
 */

package eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor;

import java.util.ArrayList;

import com.trolltech.qt.gui.QPushButton;
import com.trolltech.qt.gui.QWidget;

import eu.moonlight3d.common.ui.propertysheet.PropertySheet;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.Property;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.anim.Animation;
import eu.moonlight3d.ml3d.backend.threedim.anim.AnimationTrack;
import eu.moonlight3d.ml3d.backend.threedim.anim.KeyframeAnimationTrack;
import eu.moonlight3d.ml3d.backend.threedim.anim.KeyframeAnimationTrack.KeyValuePair;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.helper.AnimationHelper;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;

/**
 * The keyframe toggle button that appears in the property editor.
 * This button implements its own logic as far as possible
 *
 * @author gregor
 */
public class KeyframeButton extends QPushButton {
	/**
	 * The property sheet that this button is attached to
	 */
	private PropertySheet propertySheet;
	/**
	 * The property that this button is attached to
	 */
	private Property property;
	
	/**
	 * Constructor taking the parent widget, the parent property sheet and the property
	 * that it is attached to as arguments
	 * 
	 * @param parent the parent widget
	 * @param propertySheet the property sheet that this button belongs to
	 * @param property the property that this button is assigened to
	 */
	public KeyframeButton(QWidget parent, PropertySheet propertySheet, Property property) {
		super(parent);
		this.propertySheet=propertySheet;
		this.property=property;

		setText(L10N.translate("K"));
        setToolTip(L10N.translate("Set or remove keyframe"));
        clicked.connect(this,"toggleKeyframe()");
        setCheckable(true);
        setMaximumWidth(20);
		
		// register this button in the global list
        PropertyEditorView.addKeyframeButton(this);
	}

	/**
	 * A slot to create or remove a keyframe at the current time in the animation that is
	 * called when the button is toggled. 
	 */
    protected void toggleKeyframe() {
    	try {
        	Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			Selection selection=document.getSelectionManager().getSelection("animationSelection");
			
			String activeObject=selection.getActiveObject();
			if(activeObject==null) {
				State.getInstance().logger().info(L10N.translate("no animation selected"));
				setChecked(false);
				return;
			}
			Animation animation=(Animation) selection.getObjectFromSelectionString(activeObject);
			
			AnimationHelper.toggleKeyframe(animation, ((Node) propertySheet.getPropertyContainer()), property);
			updateState();
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
    }
    
    /**
     * Update the state of the button assuming that the current time is the given time to
     * reflect changes in the animation.
     * 
     * @param time the assumed current time of the animation
     */
    void updateState(Double time) {
    	try {
        	Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			Selection selection=document.getSelectionManager().getSelection("animationSelection");
			
			String activeObject=selection.getActiveObject();
			if(activeObject==null) {
				setChecked(false);
				setEnabled(false);
				return;
			}
			Animation animation=(Animation) selection.getObjectFromSelectionString(activeObject);
			
			//  look for track
			AnimationTrack track=animation.getTrackByName(((Node)propertySheet.getPropertyContainer()).getName(), property.getName());
			if(track==null) {
				// handle missing track
				setChecked(false);
				setEnabled(true);
				return;
			}

			if(track instanceof KeyframeAnimationTrack) {
				setEnabled(true);
				setChecked(false);
				// look for keyframe
				KeyframeAnimationTrack keyframeTrack=(KeyframeAnimationTrack)track;
				
				if(time==null) {
					// handle defaults case
					setEnabled(true);
					setChecked(true);
					return;
				}
				
				ArrayList<KeyValuePair> keys = keyframeTrack.getKeyList();
				for(KeyValuePair key : keys) {
					if(key.time==time) {
						setChecked(true);
						break;
					}
				}
			} else {
				// not a supported track type
				setChecked(false);
				setEnabled(false);
			}
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}    	
    }
    
    /**
     * Update the state of the button to reflect a change in the animation
     */
    void updateState() {
    	try {
        	Document document=(Document) State.getInstance().getDocumentManager().getMainDocument();
			Selection selection=document.getSelectionManager().getSelection("animationSelection");
			
			String activeObject=selection.getActiveObject();
			if(activeObject==null) {
				setChecked(false);
				setEnabled(false);
				return;
			}
			Animation animation=(Animation) selection.getObjectFromSelectionString(activeObject);
			
			updateState(animation.getCurrentTime());			
		} catch (Exception e) {
			ExceptionDialog.run(e);
		}
    }

}
