/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 13, 2005
 */
package eu.moonlight3d.ml3d.ui.mainwindow.propertyeditor;


import java.util.ArrayList;
import java.util.HashSet;

import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QComboBox;
import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QVBoxLayout;

import eu.moonlight3d.common.ui.propertysheet.PropertyEditorFactorySet;
import eu.moonlight3d.common.ui.propertysheet.PropertySheet;
import eu.moonlight3d.framework.State;
import eu.moonlight3d.framework.helper.Value;
import eu.moonlight3d.framework.localization.L10N;
import eu.moonlight3d.framework.ui.DockWidget;
import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.selection.Selection;
import eu.moonlight3d.ml3d.backend.threedim.anim.Animation;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.og.Node;
import eu.moonlight3d.ml3d.backend.threedim.og.Slot;

/**
 * This is the View implementation for the property editor plugin.
 * 
 * @author gregor
 */
public class PropertyEditorView extends DockWidget {
	/**
	 * List of all OG node property editors that are currently in existence.
	 */
    private static ArrayList<PropertyEditorView> propertyEditors;

    static {
    		propertyEditors=new ArrayList<PropertyEditorView>();
    }
    
    /**
     * Register a newly created property editor.
     * 
     * @param propertyEditor the property editor to register
     */
    public static void addPropertyEditor(PropertyEditorView propertyEditor)
    {
        propertyEditors.add(propertyEditor);
    }
    
    /**
     * Unregister a property editor
     * 
     * @param propertyEditor the property editor to unregister
     */
    public static void removePropertyEditor(PropertyEditorView propertyEditor)
    {
        propertyEditors.remove(propertyEditor);
    }
    
    /**
     * Get a list of all OG node property editors that are currently in existence.
     * @return list of all OG node property editors
     */
    public static ArrayList<PropertyEditorView> getPropertyEditors() {
    		return propertyEditors;
    }
    
    /**
     * A list of currently present keyframe buttons.
     */
    private static ArrayList<KeyframeButton> keyframeButtons;
    
    static {
    	keyframeButtons=new ArrayList<KeyframeButton>();
    	PropertySheet.addPropertyEditorFactory(new AnimatedBooleanEditorFactory());
    	PropertySheet.addPropertyEditorFactory(new AnimatedColourEditorFactory());
    	PropertySheet.addPropertyEditorFactory(new AnimatedFloatEditorFactory());
    	PropertySheet.addPropertyEditorFactory(new AnimatedIntegerEditorFactory());
    	PropertySheet.addPropertyEditorFactory(new AnimatedVectorEditorFactory());
    }
    
    /**
     * Register a keyframe button so that it can receive update notifications when
     * the animation changed.
     * 
     * @param button the keyframe button to register
     */
    static void addKeyframeButton(KeyframeButton button) {
    	keyframeButtons.add(button);
    }
    
    /**
     * Label on top of the property sheet dock
     */
    private QLabel label;
    /**
     * Dropdown with all nodes that are connected to the current OG node
     */
    private QComboBox connectedNodes;
    /**
     * Flag that is set when the connected nodes list is repopulated to prevent
     * cyclic triggering of events
     */
    private boolean populateConnectedNodes;
    /**
     * The property sheet of this dock
     */
    private PropertySheet propertySheet;
    
	/**
	 * @param parent
	 * @param style
	 */
	public PropertyEditorView() {
		super("Property editor");
		
		setDefaultArea(Qt.DockWidgetArea.RightDockWidgetArea);
		
        label=new QLabel(this);
        label.setText(L10N.translate("Selected node:"));
        connectedNodes=new QComboBox();
        connectedNodes.setEditable(false);
        connectedNodes.setEnabled(false);
        connectedNodes.setMaxVisibleItems(30);
        connectedNodes.currentIndexChanged.connect(this,"handleConnectedNodesChange()");
        populateConnectedNodes=false;
        
        QHBoxLayout topLayout=new QHBoxLayout();
        
        topLayout.addWidget(label);
        topLayout.setStretchFactor(label, 0);
        topLayout.addWidget(connectedNodes);
        topLayout.setStretchFactor(connectedNodes, 1);
        
        
        propertySheet=new PropertySheet(this);
        propertySheet.addPropertySheetListener(new PropertySheetListener());
        PropertyEditorFactorySet factorySet=PropertyEditorFactorySet.getDefaultPropertyEditorFactorySet();
        factorySet.addDefaultPropertyEditorFactory(Value.Type.Boolean, new AnimatedBooleanEditorFactory());
        factorySet.addDefaultPropertyEditorFactory(Value.Type.Colour, new AnimatedColourEditorFactory());
        factorySet.addDefaultPropertyEditorFactory(Value.Type.Float, new AnimatedFloatEditorFactory());
        factorySet.addDefaultPropertyEditorFactory(Value.Type.Integer, new AnimatedIntegerEditorFactory());
        factorySet.addDefaultPropertyEditorFactory(Value.Type.Vector3D, new AnimatedVectorEditorFactory());
        propertySheet.setPropertyEditorFactorySet(factorySet);
        
        QVBoxLayout layout=new QVBoxLayout();
        layout.addLayout(topLayout);
        layout.addWidget(propertySheet,1);
        
        layout.setMargin(0);
        
        setLayout(layout);
        
        addPropertyEditor(this);
	}
	
	/**
	 * Handle a change of the selected element in the connectedNodes combo box
	 */
	protected void handleConnectedNodesChange() {
		if(populateConnectedNodes) {
			return;
		}
		// display properties of selected node
		Manager manager=((Document) State.getInstance().getDocumentManager().getMainDocument()).getOGManager();
		
		String nodeName=connectedNodes.currentText();
		
        Node ogNode=manager.getNodeByName(nodeName);
        keyframeButtons.clear();
		propertySheet.setPropertyContainer(ogNode);
	}
	
	/**
	 * Find all nodes that connect to the input side of the given node and add them
	 * to the array that is passed as parameter. The first node in the path is not
	 * added to the list, because fillConnectedNodes() deals with it seprately.
	 * 
	 * @param node the node to process
	 * @param firstNode true if this is the first node in the path
	 * @param nodes the list of nodes that stores the result
	 * @param visitedNodes a set of already visited OG nodes used to prevent double entries of nodes
	 */
	private void findInputNodes(Node node, boolean firstNode, ArrayList<Node> nodes, HashSet<Node> visitedNodes) {
		for(Slot slot : node.getSlots()) {
			if(slot.getType()==Slot.Type.Input) {
				if(slot.getCurrentLink()!=null) {
					findInputNodes(slot.getCurrentLink().getInput().getNode(), false, nodes, visitedNodes);
				}
			}
		}

		// add node to list
		if(!firstNode && !visitedNodes.contains(node)) {
			nodes.add(node);
			visitedNodes.add(node);
		}
	}
	
	/**
	 * Find all nodes that connect to the output side of the given node and add them
	 * to the array that is passed as parameter. The first node in the path is not
	 * added to the list, because fillConnectedNodes() deals with it seprately.
	 * 
	 * @param node the node to process
	 * @param firstNode true if this is the first node in the path
	 * @param nodes the list of nodes that stores the result
	 * @param visitedNodes a set of already visited OG nodes used to prevent double entries of nodes
	 */
	private void findOutputNodes(Node node, boolean firstNode, ArrayList<Node> nodes, HashSet<Node> visitedNodes) {
		// add node to list
		if(!firstNode && !visitedNodes.contains(node)) {
			nodes.add(node);
			visitedNodes.add(node);
		}
		
		for(Slot slot : node.getSlots()) {
			if(slot.getType()==Slot.Type.Output) {
				if(slot.getCurrentLink()!=null) {
					findOutputNodes(slot.getCurrentLink().getOutput().getNode(), false, nodes, visitedNodes);
				}
			}
		}
	}
	
	/**
	 * Fill the connectedNodes combobox with the list of connected operator graph nodes
	 * 
	 * @param selectedNode the currently selected OG node
	 */
	private void fillConnectedNodes(Node selectedNode) {
		populateConnectedNodes=true;
		
		connectedNodes.clear();
		connectedNodes.setEnabled(true);
		
		HashSet<Node> visitedNodes=new HashSet<Node>();
		
		// find all nodes that are connected to the input
		ArrayList<Node> inputNodes=new ArrayList<Node>();
		findInputNodes(selectedNode, true, inputNodes, visitedNodes);
		for(int i=0;i<inputNodes.size();i++) {
			connectedNodes.addItem(inputNodes.get(i).getName());
		}
		
		// add actually selected node
		connectedNodes.addItem(selectedNode.getName());
		connectedNodes.setCurrentIndex(connectedNodes.count()-1);
		
		// find all nodes that are connected to the output
		ArrayList<Node> outputNodes=new ArrayList<Node>();
		findOutputNodes(selectedNode, true, outputNodes, visitedNodes);
		for(int i=0;i<outputNodes.size();i++) {
			connectedNodes.addItem(outputNodes.get(i).getName());
		}
		
		populateConnectedNodes=false;
	}

	/**
	 * Handle an update of the view selection by displaying the properties of the related
	 * operator graph node in the property editor.
	 */
    public void handleSelectionUpdate()
    {
        try {
            Selection selection=((Document) State.getInstance().getDocumentManager().getMainDocument()).getSelectionManager().getSelection("viewSelection");
            
        	keyframeButtons.clear();

        	String activeObjectName=selection.getActiveObject();
            if(activeObjectName==null) {
                connectedNodes.clear();
                connectedNodes.setEnabled(false);
                propertySheet.setPropertyContainer(null);
                return;
            }
            String activeObjectNameParts[]=activeObjectName.split("/");
            if(!activeObjectNameParts[0].equals("OperatorGraph")) {
                connectedNodes.clear();
                connectedNodes.setEnabled(false);
                propertySheet.setPropertyContainer(null);
                return;
            }
            Node ogNode=(Node) selection.getObjectFromSelectionString(activeObjectNameParts[0] + "/" + activeObjectNameParts[1]);
  
            // update connectedNodes combo box
            fillConnectedNodes(ogNode);
            
            propertySheet.setPropertyContainer(ogNode);

        	for(KeyframeButton button : keyframeButtons) {
        		button.updateState();
        	}
        } catch (Exception e) {
            ExceptionDialog.run(e);
        }
    }
    
    /**
     * Force an update of the displayed property values.
     */
    public void handleValueUpdate() {
   		propertySheet.updateValues();
    }
    
    /**
     * Force an update of displayed property values and keyframe button states after
     * the animation time has changed.
     */
    public void handleAnimationUpdate() {
    	propertySheet.updateValues();
    	for(KeyframeButton button : keyframeButtons) {
    		button.updateState();
    	}
    }
    
    /**
     * Force an update of displayed property values and keyframe button states after
     * the animation time has changed.
     * 
     * @param newTime the new time that the animation is being set to
     */
    public void handleAnimationUpdate(Double newTime) {
    	propertySheet.updateValues();
    	for(KeyframeButton button : keyframeButtons) {
    		button.updateState(newTime);
    	}
    }
    
    /**
     * Handle the addition of a new animation track.
     * 
     * @param newAnimation the  new animation track that was added
     */
    public void handleAnimationChange(Animation newAnimation) {
    	handleAnimationUpdate();
    }
    
    /**
     * Reset the property editior to an empty state with no OG node assigned.
     */
    public void clear() {
    	connectedNodes.clear();
    	connectedNodes.setEnabled(false);
        propertySheet.setPropertyContainer(null);    	
    }
}
