/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 31, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.view;

import javax.media.opengl.GL;

import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.ml3d.backend.threedim.sg.Instance;


/**
 *
 * @author gregor
 */
public class InstanceNode extends RenderableNode {
	/**
	 * Local transformation matrix relative to parent.
	 */
	private Matrix4 transformationMatrix;
	
	/**
	 * Used to cache accumulated transformation matrices 
	 * of all parent nodes while descending down the
	 * graph during rendering. 
	 */
	private Matrix4 parentTransformationMatrix;

	/**
	 * Creates a new RenderableGroup with the given source operator
	 * graph and scene graph nodes and the given parent RenderableGroup.
	 * 
	 * @param ogNode the source operator graph node for the contained elements
	 * @param sgNode the source scene graph node for the contained elements
	 */
	public InstanceNode(eu.moonlight3d.ml3d.backend.threedim.sg.Node sgNode) {
		if(sgNode instanceof Instance) {
			transformationMatrix=((Instance)sgNode).getTransformationMatrix();
		} else {
			transformationMatrix=new Matrix4();
			transformationMatrix.setIdentity();
		}
	}

	/**
	 * Set the transformation matrix for this group of renderables.
	 * Use this to render the same nodes at different positions in
	 * the scene.
	 * 
	 * @param transformationMatrix the group's new transformation matrix
	 */
	public void setTransformationMatrix(Matrix4 transformationMatrix) {
		this.transformationMatrix=transformationMatrix.clone();
	}
	
	/**
	 * Get the group's current transformation matrix.
	 * 
	 * @return the group's transformation matrix
	 */
	public Matrix4 getTransformationMatrix() {
		return transformationMatrix.clone();
	}
	
	/**
	 * Apply the instance transformation to the viewport's modelview matrix.
	 * 
	 * @param view the viewport to apply the transformation to
	 */
	private void applyTransformationMatrix(Viewport view) {
		GL gl=view.getGL();
		double temp[]=new double[16];		
		
		gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, temp, 0);
		
		parentTransformationMatrix=new Matrix4();
		parentTransformationMatrix.setFromOpenGLMatrix(temp);
		
		gl.glMatrixMode(GL.GL_MODELVIEW);		
		gl.glLoadMatrixd(parentTransformationMatrix.multiply(transformationMatrix).getOpenGLMatrix(),0);
	}
	
	/**
	 * Restore original modelview matrix that was in use before applyTransformationMatrix
	 * was called the last time
	 * 
	 * @param view the viewport to restore the modelview matrix in
	 */
	private void unapplyTransformationMatrix(Viewport view) {
		GL gl=view.getGL();
		gl.glLoadMatrixd(parentTransformationMatrix.getOpenGLMatrix(),0);
		parentTransformationMatrix=null;
	}
	
	/**
	 * Draw the preview pass of all contained Renderable elements. See
	 * Renderable documentation for details.
	 */
	@Override
	public void drawPreview(Viewport view, Scene scene) {
		applyTransformationMatrix(view);

		for(SceneNode child : getChildren()) {
			if(child instanceof RenderableNode) {
				((RenderableNode)child).drawPreview(view, scene);
			}
		}
		
		unapplyTransformationMatrix(view);
	}

	/**
	 * Draw the visible structure pass of all contained Renderable elements. See
	 * Renderable documentation for details.
	 */
	@Override
	public void drawVisibleStructure(Viewport view, Scene scene) {
		applyTransformationMatrix(view);
		
		for(SceneNode child : getChildren()) {
			if(child instanceof RenderableNode) {
				((RenderableNode)child).drawVisibleStructure(view, scene);
			}
		}
		
		unapplyTransformationMatrix(view);
	}

	/**
	 * Draw the occluded structure pass of all contained Renderable elements. See
	 * Renderable documentation for details.
	 */
	@Override
	public void drawOccludedStructure(Viewport view, Scene scene) {
		applyTransformationMatrix(view);
		
		for(SceneNode child : getChildren()) {
			if(child instanceof RenderableNode) {
				((RenderableNode)child).drawOccludedStructure(view, scene);
			}
		}
		
		unapplyTransformationMatrix(view);
	}

	/**
	 * Draw the overlay pass of all contained Renderable elements. See
	 * Renderable documentation for details.
	 */
	@Override
	public void drawOverlay(Viewport view, Scene scene) {
		applyTransformationMatrix(view);
		
		for(SceneNode child : getChildren()) {
			if(child instanceof RenderableNode) {
				((RenderableNode)child).drawOverlay(view, scene);
			}
		}
		
		unapplyTransformationMatrix(view);
	}

	/**
	 * Draw the 2D overlay pass of all contained Renderable elements. See
	 * Renderable documentation for details.
	 */
	@Override
	public void drawOverlay2D(Viewport view, Scene scene) {
		for(SceneNode child : getChildren()) {
			if(child instanceof RenderableNode) {
				((RenderableNode)child).drawOverlay2D(view, scene);
			}
		}
	}
}
