/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 31, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.view;

import javax.media.opengl.GL;

import eu.moonlight3d.colour.Colour;
import eu.moonlight3d.math.Vector3D;


/**
 * OpenGL light source wrapper used for lighting in the 3d view.
 * This class is used to set up light sources in the 3d view
 * based on lights in the scene graph. 
 *
 * @author gregor
 */
public class LightSource {
	/**
	 * Enumeration for supported OpenGL light source types.
	 *
	 * @author gregor
	 */
	public enum Type {
		/**
		 * Omnidirectional light
		 */
		Omni,
		/**
		 * Spotlight
		 */
		Spot
	}

	/**
	 * Light source position in world coordinates
	 */
	private Vector3D position;
	/**
	 * Light source colour.
	 */
	private Colour colour;
	/**
	 * Light source type.
	 */
	private Type type;
	/**
	 * Name of operator graph node that contains the scene
	 * graph with the light source that is represented here.
	 */
	private String ogNodeName;
	
	/**
	 * Create a new light source for OpenGL rendering.
	 * 
	 * @param type light source type
	 * @param position light source position in world coordinates
	 * @param colour light source color
	 */
	public LightSource(Type type, Vector3D position, Colour colour) {
		this.type=type;
		this.position=position;
		this.colour=colour;
	}
	
	public void setOGNodeName(String ogNodeName) {
		this.ogNodeName=ogNodeName;
	}
	
	public String getOGNodeName() {
		return ogNodeName;
	}
	
	/**
	 * Return the light position in world coordinates.
	 * 
	 * @return light source position
	 */
	public Vector3D getPosition() {
		return position;
	}

	/**
	 * Return the light colour.
	 * 
	 * @return light source colour
	 */
	public Colour getColour() {
		return colour;
	}

	/**
	 * Setup the corresponding OpenGL light source with the given number.
	 * 
	 * @param view the viewport holding the target OpenGL context
	 * @param glLightNumber the number of the OpenGL light
	 */
	public void setup(Viewport view, int glLightNumber) {
		GL gl=view.getGL();
		
        float[] ambient=new float[4];
        float[] diffuse=new float[4];
        float[] position=new float[4];

        switch(type) {
		case Omni:
	        gl.glEnable(GL.GL_LIGHT0+glLightNumber);
	        ambient[0]=(float)0.1;
	        ambient[1]=(float)0.1;
	        ambient[2]=(float)0.1;
	        ambient[3]=(float)1.0;
	        diffuse[0]=(float)colour.red;
	        diffuse[1]=(float)colour.green;
	        diffuse[2]=(float)colour.blue;
	        diffuse[3]=(float)1.0;
	        position[0]=(float)this.position.X1;
	        position[1]=(float)this.position.X2;
	        position[2]=(float)this.position.X3;
	        position[3]=(float)1.0;
	        gl.glLightfv(GL.GL_LIGHT0+glLightNumber,GL.GL_DIFFUSE,diffuse,0);
	        gl.glLightfv(GL.GL_LIGHT0+glLightNumber,GL.GL_POSITION,position,0);

	        gl.glEnable(GL.GL_COLOR_MATERIAL);
	        gl.glColorMaterial(GL.GL_FRONT_AND_BACK,GL.GL_AMBIENT_AND_DIFFUSE);
	        gl.glLightModelf(GL.GL_LIGHT_MODEL_TWO_SIDE,GL.GL_TRUE);
			break;
		case Spot:
			break;
		}
	}
}
