/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 14, 2005
 */
package eu.moonlight3d.ml3d.ui.mainwindow.view;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.ml3d.backend.threedim.document.Document;
import eu.moonlight3d.ml3d.backend.threedim.og.Manager;
import eu.moonlight3d.ml3d.backend.threedim.sg.Node;
import eu.moonlight3d.ml3d.backend.threedim.sg.SceneGraph;

/**
 * This class generates an OpenGL-renderable Scene from the data in the given 
 * operator graph. 
 * 
 * @author gregor
 */
public class OpenGLConverter {
	private static OpenGLConverter instance=null;
	private Scene scene;
	private ArrayList<Viewport> views;
	private ArrayList<SGNodeConverter> sgNodeConverters;
	
	private OpenGLConverter() {
		views=new ArrayList<Viewport>();
		sgNodeConverters=new ArrayList<SGNodeConverter>();

        scene=new Scene();
	}
	
	public void registerSGNodeConverter(SGNodeConverter converter) {
		sgNodeConverters.add(converter);
	}
	
	public void unregisterSGNodeConverter(SGNodeConverter converter) {
		sgNodeConverters.remove(converter);
	}
	
	public void clearScene() {
		scene.clearOGBasedNodes();
	}
	
    public Scene run() throws Exception
	{
//   		long startTime=System.currentTimeMillis();

    	clearScene();
   		
		List<eu.moonlight3d.ml3d.backend.threedim.og.Node> ogNodeList;
		
		// traverse operator graph
		Manager ogManager=((Document) eu.moonlight3d.framework.State.getInstance().getDocumentManager().getMainDocument()).getOGManager();
		ogNodeList = ogManager.getNodeList();
		for (int i = 0; i < ogNodeList.size(); i++) {

			if (!ogNodeList.get(i).isDrain()) { // skip nodes which aren't
												// acting drains
				continue;
			}

			convertOGNodeOutput(ogNodeList.get(i));

		}
		
//		long endTime=System.currentTimeMillis();
// State.getInstance().logger().info("conversion time [ms]:" + new
// Long(endTime-startTime).toString());
		
		return scene;
	}
    
    private void convertSGNodes(eu.moonlight3d.ml3d.backend.threedim.og.Node ogNode, SceneGraph sceneGraph) {
    	HashMap<Node, SceneNode[]> sgNodeMap=new HashMap<Node, SceneNode[]>();
		for(Node node : sceneGraph.getNodes()) {
        	for(SGNodeConverter converter : sgNodeConverters) {
        		if(converter.canConvertNode(node)) {
        			SceneNode[] nodes=converter.convert(node);
        			if(nodes!=null) {
        				for(SceneNode sceneNode : nodes) {
        					sceneNode.setOGNodeName(ogNode.getName());
        					if(node.getParents().size()==0) {
        						sceneNode.setToplevel(true);
        					} else {
        						sceneNode.setToplevel(false);
        					}
        				}
        				scene.addSceneNodes(nodes);
        				sgNodeMap.put(node, nodes);
        			}
        			
        			LightSource[] lightSources = converter.convertLights(node);
        			if(lightSources!=null) {
        				for(LightSource lightSource : lightSources) {
        					lightSource.setOGNodeName(ogNode.getName());
        				}
        				scene.addLightSources(lightSources);
        			}

        			// TODO add contents of SceneNode[] and LightSource[] to scene
        		}
        	}
    	}
		
		for(Node node : sceneGraph.getNodes()) {
			SceneNode[] parentNodes=sgNodeMap.get(node);
			for(Node child : node.getChildren()) {
				SceneNode[] childNodes=sgNodeMap.get(child);
				if(childNodes!=null) {
					for(SceneNode parentNode : parentNodes) {
						for(SceneNode childNode : childNodes) {
							parentNode.addChild(childNode);
						}
					}
				}
			}
		}
		
    }
    
    public void convertOGNodeOutput(eu.moonlight3d.ml3d.backend.threedim.og.Node node) {
    	for(eu.moonlight3d.ml3d.backend.threedim.og.Slot slot : node.getSlots()) {
    		if(slot.getType()==eu.moonlight3d.ml3d.backend.threedim.og.Slot.Type.Output) {
    			try {
    				// TODO: fix WRT to instance semantics
    				if(slot.getGraph()==null) {
    					continue;
    				}
    				convertSGNodes(node, slot.getGraph());
				} catch (Exception e) {
					ExceptionDialog.run(e);
				}
    		}
    	}
    }
    
    public void removeOGNodeOutput(eu.moonlight3d.ml3d.backend.threedim.og.Node node) {
    	scene.removeLightSources(node);
    	scene.removeSceneNodes(node);
    }
    
    public Scene getScene() {
    	return scene;
    }

	public static OpenGLConverter getInstance()
	{
		if(instance==null) {
			instance=new OpenGLConverter();
		}
		return instance;
	}
	
	public void registerView(Viewport view) 
	{
		views.add(view);
	}
	
	public void unregisterView(Viewport view) 
	{
		views.remove(view);
	}
    
    public ArrayList<Viewport> getRegisteredViews()
    {
        return views;
    }

    public void redrawViews()
    {
        for(int i=0;i<views.size();i++) {
            //views.get(i).render();
        	views.get(i).update();
        }        
    }
}
