/*
 * Moonlight|3D Copyright (C) 2005 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Jan 18, 2005
 */
package eu.moonlight3d.ml3d.ui.mainwindow.view;


import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QTabletEvent.PointerType;

/**
 * This is the base class for event handlers for input events on the 3D viewport.
 * Derive this class to define a new behaviour of the 3D view. The class also
 * contains references to the current and default event handler. While the
 * standard view event handler should be left unaltered, tools need to
 * replace the current event handler by their own one when they wish to
 * implement user interaction inside the 3D view.
 * 
 * @author gregor
 */
public class ViewEventHandler {
	/**
	 * Copy of current OpenGL modelview matrix.
	 */
    protected eu.moonlight3d.math.Matrix4 modelViewMatrix;
    /**
     * Copy of current OpenGL projection matrix.
     */
    protected eu.moonlight3d.math.Matrix4 projectionMatrix;
    /**
     * Reference to the viewport in which the event occurred
     * that is currently processed.
     */
    protected Viewport viewport;
    /**
     * Width of the viewport in which the event occurred.
     */
    protected int viewportWidth;
    /**
     * Height of the viewport in which the event occurred.
     */
    protected int viewportHeight;
    
    /**
     * Currently active event handler.
     */
    protected static ViewEventHandler currentEventHandler;
    /**
     * Standard event handler which is used if no other event handler is active.
     */
    protected static ViewEventHandler defaultEventHandler;
    
    /**
     * Set the current event handler. The current event handler gets called
     * whenever an event is recieved by one of the 3d views. This replaces
     * any event handler that has previously been set current.
     * 
     * @param eventHandler the event handler to be set as current handler
     */
    public static void setEventHandler(ViewEventHandler eventHandler)
    {
        if(currentEventHandler!=defaultEventHandler && currentEventHandler!=null) {
            currentEventHandler.onRemove();
        }
        currentEventHandler=eventHandler;
    }
    
    /**
     * Set the default event handler. The default event handler is a fallback
     * event handler that gets called when there is no other event handler registered
     * at the moment an event comes in. This replaces any event handler that has
     * previously been set as default.
     * 
     * @param eventHandler the event handler to be set as default
     */
    public static void setDefaultEventHandler(ViewEventHandler eventHandler)
    {
        defaultEventHandler=eventHandler;
    }
    
    /**
     * Return the current event handler.
     * 
     * @return the currently active event handler
     */
    public static ViewEventHandler getEventHandler()
    {
        return currentEventHandler;
    }
    
    /**
     * Return the default event handler.
     * 
     * @return the current default event handler
     */
    public static ViewEventHandler getDefaultEventHandler()
    {
        return defaultEventHandler;
    }
    
    /**
     * Remove the current event handler. This calls the
     * event handler's onRemove method.
     * 
     * @return the removed event handler
     */
    public static ViewEventHandler removeEventHandler()
    {
        ViewEventHandler old=currentEventHandler;
        currentEventHandler.onRemove();
        currentEventHandler=defaultEventHandler;
        return old;
    }

    /**
     * Constructor.
     */
    public ViewEventHandler()
    {
    }

    /**
     * This method is intended to be implemented by event handlers derived 
     * from thsi class. It gets called whenever a mouse move event is received 
     * by any 3D view. The mouse coordinates are local to the 3D view.
     * 
     * @param oldX the x coordinate of the old mouse position
     * @param oldY the y coordinate of the old mouse position
     * @param newX the x coordinate of the new mouse position 
     * @param newY the y coordinate of the new mouse position
     * @throws Exception
     */
    public void mouseMove(int oldX, int oldY, int newX, int newY) throws Exception
    {
    }

    /**
     * This method is intended to be implemented by event handlers derived 
     * from thsi class. This gets called whenever a mouse button push event 
     * is received by any 3D view
     * 
     * @param button the number of the pressed button
     * @throws Exception
     */
    public void mouseButtonDown(Qt.MouseButton button) throws Exception
    {
    }

    /**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever a mouse button release event
     * is recieved by any 3d view
     *  
     * @param button the number of the released button
     * @throws Exception
     */
    public void mouseButtonUp(Qt.MouseButton button) throws Exception
    {
    }
    
    /**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever a double click event is
     * received by any 3d view
     *  
     * @param button the number of the clicked button
     * @throws Exception
     */
    public void mouseDoubleClick(Qt.MouseButton button) throws Exception
    {
    }

    /**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever a a mouse wheel event is
     * received by any 3d view
     *  
     * @param angle the distance the wheel was turned in degrees
     * @throws Exception
     */   
	public void mouseWheelMove(double angle) throws Exception 
	{
	}

	/**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever motion was reported on any
     * attached graphics tablet
     * 
	 * @param pointerType type of pen used
	 * @param x x coordinate of pen
	 * @param y y coordinate of pen
	 * @param xTilt x tilt of pen
	 * @param yTilt y tilt of pen
	 * @param pressure pen pressure
	 * @throws Exception
	 */
	public void tabletMove(PointerType pointerType, double x, double y, double xTilt, double yTilt, double pressure) throws Exception {
	}

	/**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever a key press event
     * is recieved by any 3d view
     *  
     * @param key the key that has been pressed
     * @param state the state of the modifier keys at that time
     * @throws Exception
     */
    public void keyPress(int key, Qt.KeyboardModifiers state) throws Exception
    {
    }

    /**
     * This method is intended to be implemented by event handlers derived 
     * from this class. This gets called whenever a key release event
     * is recieved by any 3d view
     *  
     * @param key the key that has been released
     * @param state the state of the modifier keys at that time
     * @throws Exception
     */
    public void keyRelease(int key, Qt.KeyboardModifiers state) throws Exception
    {
    }

    /**
     * Internal method to set the viewport that received the event
     *  
     * @param viewport viewport that received last event
     */
    public void setViewport(Viewport viewport)
    {
        this.viewport=viewport;
    }
    
    /**
     * Returns the viewport that received the event that is currently
     * being processed
     * 
     * @return the current viewport
     */
    public Viewport getViewport()
    {
        return viewport;
    }
    
    /**
     * This method is intended to be implemented by event handlers derived
     * from this class. It gets called when this event handler gets
     * removed as current event handler.
     *
     */
    public void onRemove()
    {
    }

}
