/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on May 31, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.view;

import javax.media.opengl.GL;
import javax.media.opengl.GLContext;
import javax.media.opengl.GLDrawableFactory;
import javax.media.opengl.glu.GLU;

import com.trolltech.qt.core.QRect;
import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QCursor;
import com.trolltech.qt.gui.QFocusEvent;
import com.trolltech.qt.gui.QKeyEvent;
import com.trolltech.qt.gui.QMouseEvent;
import com.trolltech.qt.gui.QPaintEvent;
import com.trolltech.qt.gui.QResizeEvent;
import com.trolltech.qt.gui.QTabletEvent;
import com.trolltech.qt.gui.QWheelEvent;
import com.trolltech.qt.gui.QWidget;
import com.trolltech.qt.opengl.QGLFormat;
import com.trolltech.qt.opengl.QGLWidget;

import eu.moonlight3d.framework.ui.ExceptionDialog;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Vector3D;

/**
 * The OpenGL widget that used for rendering a viewport in the UI.
 * This widget contains the viewport specific state and handles
 * basic OpenGL viewport setup.
 *
 * @author gregor
 */
public class Viewport extends QGLWidget {
	/**
	 * The projections that can be used in the viewport.
	 *
	 * @author gregor
	 */
	public enum ProjectionType {
		/**
		 * Perspective projection
		 */
		Perspective,
		/**
		 * Orthographic projection
		 */
		Orthographic
	}
	
	/**
	 * The standard perspectives that the viewport can be reset to.
	 *
	 * @author gregor
	 */
	public enum StandardPerspective {
		/**
		 * Standard perspective projection - the camera is looking at the origin
		 */
		Perspective,
		/**
		 * Orthographic projection, the camera is looking at the xz plane
		 */
		Right,
		/**
		 * Orthographic projection, the camera is looking at the yz plane
		 */
		Front,
		/**
		 * Orthographic projection, the camera is looking at the xy plane
		 */
		Top
	}

	/**
	 * The geoemtry rendering mode that is used to preview the actual 
	 * object geometry.
	 *
	 * @author gregor
	 */
	public enum PreviewMode {
		/**
		 * Do not render actual object geometry
		 */
		None,
		/**
		 * Only draw a wireframe outline of the object geometry
		 */
		Wireframe,
		/**
		 * Draw the geometry shaded and lit by default lighting, ignoring any light sources in the scene
		 */
		Shaded,
		/**
		 * Draw the geometry fully shaded and lit by the light sources within the scene 
		 */
		Lit
	}
	
	/**
	 * The object structure rendering mode that is used to visualise the
	 * object structure (e.g. edges and vertices).
	 *
	 * @author gregor
	 */
	public enum StructureMode {
		/**
		 * Draw the full occluded object structure with occluded parts shown,
		 * preferably in different color
		 */
		Full,
		/**
		 * Draw an occluded object structure with occluded parts fully hidden
		 * from display 
		 */
		Occluded,
		/**
		 * Do not draw any object structure
		 */
		None
	}

	private static GLContext glContext=null;
	
    private GL gl;
    private GLU glu;
    private eu.moonlight3d.math.Matrix4 modelViewMatrix;
    private eu.moonlight3d.math.Matrix4 projectionMatrix;
    private double viewFieldOfView;
    private double matrixBuffer[];
    private boolean lockCameraOrientation;
    private double orthographicScale;
    
    private int oldX, oldY;

	private ProjectionType projectionType;
	private PreviewMode currentPreviewMode;
	private StructureMode currentStructureMode;
	
    public Viewport(QWidget parent, QGLFormat format) {
		super(format,parent);
		
		init();
	}
    
    public Viewport(QWidget parent, QGLFormat format, QGLWidget shareWidget) {
		super(format, parent, shareWidget);
		
		init();
	}

    private void init() {
		OpenGLConverter.getInstance().registerView(this);
        
		projectionType=ProjectionType.Perspective;
		currentPreviewMode=PreviewMode.Shaded;
		currentStructureMode=StructureMode.Occluded;
        matrixBuffer=new double[16];
        modelViewMatrix=new Matrix4();
        projectionMatrix=new Matrix4();
        
        setupViewingAreaInitial();
        
        setFocusPolicy(Qt.FocusPolicy.ClickFocus);
        setMouseTracking(true);
        setCursor(new QCursor(Qt.CursorShape.ArrowCursor));
    }

	public void setStandardPerspective(StandardPerspective perspective) {
		if(perspective==StandardPerspective.Perspective) {
			setProjectionType(ProjectionType.Perspective);
	        // create new modelview matrix starting with an identity matrix
	        modelViewMatrix.clear();
	        for(int i=0;i<4;i++) {
	            modelViewMatrix.X[i][i]=1.0;
	        }
	        // move camera away from origin
	        modelViewMatrix.X[2][3]=-5.0;
	        // rotate camera around x axis
	        Vector3D axis=new Vector3D();
	        axis.clear();
	        axis.X1=1.0;
	        Quaternion helper=new Quaternion();
	        helper.fromAxisAngle(axis,Math.PI/4);
	        modelViewMatrix=modelViewMatrix.multiply(helper.toRotationMatrix4());
	        // rotate camera around z axis
	        axis.clear();
	        axis.X3=1.0;
	        helper.fromAxisAngle(axis,-Math.PI/4);
	        modelViewMatrix=modelViewMatrix.multiply(helper.toRotationMatrix4());
		} else if(perspective==StandardPerspective.Right) {
			setProjectionType(ProjectionType.Orthographic);
			setOrthographicScale(3.0);
	        // create new modelview matrix as an identity matrix
	        modelViewMatrix.clear();
            modelViewMatrix.X[3][3]=1.0;
            modelViewMatrix.X[0][1]=1.0;
            modelViewMatrix.X[1][2]=1.0;
            modelViewMatrix.X[2][0]=1.0;
		} else if(perspective==StandardPerspective.Front) {
			setProjectionType(ProjectionType.Orthographic);
			setOrthographicScale(3.0);
	        // create new modelview matrix as an identity matrix
	        modelViewMatrix.clear();
            modelViewMatrix.X[3][3]=1.0;
            modelViewMatrix.X[0][0]=1.0;
            modelViewMatrix.X[1][2]=1.0;
            modelViewMatrix.X[2][1]=-1.0;
		} else if(perspective==StandardPerspective.Top) {
			setProjectionType(ProjectionType.Orthographic);
			setOrthographicScale(3.0);
	        // create new modelview matrix as an identity matrix
	        modelViewMatrix.clear();
	        for(int i=0;i<4;i++) {
	            modelViewMatrix.X[i][i]=1.0;
	        }
		}
	}
    
    /**
     * Initial setup of the OpenGL context and viewport.
     */
    protected void setupViewingAreaInitial() 
    {
        viewFieldOfView=45.0;
        orthographicScale=1;
        currentPreviewMode=PreviewMode.Shaded;
        
        makeCurrent();
        if(glContext==null) {
        	glContext=GLDrawableFactory.getFactory().createExternalGLContext();
        }
        glContext.makeCurrent();
        gl=glContext.getGL();
        glu=new GLU();
        QRect rect = rect();
        int width = rect.width();
        int height = rect.height();
        height = Math.max(height, 1);
        gl.glViewport(0, 0, width, height);     
        gl.glMatrixMode(GL.GL_PROJECTION);      // select the projection matrix
        gl.glLoadIdentity();                            // reset the projection matrix
        float fAspect = (float) width / (float) height;
        if(projectionType==ProjectionType.Perspective) {
        	glu.gluPerspective(viewFieldOfView, fAspect, 0.5f, 400.0f);
        } else {
        	gl.glOrtho(-orthographicScale,orthographicScale,-orthographicScale/fAspect,orthographicScale/fAspect,-400.0f,400.0f);
        }
        
        setStandardPerspective(StandardPerspective.Perspective);
        
        gl.glGetDoublev(GL.GL_PROJECTION_MATRIX,matrixBuffer,0);
        projectionMatrix.setFromOpenGLMatrix(matrixBuffer);
        
        gl.glDisable(GL.GL_DEPTH_TEST);
        glContext.release();
    }

    /**
	 * Per-frame setup of OpenGL context and viewport 
	 */
	protected void setupViewingArea() {
        QRect rect = rect();
        int width = rect.width();
        int height = rect.height();
        height = Math.max(height, 1);
        
        gl.glViewport(0, 0, width, height);
        
        gl.glMatrixMode(GL.GL_PROJECTION);      // select the projection matrix
        gl.glLoadIdentity();                            // reset the projection matrix
//        gl.glLoadMatrixd(projectionMatrix.getOpenGLMatrix());
        float fAspect = (float) width / (float) height;
        glu=new GLU();
        if(projectionType==ProjectionType.Perspective) {
        	glu.gluPerspective(viewFieldOfView, fAspect, 0.5f, 2000.0f);
        } else {
        	gl.glOrtho(-orthographicScale,orthographicScale,-orthographicScale/fAspect,orthographicScale/fAspect,-2000.0f,2000.0f);
        }
        
        // FIXME this is a hack to have the current projection matrix buffered
        gl.glGetDoublev(GL.GL_PROJECTION_MATRIX,matrixBuffer,0);
        projectionMatrix.setFromOpenGLMatrix(matrixBuffer);
        
        gl.glEnable(GL.GL_DEPTH_TEST);
	}

	/**
	 * Render the contents of the scene in the current viewport.
	 *
	 */
	public void render()
	{
		makeCurrent();
		glContext.makeCurrent();
		gl=glContext.getGL();
//		gl=new TraceGL(glContext.getGL(),System.out);
        
        setupViewingArea();
        
        gl.glClear(GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT);
        gl.glClearColor(0.4f,0.4f,0.4f,1.0f);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();

        if(OpenGLConverter.getInstance().getScene()!=null) {
            OpenGLConverter.getInstance().getScene().render(this);
        }

        if (hasFocus()) {
            gl.glMatrixMode(GL.GL_MODELVIEW);
			gl.glLoadIdentity();
			gl.glMatrixMode(GL.GL_PROJECTION);
			gl.glLoadIdentity();
			gl.glOrtho(0, 1, 0, 1, -1, 1);
			gl.glColor3d(0.0, 0.0, 1.0);
			gl.glLineWidth((float) 2.0);
			gl.glBegin(GL.GL_LINES);
			gl.glVertex2d(0, 0);
			gl.glVertex2d(0, 1);
			gl.glVertex2d(1, 0);
			gl.glVertex2d(1, 1);
			gl.glVertex2d(0, 0);
			gl.glVertex2d(1, 0);
			gl.glVertex2d(0, 1);
			gl.glVertex2d(1, 1);
			gl.glEnd();
			gl.glLineWidth((float) 1.0);
        }

        swapBuffers();
        glContext.release();
	}
	
	@Override
	protected void paintEvent(QPaintEvent event) {
		render();
	}
	
	@Override
	protected void resizeEvent(QResizeEvent event) {
		render();
	}
	
	@Override
	protected void focusInEvent(QFocusEvent event) {
		render();
	}
	
	@Override
	protected void focusOutEvent(QFocusEvent event) {
		render();
	}
	
	@Override
	protected void mousePressEvent(QMouseEvent event) {
        try {
            ViewEventHandler viewEventHandler=ViewEventHandler.getEventHandler();
            if(viewEventHandler==null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.mouseButtonDown(event.button());
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }		
	}
	
	@Override 
	protected void mouseReleaseEvent(QMouseEvent event) {
        try {
            ViewEventHandler viewEventHandler=ViewEventHandler.getEventHandler();
            if(viewEventHandler==null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.mouseButtonUp(event.button());
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }				
	}
	
	@Override 
	protected void mouseDoubleClickEvent(QMouseEvent event) {
        try {
            ViewEventHandler viewEventHandler=ViewEventHandler.getEventHandler();
            if(viewEventHandler==null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.mouseDoubleClick(event.button());
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }
	}
	
	@Override
	protected void mouseMoveEvent(QMouseEvent event) {
        try {
            ViewEventHandler viewEventHandler=ViewEventHandler.getEventHandler();
            if(viewEventHandler==null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.mouseMove(oldX, oldY, event.x(), event.y());
            oldX = event.x();
            oldY = event.y();
            for(Viewport viewport : OpenGLConverter.getInstance().getRegisteredViews()) {
            	viewport.update();
            }
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }
	}
	
	@Override
	protected void keyPressEvent(QKeyEvent event) {
        try {
            ViewEventHandler viewEventHandler = ViewEventHandler.getEventHandler();
            if (viewEventHandler == null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.keyPress(event.key(), event.modifiers());
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }		
	}
	
	@Override
	protected void keyReleaseEvent(QKeyEvent event) {
        try {
            ViewEventHandler viewEventHandler = ViewEventHandler.getEventHandler();
            if (viewEventHandler == null) {
                return;
            }
            viewEventHandler.setViewport(this);
            viewEventHandler.keyRelease(event.key(), event.modifiers());
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }		
	}
	
	@Override
	protected void wheelEvent(QWheelEvent event) {
        try {
            ViewEventHandler viewEventHandler = ViewEventHandler.getEventHandler();
            if (viewEventHandler == null) {
                return;
            }
            viewEventHandler.setViewport(this);
            double numDegrees=event.delta()/8.0;
            viewEventHandler.mouseWheelMove(numDegrees);
            for(Viewport viewport : OpenGLConverter.getInstance().getRegisteredViews()) {
            	viewport.update();
            }
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }		
	}
	
	@Override
	protected void tabletEvent(QTabletEvent event) {
        try {
            ViewEventHandler viewEventHandler = ViewEventHandler.getEventHandler();
            if (viewEventHandler == null) {
                return;
            }
            viewEventHandler.setViewport(this);
            double x=event.x();
            double y=event.y();
            double xTilt=event.xTilt();
            double yTilt=event.yTilt();
            double pressure=event.pressure();
            viewEventHandler.tabletMove(event.pointerType(),x,y,xTilt,yTilt,pressure);
            for(Viewport viewport : OpenGLConverter.getInstance().getRegisteredViews()) {
            	viewport.update();
            }
        } catch(Exception exception) {
            ExceptionDialog.run(exception);
        }				
	}

	@Override
	public void disposed()
	{
		OpenGLConverter.getInstance().unregisterView(this);
	}
    
    /**
     * Set the type of projection to use.
     * 
     * @param projectionType type of projection
     */
    public void setProjectionType(ProjectionType projectionType) {
    	this.projectionType=projectionType;
    }
    
    /**
     * Get the type of projection that is currently in use
     * 
     * @return type of projection
     */
    public ProjectionType getProjectionType() {
    	return projectionType;
    }
    
    /**
     * Lock camera orientation, that is direction and up vector. The
     * camera can still be panned and dollied or zoomed when the orientation
     * is locked.
     */
    public void lockCameraOrientation() {
    	this.lockCameraOrientation=true;
    }
    
    /**
     * Unlock the camera orientation. This allows the camera to be moved
     * freely again.
     */
    public void unlockCameraOrientation() {
    	this.lockCameraOrientation=false;
    }
    
    /**
     * Test if the camera orientation is currently locked.
     * 
     * @return true if camera orientation is locked, false otherwise
     */
    public boolean isCameraOrientationLocked() {
    	return lockCameraOrientation;
    }
    
    /**
     * Set the geometry preview rendering mode for the viewport. 
     * This can either be wireframe, shaded or lit.
     * 
     * @param mode the geometry rendering mode to use for the viewport
     */
    public void setCurrentPreviewMode(PreviewMode mode) {
    	currentPreviewMode=mode;
    }
    
    /**
     * Get the current geometry preview rendering mode for the viewport.
     * 
     * @return the current geometry rendering mode
     */
    public PreviewMode getCurrentPreviewMode() {
    	return currentPreviewMode;
    }
    
    /**
     * Set the structure rendering mode for the viewport. This can be either
     * full structure, occluded structure or none.
     * 
     * @param mode the structure rendering mode to use for the viewport
     */
    public void setCurrentStructureMode(StructureMode mode) {
    	currentStructureMode=mode;
    }
    
    /**
     * Get the current structure rendering mode for the viewport.
     * 
     * @return current structure rendering mode
     */
    public StructureMode getCurrentStructureMode() {
    	return currentStructureMode;
    }

    /**
     * Set the OpenGL modelview matrix. This sets the modelview
     * matrix that is used in the viewport for rendering to
     * the one that is passed to this function.
     * 
     * @param modelViewMatrix the new modelview matrix
     */
    public void setModelViewMatrix(Matrix4 modelViewMatrix)
    {
        this.modelViewMatrix=modelViewMatrix.clone();
    }

    /**
     * Return the current modelview matrix that is used
     * in this viewport.
     * 
     * @return the current modelview matrix
     */
    public Matrix4 getModelViewMatrix()
    {
        return modelViewMatrix;
    }
    
    /**
     * Set the OpenGL projection matrix. This sets the projection
     * matrix that is used in the viewport for rendering to
     * the one that is passed to this function.
     * 
     * @param projectionMatrix the new projection matrix
     */
    public void setProjectionMatrix(Matrix4 projectionMatrix) {
    	this.projectionMatrix=projectionMatrix.clone();
    }
    
    /**
     * Return the current projection matrix that is used
     * in this viewport.
     * 
     * @return the current projection matrix
     */
    public Matrix4 getProjectionMatrix() {
    	return projectionMatrix;
    }
    
    /**
     * Get scaling factor used for orthographic projection
     * 
     * @return orthographic scaling factor
     */
    public double getOrthographicScale() {
    	return orthographicScale;
    }
    
    /**
     * Set scaling factor to use for orthographic projection
     * 
     * @param orthographicScale orthographic scaling factor
     */
    public void setOrthographicScale(double orthographicScale) {
    	this.orthographicScale=orthographicScale;
    }
    
    /**
     * Return the width of the viewport in pixel
     * 
     * @return viewport width
     */
    public int getWidth() {
    	return size().width();
    }
    
    /**
     * Return the height of the viewport in pixel
     * 
     * @return viewport height
     */
    public int getHeight() {
    	return size().height();
    }
    
    /**
     * Return a reference to the GL interface for the current context.
     * Note that this is only useful while the 3D view is actually 
     * updated.
     * 
     * @return GL interface for current context
     */
    public GL getGL()
    {
        return gl;
    }

    /**
     * Check if this viewport currently owns the keyboard focus.
     * 
     * @return true if viewport has keyboard focus, false otherwise
     */
    public boolean hasKeyboardFocus() {
    	return hasFocus();
    }
}
