/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Created on Jan 9, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.widgets;

import java.util.ArrayList;

import javax.media.opengl.GL;

import com.trolltech.qt.core.Qt;

import eu.moonlight3d.math.Helper;
import eu.moonlight3d.math.Matrix4;
import eu.moonlight3d.math.Quaternion;
import eu.moonlight3d.math.Ray;
import eu.moonlight3d.math.Vector3D;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Scene;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Viewport;

/**
 * The gizmo widget, a complex 3D widget which enables the user to
 * input rotation, scaling and translation operations directly in the
 * 3D view of the scene.
 *
 * @author gregor
 */
public class Gizmo extends Widget {
	/**
	 * The distance between the individual icons on the gizmo axes.
	 */
	private final static double INSET=0.1;
	/**
	 * The radius of a sphere around the icons which is the active hit
	 * target.
	 */
	private final static double HITDISTANCE=0.05;

	/**
	 * Scale factor for translation actions.
	 */
	private final static double TRANSLATIONSCALE=1.0;
	/**
	 * Scale factor for rotation actions.
	 */
	private final static double ROTATIONSCALE=1;
	/**
	 * Scale factor for scale actions.
	 */
	private final static double SCALESCALE=1.0;

	/**
	 * Whether translation icons are enabled.
	 */
	private boolean translateEnabled;
	/**
	 * Whether rotation icons are enabled.
	 */
	private boolean rotateEnabled;
	/**
	 * Whether scale icons are enabled.
	 */
	private boolean scaleEnabled;
	/**
	 * Position in the scene where the gizmo is displayed.
	 */
	private Vector3D position;
	/**
	 * The rotation of the gizmo.
	 */
	private Quaternion rotation;
	/**
	 * The world to local transformation used to transform into the target coordinate system
	 * in which the transformation, that is entered using the gizmo, is applied.
	 */
	private Matrix4 localToWorldTransformation;
	/**
	 * Current X coordinate of the mouse cursor.
	 */
	private int currentX;
	/**
	 * Current Y coordinate of the mouse cursor.
	 */
	private int currentY;
	/**
	 * Position where the current drag operation started.
	 */
	private Vector3D dragStart;
	/**
	 * Whether the CTRL key is pressed.
	 */
	private boolean ctrlPressed;
	/**
	 * Whether the user is currently dragging on the gizmo.
	 */
	private boolean dragging;

	/**
	 * This defines various drag modes for the Gizmo.
	 *
	 * @author gregor
	 */
	public enum DragMode {
		/**
		 * Translate along X-axis.
		 */
		TranslateX,
		/**
		 * Translate along Y-axis.
		 */
		TranslateY,
		/**
		 * Translate along Z-axis.
		 */
		TranslateZ,
		/**
		 * Translate along X- and Y-axis at the same time.
		 */
		TranslateXY,
		/**
		 * Translate along X- and Z-axis at the same time.
		 */
		TranslateXZ,
		/**
		 * Translate along Y- and Z-axis at the same time.
		 */
		TranslateYZ,
		/**
		 * Translate freely in 3D.
		 */
		TranslateXYZ,
		/**
		 * Rotate along X-axis.
		 */
		RotateX,
		/**
		 * Rotate along Y-axis.
		 */
		RotateY,
		/**
		 * Rotate along Z-axis.
		 */
		RotateZ,
		/**
		 * Rotate along X- and Y-axis at the same time.
		 */
		RotateXY,
		/**
		 * Rotate along X- and Z-axis at the same time.
		 */
		RotateXZ,
		/**
		 * Rotate along Y- and Z-axis at the same time.
		 */
		RotateYZ,
		/**
		 * Scale along X-axis.
		 */
		ScaleX,
		/**
		 * Scale along Y-axis.
		 */
		ScaleY,
		/**
		 * Scale along Z-axis.
		 */
		ScaleZ,
		/**
		 * Scale along X- and Y-axis uniformly.
		 */
		ScaleXY,
		/**
		 * Scale along X- and Z-axis uniformly.
		 */
		ScaleXZ,
		/**
		 * Scale along Y- and Z-axis uniformly.
		 */
		ScaleYZ,
		/**
		 * Scale uniformly along all 3 axes.
		 */
		ScaleXYZ
	}

	/**
	 * The current drag mode of the gizmo if it's currently dragging.
	 */
	private DragMode dragMode;
	/**
	 * Axis along which the operation is executed.
	 */
	private Vector3D dragAxis;
	/**
	 * Second axis along which the operation is executed.
	 */
	private Vector3D dragAxis2;
	/**
	 * The initial position where the drag started.
	 */
	private Vector3D dragInitialPosition;

	/**
	 * The list of gizmo listeners that are notified of user actions
	 * performed through this gizmo.
	 */
	private ArrayList<GizmoListener> listeners;

	private ArrayList<GizmoSnapper> snappers;

	/**
	 * Constructor.
	 *
	 * @param viewEventHandler the parent view event handler for this widget.
	 */
	public Gizmo(WidgetViewEventHandler viewEventHandler) {
		super(viewEventHandler);
		setPriority(Widget.GIZMOPRIORITY);
		position=new Vector3D();
		position.clear();
		rotation=new Quaternion();
		rotation.clear();
		localToWorldTransformation=new Matrix4();
		localToWorldTransformation.setIdentity();
		listeners=new ArrayList<GizmoListener>();
		snappers=new ArrayList<GizmoSnapper>();
		dragging=false;
		ctrlPressed=false;
	}

	/**
	 * Add a gizmo listener.
	 *
	 * @param listener the listener to add
	 */
	public void addListener(GizmoListener listener) {
		listeners.add(listener);
	}

	/**
	 * Remove a gizmo listener
	 *
	 * @param listener the listener to remove
	 */
	public void removeListener(GizmoListener listener) {
		listeners.remove(listener);
	}

	public void addSnapper(GizmoSnapper snapper) {
		snappers.add(snapper);
	}

	public void removeSnapper(GizmoSnapper snapper) {
		snappers.remove(snapper);
	}

	/**
	 * Set the position of the gizmo in 3D.
	 *
	 * @param position the new position of the gizmo.
	 */
	public void setPosition(Vector3D position) {
		this.position=position.clone();
	}

	/**
	 * Set the local to world transformation matrix. This matrix is used
	 * to translate the world-space transformation data which the gizmo gathers
	 * into the local coordinate system of the target of the operation.
	 *
	 * @param worldToLocalTransformation the world to local transformation matrix
	 */
	public void setLocalToWorldTransformation(Matrix4 worldToLocalTransformation) {
		this.localToWorldTransformation=worldToLocalTransformation;
	}

	/**
	 * Enable or disable translation icons on the gizmo. If
	 * enabled, the user can perform translation operations on
	 * it.
	 *
	 * @param translateEnabled true if translations icons should be
	 * enabled
	 */
	public void setTranslateEnabled(boolean translateEnabled) {
		this.translateEnabled=translateEnabled;
	}

	/**
	 * Enable or disable rotation icons on the gizmo. If
	 * enabled, the user can perform rotation operations on it.
	 *
	 * @param rotateEnabled true if rotation icons should be enabled.
	 */
	public void setRotateEnabled(boolean rotateEnabled) {
		this.rotateEnabled=rotateEnabled;
	}

	/**
	 * Enable or disable scaling operations on the gizmo. If enabled,
	 * the user can perform scaling operations on it.
	 *
	 * @param scaleEnabled true if scaling icons should be enabled.
	 */
	public void setScaleEnabled(boolean scaleEnabled) {
		this.scaleEnabled=scaleEnabled;
	}

	/**
	 * Calculate the distance of the gizmo to the camera of the currently
	 * active viewport.
	 *
	 * @return distance to the camera
	 */
	private double distanceToCamera() {
		if(getViewEventHandler().getViewport().getProjectionType()==Viewport.ProjectionType.Perspective) {
			Vector3D cameraPosition=getViewEventHandler().getViewport().getModelViewMatrix().getTranslation();
			return cameraPosition.sub(position).abs();
		} else if(getViewEventHandler().getViewport().getProjectionType()==Viewport.ProjectionType.Orthographic) {
			return getViewEventHandler().getViewport().getOrthographicScale();
		} else {
			return 3;
		}
	}

	/**
	 * Calculate the distance of the gizmo to the camera of the given viewport.
	 *
	 * @param viewport the viewport to which the target camera belongs
	 * @return distance to the camera
	 */
	private double distanceToCamera(Viewport viewport) {
		if(viewport.getProjectionType()==Viewport.ProjectionType.Perspective) {
			Vector3D cameraPosition=viewport.getModelViewMatrix().getTranslation();
			return cameraPosition.sub(position).abs();
		} else if(viewport.getProjectionType()==Viewport.ProjectionType.Orthographic) {
			// TODO find scale of viewport
			return viewport.getOrthographicScale();
		} else {
			return 3;
		}
	}

	/**
	 * Draw a cone cap along the local Z axis.
	 *
	 * @param gl OpenGL context to use for drawing
	 */
	private void drawCone(GL gl) {
		Vector3D point=new Vector3D();
		point.clear();
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3d(0,0,0.12);
		for(int i=0;i<=16;i++) {
			point.X1=0.07*Math.cos(2*i/16.0*Math.PI);
			point.X2=0.07*Math.sin(2*i/16.0*Math.PI);
			gl.glVertex3d(point.X1,point.X2,point.X3);
		}
		gl.glEnd();
	}

	/**
	 * Draw a small box around the origin.
	 *
	 * @param gl OpenGL context to use for drawing
	 */
	private void drawBox(GL gl) {
		gl.glBegin(GL.GL_QUADS);
		gl.glVertex3d(0.05,0.05,0.05);
		gl.glVertex3d(-0.05,0.05,0.05);
		gl.glVertex3d(-0.05,-0.05,0.05);
		gl.glVertex3d(0.05,-0.05,0.05);

		gl.glVertex3d(0.05,0.05,-0.05);
		gl.glVertex3d(-0.05,0.05,-0.05);
		gl.glVertex3d(-0.05,-0.05,-0.05);
		gl.glVertex3d(0.05,-0.05,-0.05);

		gl.glVertex3d(0.05,0.05,0.05);
		gl.glVertex3d(0.05,0.05,-0.05);
		gl.glVertex3d(0.05,-0.05,-0.05);
		gl.glVertex3d(0.05,-0.05,0.05);

		gl.glVertex3d(-0.05,0.05,0.05);
		gl.glVertex3d(-0.05,0.05,-0.05);
		gl.glVertex3d(-0.05,-0.05,-0.05);
		gl.glVertex3d(-0.05,-0.05,0.05);

		gl.glVertex3d(0.05,0.05,0.05);
		gl.glVertex3d(0.05,0.05,-0.05);
		gl.glVertex3d(-0.05,0.05,-0.05);
		gl.glVertex3d(-0.05,0.05,0.05);

		gl.glVertex3d(0.05,-0.05,0.05);
		gl.glVertex3d(0.05,-0.05,-0.05);
		gl.glVertex3d(-0.05,-0.05,-0.05);
		gl.glVertex3d(-0.05,-0.05,0.05);
		gl.glEnd();
	}

	/**
	 * Draw a rotation symbol around the origin.
	 *
	 * @param gl OpenGL context to use for drawing
	 */
	private void drawRotationSymbol(GL gl) {
		// just draw a simple line for now:
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3d(-0.05,0.0,0.0);
		gl.glVertex3d(0.05,0.0,0.0);
		gl.glEnd();
	}

	@Override
	public void drawOverlay(Viewport view, Scene scene) {
		double inset=0.0;
		GL gl=view.getGL();

		gl.glMultMatrixd(localToWorldTransformation.getOpenGLMatrix(), 0);

		gl.glTranslated(position.X1,position.X2,position.X3);

		double distance=distanceToCamera(view);
		gl.glScaled(distance/3.5, distance/3.5, distance/3.5);

		// draw axes in different colours:
		gl.glBegin(GL.GL_LINES);
		gl.glColor3d(1.0,0.0,0.0);
		gl.glVertex3d(0.0,0.0,0.0);
		gl.glVertex3d(1.0,0.0,0.0);
		gl.glColor3d(0.0,1.0,0.0);
		gl.glVertex3d(0.0,0.0,0.0);
		gl.glVertex3d(0.0,1.0,0.0);
		gl.glColor3d(0.0,0.0,1.0);
		gl.glVertex3d(0.0,0.0,0.0);
		gl.glVertex3d(0.0,0.0,1.0);
		gl.glEnd();

		if(translateEnabled) {
			// draw cones along axes

			gl.glTranslated(1.0,0.0,0.0);
			gl.glColor3d(1.0,0.0,0.0);
			// X axis
			gl.glRotated(90,0,1,0);
			drawCone(gl);
			gl.glRotated(-90,0,1,0);

			gl.glTranslated(-1.0,1.0,0.0);
			gl.glColor3d(0.0,1.0,0.0);
			// Y axis
			gl.glRotated(-90,1,0,0);
			drawCone(gl);
			gl.glRotated(90,1,0,0);

			gl.glTranslated(0.0,-1.0,1.0);
			gl.glColor3d(0.0,0.0,1.0);
			// Z axis
			drawCone(gl);
			gl.glTranslated(0.0,0.0,-1.0);
			// center
			gl.glColor3d(1.0,1.0,0.0);
			drawBox(gl);

			inset+=INSET;
		}

		if(rotateEnabled) {
			// TODO: draw sicles along axes

			gl.glTranslated(1.0-inset,0.0,0.0);
			gl.glColor3d(1.0,0.0,0.0);
			// X axis
			gl.glRotated(90,0,1,0);
			drawRotationSymbol(gl);
			gl.glRotated(-90,0,1,0);

			gl.glTranslated(-1.0+inset,1.0-inset,0.0);
			gl.glColor3d(0.0,1.0,0.0);
			// Y axis
			drawRotationSymbol(gl);

			gl.glTranslated(0.0,-1.0+inset,1.0-inset);
			gl.glColor3d(0.0,0.0,1.0);
			// Z axis
			gl.glRotated(90,0,0,1);
			drawRotationSymbol(gl);
			gl.glRotated(-90,0,0,1);

			gl.glTranslated(0.0,0.0,-1.0+inset);
			// center if dominating
			if(!translateEnabled) {
				// TODO find a better symbol for center
				gl.glColor3d(1.0,1.0,0.0);
				drawBox(gl);
			}

			inset+=INSET;
		}

		if(scaleEnabled) {
			// draw boxes along axes

			gl.glTranslated(1.0-inset,0.0,0.0);
			gl.glColor3d(1.0,0.0,0.0);
			// X axis

			drawBox(gl);

			gl.glTranslated(-1.0+inset,1.0-inset,0.0);
			gl.glColor3d(0.0,1.0,0.0);
			// Y axis

			drawBox(gl);

			gl.glTranslated(0.0,-1.0+inset,1.0-inset);
			gl.glColor3d(0.0,0.0,1.0);
			// Z axis

			drawBox(gl);

			gl.glTranslated(0.0,0.0,-1.0+inset);
			// center if dominating
			if(!translateEnabled && !rotateEnabled) {
				gl.glColor3d(1.0,1.0,0.0);
				drawBox(gl);
			}

			inset+=INSET;
		}
	}

	@Override
	public boolean keyPress(int key, Qt.KeyboardModifiers state) {
		if(key==Qt.Key.Key_Control.value()) {
			ctrlPressed=true;
			return true;
		}
		return false;
	}

	@Override
	public boolean keyRelease(int key, Qt.KeyboardModifiers state) {
		if(key==Qt.Key.Key_Control.value()) {
			ctrlPressed=false;
			return true;
		}
		return false;
	}

	/**
	 * Calculate the distance between the projection of the given point in
	 * three dimensions onto the view plane and the current mouse cursor
	 * position.
	 *
	 * @param mouseRelative the mouse coordinates scaled to to the range of \f$[-1..1]\f$
	 * @param pos the point in 3D to calculate the projected distance to
	 * @return the distance between the projections of both points onto the view plane
	 */
	private double distanceTo(Vector3D mouseRelative, Vector3D pos) {
		pos=localToWorldTransformation.multiply(pos);
		Vector3D projected = Helper.project(getViewEventHandler().getViewport().getModelViewMatrix(),getViewEventHandler().getViewport().getProjectionMatrix(),pos);
		projected.X1=(projected.X1+1)/2.0;
		projected.X2=(-projected.X2+1)/2.0;
		projected.X3=0;
		return mouseRelative.sub(projected).abs();
	}

	/**
	 * Test if the user has selected a part of this gizmo widget.
	 *
	 * @return true if the user has selected a part of this gizmo widget, false otherwise
	 */
	private boolean select() {
		DragMode mode=null;
		double distance=-1;
		boolean firstHit=false;
		Vector3D relative=new Vector3D();
		relative.X1=currentX/(float)getViewEventHandler().getViewport().width();
		relative.X2=currentY/(float)getViewEventHandler().getViewport().height();
		relative.X3=0;

		double distanceCenter=distanceTo(relative,position);
		eu.moonlight3d.framework.State.getInstance().logger().info("select: distance to center: " + distanceCenter);
		if(distanceCenter < HITDISTANCE) {
			if(translateEnabled) {
				distance=distanceCenter;
				mode=DragMode.TranslateXYZ;
			} else if(scaleEnabled) {
				distance=distanceCenter;
				mode=DragMode.ScaleXYZ;
			}
			firstHit=true;
			eu.moonlight3d.framework.State.getInstance().logger().info("candiate: center");
		}

		double distanceScale=distanceToCamera()/3.5;
		if (translateEnabled) {
			double distanceTranslateX = distanceTo(relative, position.add(new Vector3D(distanceScale, 0, 0)));
			if (distanceTranslateX < HITDISTANCE && (!firstHit || distanceTranslateX < distance)) {
				firstHit = true;
				mode = DragMode.TranslateX;
				distance = distanceTranslateX;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: translate x");
			}
			double distanceTranslateY = distanceTo(relative, position.add(new Vector3D(0, distanceScale, 0)));
			if (distanceTranslateY < HITDISTANCE && (!firstHit || distanceTranslateY < distance)) {
				firstHit = true;
				mode = DragMode.TranslateY;
				distance = distanceTranslateY;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: translate y");
			}
			double distanceTranslateZ = distanceTo(relative, position.add(new Vector3D(0, 0, distanceScale)));
			if (distanceTranslateZ < HITDISTANCE && (!firstHit || distanceTranslateZ < distance)) {
				firstHit = true;
				mode = DragMode.TranslateZ;
				distance = distanceTranslateZ;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: translate z");
			}
		}

		double inset=0.0;
		if (rotateEnabled) {
			inset += INSET;
			double distanceRotateX = distanceTo(relative, position.add(new Vector3D((1 - inset) * distanceScale, 0, 0)));
			if (distanceRotateX < HITDISTANCE && (!firstHit || distanceRotateX < distance)) {
				firstHit = true;
				mode = DragMode.RotateX;
				distance = distanceRotateX;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: rotate x");
			}
			double distanceRotateY = distanceTo(relative, position.add(new Vector3D(0, (1 - inset) * distanceScale, 0)));
			if (distanceRotateY < HITDISTANCE && (!firstHit || distanceRotateY < distance)) {
				firstHit = true;
				mode = DragMode.RotateY;
				distance = distanceRotateY;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: rotate y");
			}
			double distanceRotateZ = distanceTo(relative, position.add(new Vector3D(0, 0, (1 - inset) * distanceScale)));
			if (distanceRotateZ < HITDISTANCE && (!firstHit || distanceRotateZ < distance)) {
				firstHit = true;
				mode = DragMode.RotateZ;
				distance = distanceRotateZ;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: rotate z");
			}
		}

		if (scaleEnabled) {
			inset += INSET;
			double distanceScaleX = distanceTo(relative, position.add(new Vector3D((1 - inset) * distanceScale, 0, 0)));
			if (distanceScaleX < HITDISTANCE && (!firstHit || distanceScaleX < distance)) {
				firstHit = true;
				mode = DragMode.ScaleX;
				distance = distanceScaleX;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: scale x");
			}
			double distanceScaleY = distanceTo(relative, position.add(new Vector3D(0, (1 - inset) * distanceScale, 0)));
			if (distanceScaleY < HITDISTANCE && (!firstHit || distanceScaleY < distance)) {
				firstHit = true;
				mode = DragMode.ScaleY;
				distance = distanceScaleY;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: scale y");
			}
			double distanceScaleZ = distanceTo(relative, position.add(new Vector3D(0, 0, (1 - inset) * distanceScale)));
			if (distanceScaleZ < HITDISTANCE && (!firstHit || distanceScaleZ < distance)) {
				firstHit = true;
				mode = DragMode.ScaleZ;
				distance = distanceScaleZ;
				eu.moonlight3d.framework.State.getInstance().logger().info("candiate: scale z");
			}
		}

		if(!firstHit) {
			// short circuit rest if no hit
			eu.moonlight3d.framework.State.getInstance().logger().info("gizmo select: no hit");
			return false;
		}

		// adjust drag mode in the presence of pressted Ctrl key
		if(ctrlPressed) {
			switch(mode) {
			case TranslateX:
				mode=DragMode.TranslateYZ;
				break;
			case TranslateY:
				mode=DragMode.TranslateXZ;
				break;
			case TranslateZ:
				mode=DragMode.TranslateXY;
				break;
			case RotateX:
				mode=DragMode.RotateYZ;
				break;
			case RotateY:
				mode=DragMode.RotateXZ;
				break;
			case RotateZ:
				mode=DragMode.RotateXY;
				break;
			case ScaleX:
				mode=DragMode.ScaleYZ;
				break;
			case ScaleY:
				mode=DragMode.ScaleXZ;
				break;
			case ScaleZ:
				mode=DragMode.ScaleXY;
				break;
			}
		}
		dragMode=mode;

		return true;
	}

	/**
	 * Project the connecting line between the given start and end points
	 * in 3D onto the view plane and normalize the connection vector.
	 *
	 * @param start the start point of the line in 3D
	 * @param end the end point of the line in 3D
	 * @return the normalized connection vector
	 */
	private Vector3D projectAxis(Vector3D start, Vector3D end) {
		start=localToWorldTransformation.multiply(start);
		end=localToWorldTransformation.multiply(end);
		Vector3D projectedStart=Helper.project(getViewEventHandler().getViewport().getModelViewMatrix(),getViewEventHandler().getViewport().getProjectionMatrix(),start);
		Vector3D projectedEnd=Helper.project(getViewEventHandler().getViewport().getModelViewMatrix(),getViewEventHandler().getViewport().getProjectionMatrix(),end);
		projectedStart.X1=(projectedStart.X1+1)/2.0;
		projectedStart.X2=(-projectedStart.X2+1)/2.0;
		projectedStart.X3=0;
		projectedEnd.X1=(projectedEnd.X1+1)/2.0;
		projectedEnd.X2=(-projectedEnd.X2+1)/2.0;
		projectedEnd.X3=0;
		return projectedEnd.sub(projectedStart).norm();
	}

	/**
	 * Initialize a dragging action.
	 */
	public void initDrag() {
		dragInitialPosition=position.clone();
		dragStart=new Vector3D();
		dragStart.X1=currentX;
		dragStart.X2=currentY;
		dragStart.X3=0;
		if(dragMode==DragMode.TranslateXYZ || dragMode==DragMode.ScaleXYZ) {
			// Nothing todo here
		} else {
			// init drag axis/drag axes
			if(dragMode==DragMode.TranslateX ||
			   dragMode==DragMode.RotateX || dragMode==DragMode.RotateYZ ||
			   dragMode==DragMode.ScaleX || dragMode==DragMode.ScaleYZ) {
				// drag axis is projected X axis
				Vector3D endPosition=position.add(new Vector3D(1,0,0));
				dragAxis=projectAxis(position,endPosition);
			} else	if(dragMode==DragMode.TranslateY ||
					   dragMode==DragMode.RotateY || dragMode==DragMode.RotateXZ ||
					   dragMode==DragMode.ScaleY || dragMode==DragMode.ScaleXZ) {
				// drag axis is projected Y axis
				Vector3D endPosition=position.add(new Vector3D(0,1,0));
				dragAxis=projectAxis(position,endPosition);
			} else	if(dragMode==DragMode.TranslateZ ||
					   dragMode==DragMode.RotateZ || dragMode==DragMode.RotateXY ||
					   dragMode==DragMode.ScaleZ || dragMode==DragMode.ScaleXY) {
				// drag axis is projected Z axis
				Vector3D endPosition=position.add(new Vector3D(0,0,1));
				dragAxis=projectAxis(position,endPosition);
			} else if(dragMode==DragMode.TranslateXY) {
				// drag axis 1 is projected X axis
				Vector3D endPosition=position.add(new Vector3D(1,0,0));
				dragAxis=projectAxis(position,endPosition);
				// drag axis 1 is projected Y axis
				endPosition=position.add(new Vector3D(0,1,0));
				dragAxis2=projectAxis(position,endPosition);
			} else if(dragMode==DragMode.TranslateXZ) {
				// drag axis 1 is projected X axis
				Vector3D endPosition=position.add(new Vector3D(1,0,0));
				dragAxis=projectAxis(position,endPosition);
				// drag axis 1 is projected Z axis
				endPosition=position.add(new Vector3D(0,0,1));
				dragAxis2=projectAxis(position,endPosition);
			} else if(dragMode==DragMode.TranslateYZ) {
				// drag axis 1 is projected Y axis
				Vector3D endPosition=position.add(new Vector3D(0,1,0));
				dragAxis=projectAxis(position,endPosition);
				// drag axis 1 is projected Z axis
				endPosition=position.add(new Vector3D(0,0,1));
				dragAxis2=projectAxis(position,endPosition);
			}
		}

		// make listeners start handling operations
		for (GizmoListener listener : listeners) {
			listener.begin(dragMode);
		}

	}

	/**
	 * Return drag distance for a drag operation that is mapped onto an axis. This is used
	 * for constrained dragging actions (translate along/perpendicular to axis, for example)
	 *
	 * @param dragAxis the axis onto which to map the drag operation
	 * @param scale the scaling to be applied to the action
	 * @return drag distance
	 */
	public double dragAlongAxisDistance(Vector3D dragAxis, boolean scale) {
		// return drag distance for a drag operation that is mapped onto an axis (constrained actions)
		Vector3D dragVector=new Vector3D();
		dragVector.X1=((currentX-dragStart.X1)/getViewEventHandler().getViewport().width());
		dragVector.X2=((currentY-dragStart.X2)/getViewEventHandler().getViewport().height());
		dragVector.X3=0;
		if(scale) {
			return dragVector.multiply(dragAxis)*distanceToCamera();
		} else {
			return dragVector.multiply(dragAxis);
		}
	}

	/**
	 * Return drag distance for a drag operation that is freehand. This is used
	 * for unconstrained dragging actions (free and uniform scale).
	 *
	 * @return drag distance
	 */
	public double dragFreeDistance() {
		Vector3D dragVector=new Vector3D();
		dragVector.X1=((currentX-dragStart.X1)/getViewEventHandler().getViewport().width());
		dragVector.X2=((currentY-dragStart.X2)/getViewEventHandler().getViewport().height());
		dragVector.X3=0;
		return dragVector.abs()*distanceToCamera();
	}

	private Vector3D snapTranslation(Vector3D start, Vector3D translation) {
		Vector3D result=null;

		for(GizmoSnapper snapper : snappers) {
			Vector3D snapperResult=snapper.snapTranslation(start, translation);
			if(result==null) {
				result=snapperResult;
			} else {
				if(snapperResult!=null) {
					if(snapperResult.sub(translation).abs() < result.sub(translation).abs()) {
						result=snapperResult;
					}
				}
			}
		}

		if(result==null) {
			result=translation;
		}

		return result;
	}

	private Quaternion snapRotation(Vector3D center, Quaternion rotation) {
		Quaternion result=null;

		for(GizmoSnapper snapper : snappers) {
			Quaternion snapperResult=snapper.snapRotation(center, rotation);
			if(result==null) {
				result=snapperResult;
			} else {
				if(snapperResult!=null) {
					if(Math.abs(snapperResult.getRotationAngle()-rotation.getRotationAngle()) < Math.abs(snapperResult.getRotationAngle()-rotation.getRotationAngle())) {
						result=snapperResult;
					}
				}
			}
		}

		if(result==null) {
			result=rotation;
		}

		return result;
	}

	private Vector3D snapScale(Vector3D center, Vector3D scale) {
		Vector3D result=null;

		for(GizmoSnapper snapper : snappers) {
			Vector3D snapperResult=snapper.snapScale(center, scale);
			if(result==null) {
				result=snapperResult;
			} else {
				if(snapperResult!=null) {
					if(snapperResult.sub(scale).abs() < result.sub(scale).abs()) {
						result=snapperResult;
					}
				}
			}
		}

		if(result==null) {
			result=scale;
		}

		return result;
	}

	/**
	 * Invoke the translate() methods on the registered {@link GizmoListener}s.
	 *
	 * @param translation the translation vector to report
	 */
	private void invokeListenersTranslate(Vector3D translation) {
		translation=snapTranslation(dragInitialPosition, translation);

		for(GizmoListener listener : listeners) {
			listener.translate(translation);
		}
	}

	/**
	 * Invoke the rotate() methods on the registered {@link GizmoListener}s.
	 *
	 * @param center center point of rotation
	 * @param rotation quaternion describing the rotation around the center point
	 */
	private void invokeListenersRotate(Vector3D center, Quaternion rotation) {
		rotation=snapRotation(center, rotation);

		for(GizmoListener listener : listeners) {
			listener.rotate(center,rotation);
		}
	}

	/**
	 * Invoke the scale() methods on the registered {@link GizmoListener}s.
	 *
	 * @param center center point of scaling operation
	 * @param scale the scale factors for the operation
	 */
	private void invokeListenersScale(Vector3D center, Vector3D scale) {
		scale=snapScale(center, scale);

		for(GizmoListener listener : listeners) {
			listener.scale(center,scale);
		}
	}

	/**
	 * Update the gizmo during the drag operation.
	 */
    private void handleDrag() {
    	double dragDistance, dragDistance2=0;
		Vector3D translation;
		Vector3D scale;
		Vector3D axis;
		Quaternion rotation;

		if(dragMode==DragMode.TranslateXYZ || dragMode==DragMode.ScaleXYZ) {
    		dragDistance=dragFreeDistance();
		} else {
			if(dragMode!=DragMode.RotateX && dragMode!=DragMode.RotateY && dragMode!=DragMode.RotateZ &&
			   dragMode!=DragMode.RotateXY && dragMode!=DragMode.RotateXZ && dragMode!=DragMode.RotateYZ) {
				   dragDistance=dragAlongAxisDistance(dragAxis,true);
			} else {
				   dragDistance=dragAlongAxisDistance(dragAxis,false);
			}
			if(dragMode==DragMode.TranslateXY || dragMode==DragMode.TranslateXZ || dragMode==DragMode.TranslateYZ) {
				dragDistance2=dragAlongAxisDistance(dragAxis2,true);
			}
		}

    	// handle all 20 possible translate/rotate/scale drag actions here (that's a lot, i know)
    	switch(dragMode) {
    	case TranslateX:
    		dragDistance*=TRANSLATIONSCALE;
    		translation=new Vector3D(dragDistance,0,0);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateY:
    		dragDistance*=TRANSLATIONSCALE;
    		translation=new Vector3D(0,dragDistance,0);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateZ:
    		dragDistance*=TRANSLATIONSCALE;
    		translation=new Vector3D(0,0,dragDistance);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateYZ:
    		dragDistance*=TRANSLATIONSCALE;
    		dragDistance2*=TRANSLATIONSCALE;
    		translation=new Vector3D(0,dragDistance,dragDistance2);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateXZ:
    		dragDistance*=TRANSLATIONSCALE;
    		dragDistance2*=TRANSLATIONSCALE;
    		translation=new Vector3D(dragDistance,0,dragDistance2);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateXY:
    		dragDistance*=TRANSLATIONSCALE;
    		dragDistance2*=TRANSLATIONSCALE;
    		translation=new Vector3D(dragDistance,dragDistance2,0);
    		position=dragInitialPosition.add(translation);
    		invokeListenersTranslate(translation);
    		break;
    	case TranslateXYZ:
    		dragDistance*=TRANSLATIONSCALE;
    		// TODO: Orthographic view doen't work
    		Ray r = Helper.unProject(getViewEventHandler().getViewport().getModelViewMatrix(),
    			getViewEventHandler().getViewport().getProjectionMatrix(), currentX, currentY,
    			getViewEventHandler().getViewport().width(), getViewEventHandler().getViewport().height());
    		Ray r2 = Helper.unProject(getViewEventHandler().getViewport().getModelViewMatrix(),
    			getViewEventHandler().getViewport().getProjectionMatrix(), (int)dragStart.X1, (int)dragStart.X2,
    			getViewEventHandler().getViewport().width(), getViewEventHandler().getViewport().height());
			Vector3D cameraPosition=getViewEventHandler().getViewport().getModelViewMatrix().getTranslation();

			double rel = cameraPosition.sub(dragInitialPosition).abs() / r2.direction.abs();
			Vector3D targetPos = r.start.add(r.direction.multiply(rel));
			position = targetPos;
			invokeListenersTranslate(targetPos.sub(dragInitialPosition));
    		break;
    	case RotateX:
    		dragDistance*=ROTATIONSCALE;
    		axis=new Vector3D(1,0,0);
    		rotation=new Quaternion();
    		rotation.fromAxisAngle(axis,dragDistance);
    		invokeListenersRotate(position, rotation);
    		break;
    	case RotateY:
    		dragDistance*=ROTATIONSCALE;
    		axis=new Vector3D(0,1,0);
    		rotation=new Quaternion();
    		rotation.fromAxisAngle(axis,dragDistance);
    		invokeListenersRotate(position, rotation);
    		break;
    	case RotateZ:
    		dragDistance*=ROTATIONSCALE;
    		axis=new Vector3D(0,0,1);
    		rotation=new Quaternion();
    		rotation.fromAxisAngle(axis,dragDistance);
    		invokeListenersRotate(position, rotation);
    		break;
    	case RotateYZ:
    		dragDistance*=ROTATIONSCALE;
    		// TODO mapping of mouse movement into 3D
    		break;
    	case RotateXZ:
    		dragDistance*=ROTATIONSCALE;
    		// TODO mapping of mouse movement into 3D
    		break;
    	case RotateXY:
    		dragDistance*=ROTATIONSCALE;
    		// TODO mapping of mouse movement into 3D
    		break;
    	case ScaleX:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1+dragDistance,1,1);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleY:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1,1+dragDistance,1);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleZ:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1,1,1+dragDistance);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleYZ:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1,1+dragDistance,1+dragDistance);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleXZ:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1+dragDistance,1,1+dragDistance);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleXY:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1+dragDistance,1+dragDistance,1);
    		invokeListenersScale(position,scale);
    		break;
    	case ScaleXYZ:
    		dragDistance*=SCALESCALE;
    		scale=new Vector3D(1+dragDistance,1+dragDistance,1+dragDistance);
    		invokeListenersScale(position,scale);
    		break;
    	}
	}

	@Override
	public boolean mouseButtonDown(Qt.MouseButton button) throws Exception {
		if(button==Qt.MouseButton.LeftButton) {
			if(select()) {
				dragging=true;
				initDrag();
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean mouseButtonUp(Qt.MouseButton button) {
		if(button==Qt.MouseButton.LeftButton) {
			if (dragging) {
				dragging = false;
				// make listeners accept changes
				for (GizmoListener listener : listeners) {
					listener.apply();
				}

				// set gizmo position to snapped position so that it starts out correctly for the next drag
				if(dragMode==DragMode.TranslateX || dragMode==DragMode.TranslateY || dragMode==DragMode.TranslateZ
				   || dragMode==DragMode.TranslateXY || dragMode==DragMode.TranslateXZ || dragMode==DragMode.TranslateYZ
				   || dragMode==DragMode.TranslateXYZ) {
					position=snapTranslation(dragInitialPosition, position);
				}

				return true;
			}
		}
		return false;
	}

	@Override
    public boolean mouseMove(int oldX, int oldY, int newX, int newY) throws Exception {
		currentX=newX;
		currentY=newY;
		if(dragging) {
			handleDrag();
			return true;
		}
        return false;
    }

	@Override
	public void onRemove() {
		for (GizmoListener listener : listeners) {
			listener.end();
		}
	}
}
