/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Mar 18, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.widgets;


import com.trolltech.qt.core.Qt;

import eu.moonlight3d.ml3d.ui.mainwindow.view.RenderableNode;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Scene;
import eu.moonlight3d.ml3d.ui.mainwindow.view.Viewport;

/**
 * This is the base class for all interactive elements in the 3D
 * view.
 *
 * @author gregor
 */
public class Widget extends RenderableNode {
	private int priority;
	private WidgetViewEventHandler widgetViewEventHandler;
	
	/**
	 * This is the minimal priority a Widget can assume.
	 */
	public final static int MINPRIORITY=0;
	/**
	 * This is the maximal priority a Widget can assume.
	 */
	public final static int MAXPRIORITY=255;
	/**
	 * This is the standard priority that a Widget receives by default.
	 */
	public final static int STANDARDPRIORITY=127;
	/**
	 * This is the priority that HandleContainer widgets get by default
	 */
	public final static int HANDLEPRIORITY=126;
	/**
	 * This is the priority that Gizmo widgets get by default
	 */
	public final static int GIZMOPRIORITY=128;
	
	/**
	 * Constructor that registers the Widget with the WidgetViewEventHandler
	 * that is propagating the events from the 3D view to the individual
	 * widgets.
	 * 
	 * @param viewEventHandler the ViewEventHandler from which this 
	 * Widget is supposed to receive its events.
	 */
	public Widget(WidgetViewEventHandler viewEventHandler) {
		this.priority=STANDARDPRIORITY;
		this.widgetViewEventHandler=viewEventHandler;
		viewEventHandler.addWidget(this);
	}
	
	/**
	 * Dummy implementation of draw function. This method
	 * needs to be overridden in a child widget that wants to
	 * draw parts of itself during the preview rendering pass.
	 * 
	 * @param view the 3d view that is being repainted
	 * @param the scene that is being drawn
	 */
	@Override
	public void drawPreview(Viewport view, Scene scene) {
	}

	/**
	 * Dummy implementation of draw function. This method
	 * needs to be overridden in a child widget that wants to
	 * draw parts of itself during the structure rendering pass.
	 * 
	 * @param view the 3d view that is being repainted
	 * @param the scene that is being drawn
	 */
	@Override
	public void drawVisibleStructure(Viewport view, Scene scene) {
	}

	/**
	 * Dummy implementation of draw function. This method
	 * needs to be overridden in a child widget that wants to
	 * draw parts of itself during the overlay rendering pass.
	 * 
	 * @param view the 3d view that is being repainted
	 * @param the scene that is being drawn
	 */
	@Override
	public void drawOverlay(Viewport view, Scene scene) {
	}
	
	/**
	 * Dummy implementation of draw function. This method
	 * needs to be overridden in a child widget that wants to
	 * draw parts of itself during the 2D overlay rendering pass.
	 * 
	 * @param view the 3d view that is being repainted
	 * @param the scene that is being drawn
	 */
	@Override
	public void drawOverlay2D(Viewport view, Scene scene) {
	}

	/**
	 * Key press event handling method. This gets called whenever
	 * this widget receives a key press event. Key press events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
	 * 
	 * @param key the code of the key that has been pressed
	 * @param state the state of the modifier keys at that time
	 * @return true if the event was handled, false otherwise
	 * @throws Exception
	 */
	public boolean keyPress(int key, Qt.KeyboardModifiers state) throws Exception {
		return false;
	}
	
	/**
	 * Key release event handling method. This gets called whenever
	 * this widget receives a key release event. Key release events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
	 * 
	 * @param key the code of the key that has been pressed
	 * @param state the state of the modifier keys at that time
	 * @return true if the event was handled, false otherwise
	 * @throws Exception
	 */
	public boolean keyRelease(int key, Qt.KeyboardModifiers state) throws Exception {
		return false;
	}

	/**
	 * Mouse press event handling method. This gets called whenever
	 * this widget receives a key release event. Mouse press events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
	 * 
	 * @param button the number of the mouse button that was pressed
	 * @return true if the event was handled, false otherwise
	 * @throws Exception
	 */
	public boolean mouseButtonDown(Qt.MouseButton button) throws Exception {
		return false;
	}

	/**
	 * Mouse release event handling method. This gets called whenever
	 * this widget receives a key release event. Mouse release events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
	 * 
	 * @param button the number of the mouse button that was pressed
	 * @return true if the event was handled, false otherwise
	 * @throws Exception
	 */
	public boolean mouseButtonUp(Qt.MouseButton button) throws Exception {
		return false;
	}

	/**
	 * Mouse double click event handling method. This gets called whenever
	 * this widget receives a key release event. Mouse double click events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
	 * 
	 * @param button the number of the mouse button that was doubleclicked
	 * @return true if the event was handled, false otherwise
	 * @throws Exception
	 */
	public boolean mouseDoubleClick(Qt.MouseButton button) throws Exception {
		return false;
	}
	
    /**
	 * Mouse move event handling method. This gets called whenever
	 * this widget receives a key release event. Mouse move events
	 * are propagated through the widgets by priority until the
	 * first widget declares itself responsible for this event.
	 * In this case the function returns true. If the event did
	 * not get handled by this widget, the function returns false
	 * and the event is sent to the next widget in line. 
     * 
     * @param oldX the x coordinate of the mouse pointer before the event
     * @param oldY the y coordinate of the mouse pointer before the event
     * @param newX the x coordinate of the mouse pointer after the event
     * @param newY the y coordinate of the mouse pointer after the event
     * @return true if the event was handled, false otherwise
     * @throws Exception
     */
	public boolean mouseMove(int oldX, int oldY, int newX, int newY) throws Exception {
		return false;
	}
	
	/**
	 * This method gets called when the widget is being removed from the
	 * view. It is meant to be overridden in child classes in order to
	 * clean up after the widget has been removed.
	 */
	public void onRemove() {
	}

	/**
	 * Return the priority of this widget
	 * 
	 * @return widget priority
	 * @see WidgetViewEventHandler
	 */
	public int getPriority() {
		return priority;
	}

	/**
	 * Set the priority of this widget
	 * 
	 * @param priority the new priority of this widget
	 * @see WidgetViewEventHandler
	 */
	public void setPriority(int priority) {
		this.priority = priority;
	}

	/**
	 * Return the WidgetViewEventHandler that is driving
	 * this widget
	 * 
	 * @return the driver for this widget
	 */
	public WidgetViewEventHandler getViewEventHandler() {
		return widgetViewEventHandler;
	}

}
