/*
 * Moonlight|3D Copyright (C) 2006 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * Created on Mar 18, 2006
 */
package eu.moonlight3d.ml3d.ui.mainwindow.widgets;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

import com.trolltech.qt.core.Qt;

import eu.moonlight3d.ml3d.ui.mainwindow.view.OpenGLConverter;
import eu.moonlight3d.ml3d.ui.mainwindow.view.StandardViewEventHandler;

/**
 * This ViewEventHandler is distributing the events sent out by the 
 * 3d view to the 3d widgets inside the view that are registered
 * with this event handler.
 * 
 * Events are distributed through the widgets by widget priority.
 * Widgets with higher priority get an earlier chance to react to
 * an event than widgets with a lower priority. The order in which
 * widgets with the same priority receive an event is undefined.
 * 
 * Events are passed to all the widgets in priority order until
 * a widget claims to have handled that event or until the event
 * has been passed to all the widgets and no widget has claimed
 * to have processed it properly.
 *
 * @author gregor
 */
public class WidgetViewEventHandler extends StandardViewEventHandler {
	private ArrayList<Widget> widgets;
	private boolean altDown;
	private LifecycleListener listener;
	
	public WidgetViewEventHandler() {
		widgets=new ArrayList<Widget>();
		altDown=false;
		
		listener=new LifecycleListener(this);
		eu.moonlight3d.framework.State.getInstance().getDocumentManager().addLifecycleListener(listener);
	}
	
	private Widget[] getPrioritySortedWidgets() {
		Widget[] widgets=new Widget[this.widgets.size()];
		widgets=this.widgets.toArray(widgets);
		
		Arrays.sort(widgets, new Comparator() {

			public int compare(Object widget1, Object widget2) {
				return ((Widget)widget1).getPriority()-((Widget)widget2).getPriority();
			}
			
		});
		
		return widgets;
	}
	
	public void addWidget(Widget widget) {
		widgets.add(widget);
		Widget[] widgets=new Widget[1];
		widgets[0]=widget;
		OpenGLConverter.getInstance().getScene().addSceneNodes(widgets);
	}
	
	public void removeWidget(Widget widget) {
		widget.onRemove();
		widgets.remove(widget);
		OpenGLConverter.getInstance().getScene().removeSceneNode(widget);
	}
	
	void clearWidgets() {
		for(Widget widget : widgets) {
			widget.onRemove();
			OpenGLConverter.getInstance().getScene().removeSceneNode(widget);
		}
		widgets.clear();
	}
	
	@Override
	public void onRemove() {
		clearWidgets();
		eu.moonlight3d.framework.State.getInstance().getDocumentManager().removeLifecycleListener(listener);
	}
	
	@Override
	public void keyPress(int key, Qt.KeyboardModifiers state) throws Exception {
		if(key==Qt.Key.Key_Alt.value()) {
			altDown=true;
			super.keyPress(key,state);
			return;
		}
		
		Widget[] widgets=getPrioritySortedWidgets();
		
		for(int i=widgets.length-1;i>=0;i--) {
			try {
				if(widgets[i].keyPress(key,state)) {
					return;
				}
			} catch (java.lang.Exception e) {
				e.printStackTrace();
			}
		}
		super.keyPress(key,state);
	}

	@Override
	public void keyRelease(int key, Qt.KeyboardModifiers state) throws Exception {
		if(key==Qt.Key.Key_Alt.value()) {
			altDown=false;
			super.keyRelease(key,state);
			return;
		}
		
		Widget[] widgets=getPrioritySortedWidgets();
		
		for(int i=widgets.length-1;i>=0;i--) {
			try {
				if(widgets[i].keyRelease(key,state)) {
					return;
				}
			} catch (java.lang.Exception e) {
				e.printStackTrace();
			}
		}
		super.keyRelease(key,state);
	}

	@Override
	public void mouseButtonDown(Qt.MouseButton button) throws Exception {
		if(altDown==true) {
			super.mouseButtonDown(button);
			return;
		}
		
		Widget[] widgets=getPrioritySortedWidgets();
		
		for(int i=widgets.length-1;i>=0;i--) {
			try {
				if(widgets[i].mouseButtonDown(button)) {
					return;
				}
			} catch (java.lang.Exception e) {
				e.printStackTrace();
			}
		}
		super.mouseButtonDown(button);
	}

	@Override
	public void mouseButtonUp(Qt.MouseButton button) throws Exception {
		if(altDown==true) {
			super.mouseButtonUp(button);
			return;
		}
		
		Widget[] widgets=getPrioritySortedWidgets();
		
		for(int i=widgets.length-1;i>=0;i--) {
			try {
				if(widgets[i].mouseButtonUp(button)) {
					return;
				}
			} catch (java.lang.Exception e) {
				e.printStackTrace();
			}
		}
		super.mouseButtonUp(button);
	}

	@Override
	public void mouseMove(int oldX, int oldY, int newX, int newY) throws Exception {
		if(altDown) {
			super.mouseMove(oldX,oldY,newX,newY);
			return;
		}
		
		Widget[] widgets=getPrioritySortedWidgets();
		
		for(int i=widgets.length-1;i>=0;i--) {
			try {
				if(widgets[i].mouseMove(oldX,oldY,newX,newY)) {
					return;
				}
			} catch (java.lang.Exception e) {
				e.printStackTrace();
			}
		}
		super.mouseMove(oldX,oldY,newX,newY);
	}
}
