/*
 * Moonlight|3D
 * Copyright (C) 2008 The Moonlight|3D team
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Created on Jul 18, 2008
 */

package eu.moonlight3d.graph.layout;

import java.util.ArrayList;
import java.util.List;

import com.trolltech.qt.core.QPointF;

import eu.moonlight3d.graph.GraphModel;
import eu.moonlight3d.graph.Link;
import eu.moonlight3d.graph.Node;
import eu.moonlight3d.graph.Slot;

public class SimpleLayouter implements Layouter {
	private int nodeDistance=10;
	
	private void placeUnlinkedNode(GraphModel model, Node node) {
		// find bottom-most node in model, set this node below to the far left
		
		double maxY=0;
		double minX=0;
		
		for(Node nodeIter : model.getNodes()) {
			if(nodeIter.getPosition().x()<minX) {
				minX=nodeIter.getPosition().x();
			}
			if(nodeIter.getRect().y()+nodeIter.getRect().height()>maxY) {
				maxY=nodeIter.getRect().y()+nodeIter.getRect().height();
			}
		}
		maxY+=nodeDistance;
		
		node.setPosition(minX, maxY);
	}
	
	private QPointF suggestPlacement(ArrayList<Node> nodes, boolean left) {
		QPointF point=new QPointF(0,0);
		
		for(Node node : nodes) {
			point.setY(point.y()+node.getRect().y()+node.getRect().height()/2);
			if(left) {
				point.setX(point.x()+node.getRect().x());				
			} else {
				point.setX(point.x()+node.getRect().x()+node.getRect().width());								
			}
		}
		
		point.setX(point.x()/nodes.size());
		if(left) {
			point.setX(point.x()-nodeDistance);
		} else {
			point.setX(point.x()+nodeDistance);
		}
		point.setY(point.y()/nodes.size());
		
		return point;
	}
	
	private void placeNode(GraphModel model, Node node, List<Node> ignoreNodes) {
		ArrayList<Node> inputNodes=new ArrayList<Node>();
		ArrayList<Node> outputNodes=new ArrayList<Node>();
		
		for(Slot slot : node.getSlots()) {
			if(slot.getLink()!=null) {
				if(slot.getType()==Slot.Type.Input) {
					Node linkedNode=slot.getLink().getStartSlot().getNode();
					if(ignoreNodes==null || ignoreNodes.contains(linkedNode)) {
						inputNodes.add(linkedNode);
					}
				} else {
					Node linkedNode=slot.getLink().getEndSlot().getNode();
					if(ignoreNodes==null || ignoreNodes.contains(linkedNode)) {
						outputNodes.add(linkedNode);
					}
				}
			}
		}
		
		if(inputNodes.size()==0 && outputNodes.size()==0) {
			placeUnlinkedNode(model, node);
		} else if(outputNodes.size()==0) {
			QPointF placement=suggestPlacement(inputNodes, false);
			// center node vertically around suggested location
			placement.setY(placement.y()-(node.getRect().height()/2));
			
			node.setPosition(placement.x(), placement.y());
		} else if(inputNodes.size()==0) {
			QPointF placement=suggestPlacement(outputNodes, true);
			// center node vertically around suggested location
			placement.setY(placement.y()-(node.getRect().height()/2));
			// align node horizontally
			placement.setX(placement.x()-node.getRect().width());
			
			node.setPosition(placement.x(), placement.y());
		} else {
			QPointF placement=suggestPlacement(inputNodes, true);
			placement.add(suggestPlacement(outputNodes,true));
			placement.divide(2);
			// center node vertically around suggested location
			placement.setY(placement.y()-(node.getRect().height()/2));
			// center node horizontally around suggested location
			placement.setX(placement.x()-(node.getRect().width()/2));
			
			node.setPosition(placement.x(), placement.y());
		}
	}
	
	private void recenter(GraphModel model) {
		QPointF center=new QPointF();
		for(Node node : model.getNodes()) {
			center.add(node.getPosition());
		}
		center.divide(model.getNodes().size());
		
		for(Node node : model.getNodes()) {
			node.setPosition(node.getPosition().subtract(center));
		}
	}
	
	public void insertLink(GraphModel model, Link link) {
//		placeNode(model,link.getStartSlot().getNode(),null);
		placeNode(model,link.getEndSlot().getNode(),null);
	}

	public void layout(GraphModel model) {
		List<Node> nodeList=model.getNodes();
		
		while(nodeList.size()>0) {
			Node node=nodeList.get(nodeList.size()-1);
			nodeList.remove(nodeList.size()-1);
			placeNode(model,node,nodeList);
		}
		
		recenter(model);
	}

	public void place(GraphModel model, Node node) {
		// TODO Auto-generated method stub
		placeNode(model, node, null);
	}

}
